function [modelH,success] = createCompositionAsModel(importerObj, compositionName, varargin)
%CREATECOMPOSITIONASMODEL creates an AUTOSAR Composition-Software-Component as a
%   Simulink model.
%
%   createCompositionAsModel(IMPORTEROBJ, 'COMPOSITIONNAME') creates the Simulink
%   model corresponding to the AUTOSAR Composition-Software-Component 'COMPOSITIONNAME'
%   described in the XML file imported by the arxml.importer object IMPORTEROBJ.
%
%   Usage:
%      [modelH, success] = createCompositionAsModel(importerObj,...
%                                                   compositionName, varargin)
%
%   Inputs:
%      importerObj:   arxml.importer object
%      compositionName: absolute Short-Name path of the Composition-Software-Component
%      varargin:      Parameter/Value pairs. Supported parameters are:
%
%         ModelPeriodicRunnablesAs: Auto (default), AtomicSubsystem,
%         or FunctionCallSubsystem. By default, createCompositionAsModel
%         attempts to model periodic runnables in the atomic software
%         components of the composition as atomic subsystems, if possible.
%         If conditions prevent use of atomic subsystems, it will model
%         them as function-call subsystems.
%
%         ComponentModels: Cell array of names of existing Simulink models
%         to use when importing the composition. The importer incorporates
%         the existing atomic component models instead of creating new ones.
%
%         DataDictionary: Character vector specifying a Simulink data
%         dictionary into which to import data objects corresponding to
%         AUTOSAR data types in the XML file. Note that if the specified
%         dictionary does not already exist, the importer creates it.
%
%         PredefinedVariant: Character vector specifying the path to a
%         PredefinedVariant defined in the AUTOSAR XML file. For example,
%         '/pkg/body/Variants/Senior'.
%
%         SystemConstValueSets: Cell array of character vectors specifying
%         paths to one or more SwSystemconstantValueSets defined in the
%         AUTOSAR XML file. For example, {'/pkg/body/SystemConstantValues/A',
%         '/pkg/body/SystemConstantValues/B'}.
%
%   Outputs:
%      modelH:  Model handle.
%      success: It is true if this function is successful. Otherwise, it is false.

%   Copyright 2016-2017 The MathWorks, Inc.

% NOTE, when updating the interface of this function (including 
% optional arguments and their possible values), also update the
% corresponding JSON file for tab completion located in
% matlabroot/toolbox/coder/autosar/functionSignatures.json and add
% relevant testing in
% matlabroot/test/toolbox/rtw/targets/AUTOSAR/API/TabCompletion
% this will help ensure keeping tab completion up to date

% convert strings to chars
for ii = 1 : length(varargin)
    if isstring(varargin{ii})
        varargin{ii} = convertStringsToChars(varargin{ii});
    end
end
% parse arguments
compositionArgParser = autosar.composition.mm2sl.private.ArgumentParser(varargin{:});

% read ARXML files
p_update_read(importerObj);
importerObj.needReadUpdate = true;

try
    % find m3iComposition
    validateattributes(compositionName, {'char','string'}, {'scalartext'},...
        '','compositionName', 2);
    compositionName = convertStringsToChars(compositionName);
    m3iTopComposition = autosar.mm.Model.findChildByName(importerObj.arModel, compositionName);
    if isempty(m3iTopComposition) || ~isa(m3iTopComposition,...
            'Simulink.metamodel.arplatform.composition.CompositionComponent')
        DAStudio.error('autosarstandard:importer:badImporterCompositionName',...
            'CompositionSwComponent', compositionName);
    end
    
    % error out if a model with the same name exist in the current folder
    autosar.mm.mm2sl.ModelBuilder.checkModelFileName(m3iTopComposition.Name, 'error');
    
    % Check if the DataDictionary exists
    autosar.mm.mm2sl.utils.checkAndCreateDD(compositionArgParser.DataDictionary);
    
    % create AUTOSAR Dictionary file if necessary
    if (slfeature('AUTOSARDictionary') > 0) && ~isempty(compositionArgParser.AutosarDictionaryFile)
        autosar.dictionary.create(compositionArgParser.AutosarDictionaryFile);
    end
    
    % get xml opts
    xmlOptsGetter = autosar.mm.util.XmlOptionsGetter(importerObj.arModel);
    
    % import all the atomic software components (bottom-up)
    componentBuilder = autosar.composition.mm2sl.AtomicSwComponentBuilder(...
        importerObj, xmlOptsGetter, compositionArgParser);
    componentBuilder.importOrUpdateAtomicComponents(m3iTopComposition);
    
    % import all the composition components (bottom-up)
    compositionBuilder = autosar.composition.mm2sl.CompositionSwComponentBuilder(...
        importerObj, xmlOptsGetter, compositionArgParser);
    modelNames = compositionBuilder.importOrUpdateCompositionComponents(...
        m3iTopComposition);
    
catch Me
    %Trim stack to hide internal composition creation functions
    autosar.mm.util.MessageReporter.throwException(Me);
end

modelH = get_param(modelNames{end}, 'Handle');
success = modelH~=-1;

% LocalWords:  metamodel arplatform autosarstandard Sw IMPORTEROBJ scalartext
% LocalWords:  COMPONENTNAME COMPOSITIONNAME Systemconstant JSON json
