classdef importer < matlab.mixin.SetGet & matlab.mixin.Copyable
%arxml.importer constructs an arxml.importer object.
%    arxml.importer(FILENAME) constructs an object for importing AUTOSAR 
%    software components described in the XML file FILENAME. 
%
%    arxml.importer methods:
%       createComponentAsModel   - create model for AUTOSAR atomic software component
%       createCompositionAsModel - create model for AUTOSAR software composition
%       getComponentNames        - get AUTOSAR software component names from arxml files
%       updateModel              - update model with arxml changes
%       updateReferences         - update model with arxml definitions of AUTOSAR reference elements

%   Copyright 2007-2017 The MathWorks, Inc.

properties (Access=private)
    arSchemaVer = '';
    file = [];
    dependencies = [];
    arModel;
    needReadUpdate logical = false;
end

methods (Access = public)
    
    % constructor
    function this = importer(filename)
        % NOTE, when updating the interface of this function (including 
        % optional arguments and their possible values), also update the 
        % corresponding JSON file for tab completion located in
        % matlabroot/toolbox/coder/autosar/functionSignatures.json and add
        % relevant testing in
        % matlabroot/test/toolbox/rtw/targets/AUTOSAR/API/TabCompletion
        % this will help ensure keeping tab completion up to date
        
        if nargin > 0
            filename = convertStringsToChars(filename);
        end
        
        if nargin < 1
            DAStudio.error('RTW:autosar:badImporterArgument');
        end
        
        autosar.spkg.loadresources;
        try
            p_importer(this, filename);
        catch Me
            %Trim stack trace to remove internal import functions
           autosar.mm.util.MessageReporter.throwException(Me);
        end
    end
    
    % methods defined in other files
    [modelH,success] = createComponentAsModel(this,ComponentName,varargin);
    [modelH,success] = createCompositionAsModel(importerObj,compositionName,varargin);
    success = createCalibrationComponentObjects(this,ComponentName,varargin);
    compList = getComponentNames(this,compKind);
    updateModel(this,modelName,varargin);
    updateReferences(this,modelName,varargin);
    display(this);
end % public methods

methods (Hidden, Access = public)
    compList = getApplicationComponentNames(this);
    compList = getSensorActuatorComponentNames(this);
    compList = getCompositionComponentNames(this);
    compList = getCalibrationComponentNames(this);
    p_importUnitsToDatabase(importerObj,filename);
    [mmChangeLoggers,slChangeLoggers] = p_updateModel(this,modelName,varargin);
    [mmChangeLogger,slChangeLogger] = p_component_updateModel(this,modelName,varargin);
    paths = find(this,rootPath,category,varargin);
    intfs = getClientServerInterfaceNames(this);
    obj = saveobj(obj);
    dependencies = getDependencies(this);
    filename = getFile(this);
    setDependencies(this,aDependencies);
    setFile(this,aFilename);
    
    function schemaVer = getSchemaVer(this)
        schemaVer = this.arSchemaVer;
    end
    function m3iModel = getM3IModel(this)
        m3iModel = this.arModel;
    end
    function setNeedReadUpdate(this, value)
        assert(islogical(value), 'needReadUpdate must be a boolean value.');
        this.needReadUpdate = value;
    end
end % hidden public methods

end  % classdef

% LocalWords:  JSON json
