function create( modelName, creationMode, varargin )
% autosar.api.create instantiate the mapping and AUTOSAR properties
% information for a Simulink model
% 
% autosar.api.create(MDLNAME, MODE) instantiates the Simulink mapping and
% AUTOSAR properties for Simulink model MDLNAME. MODE is an optional
% parameter of either 'init' (default) or 'default'. When MODE is set to
% 'init' an AUTOSAR Software component is created and is mapped to
% the Simulink model. When MODE is set to 'default' the Simulink inports 
% and outports are additionally mapped with defaults created for the 
% corresponding AUTOSAR properties.
%
% Examples:
%
% To create the default AUTOSAR properties and mapping for a model
%    autosar.api.create('rtwdemo_autosar_counter', 'default');
%
% See also autosar.api.getAUTOSARProperties autosar.api.getSimulinkMapping



%   Copyright 2012-2018 The MathWorks, Inc.
    
    if nargin > 0
        modelName = convertStringsToChars(modelName);
    end

    if nargin > 1
        creationMode = convertStringsToChars(creationMode);
    end

    if nargin < 2
        % Default creationMode
        creationMode = 'init';
    end
    
    % Create an instance of the inputParser class to handle varargin
    argParser = inputParser;
    
    % Add arguments to the schema
    supportedComponentTypes = {'AtomicComponent', 'CompositionComponent'};
    argParser.addParameter('ComponentType', 'AtomicComponent', ...
        @(x) any(validatestring(x,supportedComponentTypes)));
    argParser.parse(varargin{:});

    
    % make sure that model is loaded in memory
    systems = find_system('type', 'block_diagram', 'name', modelName);
    if isempty(systems)
        DAStudio.error('RTW:autosar:mdlNotLoaded', modelName);
    end
    
    % Is it an appropriate target?
    isCompliant = strcmp(get_param(modelName,'AutosarCompliant'), 'on');
    if ~isCompliant
        DAStudio.error('RTW:autosar:nonAutosarCompliant');
    end
    
    isCompositionSubDomain = strcmp(get_param(modelName,'SimulinkSubDomain'),...
        'AUTOSARComposition');
    
    if ~isCompositionSubDomain
        % error out for composition models
        if autosar.api.Utils.isMappedToComposition(modelName)
            DAStudio.error('autosarstandard:api:CompositionMappingNotSupported', modelName);
        end
    end
    componentType = argParser.Results.ComponentType;
    if autosar.api.Utils.isAdaptiveModelSetup(modelName)
        if slfeature('RTWAdaptiveAutosar') > 0 && ...
                ~strcmp(componentType, 'CompositionComponent') % we do not allow adaptive compositions at the moment
            componentType = 'AdaptiveApplication';
        else
            % feature flag is not set, error out
            DAStudio.error('autosarstandard:validation:adaptiveFeatureNotSet', modelName);
        end
    end
        
    % Check modeling style is valid
    try
    autosar.validation.ClientServerValidator.checkValidRunnableConfig(modelName);

    % Check for Service Component blocks
    bswServices = autosar.bsw.ServiceComponent.find(modelName);
    if ~isempty(bswServices)
        DAStudio.error('autosarstandard:bsw:createDefaultMapping', bswServices{1});
    end
    
    % Close the UI as creating a new metamodel will make the UI stale.
    autosar_ui_close(modelName);
    cp = simulinkcoder.internal.CodePerspective.getInstance;
    if cp.isInPerspective(modelName)
        editors = GLUE2.Util.findAllEditors(modelName);
        for ii = 1 : numel(editors)
            simulinkcoder.internal.CodePerspective.getInstance.togglePerspective(editors(ii));
        end
    end
    
    % get appropriate mappingType for componentType
    mappingType = i_getMappingType(componentType);
    
    switch creationMode
      case 'init'
        % Create initial valid mapping and meta-model
        i_createEmptyAutosarMapping(modelName, mappingType);

        % Add Default AtomicComponent
        componentBuilder = autosar.ui.wizard.builder.Component(modelName, modelName);
        compQName = [componentBuilder.ComponentPackage, '/', componentBuilder.ComponentName];

        dataObj = autosar.api.getAUTOSARProperties(modelName, true);
        dataObj.addComponent(compQName, 'Category', componentType);
      
      case 'default'
        % Setup stage name
        Simulink.output.Stage(...
            message('coderdictionary:mapping:StageCreateDefaultCompMsg').getString(), ...
            'ModelName', modelName, 'UIMode', false);
        
        % Override previous mapping and meta-model with default values
        
        i_createEmptyAutosarMapping(modelName, mappingType);
        
        % set defaults using wizard functionality
        componentBuilder = autosar.ui.wizard.builder.Component(modelName, modelName);

        modelH = get_param(modelName, 'Handle');
        componentBuilder.setDefaultConfiguration(modelH);
        
        % configure m3iModel
        m3iModel = autosar.api.Utils.m3iModel(modelName);
        t = M3I.Transaction(m3iModel);
        componentBuilder.populateMetamodel(m3iModel, modelName);
        t.commit();
        
        % configure mapping
        componentBuilder.buildMapping(modelName, 'cmdline');
        
      otherwise
        DAStudio.error('RTW:autosar:invalidCreationMode', creationMode);
    end
    catch Me
        %Trim stack trace to suppress internal functions
        autosar.mm.util.MessageReporter.throwException(Me);
    end

function i_createEmptyAutosarMapping(modelName, mappingType)
    % Create empty AutosarTarget mapping. 

    mmgr = get_param(modelName, 'MappingManager');
    [~] = Simulink.AutosarTarget.Component('', ''); % workaround g866593
    [~] = Simulink.AutosarTarget.Composition('', ''); % workaround g866593
    mapping = mmgr.getActiveMappingFor(mappingType);
    if ~isempty(mapping)
        mapping.unmap();
        mmgr.deleteMapping(mapping);
    end
    
    mmgr.createMapping(modelName, mappingType);
    mmgr.activateMapping(modelName);
    mapping = mmgr.getActiveMappingFor(mappingType);
    
    m3iModel = Simulink.metamodel.foundation.Factory.createNewModel();
    
    % Start m3i processing
    t = M3I.Transaction(m3iModel);
    m3iModel.Name = 'AUTOSAR';
    % Create the AUTOSAR package
    autosarPkg = Simulink.metamodel.arplatform.common.AUTOSAR(m3iModel);
    autosarPkg.Name = 'AUTOSAR';
    m3iModel.RootPackage.append(autosarPkg);    
    % End m3i processing
    t.commit();      
    
    mapping.AUTOSAR_ROOT = m3iModel;
    % Set PreSave Fcn
    b = get_param(modelName,'Object');
    if ~b.hasCallback('PreSave', 'AutosarSaveAsMdl')
        b.addCallback('PreSave','AutosarSaveAsMdl', @()autosarcore.checkSaveAsMdl(b.Handle));
    end

function mappingType = i_getMappingType(componentType)
    switch (componentType)
        case {'AtomicComponent', 'AdaptiveApplication'}
            mappingType = 'AutosarTarget';
        case 'CompositionComponent'
            mappingType = 'AutosarComposition';
        otherwise
            assert(false, 'invalid componentType: %s', componentType);
    end

% LocalWords:  MDLNAME metamodel autosarstandard cmdline rtwdemo bsw ui
% LocalWords:  coderdictionary
