classdef getSimulinkMapping < handle
% autosar.api.getSimulinkMapping API class.
%
% MAPPINGOBJ = autosar.api.getSimulinkMapping(MDLNAME) creates an object
% MAPPINGOBJ that can be used to query and set the Simulink mapping of
% Simulink model MDLNAME
% getSimulinkMapping methods:
%    getDataTransfer   - get the mapping information for a Simulink data transfer line
%    getFunction       - get the mapping information for a Simulink entry point function
%    getInport         - get the mapping information for a Simulink inport
%    getOutport        - get the mapping information for a Simulink outport
%    getFunctionCaller - get the mapping information for a Simulink function caller block
%    getLookupTable    - get the mapping information for a Simulink lookup table
%    getSignal         - get the mapping information for a Simulink block signal
%    getState          - get the mapping information for a Simulink block state
%    getParameter      - get the mapping information for a Simulink model workspace parameter   
%    mapDataTransfer   - map a Simulink data transfer line
%    mapFunction       - map a Simulink entry point function
%    mapInport         - map an inport
%    mapOutport        - map an outport
%    mapFunctionCaller - map a function caller
%    mapLookupTable    - map a lookup table
%    mapSignal         - map a Simulink block signal
%    mapState          - map a Simulink block state
%    mapParameter      - map a Simulink model workspace parameter
%
% See also autosar.api.syncModel autosar.api.create autosar.api.getAUTOSARProperties

    
%   Copyright 2012-2018 The MathWorks, Inc.
    
    properties(Access=private, Transient=true)
        ModelName;
        ChangeLogger;
    end
    
    properties(Constant, Access = private)
        ValidReceiverDAMs = {'ImplicitReceive'
                             'ExplicitReceive'
                             'ExplicitReceiveByVal'
                             'QueuedExplicitReceive'
                             'ErrorStatus'
                             'IsUpdated'
                             'EndToEndRead'
                             'EndToEndQueuedReceive'};
        ValidSenderDAMs   = {'ImplicitSend'
                             'ImplicitSendByRef'
                             'ExplicitSend'
                             'EndToEndWrite'
                             'QueuedExplicitSend'
                             'EndToEndQueuedSend'};
        ValidNvReceiverDAMs = {'ImplicitReceive'
                               'ExplicitReceive'
                               'ExplicitReceiveByVal'};
        ValidNvSenderDAMs   = {'ImplicitSend'
                               'ImplicitSendByRef'
                               'ExplicitSend'};
        ValidParameterAccessModes = {'PortParameter'
                               'PerInstance'
                               'Shared'
                               'ConstantMemory'};
        DummyServiceRequiredPortDAM = 'ImplicitReceive';
        DummyServiceProvidedPortDAM = 'ImplicitSend';
    end
    
    
    methods(Access=public)
        function this = getSimulinkMapping(modelName, changeLogger)
            % NOTE, when updating the interface of this function (including
            % optional arguments and their possible values), also update the
            % corresponding JSON file for tab completion located in
            % matlabroot/toolbox/coder/autosar/functionSignatures.json and add
            % relevant testing in
            % matlabroot/test/toolbox/rtw/targets/AUTOSAR/API/TabCompletion
            % this will help ensure keeping tab completion up to date
            
            this.ModelName = modelName;
            
            if nargin>1
                this.ChangeLogger = changeLogger;
            else
                this.ChangeLogger = [];
            end
        end
           
        function [ARIRVName, ARDataAccessMode] = getDataTransfer(this, SLDataTransferName)
           % [ARIRVNAME, ARDATAACCESSMODE] = GETDATATRANSFER(THIS, SLDATATRANSFERNAME) are
           % the corresponding AUTOSAR InterRunnableVariable ARIRVNAME and
           % DataAccessMode ARDATAACCESSMODE values mapped to the Simulink data
           % transfer line SLDATATRANSFERNAME
           %
           % See also mapDataTransfer

           if ~(ischar(SLDataTransferName) || isStringScalar(SLDataTransferName))
               DAStudio.error('autosarstandard:validation:invalidDataTransferName');
           end
           
           try
               modelMapping = autosar.api.Utils.modelMapping(this.ModelName);
           catch Me
               %Trim stack to suppress mapping internal functions
               autosar.mm.util.MessageReporter.throwException(Me);
           end
           findDataTransfer = false;
           %check whether the data transfer is mapped from Implicit
           %RTBs/signal lines
           SLDataTransfer = modelMapping.DataTransfers.findobj('SignalName', SLDataTransferName);
           if length(SLDataTransfer) == 1   % found data transfer from the signal line mapping
               findDataTransfer = true;
           else
               if ~autosar.api.getSimulinkMapping.isValidRateTransitionBlock(SLDataTransferName)
                   DAStudio.error('RTW:autosar:invalidMappingDataTransfer', SLDataTransferName);
               end
               SLDataTransfer = modelMapping.RateTransition.findobj('Block',SLDataTransferName);
               
               if length(SLDataTransfer) == 1   % found data transfer from the rtb mapping 
                   findDataTransfer = true;
               end
           end
           
           % no Data Transfer mapping is found
           if ~findDataTransfer
                DAStudio.error('RTW:autosar:invalidMappingDataTransfer', SLDataTransferName);
           end
           
           ARIRVName = SLDataTransfer.MappedTo.IrvName;
           ARDataAccessMode = SLDataTransfer.MappedTo.IrvAccessMode; 
        end
        
        function [ARRunnableName, ARRunnableSwAddrMethod, ARInternalDataSwAddrMethod] = getFunction(this, SLFcnName)
        % ARRUNNABLENAME = GETFUNCTION(THIS, SLFCNNAME) is the AUTOSAR 
        % Runnable name ARRUNNABLENAME mapped to Simulink entry point
        % function or Simulink function SLFCNNAME.
        %
        % See also mapFunction
           
            try
                modelMapping = autosar.api.Utils.modelMapping(this.ModelName);
            catch Me
                %Trim stack to suppress mapping internal functions
                autosar.mm.util.MessageReporter.throwException(Me);
            end
            SLFcnName = autosar.api.getSimulinkMapping.escapeSimulinkName(SLFcnName);
            isExportStyle = autosar.validation.ExportFcnValidator.isTopModelExportFcn(this.ModelName);
            
            isMultiTasking = length(modelMapping.StepFunctions) > 1;
            
            switch SLFcnName
                case 'InitializeFunction'
                    SLFunction = modelMapping.InitializeFunctions(1);
                case 'TerminateFunction'
                    SLFunction = modelMapping.TerminateFunctions(1);
                case 'StepFunction'
                    if isExportStyle
                        DAStudio.error('autosarstandard:validation:invalidMappingFcnForExportFcnModel', SLFcnName);
                    else
                        SLFunction = modelMapping.StepFunctions(1);
                    end
                otherwise
                    resetFunction = autosar.api.getSimulinkMapping.getResetFunction(modelMapping, SLFcnName);
                    if ~isempty(resetFunction)
                        SLFunction = resetFunction;
                    else
                        SLFunction = modelMapping.FcnCallInports.findobj( ...
                            'Block', [modelMapping.Name, '/', SLFcnName]);
                        
                        if ~isMultiTasking && length(SLFunction) ~= 1
                            SLBlock = autosar.api.Utils.getBlockNameFromFunctionName(this.ModelName,...
                                SLFcnName, false);
                            SLFunction = modelMapping.ServerFunctions.findobj('Block', SLBlock);
                        end
                        if isMultiTasking && length(SLFunction) ~= 1
                            tid = regexp(SLFcnName,'^StepFunction(\d+)','tokens');
                            if ~isempty(tid)
                                tid = str2double(tid{1});
                                if ((tid >= 1) && ...
                                        (tid < length(modelMapping.StepFunctions)) )
                                    SLFunction = modelMapping.StepFunctions(tid+1);
                                else
                                    DAStudio.error('autosarstandard:validation:invalidMappingFcnForRateBasedModel', SLFcnName);
                                end
                            else
                                DAStudio.error('autosarstandard:validation:invalidMappingFcnForRateBasedModel', SLFcnName);
                            end
                        end
                        if length(SLFunction) ~= 1
                            DAStudio.error('autosarstandard:validation:invalidMappingFcnForExportFcnModel', SLFcnName);
                        end
                    end
            end
            
            ARRunnableName = SLFunction.MappedTo.Runnable;     % Runnable Name
            ARRunnableSwAddrMethod = SLFunction.MappedTo.SwAddrMethod;   % SwAddrMethod for Runnable
            if isempty(ARRunnableSwAddrMethod)
                ARRunnableSwAddrMethod = DAStudio.message('RTW:autosar:uiUnselectOptions');
            end
            ARInternalDataSwAddrMethod = SLFunction.MappedTo.InternalDataSwAddrMethod; % SwAddrMethod for Internal Data
            if isempty(ARInternalDataSwAddrMethod)
                ARInternalDataSwAddrMethod = DAStudio.message('RTW:autosar:uiUnselectOptions');
            end
        end
        
        
        
        function [ARPortName, ARElementName, ARDataAccessMode] = getInport(this, SLPortName)
           % [ARPORTNAME, ARDATAELEMENTNAME, ARDATAACCESSMODE] = GETINPORT(THIS,
           % SLPORTNAME) are the corresponding AUTOSAR Receiver (or
           % SenderReceiver) port ARPORTNAME, DataElement ARDATAELEMENTNAME and
           % DataAccessMode ARDATAACCESSMODE values mapped to Simulink Inport
           % SLPORTNAME, SLPORTNAME should be the one given by the
           % function get_param(block, 'Name') to ensure no errors
           %
           % See also mapInport
           
               try
                   modelMapping = autosar.api.Utils.modelMapping(this.ModelName);
               catch Me
                   %Trim stack to suppress mapping internal functions
                   autosar.mm.util.MessageReporter.throwException(Me);
               end
               SLPortName = autosar.api.getSimulinkMapping.escapeSimulinkName(SLPortName);
               
               SLPort = modelMapping.Inports.findobj('Block', [modelMapping.Name, '/', SLPortName]);
           
               if length(SLPort) ~= 1
                   DAStudio.error('RTW:autosar:invalidMappingPort', SLPortName, 'Simulink Port', SLPortName);
               end

               ARPortName = SLPort.MappedTo.Port;
               ARElementName = SLPort.MappedTo.Element;
               ARDataAccessMode = SLPort.MappedTo.DataAccessMode;
           
           end
        
         function [ARPortName, ARElementName, ARDataAccessMode] = getOutport(this, SLPortName)
           % [ARPORTNAME, ARDATAELEMENTNAME, ARDATAACCESSMODE] = GETOUTPORT(THIS,
           % SLPORTNAME) are the corresponding AUTOSAR Sender (or
           % SenderReceiver) port ARPORTNAME, DataElement ARDATAELEMENTNAME and
           % DataAccessMode ARDATAACCESSMODE values mapped to Simulink Outport
           % SLPORTNAME, SLPORTNAME should be the one given by the
           % function get_param(block, 'Name') to ensure no errors
           %
           % See also mapOutport

               try
                   modelMapping = autosar.api.Utils.modelMapping(this.ModelName);
               catch Me
                   %Trim stack to suppress mapping internal functions
                   autosar.mm.util.MessageReporter.throwException(Me);
               end
               SLPortName = autosar.api.getSimulinkMapping.escapeSimulinkName(SLPortName);
               SLPort = modelMapping.Outports.findobj('Block', [modelMapping.Name, '/', SLPortName]);
           
               if length(SLPort) ~= 1
                   DAStudio.error('RTW:autosar:invalidMappingPort', SLPortName, 'Simulink Port', SLPortName);
               end

               ARPortName = SLPort.MappedTo.Port;
               ARElementName = SLPort.MappedTo.Element;
               ARDataAccessMode = SLPort.MappedTo.DataAccessMode;
                          
         end
         
         function [ARPortName, AROperationName] = getFunctionCaller(this, SLFcnName)
             % [ARPORTNAME, AROPERATIONNAME] = GETFUNCTIONCALLER(THIS,
             % SLFCNNAME) are the corresponding AUTOSAR Client
             % port ARPORTNAME, Operation AROPERATIONNAME values
             % mapped to Simulink function caller SLFCNNAME.
             % 
             % See also mapFunctionCaller
             
             try
                 modelMapping = autosar.api.Utils.modelMapping(this.ModelName);
             catch Me
                 %Trim stack to suppress mapping internal functions
                 autosar.mm.util.MessageReporter.throwException(Me);
             end
             SLFcnName = strrep(SLFcnName, newline, ' ');
             SLBlock = autosar.api.Utils.getBlockNameFromFunctionName(this.ModelName,...
                 SLFcnName, true);
             SLPort = modelMapping.FunctionCallers.findobj('Block', SLBlock);
             
             if length(SLPort) ~= 1
                 DAStudio.error('RTW:autosar:invalidFunctionCallerBlock', SLFcnName);
             end
             
             ARPortName = SLPort.MappedTo.ClientPort;
             AROperationName = SLPort.MappedTo.Operation;
             
         end
 
        function mapDataTransfer(this, SLDataTransferName, ARIRVName, ARDataAccessMode)
        % MAPDATATRANSFER(THIS, SLDATATRANSFERNAME, ARIRVNAME, ARDATAACCESSMODE) maps
        % Simulink data transfer line SLDATATRANSFERNAME to
        % InterRunnableVariable ARIRVNAME. DataAccessMODE ARDATAACCESSMODE
        % can be one of 'Implicit', 'Explicit'
        %
        % See also getDataTransfer
           
            if ~(ischar(SLDataTransferName)|| isStringScalar(SLDataTransferName))
                DAStudio.error('autosarstandard:validation:invalidDataTransferName');
            end
        
            try
                modelMapping = autosar.api.Utils.modelMapping(this.ModelName);
                dataObj = autosar.api.getAUTOSARProperties(this.ModelName, true);
            catch Me
                %Trim stack to suppress internal mapping functions
                autosar.mm.util.MessageReporter.throwException(Me);
            end
            findSignal = false;
            findRTB = false;
            
            % check the data transfers mapped from signal lines/hidden
            % RTBs.
            SLDataTransfer = modelMapping.DataTransfers.findobj('SignalName', SLDataTransferName);
            if length(SLDataTransfer) == 1  % found map from signal lines/Hidden RTBs
                findSignal = true;
                
                % check IRV path
                componentQualifiedName = dataObj.get('XmlOptions', 'ComponentQualifiedName');
                ARIRVPaths = dataObj.find(componentQualifiedName, 'IrvData', ...
                    'Name', ARIRVName, 'PathType', 'FullyQualified');
                if length(ARIRVPaths) ~= 1
                    DAStudio.error('RTW:autosar:invalidMappingDataTransferIrv', ...
                        SLDataTransferName, ARIRVName);
                end
            else % check the data transfers mapped from explicit RTBs
                if ~autosar.api.getSimulinkMapping.isValidRateTransitionBlock(SLDataTransferName)
                    DAStudio.error('RTW:autosar:invalidMappingDataTransfer', SLDataTransferName);
                end
                
                if ~(strcmp(get_param(SLDataTransferName,'Integrity'), 'on') && ...
                       strcmp(get_param(SLDataTransferName,'Deterministic'), 'off'))
                   DAStudio.error(...
                       'autosarstandard:validation:RateTranBlkForIRV',...
                       SLDataTransferName);
                end
                
                SLDataTransfer = modelMapping.RateTransition.findobj('Block',SLDataTransferName);
                
                if length(SLDataTransfer) == 1  % found map from explicit RTBs
                    findRTB = true;
                    
                    % check IRV path
                    componentQualifiedName = dataObj.get('XmlOptions', 'ComponentQualifiedName');
                    ARIRVPaths = dataObj.find(componentQualifiedName, 'IrvData', ...
                        'Name', ARIRVName, 'PathType', 'FullyQualified');
                    if length(ARIRVPaths) ~= 1
                        DAStudio.error('autosarstandard:validation:IRVNotExistForDataTransfer', ...
                            SLDataTransferName, ARIRVName);
                    end
                end 
            end
            
            if ~findSignal && ~findRTB
                DAStudio.error('RTW:autosar:invalidMappingDataTransfer', SLDataTransferName);
            end
            
            if ~isempty(this.ChangeLogger)
                if ~strcmp(SLDataTransfer.MappedTo.IrvName, ARIRVName) &&...
                        ~isempty(SLDataTransfer.MappedTo.IrvName)
                    this.ChangeLogger.logModification('Automatic', 'IRVName', ...
                        'Simulink data transfer', ...
                        SLDataTransferName, ...
                        SLDataTransfer.MappedTo.IrvName, ...
                        ARIRVName);
                end
                    
                if ~strcmp(SLDataTransfer.MappedTo.IrvAccessMode, ARDataAccessMode)
                    this.ChangeLogger.logModification('Automatic', 'DataAccessMode', ...
                        'Simulink data transfer', ...
                        SLDataTransferName, ...
                        SLDataTransfer.MappedTo.IrvAccessMode, ...
                        ARDataAccessMode);
                end
            end
            
            if findSignal
                SLDataTransfer.mapInterRunnableVariable(SLDataTransferName, ARIRVName, ARDataAccessMode, '');   
            else
                SLDataTransfer.mapInterRunnableVariable(ARIRVName, ARDataAccessMode, ''); 
            end
            
        end
        
        function mapInport(this, SLPortName, ARPortName, ARDataName, ARDataAccessModeStr)
            % MAPINPORT(THIS, SLPORTNAME, ARPORTNAME, ARDATAELEMENTNAME, ARDATAACCESSMODE)
            % maps Simulink Inport SLPORTNAME to DataElement ARDATAELEMENTNAME
            % of AUTOSAR Receiver (or SenderReceiver) port
            % ARPORTNAME. SLPORTNAME should be the one given by the
            % function get_param(block, 'Name') to ensure no errors
            % DataAccessMode ARDATAACCESSMODESTR can be one of
            % 'ImplicitReceive', 'ExplicitReceive', 'ExplicitReceiveByVal', 
            % 'QueuedExplicitReceive', 'ErrorStatus', 'IsUpdated', 
            % 'EndToEndRead'.
            %
            % See also getInport
            
            
            try
                modelMapping = autosar.api.Utils.modelMapping(this.ModelName);
                dataObj = autosar.api.getAUTOSARProperties(this.ModelName, true);
            catch Me
                %Trim stack to suppress internal mapping functions
                autosar.mm.util.MessageReporter.throwException(Me);
            end
            SLPortName = autosar.api.getSimulinkMapping.escapeSimulinkName(SLPortName);
            SLPort = modelMapping.Inports.findobj('Block', [modelMapping.Name, '/', SLPortName]);
           
            % SLPort Check
            if length(SLPort) ~= 1
                    DAStudio.error('RTW:autosar:invalidMappingPort', SLPortName, 'Simulink Port', SLPortName);
            end

            componentQualifiedName = dataObj.get('XmlOptions', 'ComponentQualifiedName');
            
            % ARPort Check
            if autosar.api.Utils.isMappedToAdaptiveApplication(this.ModelName)
                portType = 'ServiceRequiredPort';
            elseif strcmp(ARDataAccessModeStr, 'ModeReceive')
                portType = 'ModeReceiverPort';
            else
                portType = 'DataReceiverPort';
            end
            ARPortPaths = dataObj.find(componentQualifiedName, portType, ...
                 'Name', ARPortName, 'PathType', 'FullyQualified');
             
             % May be trying to map a SenderReceiverPort
             if isempty(ARPortPaths) && strcmp(portType, 'DataReceiverPort')
                 ARPortPaths = dataObj.find(componentQualifiedName, 'DataSenderReceiverPort', ...
                     'Name', ARPortName, 'PathType', 'FullyQualified');
                 if ~isempty(ARPortPaths)
                     % Found a DataSenderReceiverPort
                     portType = 'DataSenderReceiverPort';
                 end
             end
             
             % May be trying to map a NvDataReceiverPort
             if isempty(ARPortPaths) && strcmp(portType, 'DataReceiverPort')
                 ARPortPaths = dataObj.find(componentQualifiedName, 'NvDataReceiverPort', ...
                     'Name', ARPortName, 'PathType', 'FullyQualified');
                 if ~isempty(ARPortPaths)
                     % Found a NvDataReceiverPort
                     portType = 'NvDataReceiverPort';
                 end
             end
             
             % May be trying to map a NvDataSenderReceiverPort
             if isempty(ARPortPaths) && strcmp(portType, 'DataReceiverPort')
                 ARPortPaths = dataObj.find(componentQualifiedName, 'NvDataSenderReceiverPort', ...
                     'Name', ARPortName, 'PathType', 'FullyQualified');
                 if ~isempty(ARPortPaths)
                     % Found a NvDataSenderReceiverPort
                     portType = 'NvDataSenderReceiverPort';
                 end
             end
             
             if isempty(ARPortPaths) && strcmp(portType, 'ModeReceiverPort')
                 % After 3.x import SenderReceiverInterface referenced by a DataReceiverPort may have ModeDeclarationGroupElement
                 ARPortPaths = dataObj.find(componentQualifiedName, 'DataReceiverPort', ...
                     'Name', ARPortName, 'PathType', 'FullyQualified');
             end
             
            if length(ARPortPaths) ~= 1
                DAStudio.error('RTW:autosar:invalidMappingPort', SLPortName, portType, ARPortName);
            end
            
            if strcmp(portType, 'ServiceRequiredPort')
                ARDataAccessModeStr = autosar.api.getSimulinkMapping.DummyServiceRequiredPortDAM;
            end
            
            % ARDataElement Check           
            ifPath = dataObj.get(ARPortPaths{1}, 'Interface', 'PathType', 'FullyQualified');
            if strcmp(ARDataAccessModeStr, 'ModeReceive')
                metaModelClassName = 'ModeDeclarationGroupElement';
            else
                metaModelClassName = 'FlowData';
            end
            ARDataElementPaths = dataObj.find(ifPath, metaModelClassName, ...
                'Name', ARDataName, 'PathType', 'FullyQualified');
            if isempty(ifPath) || length(ARDataElementPaths) ~= 1
                DAStudio.error('RTW:autosar:invalidMappingPort', SLPortName, 'DataElement', ARDataName);
            end
            
            switch portType
              case 'DataReceiverPort'
                validDataAccessModes = this.ValidReceiverDAMs;
              case 'DataSenderReceiverPort'
                validDataAccessModes = {'ImplicitReceive', ...
                                    'ExplicitReceive', ...
                                    'ExplicitReceiveByVal', ...
                                    'ErrorStatus', ...
                                    'IsUpdated', ...
                                    'EndToEndRead'};
              case {'NvDataReceiverPort', 'NvDataSenderReceiverPort'}
                validDataAccessModes = this.ValidNvReceiverDAMs;
              case 'ModeReceiverPort'
                validDataAccessModes = {'ModeReceive'};
              case 'ServiceRequiredPort'
                  validDataAccessModes = {autosar.api.getSimulinkMapping.DummyServiceRequiredPortDAM};
              otherwise
                assert(false, 'Did not recognize portType %s', portType);
            end
                
            
            % ARDataAccessModeStr Check
            if ~any(strcmp(ARDataAccessModeStr, validDataAccessModes))
                validDataAccessModesStr = autosar.api.Utils.cell2str(validDataAccessModes);
                DAStudio.error('RTW:autosar:invalidMappingDAM', ARDataAccessModeStr, SLPortName, validDataAccessModesStr);
            end

            if ~isempty(this.ChangeLogger)
                if ~strcmp(SLPort.MappedTo.Port, ARPortName)
                    this.ChangeLogger.logModification('Automatic', 'ARPortName', ...
                                                      'Simulink inport', ...
                                                      SLPort.Block, ...
                                                      SLPort.MappedTo.Port, ...
                                                      ARPortName);
                end
                    
                if ~strcmp(SLPort.MappedTo.Element, ARDataName)
                    this.ChangeLogger.logModification('Automatic', 'ARElementName', ...
                                                      'Simulink inport', ...
                                                      SLPort.Block, ...
                                                      SLPort.MappedTo.Element, ...
                                                      ARDataName);
                end
                
                if ~strcmp(SLPort.MappedTo.DataAccessMode, ARDataAccessModeStr)
                    this.ChangeLogger.logModification('Automatic', 'DataAccessMode', ...
                                                      'Simulink inport', ...
                                                      SLPort.Block, ...
                                                      SLPort.MappedTo.DataAccessMode, ...
                                                      ARDataAccessModeStr);
                end
            end
            
            try
                if strcmp(portType, 'ServiceRequiredPort')
                    % skip adding com-spec
                else
                    % add or Update ComSpec Object according to Data Access Mode
                    autosar.mm.sl2mm.ComSpecBuilder.addOrUpdateM3IComSpec(ARPortName, ...
                        ARDataName, ARDataAccessModeStr, this.ModelName);
                end
                
                 SLPort.mapPortElement(ARPortName, ARDataName, ARDataAccessModeStr, '');
            catch Me
                %Trim stack to suppress internal comSpec mapping functions
                autosar.mm.util.MessageReporter.throwException(Me);
            end


            % Dirty the model
            set_param(this.ModelName,'Dirty','on');
        end
        
        function mapOutport(this, SLPortName, ARPortName, ARDataName, ARDataAccessModeStr)
            % MAPOUTPORT(THIS, SLPORTNAME, ARPORTNAME, ARDATAELEMENTNAME, ARDATAACCESSMODE)
            % maps Simulink Outport SLPORTNAME to DataElement ARDATAELEMENTNAME
            % of AUTOSAR Sender (or SenderReceiver) port
            % ARPORTNAME. SLPORTNAME should be the one given by the
            % function get_param(block, 'Name') to ensure no errors
            % DataAccessMode ARDATAACCESSMODESTR can be one of
            % 'ImplicitSend', 'ImplicitSendByRef', 'ExplicitSend', 
            % 'EndToEndWrite'.
            %
            % See also getOutport
            
            try
                modelMapping = autosar.api.Utils.modelMapping(this.ModelName);
                dataObj = autosar.api.getAUTOSARProperties(this.ModelName, true);
            catch Me
                %Trim stack to suppress internal mapping functions
                autosar.mm.util.MessageReporter.throwException(Me);
            end
            SLPortName = autosar.api.getSimulinkMapping.escapeSimulinkName(SLPortName);
            SLPort = modelMapping.Outports.findobj('Block', [modelMapping.Name, '/', SLPortName]);
            
            if length(SLPort) ~= 1                  
                DAStudio.error('RTW:autosar:invalidMappingPort', SLPortName, 'Simulink Port', SLPortName);               
            end
            
            if autosar.api.Utils.isMappedToAdaptiveApplication(this.ModelName)
                portType = 'ServiceProvidedPort';
            elseif strcmp(ARDataAccessModeStr, 'ModeSend')
                portType = 'ModeSenderPort';
            else
                portType = 'DataSenderPort';
            end
            
            % ARPort Check
            componentQualifiedName = dataObj.get('XmlOptions', 'ComponentQualifiedName');
            ARPortPaths = dataObj.find(componentQualifiedName, portType, ...
                'Name', ARPortName, 'PathType', 'FullyQualified');
            
            % May be trying to map a SenderReceiverPort
            if isempty(ARPortPaths)
                ARPortPaths = dataObj.find(componentQualifiedName, 'DataSenderReceiverPort', ...
                    'Name', ARPortName, 'PathType', 'FullyQualified');
            end
            
            % May be trying to map a NvDataSenderPort
            if isempty(ARPortPaths)
                ARPortPaths = dataObj.find(componentQualifiedName, 'NvDataSenderPort', ...
                    'Name', ARPortName, 'PathType', 'FullyQualified');
                if ~isempty(ARPortPaths)
                    % Found a NvDataSenderPort
                    portType = 'NvDataSenderPort';
                end
            end
            
            % May be trying to map a NvDataSenderReceiverPort
            if isempty(ARPortPaths)
                ARPortPaths = dataObj.find(componentQualifiedName, 'NvDataSenderReceiverPort', ...
                    'Name', ARPortName, 'PathType', 'FullyQualified');
                if ~isempty(ARPortPaths)
                    % Found a NvDataSenderPort
                    portType = 'NvDataSenderPort';
                end
            end
            
            if isempty(ARPortPaths) && strcmp(portType, 'ModeSenderPort')
                % After 3.x import SenderReceiverInterface referenced by
                % a DataSenderPort may have ModeDeclarationGroupElement??
                
                ARPortPaths = dataObj.find(componentQualifiedName, 'DataSenderPort', ...
                    'Name', ARPortName, 'PathType', 'FullyQualified');
            end
            
            if length(ARPortPaths) ~= 1
                DAStudio.error('RTW:autosar:invalidMappingPort', SLPortName, portType, ARPortName);
            end
            
            if strcmp(portType, 'ServiceProvidedPort')
                ARDataAccessModeStr = autosar.api.getSimulinkMapping.DummyServiceProvidedPortDAM;
            end
            
            % ARDataElement Check           
            ifPath = dataObj.get(ARPortPaths{1}, 'Interface', 'PathType', 'FullyQualified');
            if strcmp(ARDataAccessModeStr, 'ModeSend')
                metaModelClassName = 'ModeDeclarationGroupElement';
            else
                metaModelClassName = 'FlowData';
            end            
            ARDataElementPaths = dataObj.find(ifPath, metaModelClassName, ...
                'Name', ARDataName, 'PathType', 'FullyQualified');
            if isempty(ifPath) || length(ARDataElementPaths) ~= 1
                DAStudio.error('RTW:autosar:invalidMappingPort', SLPortName, 'DataElement', ARDataName);
            end
            
            % ARDataAccessModeStr Check
            switch portType
              case {'DataSenderPort', 'DataSenderReceiverPort'}
                validDataAccessModes = [this.ValidSenderDAMs; 'ModeSend']; % 3.x SenderReceiverInterface has ModeDeclGroups
              case {'NvDataSenderPort', 'NvDataSenderReceiverPort'}
                validDataAccessModes = this.ValidNvSenderDAMs;
              case 'ModeSenderPort'
                validDataAccessModes = {'ModeSend'};   
               case 'ServiceProvidedPort'
                validDataAccessModes = {autosar.api.getSimulinkMapping.DummyServiceProvidedPortDAM};
              otherwise
                assert(false, 'Did not recognize portType %s', portType);
            end
            
            
            % ARDataAccessModeStr Check
            if ~any(strcmp(ARDataAccessModeStr, validDataAccessModes))
                validDataAccessModesStr = autosar.api.Utils.cell2str(validDataAccessModes);
                DAStudio.error('RTW:autosar:invalidMappingDAM', ARDataAccessModeStr, SLPortName, validDataAccessModesStr);
            end
            
            if ~isempty(this.ChangeLogger)
                if ~strcmp(SLPort.MappedTo.Port, ARPortName)
                    this.ChangeLogger.logModification('Automatic', 'ARPortName', ...
                                                      'Simulink outport', ...
                                                      SLPort.Block, ...
                                                      SLPort.MappedTo.Port, ...
                                                      ARPortName);
                end
                    
                if ~strcmp(SLPort.MappedTo.Element, ARDataName)
                    this.ChangeLogger.logModification('Automatic', 'ARElementName', ...
                                                      'Simulink outport', ...
                                                      SLPort.Block, ...
                                                      SLPort.MappedTo.Element, ...
                                                      ARDataName);
                end
                
                % ------- Keeping Implicit send selection on update -------
                % In the AUTOSAR ARXML both ImplicitSend and 
                % ImplicitSendByRef are the expressed in the same way. When
                % updating we want to keep the previous user selection in
                % this case.
                implicitSendAccessModes = {'ImplicitSend', 'ImplicitSendByRef'};
                currIsImplicitSendAm = ismember(SLPort.MappedTo.DataAccessMode, ...
                                                implicitSendAccessModes);
                newIsImplicitSendAm = ismember(ARDataAccessModeStr, ...
                                               implicitSendAccessModes);
                if (currIsImplicitSendAm && newIsImplicitSendAm)
                    ARDataAccessModeStr = SLPort.MappedTo.DataAccessMode;
                end
                % ---------------------------------------------------------
                
                if ~strcmp(SLPort.MappedTo.DataAccessMode, ARDataAccessModeStr)
                    this.ChangeLogger.logModification('Automatic', 'DataAccessMode', ...
                                                      'Simulink outport', ...
                                                      SLPort.Block, ...
                                                      SLPort.MappedTo.DataAccessMode, ...
                                                      ARDataAccessModeStr);
                end
            end
            
            try
                if strcmp(portType, 'ServiceProvidedPort')
                    % skip adding com-spec
                else
                    % add or Update ComSpec Object according to Data Access Mode
                    autosar.mm.sl2mm.ComSpecBuilder.addOrUpdateM3IComSpec(ARPortName, ...
                        ARDataName, ARDataAccessModeStr, this.ModelName);
                end
                
                SLPort.mapPortElement(ARPortName, ARDataName, ARDataAccessModeStr, '');
            catch Me
                %Trim stack to suppress internal ComSpec mapping functions
                autosar.mm.util.MessageReporter.throwException(Me);
            end
            
            % Dirty the model
            set_param(this.ModelName,'Dirty','on');
        end
        
        
        function mapFunction(this, SLFcnName, ARRunnableName, varargin)
            % MAPFUNCTION(THIS, SLFCNNAME, ARRUNNABLENAME) maps Simulink entry point
            % function or Simulink function SLFCNNAME to AUTOSAR Runnable 
            % ARRUNNABLENAME.
            %
            % See also getFunction
            
            argParser = inputParser;
            argParser.FunctionName = 'autosar.api.getSimulinkMapping.mapFunction';
            argParser.KeepUnmatched = true;
            argParser.addRequired('this', @(x) isa(x, class(x)));
            argParser.parse(this, varargin{:});                    
            
            try
                modelMapping = autosar.api.Utils.modelMapping(this.ModelName);
            catch Me
                %Trim stack to suppress mapping internal functions
                autosar.mm.util.MessageReporter.throwException(Me);
            end
            % ARRunnableName cannot be empty.
            if isempty(ARRunnableName)
                DAStudio.error('RTW:autosar:invalidRunnableName',...
                    ARRunnableName);
            end
            
            SLFcnName = autosar.api.getSimulinkMapping.escapeSimulinkName(SLFcnName);
            isExportStyle = autosar.validation.ExportFcnValidator.isTopModelExportFcn(this.ModelName);
            isMultiTasking = length(modelMapping.StepFunctions) > 1;
            
            mappingObj = [];
            % SwAddrMethod for internal data is only configurable for
            % execution functions - step, export, async, Simulink functions
            isSwAddrMethodConfigurableForInternalData = false;
            switch SLFcnName
                case 'InitializeFunction'
                    mappingObj = modelMapping.InitializeFunctions(1);
                case 'TerminateFunction'
                    mappingObj = modelMapping.TerminateFunctions(1);
                case 'StepFunction'
                    if ~isExportStyle
                        mappingObj = modelMapping.StepFunctions(1);
                        isSwAddrMethodConfigurableForInternalData = true;
                    end
                otherwise
                    SLFcnCallInport = modelMapping.FcnCallInports.findobj( ...
                        'Block', [modelMapping.Name, '/', SLFcnName]);
                    if length(SLFcnCallInport) == 1
                        mappingObj = SLFcnCallInport;
                        isSwAddrMethodConfigurableForInternalData = true;
                    end
                    if ~isMultiTasking && isempty(mappingObj)
                        SLBlock = autosar.api.Utils.getBlockNameFromFunctionName(this.ModelName,...
                            SLFcnName, false);
                        SLServerFunction = modelMapping.ServerFunctions.findobj( ...
                            'Block', SLBlock);
                        if length(SLServerFunction) == 1
                            mappingObj = SLServerFunction;
                            isSwAddrMethodConfigurableForInternalData = true;
                        end
                    end
                    if isempty(mappingObj)
                        mappingObj = autosar.api.getSimulinkMapping.getResetFunction(modelMapping, SLFcnName);
                    end
                    if isMultiTasking && isempty(mappingObj)
                        tid = regexp(SLFcnName,'^StepFunction(\d+)','tokens');
                        if ~isempty(tid)
                            tid = str2double(tid{1});
                            if ((tid >= 1) && ...
                                    (tid < length(modelMapping.StepFunctions)) )
                                mappingObj = modelMapping.StepFunctions(tid+1);
                                isSwAddrMethodConfigurableForInternalData = true;
                            else
                                DAStudio.error('autosarstandard:validation:invalidMappingFcnForRateBasedModel', SLFcnName);
                            end
                        else
                            DAStudio.error('autosarstandard:validation:invalidMappingFcnForRateBasedModel', SLFcnName);
                        end
                    end
            end
            
            if isempty(mappingObj)
                DAStudio.error('autosarstandard:validation:invalidMappingFcnForExportFcnModel', SLFcnName);
            end

            dataObj = autosar.api.getAUTOSARProperties(this.ModelName, true);
            componentQualifiedName = dataObj.get('XmlOptions', 'ComponentQualifiedName');
            componentCategory = dataObj.get(componentQualifiedName, 'Category');
            try
                switch componentCategory
                    case 'AtomicComponent'
                        autosar.api.Utils.mapFunction(this.ModelName,...
                            mappingObj, ARRunnableName);
                        params = argParser.Unmatched;
                        if ~isempty(fields(params))
                            instSpecificPropertyNames = fieldnames(params);
                            instSpecificPropertyValues = struct2cell(params);
                            for ii = 1 : numel(instSpecificPropertyNames)
                                propertyName = instSpecificPropertyNames{ii};
                                propertyName = validatestring(propertyName,...
                                    {'SwAddrMethod', 'SwAddrMethodForInternalData'});  
                                
                                propertyValue = instSpecificPropertyValues{ii};
                                
                                % Validate property values
                                validateattributes(propertyValue, ...
                                    {'char', 'string'}, {'nonempty'}, ...
                                    'autosar.api.getSimulinkMapping.mapFunction', ...
                                    propertyName)
                                
                                if strcmpi(propertyValue, DAStudio.message('RTW:autosar:uiUnselectOptions'))
                                    propertyValue = '';
                                end
                                m3iModel = autosar.api.Utils.m3iModel(this.ModelName);
                                if strcmpi(propertyName, 'SwAddrMethod')
                                    propNameForM3I = 'Runnable';
                                    mapFcn = 'mapSwAddrMethod';
                                elseif strcmpi(propertyName, 'SwAddrMethodForInternalData')
                                    % Error out if SwAddrMethodForInternalData is configured
                                    % for non-execution functions
                                    if ~isSwAddrMethodConfigurableForInternalData
                                        DAStudio.error('autosarstandard:validation:invalidRunnableForInternalDataSwAddrMethod', SLFcnName);
                                    end
                                    propNameForM3I = 'RunnableInternalData';
                                    mapFcn = 'mapInternalDataSwAddrMethod';
                                end
                                
                                % Error out if specified SwAddrMethod is invalid
                                [allowedSwAddrMethods, validSectionTypes] = ...
                                    autosar.mm.util.SwAddrMethodHelper.findSwAddrMethodsForCategory(...
                                        m3iModel, propNameForM3I);
                                if ~isempty(propertyValue) && ...
                                        ~any(strcmp(propertyValue, allowedSwAddrMethods))                                    
                                    DAStudio.error('autosarstandard:validation:invalidSwAddrMethodForMapping', ...
                                                   SLFcnName, propertyValue, propertyName, autosar.api.Utils.cell2str(validSectionTypes));
                                end
                                mappingObj.(mapFcn)(propertyValue);
                            end
                        end
                    otherwise
                        assert(false, 'Did not recognize componentCategory %s', componentCategory);
                end
            catch Me
                %Trim stack to suppress internal functions
                autosar.mm.util.MessageReporter.throwException(Me);
            end
            
            % Dirty the model
            set_param(this.ModelName,'Dirty','on');
        end
        
        function mapFunctionCaller(this, SLFcnName, ARPortName, AROperationName)
            % MAPFUNCTIONCALLER(THIS, SLFCNNAME, ARPORTNAME, AROPERATIONNAME)
            % maps Simulink Function caller SLFCNNAME to Operation AROPERATIONNAME
            % of AUTOSAR Client port ARPORTNAME.
            %
            % See also getFunctionCaller
            
            % convert strings to chars
            if isstring(ARPortName)
                ARPortName = convertStringsToChars(ARPortName);
            end
            
            try
                modelMapping = autosar.api.Utils.modelMapping(this.ModelName);
                dataObj = autosar.api.getAUTOSARProperties(this.ModelName, true);
            catch Me
                %Trim stack to suppress mapping internal functions
                autosar.mm.util.MessageReporter.throwException(Me);
            end
                       
            SLFcnName = strrep(SLFcnName, newline, ' ');
            SLBlock = autosar.api.Utils.getBlockNameFromFunctionName(this.ModelName,...
                            SLFcnName, true);
            SLBlockMap = modelMapping.FunctionCallers.findobj('Block', SLBlock);              
            
            % SLBlockMap Check
            if length(SLBlockMap) ~= 1
                DAStudio.error('RTW:autosar:invalidCallerMapping', SLFcnName, 'function caller', SLFcnName);
            end

            componentQualifiedName = dataObj.get('XmlOptions', 'ComponentQualifiedName');

            % ARPort Check
            portType = 'ClientPort';
            ARPortPaths = dataObj.find(componentQualifiedName, portType, ...
                 'Name', ARPortName, 'PathType', 'FullyQualified');
             
            if length(ARPortPaths) ~= 1
                DAStudio.error('RTW:autosar:invalidCallerMapping', SLFcnName, portType, ARPortName);
            end
            
            % AROperationName Check           
            ifPath = dataObj.get(ARPortPaths{1}, 'Interface', 'PathType', 'FullyQualified');
            metaModelClassName = 'Operation';
            AROperationPaths = dataObj.find(ifPath, metaModelClassName, ...
                'Name', AROperationName, 'PathType', 'FullyQualified');
            if isempty(ifPath) || length(AROperationPaths) ~= 1
                DAStudio.error('RTW:autosar:invalidCallerMapping', SLFcnName, 'Operation', AROperationName);
            end                
            

            if ~isempty(this.ChangeLogger)
                if ~strcmp(SLBlockMap.MappedTo.ClientPort, ARPortName)
                    this.ChangeLogger.logModification('Automatic', 'ARPortName', ...
                                                      'Simulink Function Caller block', ...
                                                      SLBlockMap.Block, ...
                                                      SLBlockMap.MappedTo.ClientPort, ...
                                                      ARPortName);
                end
                    
                if ~strcmp(SLBlockMap.MappedTo.Operation, AROperationName)
                    this.ChangeLogger.logModification('Automatic', 'AROperationName', ...
                                                      'Simulink Function Caller block', ...
                                                      SLBlockMap.Block, ...
                                                      SLBlockMap.MappedTo.Operation, ...
                                                      AROperationName);
                end
            end
            
            componentCategory = dataObj.get(componentQualifiedName, 'Category');
            for index=1:length(modelMapping.FunctionCallers)
                if strcmp(get_param(modelMapping.FunctionCallers(index).Block, 'FunctionPrototype'),...
                        get_param(SLBlockMap.Block, 'FunctionPrototype'))
                    % Call function in utils to do the mapping
                    try
                        switch componentCategory
                          case {'AtomicComponent'} 
                            autosar.api.Utils.mapCaller(this.ModelName, modelMapping.FunctionCallers(index), ...
                                                        ARPortName, AROperationName);
                          otherwise
                            assert(false, 'Did not recognize componentCategory %s', componentCategory);
                        end
                    catch Me
                        autosar.mm.util.MessageReporter.throwException(Me);
                    end
                end
            end

            % Dirty the model
            set_param(this.ModelName,'Dirty','on');
        end
        
        function [ARParameterAccessMode, ARPortName, ARParameterData] = getLookupTable(this, SLParam)
            % [ARPARAMETERACCESSMODE, ARPORTNAME, ARPARAMETERDATA] = GETLOOKUPTABLE(THIS,
            % SLPARAM) are the corresponding ParameterAccessMode
            % ARPARAMETERACCESSMODE, AUTOSAR Parameter Receiver port
            % ARPORTNAME and Parameter Data ARPARAMETERDATA values mapped
            % to Simulink Lookup Table SLPARAM.
            %
            % See also mapLookupTable
            try
                modelMapping = autosar.api.Utils.modelMapping(this.ModelName);
            catch Me
                %Trim stack to remove internal mapping functions
                autosar.mm.util.MessageReporter.throwException(Me);
            end
            SLLut = modelMapping.LookupTables.findobj('LookupTableName', SLParam);
            
            if length(SLLut) ~= 1
                DAStudio.error('autosarstandard:api:invalidMapping', 'Lookup Table',...
                    SLParam, 'Simulink Lookup Table', SLParam);
            end
            
            ARParameterAccessMode = SLLut.MappedTo.ParameterAccessMode;
            ARPortName = SLLut.MappedTo.Port;
            ARParameterData = SLLut.MappedTo.Parameter;
        end
        
        function mapLookupTable(this, SLParam, ARParameterAccessMode, ARPortName, ARParameterData)
            % MAPLOOKUPTABLE(THIS, SLPARAM, ARPARAMETERACCESSMODE, ARPORTNAME, ARPARAMETERDATA)
            % maps Simulink Lookup Table SLPARAM to ParameterAccessMode
            % ARPARAMETERACCESSMODE, AUTOSAR Parameter Receiver Port
            % ARPORTNAME and Parameter Data ARPARAMETERDATA.
            % ParameterAccessMode ARPARAMETERACCESSMODE can be one of
            % PortParameter, Shared, PerInstance and Const.
            %
            % Examples: 
            % Map Simulink lookup table L_4_single to AUTOSAR Parameter
            % with ParameterAccessMode Shared.
            %
            % mappingObj = autosar.api.getSimulinkMapping(modelName);
            % mappingObj.mapLookupTable('L_4_single', 'Shared', '',...
            %   'L_4_single_ar');
            %
            % Map Simulink lookup table L_4_single to AUTOSAR Parameter
            % to AUTOSAR Parameter Receiver Port paramRPort and Parameter
            % paramElem1.
            %
            % mappingObj = autosar.api.getSimulinkMapping(modelName);
            % mappingObj.mapLookupTable('L_4_single', 'PortParameter',...
            %   'paramRPort', 'paramElem1');
            %
            % See also getLookupTable
            
            try
                modelMapping = autosar.api.Utils.modelMapping(this.ModelName);
                dataObj = autosar.api.getAUTOSARProperties(this.ModelName, true);
            catch Me
                %Trim to suppress mapping internal functions
                autosar.mm.util.MessageReporter.throwException(Me);
            end
            
            SLLUTMap = modelMapping.LookupTables.findobj('LookupTableName', SLParam);
            
            % SL LookupTable Check
            if length(SLLUTMap) ~= 1
                DAStudio.error('autosarstandard:api:invalidMapping', 'Lookup Table',...
                    SLParam, 'Simulink Lookup Table', SLParam);
            end
            
            if strcmp(ARParameterAccessMode, 'PortParameter')
                % ARPort Check
                portType = 'ParameterReceiverPort';
                componentQualifiedName = dataObj.get('XmlOptions', 'ComponentQualifiedName');
                ARPortPaths = dataObj.find(componentQualifiedName, portType, ...
                    'Name', ARPortName, 'PathType', 'FullyQualified');
                
                if length(ARPortPaths) ~= 1
                    DAStudio.error('autosarstandard:api:invalidMapping', 'Lookup Table', ...
                        SLParam, portType, ARPortName);
                end
                
                % ARParameterName Check
                ifPath = dataObj.get(ARPortPaths{1}, 'Interface', 'PathType', 'FullyQualified');
                metaModelClassName = 'ParameterData';
                ARParameterDataPaths = dataObj.find(ifPath, metaModelClassName, ...
                    'Name', ARParameterData, 'PathType', 'FullyQualified');
                if isempty(ifPath) || length(ARParameterDataPaths) ~= 1
                    DAStudio.error('autosarstandard:api:invalidMapping', 'Lookup Table', ...
                        SLParam, 'Parameter Data', ARParameterData);
                    
                end
            end
            % ARParameterAccessMode Check
            if ~any(strcmp(ARParameterAccessMode, this.ValidParameterAccessModes))
                validParameterDataRolesStr = autosar.api.Utils.cell2str(this.ValidParameterAccessModes);
                DAStudio.error('autosarstandard:api:invalidParameterMappingRole',...
                    ARParameterAccessMode, 'Simulink Lookup Table', SLLUTMap.LookupTableName,...
                    validParameterDataRolesStr);
            end
            SLLUTMap.mapLookupTable(SLLUTMap.LookupTableName, ARParameterAccessMode,...
                ARPortName, ARParameterData, '');
            % Dirty the model
            set_param(this.ModelName,'Dirty','on');
        end
        
        function Value = getSignal(this, PortHandle, Property)
            % VALUE = GETSIGNAL(THIS, PORTHANDLE)
            % is the signal mapping of Simulink block output port PORTHANDLE.
            %
            % VALUE = GETSIGNAL(THIS, PORTHANDLE, PROPERTY)
            % is the value for PROPERTY inside signal mapping of Simulink
            % block output port PORTHANDLE.
            %
            %   PortHandle Handle of a block output port, where signal or
            %              testpoint is specified
            %   Property   The AUTOSAR properties associated with the
            %              ArtypedPerInstanceMemory or StaticMemory. For
            %              example: ShortName, IsVolatile,
            %              Qualifier, SwAddrMethod, SwCalibrationAccess,
            %              DisplayFormat etc.
            %
            % Examples:
            % Get Simulink signal mapping for outport of block RelOpt,
            % port 1 of model rtwdemo_autosar_counter.
            %
            % mappingObj = autosar.api.getSimulinkMapping('rtwdemo_autosar_counter');
            % Value = getSignal(mappingObj, PortHandle);
            %
            % PortHandle can be obtained form block as follows:
            % PortHandles = get_param('rtwdemo_autosar_counter/RelOpt', 'PortHandles');
            % PortHandle = PortHandles.Outport;
            %
            % Get AUTOSAR property ShortName of Simulink signal mapping
            % for outport of block RelOpt, port 1 for model rtwdemo_autosar_counter.
            % mappingObj = autosar.api.getSimulinkMapping('rtwdemo_autosar_counter');
            % Value = getSignal(mappingObj, PortHandle, 'ShortName');
            
            % See also mapSignal
            
            try
                modelMapping = autosar.api.Utils.modelMapping(this.ModelName);
            catch Me
                %Trim stack to suppress mapping internal functions
                autosar.mm.util.MessageReporter.throwException(Me);
            end
            PortHandleStr = '';
            if ~ischar(PortHandle) && ~isstring(PortHandle)
                PortHandleStr = num2str(PortHandle);
            end
            
            try
                objectType = get_param(PortHandle, 'Type');
            catch ME
                DAStudio.error('coderdictionary:api:invalidMappingSimulinkObject',...
                    'signal mapping', 'port handle', PortHandleStr);
            end
            if ~strcmp(objectType, 'port')
                DAStudio.error('coderdictionary:api:invalidMappingSimulinkObject',...
                    'signal mapping', 'port handle', PortHandleStr);
            end
            
            if nargin == 3
                validateattributes(Property, {'string', 'char'}, {'nonempty'},...
                    'autosar.api.getSimulinkMapping.getSignal', 'PROPERTY', 3);
            end
            Value = '';
            
            SLSignals = modelMapping.Signals.findobj('PortHandle', PortHandle);
            if length(SLSignals) < 1
                DAStudio.error('coderdictionary:api:invalidMappingSignalPort',...
                    PortHandleStr, PortHandleStr);
            end
            SLSignal = SLSignals;
            
            mappedTo = SLSignal.MappedTo;
            if nargin > 2
                if ~strcmp(mappedTo.ArDataRole, DAStudio.message('coderdictionary:mapping:NoMapping'))
                    Property = validatestring(Property,...
                        mappedTo.getPerInstancePropertyNames(true),...
                        'autosar.api.getSimulinkMapping.getSignal', 'PROPERTY', 3);
                    Value = mappedTo.getPerInstancePropertyValue(Property);
                end
            else
                Value = mappedTo.ArDataRole;
            end
        end
        
        function mapSignal(this, PortHandle, VariableRole, varargin)
            % MAPSIGNAL(THIS, PORTHANDLE, VARIABLEROLE, 'PARAMETER1', VALUE1, 'PARAMETER2',VALUE2,...)
            % maps Simulink signal specified on block output port
            % PORTHANDLE value VARIABLEROLE.
            %
            %   PortHandle Handle of a block output port, where signal or
            %              testpoint is specified
            %   VariableRole Auto, ArtypedPerInstanceMemory, or StaticMemory
            %   varargin   The AUTOSAR properties (key,value) to be
            %              specified only when value of VariableRole is
            %              ArtypedPerInstanceMemory or StaticMemory.
            %
            % Examples:
            % Maps Simulink signal originating from block RelOpt, port 1 to
            % AUTOSAR ArTypedPerInstanceMemory.
            %
            % mappingObj = autosar.api.getSimulinkMapping('rtwdemo_autosar_counter');
            % mapSignal(mappingObj, PortHandle, 'ArTypedPerInstanceMemory');
            %
            % PortHandle can be obtained form block as follows:
            % PortHandles = get_param('rtwdemo_autosar_counter/RelOpt', 'PortHandles');
            % PortHandle = PortHandles.Outport;
            %
            % Maps Simulink signal originating from block RelOpt, port 1 to
            % AUTOSAR StaticMemory with ShortName SigMem.
            %
            % mappingObj = autosar.api.getSimulinkMapping('rtwdemo_autosar_counter');
            % mapSignal(mappingObj, PortHandle, 'StaticMemory', 'ShortName', ...
            %       'SigMem');
            %
            % See also getSignal
            argParser = inputParser;
            argParser.FunctionName = 'autosar.api.getSimulinkMapping.mapSignal';
            argParser.KeepUnmatched = true;
            argParser.addRequired('this', @(x) isa(x, class(x)));
            argParser.addRequired('VariableRole', @(x)((ischar(x) || isStringScalar(x)) && strlength(x) > 0));
            
            validateattributes(VariableRole, {'string', 'char'}, {'nonempty'},...
                'autosar.api.getSimulinkMapping.mapSignal', 'PROPERTY', 3);
            VariableRole = validatestring(VariableRole, {DAStudio.message('coderdictionary:mapping:NoMapping'),...
                'ArTypedPerInstanceMemory', 'StaticMemory'});
            argParser.parse(this, VariableRole, varargin{:});
            
            try
                modelMapping = autosar.api.Utils.modelMapping(this.ModelName);
            catch Me
                %Trim stack to suppress mapping internal functions
                autosar.mm.util.MessageReporter.throwException(Me);
            end
            PortHandleStr = '';
            if ~ischar(PortHandle) && ~isstring(PortHandle)
                PortHandleStr = num2str(PortHandle);
            end
            try
                objectType = get_param(PortHandle, 'Type');
            catch ME
                DAStudio.error('coderdictionary:api:invalidMappingSimulinkObject',...
                    'signal mapping', 'port handle', PortHandleStr);
            end
            if ~strcmp(objectType, 'port')
                DAStudio.error('coderdictionary:api:invalidMappingSimulinkObject',...
                    'signal mapping', 'port handle', PortHandleStr);
            end
            
            SLSignals = modelMapping.Signals.findobj('PortHandle', PortHandle);
            if length(SLSignals) < 1
                DAStudio.error('coderdictionary:api:invalidMappingSignalPort',...
                    PortHandleStr, PortHandleStr);
            end
            SLSignal = SLSignals;
            this.setInternalDataProperties(this.ModelName,...
                SLSignal, VariableRole, true, argParser);
            % Dirty the model
            set_param(this.ModelName,'Dirty','on');
        end
        
        function Value = getState(this, Object, StateName, Property)
            % VALUE = GETSTATE(THIS, OBJECT)
            % is the state mapping of Simulink block OBJECT.
            %
            % VALUE = GETSTATE(THIS, OBJECT, STATENAME)
            % is the state mapping of owner block OBJECT with state STATENAME.
            % STATENAME is needed for blocks with multiple states. In case
            % STATENAME is empty then mapping of first state is returned.
            %
            % VALUE = GETSTATE(THIS, OBJECT, STATENAME, PROPERTY)
            % is the value for PROPERTY inside state mapping of owner block
            % OBJ. STATENAME is needed for blocks with multiple states. In
            % case STATENAME is empty then mapping of first state is returned.
            %
            %   Object     Name or handle of a block
            %   StateName  Name of state. Should be passed empty except
            %              block with multiple states.
            %   Property   The AUTOSAR properties associated with the
            %              ArtypedPerInstanceMemory or StaticMemory. For
            %              example: ShortName, IsVolatile,
            %              Qualifier, SwAddrMethod, SwCalibrationAccess, 
            %              DisplayFormat etc.
            %
            % Examples:
            % Get state mapping for state of unit state block X of model
            % rtwdemo_autosar_counter.
            %
            % mappingObj = autosar.api.getSimulinkMapping('rtwdemo_autosar_counter');
            % getState(mappingObj, blockHandle);
            %
            % Alternatively use block name as follows:
            % getState(mappingObj, 'rtwdemo_autosar_counter/X');
            %
            % Get AUTOSAR SHORT-NAME of state mapping for state of unit
            % state block X of model rtwdemo_autosar_counter.
            %
            % mappingObj = autosar.api.getSimulinkMapping('rtwdemo_autosar_counter');
            % getState(mappingObj, 'rtwdemo_autosar_counter/X', '', 'ShortName');
            
            % See also mapState
            slroot = Simulink.Root;
            if slroot.isValidSlObject(Object) && strcmp(get_param(Object, 'Type'), 'block')
                blockH = get_param(Object, 'Handle');
                SLBlockPath = getfullname(Object);
                if nargin >= 3
                    validateattributes(StateName, {'string', 'char'}, {},...
                        'autosar.api.getSimulinkMapping.getState', 'STATENAME', 3);
                else
                    StateName = '';
                end
                if nargin == 4
                    validateattributes(Property, {'string', 'char'}, {'nonempty'},...
                        'autosar.api.getSimulinkMapping.getState', 'PROPERTY', 4);
                end
                Value = '';
                try
                    modelMapping = autosar.api.Utils.modelMapping(this.ModelName);
                catch Me
                    %Trim stack to suppress mapping internal functions
                    autosar.mm.util.MessageReporter.throwException(Me);
                end
                SLStates = modelMapping.States.findobj('OwnerBlockHandle', blockH);
                if length(SLStates) < 1
                    DAStudio.error('coderdictionary:api:invalidMappingStateBlock',...
                        StateName, SLBlockPath, SLBlockPath);
                end
                if numel(SLStates) == 1
                    SLState = SLStates;
                    if ~isempty(StateName) && ~strcmp(SLState.Name, StateName)
                        DAStudio.error('coderdictionary:api:invalidMappingStateName',...
                            StateName, SLBlockPath, StateName);
                    end
                elseif isempty(StateName)
                    % In case user has not specific the state name return
                    % the first state.
                    SLState = SLStates(1);
                else
                    stateFound = false;
                    for ii = 1 : numel(SLStates)
                        SLState = SLStates(ii);
                        if strcmp(SLState.Name, StateName)
                            stateFound = true;
                            break;
                        end
                    end
                    if ~stateFound
                        DAStudio.error('coderdictionary:api:invalidMappingStateName',...
                            StateName, SLBlockPath, StateName);
                    end
                end
                mappedTo = SLState.MappedTo;
                if nargin > 3
                    if ~strcmp(mappedTo.ArDataRole, DAStudio.message('coderdictionary:mapping:NoMapping'))
                        Property = validatestring(Property,...
                            mappedTo.getPerInstancePropertyNames(true),...
                            'autosar.api.getSimulinkMapping.getState', 'PROPERTY', 4);
                        Value = mappedTo.getPerInstancePropertyValue(Property);
                    end
                else
                    Value = mappedTo.ArDataRole;
                end
            else
                % if block handle is specified instead of block path.
                if ~ischar(Object) && ~isstring(Object)
                    Object = num2str(Object);
                end
                DAStudio.error('coderdictionary:api:invalidMappingSimulinkObject',...
                    'state mapping', 'Simulink block', Object);
            end
        end
        
        function mapState(this, Object, StateName, VariableRole, varargin)
            % MAPSTATE(THIS, OBJECT, STATENAME, VARIABLEROLE, 'PARAMETER1', VALUE1, 'PARAMETER2',VALUE2,...)
            % maps Simulink state of owner block with path SLBLOCKPATH to
            % value VARIABLEROLE.
            %
            %   Object     Name or handle of a block
            %   StateName  The name of state. Should be passed empty except
            %              block with multiple states.
            %   VariableRole Auto, ArtypedPerInstanceMemory, or StaticMemory
            %   varargin   The AUTOSAR properties (key,value) to be
            %              specified only when value of VariableRole is
            %              ArtypedPerInstanceMemory or StaticMemory.
            %
            % Examples:
            % Maps Simulink state of owner block X to AUTOSAR
            % ArTypedPerInstanceMemory.
            %
            % mappingObj = autosar.api.getSimulinkMapping('rtwdemo_autosar_counter');
            % mapState(mappingObj, blockH, '', 'ArTypedPerInstanceMemory');
            %
            % Alternatively use block name as follows:
            % mapState(mappingObj, 'rtwdemo_autosar_counter/X', '', 'ArTypedPerInstanceMemory');
            %
            % Maps Simulink state of owner block X to AUTOSAR
            % StaticMemory with ShortName StateX.
            %
            % mappingObj = autosar.api.getSimulinkMapping('rtwdemo_autosar_counter');
            % mapState(mappingObj, blockHandle, '', 'StaticMemory',...
            %   'ShortName', 'StateX');
            %
            % See also getState
            slroot = Simulink.Root;
            if slroot.isValidSlObject(Object) && strcmp(get_param(Object, 'Type'), 'block')
                blockH = get_param(Object, 'Handle');
                SLBlockPath = getfullname(Object);               
                argParser = inputParser;
                argParser.FunctionName = 'autosar.api.getSimulinkMapping.mapState';
                argParser.KeepUnmatched = true;
                argParser.addRequired('this', @(x) isa(x, class(x)));
                argParser.addRequired('StateName', @(x)((ischar(x) || isStringScalar(x))));
                argParser.addRequired('VariableRole', @(x)((ischar(x) || isStringScalar(x)) && strlength(x) > 0));
                VariableRole = validatestring(VariableRole, {DAStudio.message('coderdictionary:mapping:NoMapping'),...
                    'ArTypedPerInstanceMemory', 'StaticMemory'});
                argParser.parse(this, StateName, VariableRole, varargin{:});
                
                try
                    modelMapping = autosar.api.Utils.modelMapping(this.ModelName);
                catch Me
                    %Trim stack to suppress mapping internal functions
                    autosar.mm.util.MessageReporter.throwException(Me);
                end
                SLStates = modelMapping.States.findobj('OwnerBlockHandle', blockH);
                if length(SLStates) < 1
                    DAStudio.error('coderdictionary:api:invalidMappingStateBlock',...
                        StateName, SLBlockPath, SLBlockPath);
                end
                if numel(SLStates) == 1 % In case block only has single state.
                    SLState = SLStates;
                    if ~isempty(StateName) && ~strcmp(SLState.Name, StateName)
                        DAStudio.error('coderdictionary:api:invalidMappingStateName',...
                            StateName, SLBlockPath, StateName);
                    end
                elseif isempty(StateName) % In case block only multiple single states.
                    % In case user has not specific the state name return
                    % the first state.
                    SLState = SLStates(1);
                else
                    % In case user specified the state name
                    stateFound = false;
                    for ii = 1 : numel(SLStates)
                        SLState = SLStates(ii);
                        if strcmp(SLState.Name, StateName)
                            stateFound = true;
                            break;
                        end
                    end
                    if ~stateFound
                        DAStudio.error('coderdictionary:api:invalidMappingStateName',...
                            StateName, SLBlockPath, StateName);
                    end
                end
                this.setInternalDataProperties(this.ModelName,...
                    SLState, VariableRole, true, argParser);
                % Dirty the model
                set_param(this.ModelName,'Dirty','on');
            else
                % if block handle is specified instead of block path.
                if ~ischar(Object) && ~isstring(Object)
                    Object = num2str(Object);
                end
                DAStudio.error('coderdictionary:api:invalidMappingSimulinkObject',...
                    'state mapping', 'Simulink block', Object);
            end
        end
        
        function Value = getParameter(this, SLParam, Property)
            % VALUE = GETPARAMETER(THIS, SLPARAM)
            % is the parameter mapping of Simulink model workspace parameter
            % SLPARAM.
            %
            % VALUE = GETPARAMETER(THIS, SLPARAM, PROPERTY)
            % is the value for PROPERTY inside parameter mapping of Simulink
            % model workspace parameter SLPARAM.
            %
            %   SLParam    Name of model workspace parameter.
            %   Property   The AUTOSAR properties associated with the
            %              SharedParameter or ConstantMemory. For
            %              example: IsConst, IsVolatile,
            %              Qualifier, SwAddrMethod, SwCalibrationAccess, 
            %              DisplayFormat etc.
            %
            % Examples:
            % Get Simulink parameter mapping for model workspace parameter
            % RESET of model rtwdemo_autosar_counter.
            %
            % mappingObj = autosar.api.getSimulinkMapping('rtwdemo_autosar_counter');
            % Value = getParameter(mappingObj, 'RESET');
            %
            % Get AUTOSAR property SwCalibrationAccess of Simulink parameter mapping
            % for model workspace parameter RESET of model rtwdemo_autosar_counter.
            % mappingObj = autosar.api.getSimulinkMapping('rtwdemo_autosar_counter');
            % Value = getParameter(mappingObj, 'RESET', 'SwCalibrationAccess');
            
            % See also mapParameter
            
            try
                modelMapping = autosar.api.Utils.modelMapping(this.ModelName);
            catch Me
                %Trim stack to suppress mapping internal functions
                autosar.mm.util.MessageReporter.throwException(Me);
            end
            validateattributes(SLParam, {'string', 'char'}, {'nonempty'},...
                'autosar.api.getSimulinkMapping.getParameter', 'SLPARAM', 2);
            if nargin == 3
                validateattributes(Property, {'string', 'char'}, {'nonempty'},...
                    'autosar.api.getSimulinkMapping.getParameter', 'PROPERTY', 3);
            end
            
            Value = '';
            mappedParam = modelMapping.ModelScopedParameters.findobj('Parameter', SLParam);
            if length(mappedParam) < 1
                DAStudio.error('coderdictionary:api:invalidMappingParameterName',...
                    SLParam, SLParam);
            end
            mappedTo = mappedParam.MappedTo;
            if nargin > 2
                if ~strcmp(mappedTo.ArDataRole, DAStudio.message('coderdictionary:mapping:NoMapping'))
                    Property = validatestring(Property,...
                        mappedTo.getPerInstancePropertyNames(false),...
                        'autosar.api.getSimulinkMapping.getParameter', 'PROPERTY', 3);
                    Value = mappedTo.getPerInstancePropertyValue(Property);
                end
            else
                Value = mappedTo.ArDataRole;
            end
        end
        
        function mapParameter(this, SLParam, ParameterRole, varargin)
            % MAPPARAMETER(THIS, SLPARAM, PARAMETERROLE, 'PARAMETER1', VALUE1, 'PARAMETER2', VALUE2,...)
            % maps Simulink model workspace parameter SLPARAM to value
            % PARAMETERROLE.
            %
            %   SLParam    Name of model workspace parameter.
            %   ParameterRole Auto, SharedParameter, or ConstantMemory.
            %   varargin   The AUTOSAR properties (key,value) to be
            %              specified only when value of ParameterRole is
            %              SharedParameter or ConstantMemory.
            %
            % Examples:
            % Maps model workspace parameter RESET to AUTOSAR SharedParameter.
            %
            % mappingObj = autosar.api.getSimulinkMapping('rtwdemo_autosar_counter');
            % mapParameter(mappingObj, 'RESET', 'SharedParameter');
            %
            % Maps model workspace parameter RESET to AUTOSAR ConstantMemory
            % with read only calibration access.
            %
            % mappingObj = autosar.api.getSimulinkMapping('rtwdemo_autosar_counter');
            % mapParameter(mappingObj, 'RESET', 'ConstantMemory', 'SwCalibrationAccess', ...
            %       'ReadOnly');
            %
            % See also getParameter
            argParser = inputParser;
            argParser.FunctionName = 'autosar.api.getSimulinkMapping.mapParameter';
            argParser.KeepUnmatched = true;
            argParser.addRequired('this', @(x) isa(x, class(x)));
            argParser.addRequired('ParameterRole', @(x)((ischar(x) || isStringScalar(x)) && strlength(x) > 0));
            
            validateattributes(SLParam, {'string', 'char'}, {'nonempty'},...
                'autosar.api.getSimulinkMapping.mapParameter', 'SLPARAM', 2);
            validateattributes(ParameterRole, {'string', 'char'}, {'nonempty'},...
                'autosar.api.getSimulinkMapping.mapParameter', 'PROPERTY', 3);
            ParameterRole = validatestring(ParameterRole, {DAStudio.message('coderdictionary:mapping:NoMapping'),...
                'SharedParameter', 'ConstantMemory'});
            argParser.parse(this, ParameterRole, varargin{:});
            
            try
                modelMapping = autosar.api.Utils.modelMapping(this.ModelName);
            catch Me
                %Trim stack to suppress mapping internal functions
                autosar.mm.util.MessageReporter.throwException(Me);
            end
            
            mappedParam = modelMapping.ModelScopedParameters.findobj('Parameter', SLParam);
            if length(mappedParam) < 1
                DAStudio.error('coderdictionary:api:invalidMappingParameterName',...
                    SLParam, SLParam);
            end
            this.setInternalDataProperties(this.ModelName,...
                mappedParam, ParameterRole, false, argParser);
            % Dirty the model
            set_param(this.ModelName,'Dirty','on');
        end
    end
    
    methods(Hidden)
        
        function mapPort(this, SLPortName, ARPortName, ARDataName, ARDataAccessModeStr)
            % MAPPORT(THIS, SLPORTNAME, ARPORTNAME, ARDATAELEMENTNAME, ARDATAACCESSMODE)
            % maps Simulink port SLPORTNAME to DataElement ARDATAELEMENTNAME
            % of AUTOSAR Sender/Receiver (or SenderReceiver) port
            % ARPORTNAME. SLPORTNAME should be the one given by the
            % function get_param(block, 'Name') to ensure no errors
            % 
            % See also mapInport, mapOutport
            modelMapping = autosar.api.Utils.modelMapping(this.ModelName);
            SLPortName2 = autosar.api.getSimulinkMapping.escapeSimulinkName(SLPortName);
            SLInport = modelMapping.Inports.findobj('Block', [modelMapping.Name, '/', SLPortName2]);

            if length(SLInport) == 1
                this.mapInport(SLPortName, ARPortName,...
                    ARDataName, ARDataAccessModeStr);
            else
                this.mapOutport(SLPortName, ARPortName,...
                    ARDataName, ARDataAccessModeStr);
            end
        end
        
        function mapModelBlock(this, SLModelBlockName, ARPrototypeName, uuid)
            % MAPMODELBLOCK(THIS, SLMODELBLOCKNAME, ARPROTOTYPENAME)
            % maps Simulink model block SLMODELBLOCKNAME to component prototype
            % ARPrototypeName            
            
            modelMapping = autosar.api.Utils.modelMapping(this.ModelName);
            SLModelBlockName = autosar.api.getSimulinkMapping.escapeSimulinkName(SLModelBlockName);
            SLModelBlock = modelMapping.ModelBlocks.findobj('Block', [modelMapping.Name, '/', SLModelBlockName]);
            
            if ~isempty(this.ChangeLogger)
                if ~isempty(SLModelBlock.MappedTo) &&...
                        ~strcmp(SLModelBlock.MappedTo.PrototypeName, ARPrototypeName)
                    this.ChangeLogger.logModification(...
                        'Automatic', 'ARPrototypeName', ...
                        'Simulink model block', ...
                        SLModelBlock.Block, ...
                        SLModelBlock.MappedTo.PrototypeName, ...
                        ARPrototypeName);
                end
            end
            
            SLModelBlock.mapComponentPrototype(ARPrototypeName, uuid);
            
            % Dirty the model
            set_param(this.ModelName,'Dirty','on');
        end
        
        function ARMemoryType = getDataDefaults(this, SLData)
         % ARMEMORYTYPE = GETDATADEFAULTS(THIS, SLData) returns the AUTOSAR 
         % memory type set for the Simulink data SLData('Signals' or 'States'). 
         % The returned AUTOSAR memory type is either 'ArTypedPIM' or 'None', 
         % which only applies to signals or states whose storage classes are 
         % configured to SimulinkGlobal.
             
             try
                 modelMapping = autosar.api.Utils.modelMapping(this.ModelName);
             catch Me
                 %Trim stack to suppress mapping internal functions
                 autosar.mm.util.MessageReporter.throwException(Me);
             end
             switch SLData
                 case 'Signals'
                    ARMemoryType = modelMapping.DataDefaultsMapping.Signals.Memory;
                 case 'States'
                    ARMemoryType = modelMapping.DataDefaultsMapping.DiscreteStates.Memory;
                 otherwise
                    DAStudio.error('autosarstandard:api:invalidDataDefaultsData', SLData);
             end
        end 

        function setDataDefaults(this, modelingElementType, mappedTo, varargin)
            % SETDATADEFAULTS(THIS, MODELINGELEMENTTYPE, MAPPEDTO, VARARGIN) maps 
            % the modeling element(i.e. Internal Data, Local Parameters) to
            % the corresponding category. Internal Data
            argParser = inputParser; 
            argParser.KeepUnmatched = true;
            argParser.addRequired('ModelingElementType',  @(x) any(validatestring(x,...
                autosar.utils.mappingCategories.getDataCategoriesForDataDefaults())));
         
            if strcmp(modelingElementType, 'InternalData')
                argParser.addRequired('MappedTo',  @(x) any(validatestring(x,...
                    autosar.utils.mappingCategories.getMappedToCategoriesForInternalData())));
            elseif strcmp(modelingElementType, 'LocalParameters')
                argParser.addRequired('MappedTo',  @(x) any(validatestring(x,...
                    autosar.utils.mappingCategories.getMappedToCategoriesForParameters())));   
            else
                DAStudio.error('autosarstandard:api:invalidDataDefaultsData', modelingElementType);
            end
            
            argParser.parse(modelingElementType, mappedTo);
           
            try
                modelMapping = autosar.api.Utils.modelMapping(this.ModelName);
            catch Me
                %Trim stack to remove internal mapping functions
                Me.throwAsCaller();
            end
            
            modelMapping.DefaultsMapping.setMappedTo(modelingElementType, mappedTo);
            
        end
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%Exclusive Area%%%%%%%%%%%%%
        function mapSSToExclusiveArea(this, BlockH, ARShortName)
           % MAPEXCLUSIVEAREA(THIS, BLOCKH, SLPORTNAME) maps a
           % subsystem with block handle BlockH to an Autosar ExclusiveArea with
           % name ARShortName. 
           
           try
               modelMapping = autosar.api.Utils.modelMapping(this.ModelName);
           catch Me
               %Trim stack to suppress mapping internal functions
               autosar.mm.util.MessageReporter.throwException(Me);
           end
           
           % verify short name 
           if ~isempty(ARShortName)
               schemaVersion = get_param(this.ModelName, 'AutosarSchemaVersion');
               maxShortNameLength = get_param(this.ModelName, 'AutosarMaxShortNameLength');
               idcheckmessage = autosar.ui.utils.isValidARIdentifier(ARShortName, 'shortName', ...
                   maxShortNameLength, schemaVersion);
               if ~isempty(idcheckmessage)
                   DAStudio.error('RTW:autosar:invalidShortName', ARShortName, maxShortNameLength);
               end
           end

           modelMapping.mapSubSystem(BlockH, ARShortName, false);
        end

        function mapping = getExclusiveArea(this, BlockH) 
           % GETEXCLUSIVEAREA(THIS, BLOCKH) returns the short name of
           % the ExclusiveArea that the Subsystem with block handle BlockH 
           % is mapped to. 
           
           try
               modelMapping = autosar.api.Utils.modelMapping(this.ModelName);
           catch Me
               %Trim stack to suppress mapping internal functions
               autosar.mm.util.MessageReporter.throwException(Me);
           end
         
           sid = get_param(BlockH, 'SID');
           SLExclusiveArea = modelMapping.SubSystemMappings.findobj('SID', sid);
           if ~isempty(SLExclusiveArea)
                mapping = SLExclusiveArea.MappedTo;
           else 
                DAStudio.error('autosarstandard:api:noExclusiveAreaMapping', getfullname(BlockH));
           end
        
        end
        
        
        function mappings = getExclusiveAreas(this)
           % GETEXCLUSIVEAREA(THIS) returns a table of Subsystem to
           % ExclusiveArea Mapping.
           
           try
               modelMapping = autosar.api.Utils.modelMapping(this.ModelName);
           catch Me
               %Trim stack to suppress mapping internal functions
               autosar.mm.util.MessageReporter.throwException(Me);
           end
           
           Subsystems = [];
           ExclusiveAreas = [];
           for i = 1:length(modelMapping.SubSystemMappings)
                block = find_system(this.ModelName, 'SID', modelMapping.SubSystemMappings(i).SID);
                Subsystems = [Subsystems; block];
                ExclusiveAreas = [ExclusiveAreas; modelMapping.SubSystemMappings(i).MappedTo];
           end
           mappings = table(Subsystems, ExclusiveAreas);
        end
   
    end
    
    methods(Static, Hidden)
        
        function validNvReceiverDAMs = getValidNvReceiverDAMs()
            % VALIDNVRECEIVERDAMS = GETVALIDNVRECEIVERDAMS returns a cell-array of valid
            % NvReceiverPort Data Access Modes.
            validNvReceiverDAMs = autosar.api.getSimulinkMapping.ValidNvReceiverDAMs;
        end
        
        function validNvSenderDAMs = getValidNvSenderDAMs()
            % VALIDNVSENDERDAMS = GETVALIDNVSENDERDAMS returns a cell-array of valid
            % NvSenderPort Data Access Modes.
            validNvSenderDAMs = autosar.api.getSimulinkMapping.ValidNvSenderDAMs;
        end
        
    end
    
    methods(Static, Access=private)
        function res = escapeSimulinkName(pathStr)
        % Convert new line to space as this is what
        % BlockMapping.getBlockPath() does, and change / into // as this is
        % what getFullName() does when a port name is obtained
        
           %convert strings to chars
           if isstring(pathStr)
               pathStr = convertStringsToChars(pathStr);
           end
           pathStr = strrep(pathStr, newline, ' ');
           res = strrep(pathStr, '/', '//');
        end      
        
        function isValid = isValidRateTransitionBlock(blkPath)
            % Check if input string is a valid rate transition block
            isValid = false;
            try
               if Simulink.ID.isValid(Simulink.ID.getSID(blkPath))
                   if strcmp(get_param(blkPath,'BlockType'), 'RateTransition')
                        isValid = true;                      
                   end
               end
            catch ex %#ok<NASGU>
            end
        end
        
        function resetFunction = getResetFunction(modelMapping, SLResetFcnName)
            for idx = 1:length(modelMapping.ResetFunctions)
                currResetFcn = modelMapping.ResetFunctions(idx);
                if strcmp(SLResetFcnName, currResetFcn.Name)
                    resetFunction = currResetFcn;
                    return;
                end
            end
            resetFunction = [];
            return;
        end
        
        function setInternalDataProperties(modelName, mappingObj, DataRole, isVariable, argParser)            
            mappingObj.map(DataRole);
            mappedTo = mappingObj.MappedTo;
            if ~strcmp(DataRole, DAStudio.message('coderdictionary:mapping:NoMapping'))
                params = argParser.Unmatched;
                if ~isempty(fields(params))
                    instSpecificPropertyNames = fieldnames(params);
                    instSpecificPropertyValues = struct2cell(params);
                    modelH = get_param(modelName, 'Handle');
                    for ii = 1 : numel(instSpecificPropertyNames)
                        propertyName = instSpecificPropertyNames{ii};
                        propertyName = validatestring(propertyName,...
                        mappedTo.getPerInstancePropertyNames(isVariable));
                        propertyValue = instSpecificPropertyValues{ii};
                        % Validation check for instance specific properties
                        if isa(mappingObj, 'Simulink.AutosarTarget.ModelScopedParameterMapping')
                            source = mappingObj.Parameter;
                        elseif isa(mappingObj, 'Simulink.AutosarTarget.SignalMapping')
                            portHandles = get_param(mappingObj.OwnerBlockPath, 'PortHandles');
                            % Append the port number
                            if numel(portHandles.Outport) > 1
                                source = [mappingObj.OwnerBlockPath ':' num2str(get_param(mappingObj.PortHandle, 'PortNumber'))];
                            else
                                source = mappingObj.OwnerBlockPath;
                            end
                        else
                            source = mappingObj.OwnerBlockPath;
                        end
                        [result, errorMessage, propertyValue] = autosar.validation.AutosarUtils.isValidPerInstanceProperty(...
                            modelName, source, propertyName, propertyValue);
                        if ~result
                            error(errorMessage);
                        end
                        Simulink.CodeMapping.setPerInstancePropertyValue(modelH, mappingObj,'MappedTo', propertyName, propertyValue);
                    end
                end
            end
        end    
    end
    
end

% LocalWords:  MAPPINGOBJ MDLNAME ARIRVNAME ARDATAACCESSMODE GETDATATRANSFER Bs
% LocalWords:  SLDATATRANSFERNAME ARRUNNABLENAME GETFUNCTION SLFCNNAME utils
% LocalWords:  ARPORTNAME ARDATAELEMENTNAME GETINPORT SLPORTNAME GETOUTPORT RTB
% LocalWords:  MAPDATATRANSFER MAPINPORT ARDATAACCESSMODESTR MAPOUTPORT MAPPORT
% LocalWords:  MAPFUNCTION AROPERATIONNAME GETFUNCTIONCALLER MAPFUNCTIONCALLER
% LocalWords:  ARCOMPUMETHODNAME SLDATATYPENAME Compu MAPDATATYPE Nv RPort rtb
% LocalWords:  autosarstandard ARLOOKUPTABLENAME ARPARAMETERACCESSMODE GETLOOKUPTABLE
% LocalWords:  ARPARAMETERDATANAME MAPLOOKUPTABLE SLPARAM ARPARAMETERDATA JSON
% LocalWords:  VALIDNVRECEIVERDAMS GETVALIDNVRECEIVERDAMS VALIDNVSENDERDAMS
% LocalWords:  GETVALIDNVSENDERDAMS MAPMODELBLOCK SLMODELBLOCKNAME Specsfor
% LocalWords:  ARPROTOTYPENAME ARCOMPONENTNAME GETSIGNALMEMORY MAPSIGNALMEMORY
% LocalWords:  SLMEMORYTYPE ARMEMORYTYPE GETDATADEFAULTS MAPDATADEFAULTS json
% LocalWords:  GETSIGNAL SLBLOCKPATH PORTINDEX rtwdemo coderdictionary Decl
% LocalWords:  MAPSIGNAL MAPPEDTO STATEID MAPSTATE SETDATADEFAULTS PORTHANDLE
% LocalWords:  MODELINGELEMENTTYPE Decl PORTHANDLE Artyped Sw Addr STATENAME
% LocalWords:  Decl PORTHANDLE Artyped Sw Addr STATENAME MAPPARAMETER
% LocalWords:  VARIABLEROLE PARAMETERROLE
