classdef ServiceImplementation < handle
    % Abstract class for BSW block implementation
    
    %   Copyright 2016-2017 The MathWorks, Inc.
    
    methods (Static, Abstract)
        % methods must be implemented by sub-classes
        deepCopyInterface(mdlName, dstInterfacePath);
        defaultClientPortName = getDefaultClientPortName();
        interfaceName = getInterfaceName();
        operations = getOperations();
        desc = getDescription();
        type = getType();
        help = getHelp();
        hastypeParameter = hasDatatypeParameter();
        dataTypeCallback(blkPath);
        operationCallback(blkPath);
    end
    
    methods (Static)
        % Methods to be over-ridden by sub-classes if necessary
        function hasArgSpecParameter = hasArgumentSpecificationParameter()
            hasArgSpecParameter = false;
        end            
    end
    
    
    methods (Static, Access = private)
        % returns true if the Enum datatype used by the block is a
        % non-default one
        function isCustomDt = isCustomEnumDatatype(enumDatatype, allEnumDataTypes)
            isCustomDt = ~isempty(enumDatatype) &&...
                isempty(find(strcmp(enumDatatype, allEnumDataTypes), 1));
        end
        
        % check validity of enum datatype parameter
        function isValid = isEnumDataTypeValid(dataType)
            isValid = false;
            if isempty(dataType)
                return;
            end
                        
            % Error out if the specified enum is not in the expected form
            dataType = strtrim(dataType);
            if ~strncmp(dataType, 'Enum:', length('Enum:'))
               return
            end
            
            % Error out if the enum is not a valid class name
            enumName = regexp(dataType, 'Enum:\s*(\w*)', 'tokens');
            enumName = enumName{1}{1};            
            mprops = Simulink.getMetaClassIfValidEnumDataType(enumName);
            if isempty(mprops)
                return;
            end
            isValid = true;
        end
    end
    
    methods (Static, Access=public)
        function blockIconType = getBlockIconType()
            blockIconType = '';
        end
    end
    
    methods (Static, Sealed, Access=protected)
        function updateReferences(mdlName, arxmlFiles)
            obj = arxml.importer(arxmlFiles);
            obj.updateReferences(mdlName,...
                'BackupModel', false,...
                'CreateReport', false,...
                'DisplayMessages', false);
        end
        
        % gets called when user changes the datatype value in the dialog
        % and when the initialization command is run
        function dataTypeCallbackImpl(blkPath)
            mo = get_param(blkPath, 'MaskObject');
            paramIdx = strcmp({mo.Parameters.Name}, 'Datatype');
            
            dataType = get_param(blkPath, 'Datatype');
            % Error out if the specified enum is not in the expected form
            if ~autosar.bsw.ServiceImplementation.isEnumDataTypeValid(dataType)
                DAStudio.error('Simulink:DataType:UdtInvEnumName',...
                    dataType, mo.Parameters(paramIdx).Name);
            end
        end
        
        % gets called when user changes the operation value in the dialog
        % and when the initialization command is run
        function operationCallbackImpl(blkPath, defaultEnumDatatype, allEnumDataTypes,...
                datatypeMaskPromptNameMap)
            maskObject = get_param(blkPath, 'MaskObject');
            operationValue = get_param(blkPath, 'Operation');
            
            datatypeIdx = strcmp({maskObject.Parameters.Name}, 'Datatype');
            if isempty(defaultEnumDatatype)
                maskObject.Parameters(datatypeIdx).Visible = 'off';
                set_param(blkPath, 'Datatype', '');
            else
                maskObject.Parameters(datatypeIdx).Visible = 'on';
                enumDataType =  get_param(blkPath, 'Datatype');
                % don't override the user specified datatype
                if ~autosar.bsw.DemDiagnosticInfo.isCustomEnumDatatype(enumDataType, allEnumDataTypes)
                    set_param(blkPath, 'Datatype', defaultEnumDatatype);
                end
                maskObject.Parameters(datatypeIdx).Prompt =...
                    datatypeMaskPromptNameMap(operationValue);
            end
            
            % g1646618 Ensure ports on block get updated on canvas when
            % operation changes
            portName = get_param(blkPath, 'PortName');
            operationName = get_param(blkPath, 'Operation');
            serviceImpl = eval(get_param(blkPath, 'ServiceImpl'));
            serviceImpl.updateFunctionCaller(blkPath, portName, operationName);
        end
        
        % gets called when user changes the operation value in the dialog
        % and when the initialization command is run
        function argumentSpecificationOperationCallbackImpl(blkPath, defaultDatatype,...
                datatypeMaskPromptNameMap)
            maskObject = get_param(blkPath, 'MaskObject');
            operationValue = get_param(blkPath, 'Operation');
            
            argSpecIdx = strcmp({maskObject.Parameters.Name}, 'ArgumentSpecification');
            if isempty(defaultDatatype)
                maskObject.Parameters(argSpecIdx).Visible = 'off';
            else
                maskObject.Parameters(argSpecIdx).Visible = 'on';
                maskObject.Parameters(argSpecIdx).Prompt =...
                    datatypeMaskPromptNameMap(operationValue);
            end
            
            % g1646618 Ensure ports on block get updated on canvas when
            % operation changes
            portName = get_param(blkPath, 'PortName');
            operationName = get_param(blkPath, 'Operation');
            serviceImpl = eval(get_param(blkPath, 'ServiceImpl'));
            serviceImpl.updateFunctionCaller(blkPath, portName, operationName);
        end
        
        
    end
end

% LocalWords:  BSW Udt
