classdef MappingHelper
    %MAPPINGHELPER - Helper class for ease-of-use functionality for Mapping

%   Copyright 2018 The MathWorks, Inc.
    
    methods (Static)
        
        function syncStatesAndSignals(modelName)
            %SYNCSTATESANDSIGNALS will find the definitions of ArTypedPIM and StaticMemory
            % in the metamodel and map and Signals or states with matching names
            % to the relevant class
            
            slMapping = autosar.api.getSimulinkMapping(modelName);
            m3iComp = autosar.api.Utils.m3iMappedComponent(modelName);
            m3iModel = m3iComp.rootModel;
            
            changeLogger = autosar.updater.ChangeLogger;
            sysConstsValueMap = autosar.api.Utils.createSystemConstantMap(m3iModel, [], []);
            
            slTypeBuilder = autosar.mm.mm2sl.TypeBuilder(m3iModel, false, 'base', changeLogger, sysConstsValueMap);
            slTypeBuilder.buildDataTypeMappings(m3iComp);
            
            % Find all the ArTypedPIM and StaticMemory definitions in the metamodel
            arTypedPIMSeq = m3iComp.Behavior.ArTypedPIM;
            staticMemorySeq = m3iComp.Behavior.StaticMemory;
            
            % Look for named states/signals in the model and map them
            autosar.utils.MappingHelper.mapStatesAndSignals(arTypedPIMSeq,modelName, 'ArTypedPerInstanceMemory', slMapping, m3iModel, slTypeBuilder);
            autosar.utils.MappingHelper.mapStatesAndSignals(staticMemorySeq,modelName, 'StaticMemory', slMapping, m3iModel, slTypeBuilder);
        end
        
        function type = getVariableRoleFromM3IData(m3iData)
            m3iBehavior = m3iData.containerM3I;
            type = '';
            
            m3iArTypedPIM = m3iBehavior.ArTypedPIM;
            idx = autosar.mm.Model.findObjectIndexInSequence(m3iArTypedPIM, m3iData);
            if idx ~= -1
                type = 'ArTypedPerInstanceMemory';
                return;
            end
            
            m3iStaticMemory = m3iBehavior.StaticMemory;
            idx = autosar.mm.Model.findObjectIndexInSequence(m3iStaticMemory, m3iData);
            if idx ~= -1
                type = 'StaticMemory';
                return;
            end
        end
    end
    
    methods (Static, Access = private)
        function mapStatesAndSignals(m3iSeq, modelName, type, slMapping, m3iModel, slTypeBuilder)
            if ~ischar(modelName)
                modelName = get_param(modelName,'Name');
            end
            
            for ii = 1:m3iSeq.size()
                m3iData = m3iSeq.at(ii);
                name = m3iData.Name;
                
                [doesExist,sigObj,inModelWorkSpace] = autosar.utils.Workspace.objectExistsInModelScope(modelName, name);
                
                if doesExist && isa(sigObj,'AUTOSAR.Signal') && ~inModelWorkSpace
                    % Classic ArTypedPIM mapped to DSM in model
                    continue;
                end
                
                modelMapping = autosar.api.Utils.modelMapping(modelName);
                states = modelMapping.States;
                signals = modelMapping.Signals;
                matchedState = [];
                matchedSignal = [];
                stateName = name;
                for stateIdx = 1:length(states)
                    state = states(stateIdx);
                    mappedName = slMapping.getState(state.OwnerBlockHandle,state.Name,'ShortName');
                    if strcmp(mappedName,name)
                        stateName = state.Name;
                        matchedState = state;
                        break;
                    end
                end
                if isempty(matchedState)
                    for sigIdx = 1:length(signals)
                        sig = signals(sigIdx);
                        mappedName = slMapping.getSignal(sig.PortHandle,'ShortName');
                        if strcmp(mappedName,name)
                            matchedSignal = sig;
                            break;
                        end
                    end
                end
                if ~(isempty(matchedState) && isempty(matchedSignal))
                    assert(length(matchedState)+length(matchedSignal)==1,'Expected only one signal or state to be mapped');
                    if isempty(matchedState)
                        % Use the signal mapping
                        srcPort = matchedSignal.PortHandle;
                        blockPath = [];
                    else
                        % Use the state mapping
                        blockPath = matchedState.OwnerBlockPath;
                        srcPort = [];
                    end
                else
                    namedSignal = find_system(modelName,...
                        'FindAll','on',...
                        'FollowLinks',1,...
                        'LookUnderMasks','all',...
                        'type','line',...
                        'Name',name);

                    blockPath = find_system(modelName,...
                        'FollowLinks',1,...
                        'LookUnderMasks','all',...
                        'StateName',name);

                    if isempty(namedSignal) && isempty(blockPath)
                        % Not used in the model
                        continue;
                    end

                    % block is a state owner but has no state parameter (e.g. StateWriter)
                    ignoreStateBlkIdxs = cellfun(@(x) ~isfield(get_param(x, 'DialogParameters'), 'StateName'), blockPath);
                    blockPath(ignoreStateBlkIdxs) = [];

                    if (length(namedSignal) + length(blockPath))>1
                        % More than one use of the same name
                        autosar.mm.util.MessageReporter.createWarning('autosarstandard:validation:cannotMapMultipleSignalOrState',name);
                        continue;
                    end
                    
                    srcPort = get_param(namedSignal,'SrcPortHandle');
                end
                
                m3iType = autosar.utils.MappingHelper.getM3IImpType(m3iModel,m3iData, slTypeBuilder);
                
                % Bus Type
                if strcmp(m3iType.MetaClass.qualifiedName,'Simulink.metamodel.types.Structure')
                    IsVolatile = m3iType.Elements.at(1).Type.IsVolatile;
                    Qualifier = m3iType.Elements.at(1).Type.Qualifier;
                else
                    IsVolatile = m3iType.IsVolatile;
                    Qualifier = m3iType.Qualifier;
                end
                
                % Convert isVolatile to char representation
                if IsVolatile
                    IsVolatile = 'true';
                else
                    IsVolatile = 'false';
                end
                
                if isempty(m3iData.SwAddrMethod)
                    swAddrMethodName = '';
                else
                    swAddrMethodName = m3iData.SwAddrMethod.Name;
                end
                
                extraArgs = {...
                    'ShortName', name,...
                    'IsVolatile', IsVolatile,...
                    'SwAddrMethod', swAddrMethodName,...
                    'DisplayFormat', m3iData.DisplayFormat,...
                    'SwCalibrationAccess', m3iData.SwCalibrationAccess.toString,...
                    'Qualifier', Qualifier...
                    };
                
                if ~isempty(srcPort)
                    if srcPort==-1
                        % Unconnected signal
                        continue;
                    end
                    
                    try
                        signalMapping = getSignal(slMapping,srcPort);
                    catch ME
                        if strcmp(ME.identifier,'coderdictionary:api:invalidMappingSignalPort')
                            % We cannot map a signal if it is not supported for mapping
                            continue;
                        else
                            ME.rethrow();
                        end
                    end
                    
                    if strcmp(signalMapping,'Auto')
                            mapSignal(slMapping,...
                                srcPort,...
                                type,...
                                extraArgs{:});
                    end
                else
                    if iscell(blockPath), blockPath = blockPath{1}; end
                    if strcmp(getState(slMapping,blockPath,stateName),'Auto')
                        mapState(slMapping,...
                            blockPath, stateName,...
                            type,...
                            extraArgs{:});
                    end
                end
            end
        end
        
        function m3iImpType = getM3IImpType(m3iModel,m3iVarData, slTypeBuilder)
            m3iType = m3iVarData.Type;
            if m3iType.IsApplication
                impQName = slTypeBuilder.getImplementationDataType(m3iType.qualifiedName);
                m3iSeq = autosar.mm.Model.findObjectByName(m3iModel, impQName);
                m3iImpType = m3iSeq.at(1);
            else
                m3iImpType = m3iType;
            end
        end
    end
end

% LocalWords:  metamodel Idxs Sigs Sw Addr SYNCSTATESANDSIGNALS autosarstandard
% LocalWords:  coderdictionary
