classdef AdaptiveValidator < autosar.validation.PhasedValidator
    % AdaptiveValidator
    
    %   Copyright 2018 The MathWorks, Inc.
    
    methods (Access=protected)
        
        function verifyInitial(this, hModel)
            modelName = get_param(hModel, 'Name');
            this.verifyAdaptiveModelSetup(modelName);
            this.verifyAdaptiveFeatureFlag(modelName);
            this.verifyAdaptiveMetaModel(modelName);
            this.verifyAdaptiveEventDataAccessMode(hModel);
        end
        
    end
    
    methods (Static, Access=private)
        
        function verifyAdaptiveModelSetup(modelName)
            % verify that the model is setup for adaptive, i.e. TargetLang
            % is set to C++ and CodeInterfacePackaging is set to C++ class
            
            if ~autosar.api.Utils.isAdaptiveModelSetup(modelName)
                % determine what to do here with g1773251
                % we might error, but metamodel and mapping might need
                % trashing too.
                DAStudio.error('autosarstandard:validation:invalidAdaptiveMapping', modelName);
            end
        end
        
        function verifyAdaptiveFeatureFlag(modelName)
            % verify that the feature flag is set correctly
            if slfeature('RTWAdaptiveAutosar') == 0
                DAStudio.error('autosarstandard:validation:adaptiveFeatureNotSet', modelName);
            end
        end
        
        function verifyAdaptiveMetaModel(modelName)
        % verify in metamodel that the service interface name event name are unique.
        % Otherwise it leads to C++ compilation errors.
            mapping = autosar.api.Utils.modelMapping(modelName);
            dataObj = autosar.api.getAUTOSARProperties(modelName, true);
            componentQName = dataObj.get('XmlOptions', 'ComponentQualifiedName');
            
            for dataIdx = 1:length(mapping.Inports)
                inport = mapping.Inports(dataIdx);
                apPortName = inport.MappedTo.Port;
                apEventName = inport.MappedTo.Element;
                
                interfaceName = dataObj.get([componentQName, '/', apPortName], ...
                                            'Interface');
                if strcmp(apEventName, interfaceName)
                    DAStudio.error('autosarstandard:validation:invalidAdaptiveInterfaceEventName', ...
                                   modelName, interfaceName, apEventName);
                end
            end
        end


        function verifyAdaptiveEventDataAccessMode(hModel)
            mapping = autosar.api.Utils.modelMapping(hModel);
            % verify data access mode for adaptive events
            % inports
            for dataIdx = 1:length(mapping.Inports)
                inport = mapping.Inports(dataIdx);
                if ~ismember(inport.MappedTo.DataAccessMode, {'QueuedExplicitReceive', 'ImplicitReceive'})
                    DAStudio.error('autosarstandard:validation:adaptiveInvalidDataAccessMode', ...
                                   inport.MappedTo.Port, ...
                                   inport.MappedTo.DataAccessMode, ...
                                   'QueuedExplicitReceive');
                end
            end
            % outports
            for dataIdx = 1:length(mapping.Outports)
                outport = mapping.Outports(dataIdx);
                if ~ismember(outport.MappedTo.DataAccessMode, {'QueuedExplicitSend', 'ImplicitSend'})
                    DAStudio.error('autosarstandard:validation:adaptiveInvalidDataAccessMode', ...
                                   outport.MappedTo.Port, ...
                                   outport.MappedTo.DataAccessMode, ...
                                   'QueuedExplicitSend');
                end
            end
        end
    end
    
end

% LocalWords:  metamodel autosarstandard
