classdef AutosarUtils
% autosar specific checks
    
%   Copyright 2013-2018 The MathWorks, Inc.

    methods(Static, Access=public)
        function [result, errorMessage, propertyValue] = isValidPerInstanceProperty(...
                modelName, source, propertyName, propertyValue)
            result =  true;
            errorMessage = '';
            switch propertyName
                case 'DisplayFormat'
                    if strlength(propertyValue) > 0
                        [result, errorMessage] = autosar.validation.AutosarUtils.checkDisplayFormat(propertyValue, source);
                    end
                case {'IsVolatile', 'IsConst'}
                    if islogical(propertyValue)
                        validateattributes(propertyValue, {'logical'}, {'nonempty'});
                    else
                        propertyValue = validatestring(propertyValue,...
                            {'true', 'false'});
                    end
                case 'ShortName'
                    if strlength(propertyValue) > 0
                        schemaVersion = get_param(modelName, 'AutosarSchemaVersion');
                        maxShortNameLength = get_param(modelName, 'AutosarMaxShortNameLength');
                        idcheckmessage = autosar.ui.utils.isValidARIdentifier(propertyValue, 'shortName', ...
                            maxShortNameLength, schemaVersion);
                        if ~isempty(idcheckmessage)
                            result = false;
                            errorMessage = DAStudio.message('RTW:autosar:invalidShortName', propertyValue, maxShortNameLength);
                        end
                    end
                case 'SwCalibrationAccess'
                    propertyValue = validatestring(propertyValue,...
                        {'NotAccessible','ReadOnly', 'ReadWrite'});
                case 'Qualifier'
                    validateattributes(propertyValue, {'char', 'string'},...
                        {'scalartext'}, propertyName);
                    % Qualifier should also be a valid C identifier or
                    % multiple ones
                    if ~isempty(propertyValue)
                        % separate identifiers by whitespaces
                        validCIdentifierPattern = '^[_a-zA-Z][_a-zA-Z0-9]{0,30}$';
                        identifierCellStr = strsplit(propertyValue);
                        for ii = 1 : length(identifierCellStr)
                            result = ~isempty( regexp( identifierCellStr{ii} , validCIdentifierPattern, 'ONCE' ) );
                            if ~result
                                % generate error message and return if an
                                % identifier is not valid
                                errorMessage = DAStudio.message('autosarstandard:validation:invalidCQualifier', propertyValue, source);
                                return;
                            end
                        end
                    end
            end
        end
        
       function msg = isValidIdentifier(cellstrToTest, idType, maxShortNameLength)
       % Function: isValidIdentifier =========================================
       % Abstract:
       %   Validate cell array of AUTOSAR identifier strings

           msg = '';

           for idx = 1:length(cellstrToTest)
               str = cellstrToTest{idx};

               [isvalid, msg] = autosarcore.checkIdentifier(str, idType, maxShortNameLength);
               if ~isvalid || ~isempty(msg)
                   return;
               end

           end

       end

       function checkShortNameCaseClash( shortNames )
       % Function: checkShortNameCaseClash =========================================
       % Abstract: Check for case insensitive clashes among shortNames
       %           (e.g. "RPORT1" and "rport1" clash)

           if isempty(shortNames)
               return
           end
           
           shortNameMap = containers.Map( lower(shortNames), shortNames );
           
           for i=1:length(shortNames)
               shortName1 = shortNames{i};
               shortName2 = shortNameMap( lower(shortName1 ) );
               if ~strcmp( shortName1, shortName2 )
                   DAStudio.error('RTW:autosar:shortNameCaseClash', shortName1, shortName2);
               end
           end
           
       end

       function isNotCKeyword = isNotCKeyword( name )
       % Function: isNotCKeyword ===============================================
       % Abstract:
       %   check NAME is not a c keyword

           dummyConfigEntry = RTW.FcnArgSpec;
           dummyConfigEntry.ArgName = name;

           isNotCKeyword = dummyConfigEntry.isValidIdentifier();
       end
              
       function isErrArg = isErrorArgument(mdlName, Identifier)
       % Function: isErrorArgument ===============================================
       % Abstract:
       %   check NAME is error argument for a function caller or simulink
       %   function.
           isErrArg = false;
           [~, opName, intfQName] = autosar.validation.AutosarUtils.getInterface(mdlName, Identifier);
           if ~isempty(intfQName)
               argName = get_param(Identifier, 'ArgumentName');
               dataobj = autosar.api.getAUTOSARProperties(mdlName, true); 
               direction = dataobj.get([intfQName '/' opName '/' argName], 'Direction');
               isErrArg = strcmp(direction, 'Error');
           end
       end
       
       function [intfName, opName, intfQName] = getInterface(mdlName, Identifier)
       % Function: getInterface ===============================================
       % Abstract:
       %   Determines autosar interface and operation name corresponding to a Simulink
       %   function arg
      
           intfName = '';
           opName = '';
           intfQName = '';
           [parentBlk,~,~] = fileparts(Identifier);
           [~,~,fcnName]  = autosar.validation.ClientServerValidator.getBlockInOutParams(...
               parentBlk);
           mapping = autosar.api.getSimulinkMapping(mdlName);
           runName = mapping.getFunction(fcnName);
           
           dataobj = autosar.api.getAUTOSARProperties(mdlName, true);  
           swc = dataobj.get('XmlOptions', 'ComponentQualifiedName');
           behavior = dataobj.get(swc, 'Behavior', 'PathType', 'FullyQualified');
           runnablePath = [behavior '/' runName];
           opEvents = dataobj.find(behavior, 'OperationInvokedEvent', 'StartOnEvent', runnablePath, 'PathType', 'FullyQualified');
           if length(opEvents) ~= 1
               return;
           end
           trigger = dataobj.get(opEvents{1},'Trigger');
           if isempty(trigger)
               return;
           end
           trigger = strsplit(trigger,'.');
           
           %get server port for operation invoked event
           serverPorts = dataobj.find(swc,'ServerPort','PathType','FullyQualified');
           serverPath = serverPorts(cellfun(@(x)~isempty(x),regexp(serverPorts,['/' trigger{1} '$'])));
           
           %get server port interface
           intfQName = dataobj.get(serverPath{1},'Interface', 'PathType', 'FullyQualified');
           [~, intfName, ~] = fileparts(intfQName);
           opName = trigger{2};
       end
       
       function [propertyValue, interfaceName] = getInterfaceElementProperty(mdlName, portName, elementName, propertyName)
           % Function: getInterfaceElementProperty =============================
           % Abstract:
           %   Returns the value of a port element property.
           
           dataObj = autosar.api.getAUTOSARProperties(mdlName, true);
           swc = dataObj.get('XmlOptions', 'ComponentQualifiedName');
           port = dataObj.find(swc, 'ProvidedPort', 'Name', portName, 'PathType', 'FullyQualified');           
           if iscell(port)
               assert(length(port) == 1);
               port = port{1};
           end
           interfaceName = dataObj.get(port, 'Interface','PathType', 'FullyQualified');
           dataElem = dataObj.find(interfaceName, 'FlowData', 'Name', elementName, 'PathType', 'FullyQualified');
           if iscell(dataElem)
               assert(length(dataElem) == 1);
               dataElem = dataElem{1};
           end
           propertyValue = dataObj.get(dataElem, propertyName);
       end       
       
       function checkDataTypeForErrArg(mdlName, blkPath, ~, ~)
        % Function: checkDataTypeForErrArg ====================================
        % Abstract:
        %   Check header file, name and various settings of error argument

            portHandles = get_param(blkPath, 'PortHandles');
            errPortH = portHandles.Inport(1);

            [~, ~, intfQName] = autosar.validation.AutosarUtils.getInterface(mdlName, blkPath);
            m3iModel = autosar.api.Utils.m3iModel(mdlName);
            arRoot = m3iModel.RootPackage.front();
            intfSeq = autosar.mm.Model.findObjectByNameAndMetaClass(arRoot,...
                intfQName, ...
                Simulink.metamodel.arplatform.interface.ClientServerInterface.MetaClass());
            assert(intfSeq.size() == 1, 'Did not find client server interface');
            m3iInterface = intfSeq.at(1);
            
            autosar.validation.AutosarUtils.checkDataTypeForErrArgPort(errPortH, m3iInterface);
            
        end
            
       function checkDataTypeForErrArgPort(portH, m3iInterface)
        % Function: checkDataTypeForErrArg ====================================
        % Abstract:
        %   Check header file, name and various settings of error argument

            blkPath = get_param(portH, 'Parent');
            
            width = get_param(portH, 'CompiledPortWidth');
            
            if width > 1
               DAStudio.error('autosarstandard:validation:errorArgInvalidWidth',...
                   blkPath);
            end

            dataType = get_param(portH, 'CompiledPortAliasedThruDataType');
            
            aliasDataType = get_param(portH, 'CompiledPortDataType');                       

            if strcmp(dataType, 'uint8')
                % Okay
                return
            elseif Simulink.data.isSupportedEnumClass(dataType)
                if slfeature('ControlStorageTypeForEnums')
                   storageType = Simulink.data.getEnumTypeInfo(dataType, 'StorageType');
                   if ~strcmp( storageType, 'uint8' )
                       DAStudio.error('autosarstandard:validation:errorArgInvalidDataType', blkPath, aliasDataType);
                   end
                end
            else
                DAStudio.error('autosarstandard:validation:errorArgInvalidDataType', blkPath, aliasDataType);
            end

            % Enumeration checks            
            
            % Valid enum check
            errCodes = enumeration(dataType);
            if isempty(errCodes)
                msg = DAStudio.message('RTW:autosar:invalidErrorEnum', dataType);
                DAStudio.error('RTW:fcnClass:finish', msg);
            end
            
            % Range check            
            for ii=1:length(errCodes)
                errCodeValue = uint8(errCodes(ii));                
                if errCodeValue < 0 || errCodeValue > 63
                    msg = DAStudio.message('autosarstandard:validation:invalidAppErrValue', ...
                                           dataType, ...
                        num2str(errCodes(ii)));
                    DAStudio.error('RTW:fcnClass:finish', msg);
                end
            end
            
        end

        
        function checkDataType(mdlName, Identifier, dataTypeName, maxShortNameLength, supportMatrixIOAsArray)
        % Function: checkDataType ====================================
        % Abstract:
        %   Check header file and name settings of Simulink types

            % strip off leading ? or Enum: for enum types
            % strip off leading Bus: for Bus types
            dataTypeName = autosar.utils.StripPrefix(dataTypeName);
            Identifier = strtrim(Identifier);            

            % Data name check
            if length(dataTypeName) > maxShortNameLength
                msg = DAStudio.message('RTW:autosar:invalidDataTypeName', Identifier, dataTypeName, maxShortNameLength, maxShortNameLength);
                DAStudio.error('RTW:fcnClass:finish', msg);
            end

            [objExists,slObj] = autosar.utils.Workspace.objectExistsInModelScope(mdlName, dataTypeName);
            if objExists
                % Check for header file at root level
                if isa(slObj, 'Simulink.AliasType') || ...
                   isa(slObj, 'Simulink.NumericType') || ...
                   isa(slObj, 'Simulink.StructType')
                    if ~strcmp( slObj.HeaderFile, 'Rte_Type.h' )
                        msg = DAStudio.message('RTW:autosar:invalidDataTypeHeader', Identifier, dataTypeName);
                        DAStudio.error('RTW:fcnClass:finish', msg);
                    end
                    if ~isempty(strmatch('DataScope', fieldnames(slObj))) && ...
                            strcmp( slObj.DataScope, 'Exported' ) %#ok<MATCH2>
                        msg = DAStudio.message('RTW:autosar:invalidDataTypeScope', Identifier, dataTypeName);
                        DAStudio.error('RTW:fcnClass:finish', msg);
                    end
                else
                    % okay to fall through
                end

                % now attempt to recurse through datatype elements/base types etc.
                if isa(slObj, 'Simulink.AliasType')
                    % check base type
                    if strcmp(dataTypeName, slObj.BaseType)
                        % Skip check for self referencing aliastypes
                    else
                        autosar.validation.AutosarUtils.checkDataType(mdlName, Identifier, slObj.BaseType, maxShortNameLength, supportMatrixIOAsArray);
                    end
                elseif isa(slObj, 'Simulink.NumericType') && ~strcmp(slObj.DataTypeMode, 'Double')
                    autosar.validation.AutosarUtils.doWordSizeCheck(mdlName, dataTypeName, slObj.WordLength);
                elseif isa(slObj, 'Simulink.StructType')

                    % make sure StructType Elements are arrays rather than matrices
                    if ~supportMatrixIOAsArray
                        autosar.validation.AutosarUtils.checkStructTypeDimension(mdlName, slObj, Identifier, dataTypeName);
                    end

                    % recurse through elements
                    for elemIdx=1:length(slObj.Elements)
                        elem = slObj.Elements(elemIdx);

                        % check each element
                        autosar.validation.AutosarUtils.checkDataType(mdlName, Identifier, elem.DataType, maxShortNameLength, supportMatrixIOAsArray);
                    end
                else
                    % okay to fall through
                end
            end

            if isvarname(dataTypeName)
                if Simulink.data.isSupportedEnumClass(dataTypeName)
                    headerFile = Simulink.data.getEnumTypeInfo(dataTypeName, 'HeaderFile');
                    if ~strcmp( headerFile, 'Rte_Type.h' )
                        msg = DAStudio.message('RTW:autosar:invalidEnumHeader', dataTypeName);
                        DAStudio.error('RTW:fcnClass:finish', msg);
                    end
                    dataScope = Simulink.data.getEnumTypeInfo(dataTypeName, 'DataScope');
                    if strcmp( dataScope, 'Exported' )
                        msg = DAStudio.message('RTW:autosar:invalidEnumScope', dataTypeName);
                        DAStudio.error('RTW:fcnClass:finish', msg);
                    end
                else
                    if strncmp(dataTypeName, 'ufix', 4) || strncmp(dataTypeName, 'sfix', 4)
                        % extract WordLength
                        output = regexp(dataTypeName, 'fix(?<WordLength>\d+)', 'names');
                        wordLength = str2double(output.WordLength);
                        autosar.validation.AutosarUtils.doWordSizeCheck(mdlName, dataTypeName, wordLength);   
                    end
                end
            end

        end

        function str = removeNewLine( str )
        % Function: removeNewLine =================================================
        % Abstract:
            str = strrep(str, newline, ' ');
        end
        
        function str = escapeNewLine(str)
            str = strrep(str, newline, '\n');
        end
        
        function [isValid, errmsg] = checkSymbol(str)
        % Function: checkSymbol =================================================
        % Abstract: Check if the input string is a valid AUTOSAR symbol
            
            % Check if symbol is conflict with C keywords
            % and check if symbol is a valid C identifier
            isValid = ...
                autosar.validation.AutosarUtils.isNotCKeyword(str) && ...
                all(regexp(str, '[_a-zA-Z][_a-zA-Z0-9]*'));
            % Generate error message
            if isValid == true
                errmsg = [];
            else
                errmsg = DAStudio.message('RTW:autosar:invalidSymbol', str);
            end
        end
        
        function  [isValid, errmsg] = checkDisplayFormat(str, identifier)
            isValid = false;
            errmsg = [];
            if ~isempty(str)
                % % [flags] [width] [.prec] type character
                expression = '%[ \-+#]?[0-9]*(\.[0-9])?[diouxXfeEgGcs]';
                [startRegex, endRegex] = regexp(str, expression, 'ONCE');
                if ~isempty(startRegex) && startRegex == 1 && numel(str) == endRegex
                    isValid = true;
                else
                    errmsg = DAStudio.message('autosarstandard:ui:validateDisplayFormat', identifier);
                end
            else
              isValid = true;
            end
        end
        
        function doWordSizeCheck(ModelName, ObjectName, WordLength)
        % AUTOSAR 4 supports 64 bit integers, this requires long long
        % support to be enabled. This helper function checks the 
        % autosar version and model settings and issues the respective
        % error message
            if str2double(get_param(ModelName, 'AutosarSchemaVersion')) >= 4.0
                if WordLength < 0 || WordLength > 64
                    DAStudio.error('autosarstandard:validation:incorrectExportedWordSize', ...
                        ObjectName, int2str(WordLength));
                elseif strcmp(get_param(ModelName, 'TargetLongLongMode'), 'off') && WordLength > 32
                    DAStudio.error('autosarstandard:validation:incorrectTargetLongLongMode');
                end
            elseif WordLength > 32
                DAStudio.error('autosarstandard:validation:incorrectExportedWordSize', ...
                    ObjectName, int2str(WordLength));
            end        
        end
        
        function reportErrorWithFixit(errId, varargin)
            % this function is used to throw a validation error containing
            % a fixit. Fixits do not work properly with DAStudio error
            % messages as they will print raw XML into the commandline when
            % the validation is triggered via the command line
            msg = message(errId, varargin{:});
            MSLException([], msg).throw();
        end
        
        function checkAdaptiveModelSetup(modelName)
            %this function checks whether a model is setup for adaptive
            % and if so, it checks whether the feature flag is set. If not,
            % we error out to guard the entry points
            if autosar.api.Utils.isMappedToAdaptiveApplication(modelName)
                % if the model is mapped to an adaptive application, ensure
                % that the model is setup correctly and that the feature flag
                % is set
                if ~autosar.api.Utils.isAdaptiveModelSetup(modelName)
                    % determine what to do here with g1773251
                    % we might error, but metamodel and mapping might need
                    % trashing too.
                    DAStudio.error('autosarstandard:validation:invalidAdaptiveMapping', modelName);
                elseif slfeature('RTWAdaptiveAutosar') == 0
                    DAStudio.error('autosarstandard:validation:adaptiveFeatureNotSet', modelName);
                end
            end
        end
    end        
    
    methods(Static, Access=private)
        
        function checkStructTypeDimension(mdlName, slObj, Identifier, dataTypeName)
        % Function: checkStructTypeDimension ====================================
        % Abstract:
        %   Check dimension of struct type is n, [1xn] or [nx1]

            for elemIdx=1:length(slObj.Elements)
                elem = slObj.Elements(elemIdx);
                if length(elem.Dimensions) > 1
                    if (length(elem.Dimensions) == 2 && (elem.Dimensions(1) == 1 || ...
                                                         elem.Dimensions(2) == 1))
                        % okay
                    else
                        msg = DAStudio.message('RTW:autosar:invalidStructTypeDimension', Identifier, dataTypeName);
                        DAStudio.error('RTW:fcnClass:finish', msg);
                    end
                end

                % recurse through elements
                [objExists,elemObj] = autosar.utils.Workspace.objectExistsInModelScope(mdlName, elem.DataType);
                if objExists
                    if isa(elemObj, 'Simulink.StructType')
                        autosar.validation.AutosarUtils.checkStructTypeDimension(mdlName, elemObj, Identifier, elem.DataType);
                    else
                        % do nothing for everything else
                    end
                end
            end
        end

    end
    
end

% LocalWords:  CKeyword xn nx RPORT rport aliastypes autosarstandard prec dioux
% LocalWords:  Xfe Eg Compu Thru ui fixit Fixits commandline metamodel Sw Addr
% LocalWords:  scalartext CQualifier
