classdef InitValueChecker < handle
    %INITVALUECHECKER checks user input for InitValue against the Simulink
    %port
    %   This checking includes datatype as well as min/max checking

%   Copyright 2017-2018 The MathWorks, Inc.
    
    methods (Static, Access=public)
        
        function checkInitValue(hModel, userInitValueStr, portDataType, portName)
            % this methods extracts the data type and calls the respective
            % methods to verify the user input can be converted to the
            % port's datatype and lies within the specified min/max values
            portDataType = autosar.utils.StripPrefix(portDataType);            
            if strcmpi(portDataType, 'boolean')
                portDataType = 'logical'; % conversion for casting
            end
            if Simulink.data.isSupportedEnumClass(portDataType)
                autosar.validation.InitValueChecker.checkInitValueEnumType(...
                    hModel, userInitValueStr, portDataType, portName);
                return;
            end
            [dtExists,dtObj] = autosar.utils.Workspace.objectExistsInModelScope(hModel, portDataType);
            if dtExists && isa(dtObj, 'Simulink.Bus')
                autosar.validation.InitValueChecker.checkInitValueBusType(...
                    hModel, userInitValueStr, portDataType, portName);
            else
                autosar.validation.InitValueChecker.checkDataTypeCompatibility(...
                    hModel, userInitValueStr, portDataType, portName)
                % also check against min/max of the port
                minValue = eval(get_param(portName, 'OutMin'));
                maxValue = eval(get_param(portName, 'OutMax'));
                autosar.validation.InitValueChecker.checkInitValueMinMax(...
                    userInitValueStr, minValue, maxValue, portName);
            end
        end
        
    end
    
    methods(Static, Access = private)
        function checkInitValueBusType(hModel, userInitValueStr, portDataType, portName)
            % for bus types, we need to loop through the bus elements and
            % verify that the value can be casted and adheres to possible
            % min/max settings
            % get bus
            [~,bus] = autosar.utils.Workspace.objectExistsInModelScope(hModel, portDataType);
            for ii = 1 : length(bus.Elements)
                busElement = bus.Elements(ii);
                elementDataType = busElement.DataType;
                autosar.validation.InitValueChecker.checkInitValue(...
                    hModel, userInitValueStr, elementDataType, portName);
                elementMin = busElement.Min;
                elementMax = busElement.Max;
                autosar.validation.InitValueChecker.checkInitValueMinMax(...
                    userInitValueStr, elementMin, elementMax, portName);
            end
        end
        
        function checkInitValueEnumType(hModel, userInitValueStr, portDataType, portName)
            % for enum types, we need to verify that the value specified by
            % the user is a enumeration value
            assert(Simulink.data.isSupportedEnumClass(portDataType), ...
                [portDataType ' is not a supported enum class']);
            userInitValue = str2double(userInitValueStr);
            enumProps = Simulink.getMetaClassIfValidEnumDataType(portDataType);
            enumLiterals = {enumProps.EnumerationMemberList.Name};
            valueFoundInEnum = false;
            for ii = 1 : length(enumLiterals)
                if evalinGlobalScope(hModel, ...
                        [portDataType '.' enumLiterals{ii} '.double']) ...
                        == userInitValue
                    % value found
                    valueFoundInEnum = true;
                    break;
                end
            end
            if ~valueFoundInEnum
                % get default value for fixit
                defaultValue = Simulink.data.getEnumTypeInfo(portDataType, 'DefaultValue');
                defaultValueStr = num2str(defaultValue.double);
                 autosar.validation.AutosarUtils.reportErrorWithFixit('autosarstandard:validation:valueNotFoundInEnum', ...
                    userInitValueStr, portDataType, portName, defaultValueStr);
            end
        end
        
        function checkInitValueMinMax(userInitValueStr, minValue, maxValue, portName)
            % check minimum value
            userInitValue = str2double(userInitValueStr);
            if ~isempty(minValue) && userInitValue < minValue
                 autosar.validation.AutosarUtils.reportErrorWithFixit(...
                     'autosarstandard:validation:initValueLessThanMin', ...
                    userInitValueStr, num2str(minValue), portName);
            end
            % check maximum value
            if ~isempty(maxValue) && userInitValue > maxValue
                 autosar.validation.AutosarUtils.reportErrorWithFixit(...
                     'autosarstandard:validation:initValueGreaterThanMax', ...
                    userInitValueStr, num2str(maxValue), portName);
            end
        end
        
        function checkDataTypeCompatibility(hModel, userInitValueStr, portDataType, portName)
            % for other types, try casting to the base
            % type and see if the result is the same as the
            % input
            userInitValue = str2double(userInitValueStr);
            try
                castedValue = cast(userInitValue, portDataType);
            catch
                % dataType could be defined in the model scope
                if existsInGlobalScope(hModel, portDataType)
                    slObj = evalinGlobalScope(hModel, portDataType);
                    if isa(slObj, 'Simulink.AliasType')
                        baseType = evalinGlobalScope(hModel, [portDataType '.BaseType']);
                        autosar.validation.InitValueChecker.checkInitValue(...
                            hModel, userInitValueStr, baseType, portName);
                        return;
                    elseif isa(slObj, 'Simulink.NumericType')
                        if slObj.isfixed()
                            castedValue = fi(userInitValue, slObj);
                        else
                            % determine datatype and recurse
                            portDataType = slObj.DataTypeMode;
                            autosar.validation.InitValueChecker.checkInitValue(...
                                hModel, userInitValueStr, portDataType, portName);
                            return;
                        end
                    end
                elseif strncmp(portDataType, 'ufix', 4) || strncmp(portDataType, 'sfix', 4)
                    % numeric type with a fixed point notation
                    tmpFiObj = fixdt(portDataType);
                    castedValue = fi(userInitValue, tmpFiObj);
                else
                    assert(false, ['Datatype ' portDataType ' not recognized']);
                end
            end
            if abs(userInitValue - double(castedValue)) >= eps(userInitValue)
                 autosar.validation.AutosarUtils.reportErrorWithFixit(...
                     'autosarstandard:validation:valueNotRepresentableInType', ...
                    userInitValueStr, portDataType, portName, ...
                    num2str(double(castedValue)));
            end
        end
    end
end

% LocalWords:  fixit autosarstandard
