classdef MetaModelCommonValidator < m3i.Validator
% MetaModel validation checks for common elements
    
%   Copyright 2013-2018 The MathWorks, Inc.

    properties (Access=private, Transient)
        ModelName
        PropertiesToVerify;
    end
    
    
    methods (Access=public)
        
        function self = MetaModelCommonValidator(modelName)
            self.ModelName = modelName;
            self.PropertiesToVerify = {};

            % bind visitor
            self.dispatcher.bind('Simulink.metamodel.foundation.NamedElement',  @verifyIdentifiable);
            self.dispatcher.bind('Simulink.metamodel.arplatform.common.AUTOSAR', @verifyARRoot);
            self.dispatcher.bind('Simulink.metamodel.arplatform.common.ApplicationError', @verifyApplicationError);
            self.dispatcher.bind('Simulink.metamodel.arplatform.interface.NvDataInterface', @verifyNvDataInterface);
            self.dispatcher.bind('Simulink.metamodel.arplatform.interface.TriggerInterface', @verifyTriggerInterface);
            self.dispatcher.bind('Simulink.metamodel.arplatform.port.ProvidedRequiredPort', @verifyProvideRequirePort);
            self.dispatcher.bind('Simulink.metamodel.arplatform.behavior.IncludedDataTypeSet', @verifyIncludedDataTypeSet);
            self.dispatcher.bind('Simulink.metamodel.arplatform.interface.FlowData', @verifySwAddrMethod);
            self.dispatcher.bind('Simulink.metamodel.arplatform.interface.ArgumentData', @verifySwAddrMethod);
            self.dispatcher.bind('Simulink.metamodel.arplatform.interface.ParameterData', @verifySwAddrMethod);
            self.dispatcher.bind('Simulink.metamodel.arplatform.behavior.IrvData', @verifySwAddrMethod);
            
            % Schema version is needed for verifyIdentifiable
            schemaVersionHandler = autosar.SchemaVersionHandler.instance();
            schemaVersionHandler.setSchemaVersion(get_param(modelName, 'AutosarSchemaVersion'));
        end

        
        function verifyARRoot(this, arRoot)
            import autosar.mm.util.XmlOptionsAdapter;
            
            % Verify qualified names
            dataObj = autosar.api.getAUTOSARProperties(this.ModelName);
            compQName = dataObj.get('XmlOptions', 'ComponentQualifiedName');
            intBehQName = dataObj.get('XmlOptions', 'InternalBehaviorQualifiedName');
            impQName = dataObj.get('XmlOptions', 'ImplementationQualifiedName');
            
            this.checkQualifiedNameStr(compQName, 'ComponentQualifiedName');
            this.checkQualifiedNameStr(intBehQName, 'InternalBehaviorQualifiedName');
            this.checkQualifiedNameStr(impQName, 'ImplementationQualifiedName');
            
            duplicates = this.findDuplicates({compQName, intBehQName, impQName});

            if ~isempty(duplicates)
                DAStudio.error('RTW:autosar:duplicatePkgPath', duplicates{1});
            end

            this.checkPathStr(arRoot, 'DataTypePackage');
            this.checkPathStr(arRoot, 'InterfacePackage');
            
            xmlProps = XmlOptionsAdapter.getValidProperties();
            for ii=1:length(xmlProps)
                propertyName = xmlProps{ii};
                switch propertyName
                    case {'CompuMethodDirection', ...
                          'ImplementationTypeReference', ...
                          'SwCalibrationAccessDefault', ...
                          'InternalDataConstraintExport',...
                          'MoveElements',...
                          'CanBeInvokedConcurrentlyDiagnostic', ...
                          'ExportPropagatedVariantConditions'}
                        % not paths
                    case XmlOptionsAdapter.ComponentSpecificXmlOptions
                        m3iComp = autosar.api.Utils.m3iMappedComponent(this.ModelName);
                        this.checkPathStr(m3iComp, propertyName);
                    otherwise
                        this.checkPathStr(arRoot, propertyName);
                end
            end
            
        end
        
        function verifyProperties(this, m3iParent, propertyNames)
            % cache property names that we want to verify
            if ischar(propertyNames) || isStringScalar(propertyNames)
                propertyNames = {propertyNames};
            end
            this.PropertiesToVerify = propertyNames;
            
            % call verify
            this.verify(m3iParent);
            
            % make sure we don't drool PropertiesToVerify
            this.PropertiesToVerify = {};
        end
        
        function verifyIdentifiable(this, m3iIdentifiable)
            
            maxShortNameLength = get_param(this.ModelName, 'AutosarMaxShortNameLength');
            idcheckmessage = autosar.validation.AutosarUtils.isValidIdentifier(...
                {m3iIdentifiable.Name}, 'shortName', maxShortNameLength);
            if ~isempty(idcheckmessage) && ...
                    ~isa(m3iIdentifiable, 'Simulink.metamodel.types.EnumerationLiteral')
                DAStudio.error('RTW:fcnClass:finish', idcheckmessage);
            end             
            
            % find children with duplicate names
            this.verifyUniqueNames(m3iIdentifiable.containeeM3I);
        end
        
        function verifyUniqueNames(this, objSeq)
            
            objCell = m3i.filter(@(x) isa(x, 'Simulink.metamodel.foundation.NamedElement'), ...
                                 objSeq);
            
            nameCell = m3i.mapcell(@(x) x.Name, objCell);
            
            % Check for case insensitive clashes, e.g. RPort1, rPort1
            duplicates = this.findDuplicates(lower(nameCell));
            if ~isempty(duplicates)
                dupIdx = find(strcmpi(duplicates{1}, nameCell));
                object1 = objCell{dupIdx(1)};
                object2 = objCell{dupIdx(2)};

                DAStudio.error('autosarstandard:validation:shortNameCaseClash', ...
                    object1.Name, sprintf('%s %s', object1.MetaClass.name, autosar.api.Utils.getQualifiedName(object1)), ...
                    object2.Name, sprintf('%s %s', object2.MetaClass.name, autosar.api.Utils.getQualifiedName(object2)));
            end
            
        end

        function verifyApplicationError(~, m3iAppErr)
           
            errCodeValue = m3iAppErr.errorCode;
            if errCodeValue < 0 || errCodeValue > 63
                DAStudio.error('autosarstandard:validation:invalidAppErrValue', ...
                                       autosar.api.Utils.getQualifiedName(m3iAppErr), num2str(errCodeValue));
            end
            
        end
        
        function verifyNvDataInterface(this, m3iNvDataInterface)
           
            % NvDataInterface not supported for AUTOSAR Schema version less than 3.2
            schemaVersion = get_param(this.ModelName, 'AutosarSchemaVersion');
            if str2double(schemaVersion) < 3.2
                DAStudio.error('autosarstandard:validation:NvDataInterfaceNotAllowed', ...
                               autosar.api.Utils.getQualifiedName(m3iNvDataInterface));
            end
        end
        
        function verifyTriggerInterface(this, m3iTriggerInterface)
            % TriggerInterface not supported for AUTOSAR Schema version
            % less than 4.0
            schemaVersion = get_param(this.ModelName, 'AutosarSchemaVersion');
            
            if ~AR4.isCompatibleSchemaVer(schemaVersion)
                DAStudio.error('autosarstandard:validation:TriggerInterfaceNotAllowed', ...
                               autosar.api.Utils.getQualifiedName(m3iTriggerInterface));
            end
        end
        
        function verifyProvideRequirePort(this, m3iPRPort)
           
            schemaVersion = get_param(this.ModelName, 'AutosarSchemaVersion');
            [messageID, message] = autosar.validation.MetaModelValidator.verifyPRPort(autosar.api.Utils.getQualifiedName(m3iPRPort), schemaVersion);
            if ~isempty(messageID)
                mException = MException(messageID, message);
                throw(mException);
            end
        
        end
        
        function verifyIncludedDataTypeSet(this, includedDataTypeSet)
            if isempty(this.PropertiesToVerify)
                % by default verify all properties on IncludedDataTypeSet
                propertiesToVerify = {'LiteralPrefix'};
            else
                propertiesToVerify = this.PropertiesToVerify;
            end
            
            for idx=1:length(propertiesToVerify)
                propertyName = propertiesToVerify{idx};
                switch (propertyName)
                    case 'LiteralPrefix'
                        autosar.validation.MetaModelCommonValidator.verifyLiteralPrefix(includedDataTypeSet.LiteralPrefix);
                    otherwise
                        assert(false, 'Unrecognized property name "%s".', propertyName);
                end
            end
        end
        
        function verifySwAddrMethod(this, m3iData)
            % ensure MetaModelCommonValidator is setup correctly
            if ~isempty(this.PropertiesToVerify)
                assert(strcmp(this.PropertiesToVerify, 'SwAddrMethod'), ...
                    'Unexpected property');
            end
            % if SwAddrMethod is empty, nothing to verify
            if isempty(m3iData.SwAddrMethod)
                return;
            end
            selectedSwAddrMethod = m3iData.SwAddrMethod;
            if isempty(selectedSwAddrMethod.SectionType)
                % cannot check empty SectionType
                return;
            end
            selectedSectionTypeStr = selectedSwAddrMethod.SectionType.toString;
            % get category of SwAddrMethods from m3iData
            category = autosar.mm.util.SwAddrMethodHelper.getSwAddrMethodCategoryFromM3IObject(m3iData);
            % get accepted section types
            [~, acceptedSectionTypesStr] = autosar.mm.util.SwAddrMethodHelper.getAcceptedSectionTypesForCategory(category);
            
            if ~ismember(selectedSectionTypeStr, acceptedSectionTypesStr)
                % the SwAddrMethod does not have a valid Section Type.  
                DAStudio.error('autosarstandard:validation:invalidSwAddrMethodSectionType', ...
                    selectedSwAddrMethod, category, selectedSectionType.toString(), autosar.api.Utils.cell2str(acceptedSectionTypes));
            end
        end
    end
    
    methods(Access=private)
    
        function checkQualifiedNameStr(this, qname, propertyName)
            
            maxShortNameLength = get_param(this.ModelName, 'AutosarMaxShortNameLength');
            [isvalid, errMsg, errId] = autosarcore.checkIdentifier(qname, 'absPathShortName', maxShortNameLength);
            if ~isvalid
                errMsg = sprintf('XmlOptions %s %s', propertyName, errMsg);
                me = MException(errId, errMsg);
                throw(me);
            end
        end
            
        function checkPathStr(this, m3iObj, propertyName)        
            import autosar.mm.util.XmlOptionsAdapter;

            if XmlOptionsAdapter.isProperty(propertyName)
                pathStr = XmlOptionsAdapter.get(m3iObj, propertyName);
                if isempty(pathStr)
                    % XmlOptionAdapter paths can be empty
                    return
                end
            else
                pathStr = m3iObj.(propertyName);
            end

            maxShortNameLength = get_param(this.ModelName, 'AutosarMaxShortNameLength');

            [isvalid, errMsg, errId] = autosarcore.checkIdentifier(pathStr, 'absPath', maxShortNameLength);
            if ~isvalid
                errMsg = sprintf('XmlOptions %s %s', propertyName, errMsg);
                me = MException(errId, errMsg);
                throw(me);
            end
        end
        
    end
    
    methods(Static, Access=private)

        function duplicates = findDuplicates(names)
        
            % Find duplicates
            numNames = length(names);
            [~, unique_name_indices] = unique(names);
            duplicates = unique(names(setdiff(1:numNames, unique_name_indices)));
            
        end
        
        function verifyLiteralPrefix(propertyValue)
            % Concept of IncludedDataTypeSet (and thus LiteralPrefix)
            % did not exist prior to AR4, so no need to tie this
            % check to the AutosarMaxShortNameLength in the model
            %
            % Note that LiteralPrefix is optional and so may not be
            % specified in IncludedDataTypeSet. This translates to having
            % an empty value in the meta-model.
            if ~isempty(propertyValue)
                [isValid, errmsg] = autosarcore.checkIdentifier(propertyValue, 'shortName', 128);
                if ~isValid
                    error('RTW:autosar:invalidShortName', errmsg);
                end
            end
        end
    end
end

% LocalWords:  metamodel arplatform RPort Compu Sw autosarstandard Nv autosarstandard
% LocalWords:  Addr

