classdef CodeInfoBuilder < m3i.Visitor
    % CODEINFOBUILDER
    %
    %   This class implements an m3i Visitor that uses a Simulink
    %   AUTOSAR meta-model to create a CodeInfo object using the information
    %   from the AUTOSAR m3i meta-model.
    %
    % TODO:
    %       * What happens if the input/output was stubbed in the old
    %       release model, and has no DataRead/DataWrite associated with
    %       it?
    %
    %       * Can an IRV be queued?
    %
    % Copyright 2017-2018 The MathWorks, Inc.
    %
    properties
        CodeInfo;       % The CodeInfo object to be worked upon
        M3iModel;       % The top level M3i model
        M3iComp;        % The M3i component object providing the source data
        
        SourceFile = '';
        InitRunnableSymbol = '';
        TerminateRunnableSymbol = '';
        ExtraMappingInfo;
    end
    
    properties (Access = private)
       TypeBuilder;
       OperationBuilder;
       TimingInterfaceBuilder;
       PIMBuilder;
       
       PrototypeInports = coder.internal.xrel.AUTOSAR.mm2CodeInfo.DataInterfaceFactory.empty;
       PrototypeOutports = coder.internal.xrel.AUTOSAR.mm2CodeInfo.DataInterfaceFactory.empty;
       
       IsReusable;
       RteInstanceDataInterface;
    end
    
    properties (Constant, Access = private)
        AccessKindMap = containers.Map(...
            {...
            Simulink.metamodel.arplatform.behavior.DataAccessKind.ImplicitRead.toString(),...
            Simulink.metamodel.arplatform.behavior.DataAccessKind.ImplicitWrite.toString(),...
            Simulink.metamodel.arplatform.behavior.DataAccessKind.ExplicitReadByArg.toString(),...
            Simulink.metamodel.arplatform.behavior.DataAccessKind.ExplicitReadByValue.toString(),...
            Simulink.metamodel.arplatform.behavior.DataAccessKind.ExplicitWrite.toString()...
            },...
            {...
            'ImplicitReceive',...
            'ImplicitSend',...
            'ExplicitReceive',...
            'ExplicitReceiveByVal',...
            'ExplicitSend'...
            }...
            );
        
        ReadAccessKinds = {'ImplicitRead', 'ExplicitReadByArg', 'ExplicitReadByValue'};
        WriteAccessKinds = {'ImplicitWrite', 'ExplicitWrite'};
        QueuedAccessKinds = {'QueuedExplicitSend', 'QueuedExplicitReceive'};
        SenderReceiverModes = {'ExplicitSend', 'ExplicitReceive', 'ImplicitSend', 'ImplicitSendByRef', ...
            'QueuedExplicitReceive', 'ImplicitReceive', 'ModeReceive', ...
            'ExplicitReceiveByVal', 'ModeSend', 'QueuedExplicitSend', ...
            'EndToEndQueuedReceive', 'EndToEndQueuedSend'};
        IRTModes = {'INITIALIZE', 'RESET', 'TERMINATE'};
    end
      
    
    methods
        
        function obj = CodeInfoBuilder(lCodeInfo, m3iModel, m3iComp, xrelInfo)
            % CODEINFOBUILDER
            
            obj.CodeInfo = lCodeInfo;
            obj.M3iModel = m3iModel;
            obj.M3iComp = m3iComp;
            obj.IsReusable = m3iComp.Behavior.isMultiInstantiable;
            obj.ExtraMappingInfo = coder.internal.xrel.AUTOSAR.XRelArtifactUtils(xrelInfo);
            obj.InitRunnableSymbol = obj.ExtraMappingInfo.getAutosarInitRunnableSymbol;
            obj.TerminateRunnableSymbol = obj.ExtraMappingInfo.getAutosarTerminateRunnableSymbol;
            
            % Type builder helper
            obj.TypeBuilder = coder.internal.xrel.AUTOSAR.mm2CodeInfo.CodeInfoTypeBuilder(obj.CodeInfo, obj.ExtraMappingInfo, obj.M3iComp);
            obj.OperationBuilder = coder.internal.xrel.AUTOSAR.mm2CodeInfo.OperationBuilder(obj.SourceFile, obj.CodeInfo, obj.TypeBuilder, obj.ExtraMappingInfo);
            obj.TimingInterfaceBuilder = coder.internal.xrel.AUTOSAR.mm2CodeInfo.TimingInterfaceBuilder(obj.CodeInfo, obj.ExtraMappingInfo, obj.OperationBuilder);
            obj.PIMBuilder = coder.internal.xrel.AUTOSAR.mm2CodeInfo.PIMBuilder(obj.M3iModel, obj.CodeInfo, obj.TypeBuilder, obj.TimingInterfaceBuilder, obj.ExtraMappingInfo);
            
            % Register visitor for the AUTOSAR side
            obj.registerVisitor('mmVisit', 'mmVisit');
            
            obj.bind('Simulink.metamodel.arplatform.behavior.Runnable', @mmWalkRunnable, 'mmVisit');
            obj.bind('Simulink.metamodel.arplatform.behavior.PortParameterAccess', @mmWalkPortParameterAccess, 'mmVisit');
            obj.bind('Simulink.metamodel.arplatform.behavior.ComponentParameterAccess', @mmWalkComponentParameterAccess, 'mmVisit');
            
            obj.bind('Simulink.metamodel.arplatform.behavior.IrvAccess', @mmWalkIrvAccess, 'mmVisit');
            obj.bind('Simulink.metamodel.arplatform.behavior.FlowDataAccess', @mmWalkFlowDataAccess, 'mmVisit');            
        end
        
        function reorderPorts(obj, lCodeInfoProp, lArtifactNames)
            % REORDERPORTS
            if ~isempty(lArtifactNames)
                lCodeInfoNames = {obj.CodeInfo.(lCodeInfoProp).GraphicalName};
                assert(numel(lArtifactNames) == numel(lCodeInfoNames));
                lRemap = zeros(1, numel(lCodeInfoNames));
                for kPort = 1:numel(lArtifactNames)
                    % Find the position in CodeInfo of the current port in
                    % from the cross release artifact
                    lIdx = find(strcmp(lArtifactNames{kPort}, lCodeInfoNames));
                    assert(numel(lIdx) == 1);
                    lRemap(kPort) = lIdx;
                end
                % Finally, re-order
                obj.CodeInfo.(lCodeInfoProp) = obj.CodeInfo.(lCodeInfoProp)(lRemap);
            else
                return;
            end
        end
        
        function lExpInports = getExpInports(obj)
            % Use the cross-release artifact to get the number of input
            % ports
            lCodeInfoOutputFunctionSymbols = arrayfun(@(x)x.Prototype.Name, obj.CodeInfo.OutputFunctions, 'UniformOutput', false);
            lCodeInfoInportNames = arrayfun(@(x)x.GraphicalName, obj.CodeInfo.Inports, 'UniformOutput', false);
            
            nExpInports = obj.ExtraMappingInfo.getNumPorts('Inport');
            
            % First, check if the structure is needed and return empty if
            % not
            if ~obj.needsExpInports
                lExpInports = [];
                return;
            end
            
            % If we have anything that looks asynchronous then we should
            % make sure that reusable is false.  Code Generation does not
            % support reusable interface for models with function call
            % inputs, or Simulink Functions.  This assertion serves as a
            % cross-check in case that situation changes in the future.
            assert(~obj.IsReusable, ...
                'Did not expect asynchronous, explicit timing, or Simulink functions if we have reusable code interface');
            
            lExpInports = struct('PortType', {}, ...
                'IsLatched', {}, ...
                'EnableFcn', {}, ...
                'DisableFcn', {}, ...
                'IsFcnCall', {}, ...
                'Index', {}, ...
                'FcnType', {}, ...
                'TaskName', {});
            
            for kPort = 1:nExpInports
                if obj.ExtraMappingInfo.isFunctionCall('Inport', kPort)
                    lExpInports(kPort).PortType = 'fcn_call';
                else
                    lExpInports(kPort).PortType = 'input';
                end
                lExpInports(kPort).IsLatched = false;
                lExpInports(kPort).EnableFcn = [];
                lExpInports(kPort).DisableFcn = [];
                
                % Need to find the index of the output function
                if obj.ExtraMappingInfo.isFunctionCall('Inport', kPort)
                    % For function calls, the index is the output function
                    % index
                    lCurrSymbol = obj.ExtraMappingInfo.getPortSymbol('Inport', kPort);
                    lIndex = find(strcmp(lCurrSymbol, lCodeInfoOutputFunctionSymbols));
                else
                    % For inputs, the index is the Inport
                    lGraphicalName = obj.ExtraMappingInfo.getGraphicalName('Inport', kPort);
                    lIndex = find(strcmp(lGraphicalName, lCodeInfoInportNames));
                end
                
                assert(numel(lIndex) == 1);
                lExpInports(kPort).Index = lIndex;
                lExpInports(kPort).FcnType = '';
                lExpInports(kPort).TaskName = '';
            end
        end
            
        function updateCodeInfo(obj)
            % UPDATECODEINFO
            %   Triggers walking of the arModel and population of the
            %   skeleton CodeInfo object.
            
            lBehavior = obj.M3iComp.Behavior;
            
            % Set component name
            obj.CodeInfo.Name = obj.M3iComp.Name;
            
            % If reusable interface, add the reusable interface DWork
            if obj.IsReusable
                obj.processReusableDWork;
            end
            
            % Stub per-instance memories.  The ARXML does not reflect which
            % runnables access them.
            obj.PIMBuilder.processPims(obj.M3iComp.Behavior.ArTypedPIM);
            obj.PIMBuilder.processPims(obj.M3iComp.Behavior.PIM);
            
            % Initial condition for PIMs might be stored in a parameter -
            % the access is not necessarily captured by any runnable.
            lPIMInitCondParameterNames = unique(obj.PIMBuilder.NVRAMPIMInitCondMap.values);
            for kParamData = 1:lBehavior.Parameters.size()
                lM3iParamData = lBehavior.Parameters.at(kParamData);
                if ismember(lM3iParamData.Name, lPIMInitCondParameterNames)
                    obj.processInternalCalPrmParameterData(lM3iParamData);
                end
            end
            
            % We need to walk the runnables in two passes because we need
            % to gather all timing information before creating
            % RTW.DataImplementation objects.
            lFirstPass = true;
            obj.processRunnables(lBehavior.Runnables, lFirstPass);
            
            % Materialize the RTW.DataInterface objects for Inports/Outports 
            % from their prototypes, now we have the timing information in place.
            obj.materializePorts;
            
            % Do the second pass over the runnables.
            lFirstPass = false;
            obj.processRunnables(lBehavior.Runnables, lFirstPass);
            
            % TODO: 
            % In theory, a user might have generated code in an old release
            % with an input port or an output port stubbed with a
            % terminator or a ground block.  In this case, we won't have
            % visited a FlowDataAccess that referenced the port.
                        
            % Now we need to make sure that input and output ports are
            % ordered correctly
            
            % Assume Inports and Outports not empty
            lInportNames = obj.ExtraMappingInfo.getDataInportGraphicalNames;
            lOutportNames = obj.ExtraMappingInfo.getDataOutportGraphicalNames;
            obj.reorderPorts('Inports', lInportNames);
            obj.reorderPorts('Outports', lOutportNames);
            
            % Symbolic dimensions
            lUsedSystemConstants = unique(obj.TypeBuilder.SystemConstants);
            lSysConstsValueMap = autosar.api.Utils.createSystemConstantMap(obj.M3iModel, {}, {});
            
            if ~isempty(lUsedSystemConstants)
                lSymbolicDimensionDefinitions = ...
                    cellfun(@(x)struct('Name', x, 'Value', lSysConstsValueMap(x)), lUsedSystemConstants);
                obj.CodeInfo.SymbolicDimensionDefinitions = lSymbolicDimensionDefinitions;
            end
            
        end
        
        function ret = getGeneralDataFromTLCPatch(obj)
            % GETGENERALDATAFROMTLCPATCH
            ret = struct;
            ret.AutosarPIMDSM = obj.PIMBuilder.AutosarPIMDSM;
        end
        
        function ret = getNVRAMPIMInitCondMap(obj)
            % GETNVRAMPIMINITCONDMAP
            ret = obj.PIMBuilder.NVRAMPIMInitCondMap;
        end
    end
    
    methods (Access = public)
        function ret = mmVisitM3IObject(~, m3iObj, varargin)
            % MMVISITM3IOBJECT
            %
            % Note: this cannot be a true static method, because M3I is
            % expecting a particular function signature.  Default visitor
            % must be public.
            %
            ret = [];
            assert(false, 'Failure in: %s\n Unsupported M3I object class %s', mfilename('-class'), class(m3iObj));
        end
    end
    
    methods (Access = private)
        %
        % This methods section contains the methods that are associated
        % with the M3I Visitor pattern
        %        
        function ret = mmWalkRunnable(obj, m3iRunnable, lDoInit, lFirstPass)
            % MMWALKRUNNABLE
            %
            %   This method ensures that data reads and writes for the
            %   runnable are correctly registered in CodeInfo.
            %
            ret = [];
            
            lIsInitRunnable = strcmp(m3iRunnable.symbol, obj.InitRunnableSymbol);
            lIsTerminateRunnable = ~isempty(obj.TerminateRunnableSymbol) && strcmp(m3iRunnable.symbol, obj.TerminateRunnableSymbol);
            if (lIsInitRunnable == lDoInit)
                
                lCodeInfoFcn = obj.findFunctionByName(m3iRunnable.symbol);
                
                if lFirstPass
                    % On the first pass, create the timing interface and
                    % RTW.FunctionInterface objects for the runnables.
                    if isempty(lCodeInfoFcn)
                        % Need to create the function prototype
                        %
                        % Single use  AUTOSAR runnables are void/void if
                        % periodically triggered.
                        %
                        % Periodically triggered reusable runnables take an
                        % input argument "self" of type Rte_Instance
                        %
                        % Arguments and return values can be added
                        % if associated with an event.
                        lPrototype = coder.types.Prototype;
                        lPrototype.Name = m3iRunnable.symbol;
                        lPrototype.SourceFile = obj.SourceFile;
                        lCodeInfoFcn = RTW.FunctionInterface;
                        
                        if obj.IsReusable
                            % Add Rte_Instance for reusable interface.
                            %
                            % The Rte_Instance is passed in, even when the 
                            % model doesn't have a state DWork.
                            %
                            % In this case, we don't need to worry about
                            % Simulink functions because the function-call
                            % based modeling style is not permitted.
                            lArgument = coder.types.Argument;
                            lArgument.Type = obj.TypeBuilder.getRTEInstance;
                            lArgument.Name = 'self';
                            lArgument.IOType = 'INPUT';
                            lPrototype.Arguments(end+1) = lArgument;
                            lCodeInfoFcn.Prototype = lPrototype;
                            lCodeInfoFcn.ActualArgs(end+1) = obj.RteInstanceDataInterface;
                        else
                            lCodeInfoFcn.Prototype = lPrototype;
                        end
                        
                        if lIsInitRunnable
                            obj.TimingInterfaceBuilder.createTimingInterfaceForInitRunnable(m3iRunnable, lCodeInfoFcn);
                            obj.CodeInfo.InitializeFunctions(end+1) = lCodeInfoFcn;
                        elseif lIsTerminateRunnable
                            obj.TimingInterfaceBuilder.createTimingInterfaceForTerminateRunnable(m3iRunnable, lCodeInfoFcn);
                            obj.CodeInfo.TerminateFunctions(end+1) = lCodeInfoFcn;
                        else
                            obj.TimingInterfaceBuilder.createTimingInterfaceForOutputRunnable(m3iRunnable, lCodeInfoFcn);
                            obj.CodeInfo.OutputFunctions(end+1) = lCodeInfoFcn;
                        end
                    end
                end
                
                % Get the timing interface for the function
                
                % Ports associated with the initialize function use the
                % INITIALIZE timing mode even though the runnable has the
                % ONESHOT timing mode
                if lIsInitRunnable
                    assert(~isempty(obj.TimingInterfaceBuilder.TimingForInitialize), 'No INITIALIZE Timing Interface created');
                    lTimingInterface = obj.TimingInterfaceBuilder.TimingForInitialize;
                else
                    lTimingInterface = lCodeInfoFcn.Timing;
                end
                    
                % Make sure that data access is registered
                m3iDataAccess = m3iRunnable.dataAccess;
                obj.applySeq('mmVisit', m3iDataAccess, lCodeInfoFcn, lTimingInterface, lFirstPass);   %  mmWalkFlowDataAccess
                
                if ~lFirstPass
                    % Make sure port parameters are registered
                    m3iPortParam = m3iRunnable.portParamRead;
                    obj.applySeq('mmVisit', m3iPortParam, lCodeInfoFcn);
                    
                    m3iCompParam = m3iRunnable.compParamRead;
                    obj.applySeq('mmVisit', m3iCompParam, lCodeInfoFcn);
                    
                    % Make sure irv writes are registered
                    m3iIrvWrites = m3iRunnable.irvWrite;
                    lIsWrite = true;
                    obj.applySeq('mmVisit', m3iIrvWrites, lCodeInfoFcn, lIsWrite);
                    
                    % Make sure irv reads are registered
                    m3iIrvReads = m3iRunnable.irvRead;
                    lIsWrite = false;
                    obj.applySeq('mmVisit', m3iIrvReads, lCodeInfoFcn, lIsWrite);
                    
                    % Look for function calls
                    m3iBlockingCall = m3iRunnable.operationBlockingCall;
                    obj.OperationBuilder.processOperationBlockingAccess(m3iBlockingCall, lTimingInterface)
                end
                
            end
        end
                
        function ret = mmWalkFlowDataAccess(obj, m3iDataAccess, lCodeInfoFcn, lTimingInterface, lFirstPass)
            % MMWALKFLOWDATAACCESS
            %
            %   This method is used to populate DirectReads and
            %   DirectWrites on functions for inputs/outputs.
            %
            ret = [];
            lAccessKind = m3iDataAccess.Kind;
            m3iInstanceRef = m3iDataAccess.instanceRef;
            m3iPort = m3iInstanceRef.Port;
            m3iData = m3iInstanceRef.DataElements;
            switch lAccessKind.toString
                case obj.ReadAccessKinds
                    lAccessProp = 'DirectReads';
                    lPortType = 'Inports';
                    lPrototypePortType = 'PrototypeInports';
                case obj.WriteAccessKinds
                    lAccessProp = 'DirectWrites';
                    lPortType = 'Outports';
                    lPrototypePortType = 'PrototypeOutports';
                otherwise
                    assert(false, 'Invalid access kind');
            end            
            
            % Determine whether this port has an "IsUpdated" input
            lIsUpdated = autosar.api.Utils.isUpdatedPortElement(m3iPort, m3iData);
            
            % Determine whether there should be an "ErrorStatus" port as
            % well as a regular data port.
            lHasErrorStatus = obj.ExtraMappingInfo.hasErrorStatusPort(lPortType, m3iPort.Name, m3iData.Name);
            
            if lFirstPass
                % On the first pass we can create the object prototype
                obj.walkFlowDataAccess_FirstPass(lPrototypePortType, lPortType, lTimingInterface, lAccessKind, m3iPort, m3iData, lIsUpdated, lHasErrorStatus);
            else
                % On the second pass, we want the actual object, not the
                % prototype
                obj.walkFlowDataAccess_SecondPass(lPortType, lCodeInfoFcn, m3iPort, m3iData, lAccessProp, lIsUpdated, lHasErrorStatus);
            end
        end
        
        function ret = mmWalkIrvAccess(obj, m3iIrvAccess, lCodeInfoFcn, lIsWrite)
            % MMWALKIRVACCESS
            %
            %   This method is used to register DirectWrites and
            %   DirectReads for interrunnable variables
            %
            ret = [];
            m3iIrvData = m3iIrvAccess.instanceRef.DataElements;
            lVariableName = m3iIrvData.Name;
            lCodeInfoInternalData = obj.findInternalDataByVariableName(lVariableName);
            
            if isempty(lCodeInfoInternalData)
                % Find or create type object
                m3iType = m3iIrvData.Type;
                lCodeInfoType = obj.TypeBuilder.processType(m3iType);
                
                if strcmp(m3iIrvData.Kind.toString, 'Explicit')
                    lImplementation = RTW.AutosarInterRunnable('ExplicitInterRunnable');
                elseif strcmp(m3iIrvData.Kind.toString, 'Implicit')
                    lImplementation = RTW.AutosarInterRunnable('ImplicitInterRunnable');
                else
                    assert(false, 'm3iIrvData kind should be Explicit or Implicit');
                end
                
                lImplementation.VariableName = m3iIrvData.Name;
                lImplementation.Type = lCodeInfoType;
                lSID = 'NULLVALUE:0';
                lGraphicalName = m3iIrvData.Name;
                lTimingInterface = RTW.TimingInterface.empty;
                lIsQueued = false;
                lQueueLength = 1;
                lCodeInfoInternalData = RTW.DataInterface(lSID, lGraphicalName, ...
                    lImplementation, lTimingInterface);
                lCodeInfoInternalData.Type = coder.types.Type.empty;  % No type for IRV except on implementation
                lCodeInfoInternalData.IsQueued = lIsQueued;
                lCodeInfoInternalData.QueueLength = lQueueLength;
                obj.CodeInfo.InternalData(end+1) = lCodeInfoInternalData;                
            end
            
            if lIsWrite
                lProp = 'DirectWrites';
            else
                lProp = 'DirectReads';
            end
            lDirectAccess = lCodeInfoFcn.(lProp);
            lDirectIRVs = obj.findObjByImplementationClass(lDirectAccess, 'RTW.AutosarInterRunnable');
            lDirectIRVNames = arrayfun(@(x)x.Implementation.VariableName, ...
                lDirectIRVs, ...
                'UniformOutput', false);    
            if ~ismember(lVariableName, lDirectIRVNames)
                lCodeInfoFcn.(lProp)(end+1) = lCodeInfoInternalData;
            end
        end
                        
        function ret = mmWalkComponentParameterAccess(obj, m3iComponentParameterAccess, lCodeInfoFcn)
            % MMWALKCOMPONENTPARAMETERACCESS
            ret = [];
            m3iParameterData = m3iComponentParameterAccess.instanceRef.DataElements;
            lCodeInfoParam = obj.processInternalCalPrmParameterData(m3iParameterData);
            
            % Append to the param to the direct reads of the runnable
            lCodeInfoFcn.DirectReads(end+1) = lCodeInfoParam;
        end
        
        function lCodeInfoParam = processInternalCalPrmParameterData(obj, m3iParameterData)
            % PROCESSINTERNALCALPRMPARAMETERDATA
            lGraphicalName = m3iParameterData.Name;
            lPort = m3iParameterData.Name;
            if strcmp(m3iParameterData.Kind.toString, 'Shared')
                lElement = 'true';
            else
                lElement = 'false';
            end
            lIsInternal = true;
            lInterfacePath = '';
            lCodeInfoParam = obj.findOrCreateAutosarCalibration(lGraphicalName, lPort, lElement, lInterfacePath, m3iParameterData, lIsInternal);
        end
        
        function ret = mmWalkPortParameterAccess(obj, m3iPortParameterAccess, lCodeInfoFcn)
            % MMWALKPORTPARAMETERACCESS
            %
            %   This method registers DirectReads for parameters with the
            %   CodeInfo function.
            ret = [];
            m3iParameterData = m3iPortParameterAccess.instanceRef.DataElements;
            m3iPortData = m3iPortParameterAccess.instanceRef.Port;
            lPort = m3iPortData.Name;
            lElement = m3iParameterData.Name;
            lIsInternal = false;
            lGraphicalName = [lPort, '_', m3iParameterData.Name];
            
            % But a MATLAB identifier can't be > 63 chars - so truncate.
            if numel(lGraphicalName) > 63
                lGraphicalName = lGraphicalName(1:63);
            end
            
            lQualifiedName = m3iPortData.Interface.qualifiedNameWithSeparator('/');
            lInterfacePath = regexprep(lQualifiedName, '^AUTOSAR', '');
            lCodeInfoParam = obj.findOrCreateAutosarCalibration(lGraphicalName, lPort, lElement, lInterfacePath, m3iParameterData, lIsInternal);
            
            % Append to the param to the direct reads of the runnable
            lCodeInfoFcn.DirectReads(end+1) = lCodeInfoParam;
        end
    end
  
    methods (Access = private)
        %
        % PRIVATE UTILITY METHODS
        %
        % Deliberately a separate methods section to create separation from
        % mmWalk* methods.
        %
        function processRunnables(obj, m3iRunnables, lFirstPass)
            % PROCESSRUNNABLES
            %
            % Need to walk twice to make sure the init runnable timing is
            % at the end.  Process regular runnables first.
            lDoInit = false;
            obj.applySeq('mmVisit', m3iRunnables, lDoInit, lFirstPass);
            
            % Now process the init runnable
            lDoInit = true;
            obj.applySeq('mmVisit', m3iRunnables, lDoInit, lFirstPass);
        end
        
        function materializePorts(obj)
            % MATERIALIZEPORTS
            %
            %   Materialize the prototype RTW.DataInterface objects for
            %   Inports and Outports.
            for kInport = 1:numel(obj.PrototypeInports)
                obj.CodeInfo.Inports(end+1) = obj.PrototypeInports(kInport).build;
            end
            for kOutport = 1:numel(obj.PrototypeOutports)
                obj.CodeInfo.Outports(end+1) = obj.PrototypeOutports(kOutport).build;
            end
        end
        
        function walkFlowDataAccess_FirstPass(obj, lPrototypePortType, lPortType, lTimingInterface, lAccessKind, m3iPort, m3iData, lIsUpdated, lHasErrorStatus)
            % WALKFLOWDATAACCESS_FIRSTPASS
            % 
            % Create an object prototype and cache it.
            %
            % If an IsUpdated port is needed, create an object prototype
            % and cache that as well.
            obj.cacheSenderReceiverDataInterface(lPrototypePortType, ...
                lPortType, lTimingInterface, lAccessKind, m3iPort, m3iData);
            if lIsUpdated
                % If we also have an IsUpdated input, then we should
                % also create an IsUpdated port in CodeInfo
                obj.cacheIsUpdatedDataInterface(lPrototypePortType, lPortType, lTimingInterface, m3iPort, m3iData);
            end
            
            if lHasErrorStatus
                % If we need an error status port, then we need to drop one
                % into CodeInfo
                obj.cacheErrorStatusDataInterface(lPrototypePortType, lPortType, lTimingInterface, m3iPort, m3iData);
            end
        end
        
        function walkFlowDataAccess_SecondPass(obj, lPortType, lCodeInfoFcn, m3iPort, m3iData, lAccessProp, lIsUpdated, lHasErrorStatus)
            % WALKFLOWDATAACCESS_SECONDPASS
            %
            m3iPortName = m3iPort.Name;
            m3iDataElementName = m3iData.Name;
            lCodeInfoDataInterface = obj.findObjByPortAndDataElement(lPortType, ...
                m3iPortName, m3iDataElementName, obj.SenderReceiverModes);
            
            % Per CodeInfo conventions, don't add queued interfaces to
            % direct reads/writes.  Note that a Queued AUTOSAR 
            % DataAccessMode might not necessarily resolve to a Queued 
            % Simulink port type.
            %
            % We care about the Simulink scenario at this point.
            if lCodeInfoDataInterface.IsQueued
                return;
            end
            
            % Add the data interface to DirectReads or DirectWrites if it
            % is not already there.
            %
            % Don't attach QueuedExplicitSend or QueuedExplicitReceive to the function interface
            lDirectAccess = lCodeInfoFcn.(lAccessProp);
            lAutosarSenderReceivers = obj.findObjByImplementationClass(lDirectAccess, 'RTW.AutosarSenderReceiver');
            lPortNames = arrayfun(@(x)x.Implementation.Port, lAutosarSenderReceivers, 'UniformOutput', false);
            lDataElementNames = arrayfun(@(x)x.Implementation.DataElement, lAutosarSenderReceivers, 'UniformOutput', false);
            if ~any(strcmp(lPortNames, m3iPortName) & strcmp(lDataElementNames, m3iDataElementName))
                lCodeInfoFcn.(lAccessProp)(end+1) = lCodeInfoDataInterface;
            end
            
            if lIsUpdated
                lIsUpdatedDataInterface = obj.findObjByPortAndDataElement(lPortType, m3iPortName, m3iDataElementName, 'IsUpdated');
                % And we should also make sure that
                % DirectReads/DirectWrites are registered.
                if ~isempty(lIsUpdatedDataInterface)
                    lIsUpdatedImp = obj.findObjByImplementationClass(lDirectAccess, 'RTW.AutosarIsUpdated');
                    lPortNames = arrayfun(@(x)x.Implementation.Port, lIsUpdatedImp, 'UniformOutput', false);
                    lDataElementNames = arrayfun(@(x)x.Implementation.DataElement, lIsUpdatedImp, 'UniformOutput', false);
                    if ~any(strcmp(lPortNames, m3iPortName) & strcmp(lDataElementNames, m3iDataElementName))
                        lCodeInfoFcn.(lAccessProp)(end+1) = lIsUpdatedDataInterface;
                    end
                end
            end
            
            if lHasErrorStatus
                % The error status data interface only references the
                % original data port by its graphical port number!
                %
                % The XRelArtifactUtils class wraps the cross-release
                % artifact and allows us to de-reference this
                lXrelErrorStatusPortStruct = obj.ExtraMappingInfo.getXRelInfoFromPortAndDataElement( ...
                    lPortType, m3iPort.Name, m3iData.Name, 'ErrorStatus');
                lReceiverPortNumber = lXrelErrorStatusPortStruct.AutosarReceiverPortNumber;
                lErrorStatusDataInterface = obj.findObjByReceiverPortNumber(lPortType, lReceiverPortNumber, 'ErrorStatus');
                if ~isempty(lErrorStatusDataInterface)
                    lErrorStatusImp = obj.findObjByImplementationClass(lDirectAccess, 'RTW.AutosarErrorStatus');
                    lReceiverPortNumbers = arrayfun(@(x)x.Implementation.ReceiverPortNumber, lErrorStatusImp, 'UniformOutput', false);
                    if ~any(strcmp(lReceiverPortNumbers, lXrelErrorStatusPortStruct.AutosarReceiverPortNumber))
                        lCodeInfoFcn.(lAccessProp)(end+1) = lErrorStatusDataInterface;
                    end
                end
            end
        end
        
        function cacheSenderReceiverDataInterface(obj, lPrototypePortType, ...
                lPortType, lTimingInterface, lAccessKind, m3iPort, m3iData)
            % CACHESENDERRECEIVERDATAINTERFACE
            m3iPortName = m3iPort.Name;
            m3iDataElementName = m3iData.Name;
            lCodeInfoDataInterface = obj.findProtoObjByPortAndDataElement(lPrototypePortType, m3iPortName, m3iDataElementName, obj.SenderReceiverModes);
            if isempty(lCodeInfoDataInterface)
                obj.(lPrototypePortType)(end+1) = obj.createSenderReceiverDataInterface(m3iPort, m3iData, lAccessKind, lPortType, lTimingInterface);
            else
                % If the data interface already exists, then we need to check
                % the timing interface.
                obj.appendTimingInterface(lCodeInfoDataInterface, lTimingInterface);
            end
        end
        
        function cacheIsUpdatedDataInterface(obj, lPrototypePortType, lPortType, lTimingInterface, m3iPort, m3iData)
            % CACHEISUPDATEDDATAINTERFACE
            m3iPortName = m3iPort.Name;
            m3iDataElementName = m3iData.Name;
            lIsUpdatedDataInterface = obj.findProtoObjByPortAndDataElement(lPrototypePortType, ...
                m3iPortName, m3iDataElementName, 'IsUpdated');
            if isempty(lIsUpdatedDataInterface)
                lIsUpdatedDataInterface = obj.createIsUpdatedDataInterface(m3iPort, m3iData, lPortType, lTimingInterface);
                if isempty(lIsUpdatedDataInterface)
                    % g1672968
                    %   EnableUpdate might be true on the comspec even
                    %   though an IsUpdated port is not required.
                else
                    obj.(lPrototypePortType)(end+1) = lIsUpdatedDataInterface;
                end
            else
                obj.appendTimingInterface(lIsUpdatedDataInterface, lTimingInterface);
            end
        end
        
        function cacheErrorStatusDataInterface(obj, lPrototypePortType, lPortType, lTimingInterface, m3iPort, m3iData)
            % CACHEERRORSTATUSDATAINTERFACE
            
            % Find information for the ErrorStatus port
            lXrelPortStruct = obj.ExtraMappingInfo.getXRelInfoFromPortAndDataElement( ...
                lPortType, m3iPort.Name, m3iData.Name, 'ErrorStatus');
            
            % Should not be empty because this code path is conditional on
            % it not being empty!
            assert(~isempty(lXrelPortStruct));
            
            lReceiverPortNumber = lXrelPortStruct.AutosarReceiverPortNumber;
            lGraphicalName = lXrelPortStruct.GraphicalName;
            
            lDI = obj.findProtoObjByReceiverPortNumber(lPrototypePortType, ...
                lReceiverPortNumber, 'ErrorStatus');
            if isempty(lDI)
                lDI = obj.createErrorStatusDataInterface(m3iPort, m3iData, lReceiverPortNumber, lGraphicalName, lTimingInterface);
                obj.(lPrototypePortType)(end+1) = lDI;
            else
                obj.appendTimingInterface(lDI, lTimingInterface);
            end
        end
        
        
        function appendTimingInterface(obj, lCodeInfoDataInterface, lTimingInterface)
            % APPENDTIMINGINTERFACE
            %
            % Don't do anything if the timing interface is ONESHOT
            if strcmp(lTimingInterface.TimingMode, 'ONESHOT')
                return;
            end
            
            lExistingTiming = lCodeInfoDataInterface.Timing;
            
            % If the existing timing interface is INITIALIZE, RESET, OR
            % TERMINATE and the new one is periodic, then we should *replace*
            % the interface with the new one.
            if ismember(lExistingTiming.TimingMode, obj.IRTModes)
                if strcmp(lTimingInterface.TimingMode, 'PERIODIC') || strcmp(lTimingInterface.TimingMode, 'ASYNCHRONOUS')
                    lCodeInfoDataInterface.Timing = lTimingInterface;
                else
                    % Do nothing - leave what we have - we don't want a
                    % union between IRT modes
                end
            elseif ( strcmp(lExistingTiming.TimingMode, 'PERIODIC') || strcmp(lExistingTiming.TimingMode, 'ASYNCHRONOUS') ) ...
                    && ismember(lTimingInterface.TimingMode, obj.IRTModes)
                % Do nothing - don't union IRT modes with an existing
                % periodic timing mode.
            elseif isempty(lExistingTiming.UnionTimingInfo)
                % Otherwise, see if the timing interface is the same as the
                % one already there
                lIsMatch = obj.TimingInterfaceBuilder.isTimingEqual(lTimingInterface, lExistingTiming);
                if ~lIsMatch
                    % If the timing interface doesn't match, we need to
                    % create a union timing
                    lNewTiming = obj.TimingInterfaceBuilder.createUnion(lExistingTiming, lTimingInterface);
                    lCodeInfoDataInterface.Timing = lNewTiming; % Note: Handle classes
                end
            else
                lIsMember = obj.TimingInterfaceBuilder.isUnionMember(lTimingInterface, lExistingTiming);
                if ~lIsMember
                    lExistingTiming.UnionTimingInfo(end+1) = lTimingInterface;
                end
            end
        end
        
        function lIsUpdatedDataInterface = createIsUpdatedDataInterface(obj, m3iPort, m3iData, lPortType, lTimingInterface)
            % CREATEISUPDATEDDATAINTERFACE
            
            % Non-empty stub SID value
            lSID = 'NULLVALUE:0';
            
            % Find information for the IsUpdated type
            lXrelPortStruct = obj.ExtraMappingInfo.getXRelInfoFromPortAndDataElement( ...
                lPortType, m3iPort.Name, m3iData.Name, 'IsUpdated');
            
            if isempty(lXrelPortStruct)
                % Even though the ComSpec says we might have an Isupdated
                % Port, one might not be present. (g1672968)
                %
                % It turns out in this case that we don't need to present the IsUpdated
                % port for this model because the model in the old release
                % did not expose such a port.
                lIsUpdatedDataInterface = [];
                return;
            end
            
            % Get the index of the matching data element
            lDataPortStruct = obj.ExtraMappingInfo.getXRelInfoFromPortAndDataElement( ...
                lPortType, m3iPort.Name, m3iData.Name, obj.SenderReceiverModes);
            assert(~isempty(lDataPortStruct), 'Could not find port');
            
            % lIdx is the Inports/Outports index.  If this is an inport
            % then we need to map to the actual input ports (including
            % function caller inputs).
            if strcmp(lPortType, 'Inports')
                lExpIdx = obj.ExtraMappingInfo.getExpInportIdxFromGraphicalName(lDataPortStruct.GraphicalName);
            end
            
            lIsQueued = lXrelPortStruct.IsQueued;
            lQueueLength = lXrelPortStruct.QueueLength;
            lCodeInfoType = obj.TypeBuilder.getBoolean;
            lDataAccessMode = 'IsUpdated';
            
            lImplementation = RTW.AutosarIsUpdated(lDataAccessMode);
            lImplementation.Port = m3iPort.Name;
            lImplementation.DataElement = m3iData.Name;
            lImplementation.ReceiverPortNumber = num2str(lExpIdx);
            lImplementation.Type = lCodeInfoType;
            
            lIsUpdatedDataInterface = coder.internal.xrel.AUTOSAR.mm2CodeInfo.DataInterfaceFactory( ...
                lSID, ...
                lXrelPortStruct.GraphicalName, ...
                lImplementation, ...
                lTimingInterface);
            lIsUpdatedDataInterface.Type = lCodeInfoType;
            lIsUpdatedDataInterface.IsQueued = lIsQueued;
            lIsUpdatedDataInterface.QueueLength = lQueueLength;
        end
        
        function lCodeInfoType = getTypeForErrorStatus(obj, m3iPort, m3iData)
            % GETTYPEFORERRORSTATUS
            
            lIsNvPort = autosar.api.Utils.isNvPort(m3iPort);
            lUsesEndToEndProtection = ~lIsNvPort && ...
                coder.internal.xrel.AUTOSAR.AutosarImportModelManager.isE2E(m3iPort, m3iData);
            if lUsesEndToEndProtection
                lCodeInfoType = obj.TypeBuilder.getUInt32;
            else
                lCodeInfoType = obj.TypeBuilder.getUInt8;
            end
        end
        
        function lErrorStatusDataInterface = createErrorStatusDataInterface(obj, m3iPort, m3iData, ...
                lReceiverPortNumber, lGraphicalName, lTimingInterface)
            % CREATEERRORSTATUSDATAINTERFACE
            
            % Non-empty stub SID value
            lSID = 'NULLVALUE:0';
            
            % Type should be uint8 except if end-to-end protected, then it
            % should be uint32
            lCodeInfoType = obj.getTypeForErrorStatus(m3iPort, m3iData);
            
            
            lDataAccessMode = 'ErrorStatus';
            lImplementation = RTW.AutosarErrorStatus(lDataAccessMode);
            lImplementation.ReceiverPortNumber = lReceiverPortNumber;
            lImplementation.Type = lCodeInfoType;
            lErrorStatusDataInterface = coder.internal.xrel.AUTOSAR.mm2CodeInfo.DataInterfaceFactory( ...
                lSID, ...
                lGraphicalName, ...
                lImplementation, ...
                lTimingInterface);
            lErrorStatusDataInterface.Type = lCodeInfoType;
            lErrorStatusDataInterface.IsQueued = false;
            lErrorStatusDataInterface.QueueLength = 1;
        end
        
        
        
        function lCodeInfoDataInterface = createSenderReceiverDataInterface(obj, m3iPort, m3iData, lAccessKind, lPortType, lTimingInterface)
            % CREATESENDERRECEIVERDATAINTERFACE
            
            % Find or create type object
            m3iType = m3iData.Type;
            lXrelPortStruct = obj.ExtraMappingInfo.getXRelInfoFromPortAndDataElement( ...
                lPortType, m3iPort.Name, m3iData.Name, obj.SenderReceiverModes);
            assert(~isempty(lXrelPortStruct), 'Could not find port');
            
            lCodeInfoBaseType = obj.TypeBuilder.processType(m3iType, 'Dimensions', lXrelPortStruct.Dimensions);
            
            lIsQueued = lXrelPortStruct.IsQueued;
            lQueueLength = lXrelPortStruct.QueueLength;
            if numel(lXrelPortStruct.Dimensions) > 1 && prod(lXrelPortStruct.Dimensions) == 1
                lCodeInfoType = obj.TypeBuilder.wrapScalarWithMatrix(lCodeInfoBaseType, m3iType.Name);
            else
                lCodeInfoType = lCodeInfoBaseType;
            end
            lDataAccessMode = obj.AccessKindMap(lAccessKind.toString);
            
            % DataAccessMode might or might not be correct
            lIsRead = strcmp(lPortType, 'Inports');
            lComSpec = autosar.mm.Model.findComSpecForDataElement(m3iPort, m3iData.Name, lIsRead);
            lArIsQueued = false;
            if ~isempty(lComSpec)
                lArIsQueued = isa(lComSpec, 'Simulink.metamodel.arplatform.port.DataSenderQueuedPortComSpec') || ...
                    isa(lComSpec, 'Simulink.metamodel.arplatform.port.DataReceiverQueuedPortComSpec');
            end
            if lArIsQueued
                lDataAccessMode = sprintf('%s%s', 'Queued', lDataAccessMode);
            end
            
            % Patch ImplicitSendByRef
            if strcmp(lDataAccessMode, 'ImplicitSend') && ...
                    strcmp(lXrelPortStruct.DataAccessMode, 'ImplicitSendByRef')
                lDataAccessMode = 'ImplicitSendByRef';
            end
            assert(strcmp(lDataAccessMode, lXrelPortStruct.DataAccessMode));
            
            % Non-empty stub SID value
            lSID = 'NULLVALUE:0';
            
            % Create Implementation object
            lImplementation = RTW.AutosarSenderReceiver(lDataAccessMode);
            lImplementation.Port = m3iPort.Name;
            lImplementation.DataElement = m3iData.Name;
            lImplementation.Type = lCodeInfoType;
            
            %lCodeInfoDataInterface = RTW.DataInterface(lSID, lXrelPortStruct.GraphicalName, ...
            %    lImplementation, lTimingInterface);
            lCodeInfoDataInterface = coder.internal.xrel.AUTOSAR.mm2CodeInfo.DataInterfaceFactory( ...
                lSID, ...
                lXrelPortStruct.GraphicalName, ...
                lImplementation, ...
                lTimingInterface);
            lCodeInfoDataInterface.Type = lCodeInfoType;
            lCodeInfoDataInterface.IsQueued = lIsQueued;
            lCodeInfoDataInterface.QueueLength = lQueueLength;
        end

        function processReusableDWork(obj)
            % PROCESSREUSABLEDWORK
            %
            % Need to add the "Self" DWork first
            fhIsSelf = @(x)~isempty(x.Implementation) && strcmp(x.Implementation.Identifier, 'self');
            bIsSelf = arrayfun(fhIsSelf, obj.CodeInfo.InternalData);
            if ~any(bIsSelf)
                % If there isn't an Rte_Instance DataInterface, add one.
                % Even if the model does not have a block state DWork, this
                % DataInterface should be present.
                lImplementation = RTW.Variable;
                lImplementation.Identifier = 'self';
                lImplementation.Type = obj.TypeBuilder.getRTEInstance;
                lDataInterface = RTW.DataInterface;
                lDataInterface.Implementation = lImplementation;
                obj.CodeInfo.InternalData(end+1) = lDataInterface;
                obj.RteInstanceDataInterface = lDataInterface;  % To avoid having to search for it
            end
            
            % Now need to add the Block States DWork
            lStateTypes = {'Block states', ...
                'Block signals', ...
                'Zero crossing states'};
            
            for kState = 1:numel(lStateTypes)
                lCurrStateType = lStateTypes{kState};
                fhIsXStates = @(x)strcmp(x.GraphicalName, lCurrStateType);
                bIsXStates = arrayfun(fhIsXStates, obj.CodeInfo.InternalData); 
                if ~any(bIsXStates)  && ~isempty(obj.ExtraMappingInfo.getStatesDWorkName(lCurrStateType))
                    % Assuming a Block states DWork is not already present AND
                    % assuming that the original code has a block states DWork
                    % (might not if there is no state, even if reusable)
                    lImplementation = RTW.Variable;
                    lImplementation.Identifier = obj.ExtraMappingInfo.getStatesDWorkName(lCurrStateType);
                    lImplementation.Type = obj.TypeBuilder.getStatesType(...
                        obj.ExtraMappingInfo.getStatesDWorkType(lCurrStateType), ...
                        obj.ExtraMappingInfo.getStatesDWorkTag(lCurrStateType));
                    lDataInterface = RTW.DataInterface( ...
                        '', ...  % Empty SID
                        lCurrStateType, ...
                        lImplementation, ...
                        RTW.TimingInterface.empty);
                    obj.CodeInfo.InternalData(end+1) = lDataInterface;
                end
            end
        end
        
        
        function lCodeInfoParam = findOrCreateAutosarCalibration(obj, lGraphicalName, lPortName, lElementName, lInterfacePath, m3iParameterData, lIsInternal)
            % FINDORCREATEAUTOSARCALIBRATION
            
            lCodeInfoParam = obj.findParameterByPortAndElement(lPortName, lElementName);
            if isempty(lCodeInfoParam)
                % Deal with type
                m3iType = m3iParameterData.Type;
                
                if lIsInternal
                    lDataAccess = 'InternalCalPrm';
                else
                    lDataAccess = 'Calibration';
                end
                lArtifact = obj.ExtraMappingInfo.getParameterByImplementation(lPortName, lElementName, lDataAccess);
                if isempty(lArtifact)
                    % If we insert a parameter for NVRAM init conditions,
                    % these won't be in the cross-release artifact.
                    %
                    % Flatten and make column vector
                    if isa(m3iType, 'Simulink.metamodel.types.Matrix')
                        nDims = m3iType.Dimensions.size;
                        lDims = ones(1, nDims);
                        for kDim = 1:nDims
                            lDims(kDim) = m3iType.Dimensions.at(kDim);
                        end
                        lDimensions = [prod(lDims),1];
                    else
                        lDimensions = [1,1];
                    end
                    lCodeInfoBaseType = obj.TypeBuilder.processType(m3iType, 'Dimensions', lDimensions);
                else
                    lCodeInfoBaseType = obj.TypeBuilder.processType(m3iType, 'Dimensions', lArtifact.Dimensions);
                end
                    
                
                % If not a matrix, then create a matrix wrapper type
                lCodeInfoType = obj.TypeBuilder.wrapScalarWithMatrix(lCodeInfoBaseType, m3iType.Name);
                
                % Create an implementation object
                lImplementation = RTW.AutosarCalibration(lDataAccess);
                lImplementation.Port = lPortName;
                lImplementation.ElementName = lElementName;
                lImplementation.Shared = lElementName;
                
                if isa(m3iType, 'Simulink.metamodel.types.LookupTableType') && ~lIsInternal
                    lImplementation.InterfacePath = 'UNDEFINED';
                else
                    lImplementation.InterfacePath = lInterfacePath;
                end
                lImplementation.Type = lCodeInfoType;
                
                % Now create the parameter object
                lSID = '';
                lCodeInfoParam = RTW.DataInterface(lSID, lGraphicalName, ...
                    RTW.Variable.empty, RTW.TimingInterface.empty);
                lCodeInfoParam.Type = lCodeInfoType;
                lCodeInfoParam.Implementation = lImplementation;
                obj.CodeInfo.Parameters(end+1) = lCodeInfoParam;
            end
        end
        
        function lInternalData = findInternalDataByVariableName(obj, lVariableName)
            % FINDINTERNALDATABYVARIABLENAME
            lInternalData = [];
            for kObj = 1:numel(obj.CodeInfo.InternalData)
                if strcmp(lVariableName, obj.CodeInfo.InternalData(kObj).Implementation.VariableName)
                    lInternalData = obj.CodeInfo.InternalData(kObj);
                    break;
                end
            end
        end
        
        function lDIFactory = findProtoObjByPortAndDataElement(obj, lProp, lPort, lDataElement, lDataAccess)
            % FINDPROTOOBJBYPORTANDDATAELEMENT
            lDIFactory = [];
            lObjArray = obj.(lProp);
            if ~iscell(lDataAccess)
                lDataAccess = {lDataAccess};
            end
            for kObj = 1:numel(lObjArray)
                if ismember(lObjArray(kObj).Implementation.DataAccessMode, lDataAccess) && ...
                        strcmp(lPort, lObjArray(kObj).Implementation.Port) && ...
                        strcmp(lDataElement, lObjArray(kObj).Implementation.DataElement)
                    lDIFactory = lObjArray(kObj);
                    break;
                end
            end
        end
        
        function lDIFactory = findProtoObjByReceiverPortNumber(obj, lProp, lReceiverPortNumber, lDataAccess)
            % FINDPROTOOBJBYRECEIVERPORTNUMBER
            lDIFactory = [];
            lObjArray = obj.(lProp);
            if ~iscell(lDataAccess)
                lDataAccess = {lDataAccess};
            end
            for kObj = 1:numel(lObjArray)
                if ismember(lObjArray(kObj).Implementation.DataAccessMode, lDataAccess) && ...
                        strcmp(lReceiverPortNumber, lObjArray(kObj).Implementation.ReceiverPortNumber)
                    lDIFactory = lObjArray(kObj);
                    break;
                end
            end
        end
            
        
        function lDataInterface = findObjByPortAndDataElement(obj, lPortProp, lPort, lDataElement, lDataAccess)
            % FINDOBJBYPORTANDDATALEMENT
            lDataInterface = [];
            lObjArray = obj.CodeInfo.(lPortProp);
            if ~iscell(lDataAccess)
                lDataAccess = {lDataAccess};
            end
            for kObj = 1:numel(lObjArray)
                if ismember(lObjArray(kObj).Implementation.DataAccessMode, lDataAccess) && ...
                        strcmp(lPort, lObjArray(kObj).Implementation.Port) && ...
                        strcmp(lDataElement, lObjArray(kObj).Implementation.DataElement)
                    lDataInterface = lObjArray(kObj);
                    break;
                end
            end
        end
        
        function lDataInterface = findObjByReceiverPortNumber(obj, lPortProp, lReceiverPortNumber, lDataAccess)
            % FINDOBJBYRECEIVERPORTNUMBER
            lDataInterface = [];
            lObjArray = obj.CodeInfo.(lPortProp);
            if ~iscell(lDataAccess)
                lDataAccess = {lDataAccess};
            end
            for kObj = 1:numel(lObjArray)
                if ismember(lObjArray(kObj).Implementation.DataAccessMode, lDataAccess) && ...
                        strcmp(lReceiverPortNumber, lObjArray(kObj).Implementation.ReceiverPortNumber)
                    lDataInterface = lObjArray(kObj);
                    break;
                end
            end
        end
        
        
        function lDataInterface = findParameterByPortAndElement(obj, lPort, lElement)
            % FINDPARAMETERBYPORTANDELEMENT
            lDataInterface = [];
            lAllParams = obj.CodeInfo.Parameters;
            for kObj = 1:numel(lAllParams)
                if strcmp(lPort, lAllParams(kObj).Implementation.Port) && ...
                        strcmp(lElement, lAllParams(kObj).Implementation.ElementName)
                    lDataInterface = lAllParams(kObj);
                    break;
                end
            end
        end
        
        function lFunction = findFunctionByName(obj, lFunctionName)
            % FINDFUNCTIONBYNAME
            lFunction = [];
            lAllFunctions = [obj.CodeInfo.InitializeFunctions, obj.CodeInfo.OutputFunctions, obj.CodeInfo.TerminateFunctions];
            for kFcn = 1:numel(lAllFunctions)
                lName = lAllFunctions(kFcn).Prototype.Name;
                if strcmp(lName, lFunctionName)
                    lFunction = lAllFunctions(kFcn);
                    break;
                end
            end
        end
        
        function lNeedsExpInports = needsExpInports(obj)
            % NEEDSEXPINPORTS
            lNeedsExpInports = false;
            nTiming = numel(obj.CodeInfo.TimingProperties);
            for kTiming = 1:nTiming
                lHasAsyncTiming = strcmp(obj.CodeInfo.TimingProperties(kTiming).TimingMode, 'ASYNCHRONOUS');
                lHasPeriodicTiming = strcmp(obj.CodeInfo.TimingProperties(kTiming).TimingMode, 'PERIODIC');  % (Not opposite of above - could be UNION/RESET/TERMINATE)
                lHasExplicitTiming = strcmp(obj.CodeInfo.TimingProperties(kTiming).TaskingMode, 'EXPLICIT_TASKING');
                lHasExplicitPeriodicTiming = (lHasPeriodicTiming && lHasExplicitTiming);
                if lHasAsyncTiming || lHasExplicitPeriodicTiming
                    lNeedsExpInports = true;
                    break;
                end
            end
        end
    end
    
    methods (Access = private, Static)
        function lObjectArrayOut = findObjByImplementationClass(lObjectArray, lClass)
            % FINDOBJBYIMPLEMENTATIONCLASS
            lMatch = arrayfun(@(x)isa(x, lClass), [lObjectArray.Implementation]);
            lObjectArrayOut = lObjectArray(lMatch);
        end
    end
end
