classdef PIMBuilder < m3i.Visitor
    % PIMBUILDER

%   Copyright 2017-2018 The MathWorks, Inc.
    
    properties (Access = private)
        M3iModel;
        CodeInfo
        TypeBuilder;
        TimingInterfaceBuilder;
        XRelArtifactUtils
        
    end

    properties (SetAccess = private)
        NVRAMPIMInitCondMap;
    end
    
    methods (Access = public)
        function obj = PIMBuilder(m3iModel, lCodeInfo, lTypeBuilder, lTimingInterfaceBuilder, lXRelArtifactUtils)
            obj.M3iModel = m3iModel;
            obj.CodeInfo = lCodeInfo;
            obj.TypeBuilder = lTypeBuilder;
            obj.TimingInterfaceBuilder = lTimingInterfaceBuilder;
            obj.XRelArtifactUtils = lXRelArtifactUtils;
            obj.NVRAMPIMInitCondMap = containers.Map;
            
            % Register visitor for the AUTOSAR side
            obj.registerVisitor('mmVisit', 'mmVisit');
            obj.bind('Simulink.metamodel.arplatform.behavior.PerInstanceMemory', @mmWalkPerInstanceMemory, 'mmVisit');
            obj.bind('Simulink.metamodel.arplatform.interface.VariableData', @mmWalkArTypePerInstanceMemory, 'mmVisit');
        end
        
        function processPims(obj, m3iPim)
            % PROCESSPIM
            obj.applySeq('mmVisit', m3iPim);
        end
    end
    
    methods (Access = public)
        function ret = mmVisitM3IObject(~, m3iObj, varargin)
            % MMVISITM3IOBJECT
            %
            % Note: this cannot be a true static method, because M3I is
            % expecting a particular function signature.  Default visitor
            % must be public.
            %
            ret = [];
            assert(false, 'Failure in: %s\n Unsupported M3I object class %s', mfilename('-class'), class(m3iObj));
        end
    end
    
    methods (Access = private)
        
        function ret = mmWalkArTypePerInstanceMemory(obj, m3iPim)
            % MMWALKARTYPEPERINSTANCEMEMORY
            ret = [];
            lName = m3iPim.Name;
            m3iType = m3iPim.Type;
            lCodeInfoType = obj.TypeBuilder.processType(m3iType);
            
            % The arxml.importer does not distinguish between "global" data
            % store memories with a block and those without a block.
            %
            lSid = 'NULLVALUE:0';   % Since it creates a block for every DSM
            
            % We need to check if there is a service dependency (e.g., for NVRAM access)
            % and if so, whether there is a parameter we need to register.
            m3iComp = m3iPim.containerM3I.containerM3I;
            m3iBehavior = m3iComp.Behavior;
            lParamName = coder.internal.xrel.AUTOSAR.PIMNVRAMUtils.getParamNameFromServiceDepForArTypedPimByName(m3iBehavior, lName);
            if ~isempty(lParamName)
                % Add the parameter name to a list if not already on
                % it.
                obj.NVRAMPIMInitCondMap(lName) = lParamName;
            end
            
            lImplementation = RTW.Variable;
            lTiming = obj.TimingInterfaceBuilder.getInheritedTiming;
            lVariantInfo = RTW.VariantInfo.empty;
            lIsQueued = false;
            lQueueLength = 1;
            lImplementation.Type = lCodeInfoType;
            lImplementation.Identifier = lName;
            lImplementation.Owner = obj.CodeInfo.Name;
            lImplementation.StorageSpecifier = 'extern';
            lCodeInfoObj = RTW.ReadWriteDataInterface(lSid, ...
                lName, ...
                lImplementation, ...
                lTiming, ...
                lVariantInfo);
            lCodeInfoObj.IsQueued = lIsQueued;
            lCodeInfoObj.QueueLength = lQueueLength;
            lCodeInfoObj.Type = lCodeInfoType;
            obj.CodeInfo.DataStores(end+1) = lCodeInfoObj;
        end
        
        function ret = mmWalkPerInstanceMemory(obj, m3iPim)
            % MMWALKPERINSTANCEMEMORY
            ret = [];
            lName = m3iPim.Name;
            
            % We can't get the MATLAB type from the m3i Model in this case,
            % so we have to resort to the cross-release artifact
            lDataStoreInfo = obj.XRelArtifactUtils.getDataStoreByGraphicalName(lName);
            
            % Ignore any PIMs that are the Block States, Signal States or
            % Zero crossing States.  These won't be in the cross-release
            % artifact.
            if isempty(lDataStoreInfo)
                return;
            end
            
            % We need to reconstruct the data type - start with built-in
            % types.
            lDim = lDataStoreInfo.Dimensions;
            lCodeInfoType = obj.TypeBuilder.getTypeFromArtifact(lDataStoreInfo.TypeName, lDim);
            
            
            % Start with built-in types only
            lSid = 'NULLVALUE:0';   % Since it creates a block for every DSM
            lImplementation = RTW.Variable;
            lTiming = obj.TimingInterfaceBuilder.getInheritedTiming;
            lVariantInfo = RTW.VariantInfo.empty;
            lIsQueued = false;
            lQueueLength = 1;
            lImplementation.Type = lCodeInfoType;
            lImplementation.Identifier = lName;
            lImplementation.Owner = obj.CodeInfo.Name;
            lImplementation.StorageSpecifier = 'extern';
            lCodeInfoObj = RTW.ReadWriteDataInterface(lSid, ...
                lName, ...
                lImplementation, ...
                lTiming, ...
                lVariantInfo);
            lCodeInfoObj.IsQueued = lIsQueued;
            lCodeInfoObj.QueueLength = lQueueLength;
            lCodeInfoObj.Type = lCodeInfoType;
            obj.CodeInfo.DataStores(end+1) = lCodeInfoObj;
            
        end
    end
    
end
