classdef TimingInterfaceBuilder < m3i.Visitor
    % TIMINGINTERFACEBUILDER
    % Copyright 2017-2018 The MathWorks, Inc.
    properties (Access = private)
        CodeInfo;
        ExtraMappingInfo;
        OperationBuilder;
        TimingForRunnable;
    end
    
    properties (SetAccess = private)
        TimingForInitialize;
    end
    
    methods (Access = public)
        function obj = TimingInterfaceBuilder(lCodeInfo, lExtraMappingInfo, lOperationBuilder)
            % TIMINGINTERFACEBUILDER
            obj.CodeInfo = lCodeInfo;
            obj.ExtraMappingInfo = lExtraMappingInfo;
            obj.OperationBuilder = lOperationBuilder;
            obj.TimingForRunnable = containers.Map;
            
            % Register visitor for the AUTOSAR side
            obj.registerVisitor('mmVisit', 'mmVisit');
            obj.bind('Simulink.metamodel.arplatform.behavior.OperationInvokedEvent', @mmWalkOperationInvokedEvent, 'mmVisit');
            obj.bind('Simulink.metamodel.arplatform.behavior.TimingEvent', @mmWalkTimingEvent, 'mmVisit');
            obj.bind('Simulink.metamodel.arplatform.behavior.DataReceivedEvent', @mmWalkDataReceivedEvent, 'mmVisit');
            obj.bind('Simulink.metamodel.arplatform.behavior.DataReceiveErrorEvent', @mmWalkDataReceiveErrorEvent, 'mmVisit');
            obj.bind('Simulink.metamodel.arplatform.behavior.ExternalTriggerOccurredEvent', @mmWalkExternalTriggerOccurredEvent, 'mmVisit');
            obj.bind('Simulink.metamodel.arplatform.behavior.ModeSwitchEvent', @mmWalkModeSwitchEvent, 'mmVisit');
        end
        
        function createTimingInterfaceForInitRunnable(obj, m3iRunnable, lCodeInfoFcn)
            % CREATETIMINGINTERFACEFORINITRUNNABLE
            
            % Add the oneshot interface
            lTimingInterface = RTW.TimingInterface;
            lTimingInterface.TimingMode = 'ONESHOT';
            obj.TimingForRunnable(m3iRunnable.symbol) = lTimingInterface;
            obj.CodeInfo.TimingProperties(end+1) = lTimingInterface;
            lCodeInfoFcn.Timing = lTimingInterface;
            
            % Add init event
            lTimingInterface2 = RTW.TimingInterface;
            lTimingInterface2.TimingMode = 'INITIALIZE';
            lTimingInterface2.NonFcnCallPartitionName = '';
            lTimingInterface2.SamplePeriod = inf;
            lTimingInterface2.SampleOffset = 1;
            lTimingInterface2.Priority = -32768;
            obj.CodeInfo.TimingProperties(end+1) = lTimingInterface2;
            obj.TimingForInitialize = lTimingInterface2;
        end
        
        function createTimingInterfaceForTerminateRunnable(obj, m3iRunnable, lCodeInfoFcn)
            % CREATETIMINGINTERFACEFORTERMINATERUNNABLE
            lTimingInterface = RTW.TimingInterface;
            lTimingInterface.TimingMode = 'TERMINATE';
            lTimingInterface.NonFcnCallPartitionName = '';
            lTimingInterface.SamplePeriod = Inf;
            lTimingInterface.SampleOffset = 2;
            lTimingInterface.Priority = -32768;
            lTimingInterface.TaskingMode = 'IMPLICIT_TASKING';
            obj.CodeInfo.TimingProperties(end+1) = lTimingInterface;
            lCodeInfoFcn.Timing = lTimingInterface;
            obj.TimingForRunnable(m3iRunnable.symbol) = lTimingInterface;
        end
        
        function createTimingInterfaceForOutputRunnable(obj, m3iRunnable, lCodeInfoFcn)
            % CREATETIMINGINTERFACEFOROUTPUTRUNNABLE
            lSymbol = m3iRunnable.symbol;
            lAutosarResetRunnableSymbols    = obj.ExtraMappingInfo.getAutosarResetRunnableSymbols;
            if ~isempty(lAutosarResetRunnableSymbols) && ismember(lSymbol, lAutosarResetRunnableSymbols)
                if obj.TimingForRunnable.isKey(lSymbol)
                    lTimingInterface =  obj.TimingForRunnable(lSymbol);
                else
                    lTimingInterface = obj.createResetTimingInterface(lSymbol);
                end
            elseif m3iRunnable.Events.size == 1
                m3iEvent = m3iRunnable.Events.at(1);
                if obj.TimingForRunnable.isKey(lSymbol)
                    lTimingInterface =  obj.TimingForRunnable(lSymbol);
                else
                    lTimingInterface = obj.apply('mmVisit', m3iEvent, lSymbol, lCodeInfoFcn);
                end        
            else
                % If we have more than one Event, choose the timing event.
                for kEvent = 1:m3iRunnable.Events.size
                    m3iEvent = m3iRunnable.Events.at(kEvent);
                    if isa(m3iEvent, 'Simulink.metamodel.arplatform.behavior.TimingEvent')
                        if obj.TimingForRunnable.isKey(lSymbol)
                            lTimingInterface =  obj.TimingForRunnable(lSymbol);
                        else
                            lTimingInterface = obj.apply('mmVisit', m3iEvent, lSymbol, lCodeInfoFcn);
                        end
                        break;
                    end
                    assert(false, 'More than one event but no TimingEvent');
                end
                
            end
            lCodeInfoFcn.Timing = lTimingInterface;
        end
        
        function ret = getInheritedTiming(obj)
            % GETINHERITEDTIMING
            for kElem = 1:numel(obj.CodeInfo.TimingProperties)
                if strcmp(obj.CodeInfo.TimingProperties(kElem).TimingMode, 'INHERITED')
                    ret = obj.CodeInfo.TimingProperties(kElem);
                    return;
                end
            end
            
            % If we didn't already find one, create it
            ret = RTW.TimingInterface;
            ret.TaskingMode = 'IMPLICIT_TASKING';
            ret.TimingMode = 'INHERITED';
            ret.NonFcnCallPartitionName = '';
            ret.SamplePeriod = 0;
            ret.SampleOffset = 0;
            ret.Priority = 0;
            obj.CodeInfo.TimingProperties(end+1) = ret;
        end
        
    end
    
    methods (Access = public)
        function ret = mmVisitM3IObject(~, m3iObj, varargin)
            % MMVISITM3IOBJECT
            %
            % Note: this cannot be a true static method, because M3I is
            % expecting a particular function signature.  Default visitor
            % must also be public.
            %
            ret = [];
            assert(false, 'Failure in: %s\n Unsupported M3I object class %s', mfilename('-class'), class(m3iObj));
        end
        
        function ret = isTimingEqual(obj, lTiming1, lTiming2)
            % ISTIMINGEQUAL
            lLocalEqual = isequal(lTiming1.TimingMode, lTiming2.TimingMode) && ...
                isequal(lTiming1.NonFcnCallPartitionName, lTiming2.NonFcnCallPartitionName) && ...
                isequal(lTiming1.SamplePeriod, lTiming2.SamplePeriod) && ...
                isequal(lTiming1.SampleOffset, lTiming2.SampleOffset) && ...
                isequal(lTiming1.Priority, lTiming2.Priority) && ...
                isequal(lTiming1.TaskingMode, lTiming2.TaskingMode) && ...
                numel(lTiming1.UnionTimingInfo) == numel(lTiming2.UnionTimingInfo);
            
            ret = lLocalEqual;
            if lLocalEqual && strcmp(lTiming1.TimingMode, 'UNION')
                % Only traverse union timing info if everything else is
                % already in place
                for kTiming = 1:numel(lTiming1.UnionTimingInfo)
                    lSubEqual = obj.isTimingEqual(lTiming1.UnionTimingInfo(kTiming), ...
                        lTiming2.UnionTimingInfo(kTiming));
                    ret = ret && lSubEqual;
                end
            end
        end
        
        function ret = isUnionMember(obj, lTiming1, lTiming2)
            % ISUNIONMEMBER
            %   Is lTiming1 a member of lTiming2.UnionTimingInfo
            ret = false;
            for kTiming = 1:numel(lTiming2.UnionTimingInfo)
                lSubEqual = obj.isTimingEqual(lTiming1, ...
                    lTiming2.UnionTimingInfo(kTiming));
                if lSubEqual
                    ret = true;
                    return;
                end
            end
        end
        
        function lTimingInterface = createUnion(obj, lTiming1, lTiming2)
            % CREATEUNION
            lTimingInterface = RTW.TimingInterface;
            lTimingInterface.TimingMode = 'UNION';
            lTimingInterface.TaskingMode = 'EXPLICIT_TASKING';
            lTimingInterface.UnionTimingInfo = [lTiming1, lTiming2];
            obj.CodeInfo.TimingProperties(end+1) = lTimingInterface;
        end
    end
    
    methods (Access = private)
        %
        % M3I Walk methods
        %
        
        function lTimingInterface = mmWalkOperationInvokedEvent(obj, m3iEvent, lSymbol, lCodeInfoFcn)
            % MMWALKOPERATIONINVOKEDEVENT
            lTimingInterface = RTW.TimingInterface;
            lTimingInterface.TimingMode = 'ASYNCHRONOUS';
            lTimingInterface.NonFcnCallPartitionName = '';
            lTimingInterface.SamplePeriod = -1;
            lTimingInterface.SampleOffset = -2;
            lTimingInterface.TaskingMode = 'EXPLICIT_TASKING';
            obj.injectInfoFromXRelArtifact(lTimingInterface, lSymbol);
            obj.CodeInfo.TimingProperties(end+1) = lTimingInterface;
            obj.TimingForRunnable(lSymbol) = lTimingInterface;
            obj.OperationBuilder.processOperation(m3iEvent.instanceRef.Operations, lCodeInfoFcn);
        end
        
        function lTimingInterface = mmWalkTimingEvent(obj, m3iEvent, lSymbol, lCodeInfoFcn) %#ok<INUSD>  % Need consistent API for mmWalk
            % MMWALKTIMINGEVENT
            lPeriod = m3iEvent.Period;
            lTimingInterface = RTW.TimingInterface;
            lTimingInterface.TimingMode = 'PERIODIC';
            lTimingInterface.SamplePeriod = lPeriod;
            
            % Is this implicit or explicit?
            if obj.ExtraMappingInfo.isExplicitTaskingForRunnable(lSymbol)
                lTimingInterface.TaskingMode = 'EXPLICIT_TASKING';
            else
                lTimingInterface.TaskingMode = 'IMPLICIT_TASKING';
            end
            obj.injectInfoFromXRelArtifact(lTimingInterface, lSymbol);            
            obj.CodeInfo.TimingProperties(end+1) = lTimingInterface;
            obj.TimingForRunnable(lSymbol) = lTimingInterface;
        end
        
        function lTimingInterface = mmWalkModeSwitchEvent(obj, m3iEvent, lSymbol, lCodeInfoFcn) %#ok<INUSD,INUSL>   % Need common API for M3i walking
            % MMWALKMODESWITCHEVENT
            lTimingInterface = obj.getAsyncTimingEvent(lSymbol);
        end
        
        function lTimingInterface = mmWalkExternalTriggerOccurredEvent(obj, m3iEvent, lSymbol, lCodeInfoFcn) %#ok<INUSD,INUSL> % Need common API for M3i walking
            % MMWALKEXTERNALTRIGGEROCCURREDEVENT
            lTimingInterface = obj.getAsyncTimingEvent(lSymbol);
        end
        
        function lTimingInterface = mmWalkDataReceivedEvent(obj, m3iEvent, lSymbol, lCodeInfoFcn) %#ok<INUSD,INUSL> % Need common API for M3i walking
            % MMWALKDATARECEIVEDEVENT
            lTimingInterface = obj.getAsyncTimingEvent(lSymbol);
        end
        
        function lTimingInterface = mmWalkDataReceiveErrorEvent(obj, m3iEvent, lSymbol, lCodeInfoFcn) %#ok<INUSD,INUSL> % Need common API for M3i walking
            % MMWALKDATARECEIVEDERROREVENT
            lTimingInterface = obj.getAsyncTimingEvent(lSymbol);
        end        
    end
    
    
    methods (Access = private)
        %
        % Remaining private methods
        %
        
        function lTimingInterface = createResetTimingInterface(obj, lSymbol)
            % CREATERESETTIMINGINTERFACE
            lTimingInterface = RTW.TimingInterface;
            lTimingInterface.TimingMode = 'RESET';
            lTimingInterface.NonFcnCallPartitionName = '';
            lTimingInterface.SamplePeriod = Inf;
            lTimingInterface.SampleOffset = obj.ExtraMappingInfo.getSampleOffsetForRunnable(lSymbol);
            lTimingInterface.Priority = -32768;
            lTimingInterface.TaskingMode = 'EXPLICIT_TASKING';
            obj.CodeInfo.TimingProperties(end+1) = lTimingInterface;
            obj.TimingForRunnable(lSymbol) = lTimingInterface;
        end
        
        function injectInfoFromXRelArtifact(obj, lTimingInterface, lSymbol)
            % INJECTINFOFROMXRELARTIFACT
            % Set sample offset from cross-release artifact
            lTimingInterface.SampleOffset = obj.ExtraMappingInfo.getSampleOffsetForRunnable(lSymbol);
            
            % Need to get priority from the cross-release artifact
            lTimingInterface.Priority = obj.ExtraMappingInfo.getPriorityForRunnable(lSymbol);
        end
            
        function lTimingInterface = getAsyncTimingEvent(obj, lSymbol)
            % GETASYNCTIMINGEVENT
            lTimingInterface = RTW.TimingInterface;
            lTimingInterface.TimingMode = 'ASYNCHRONOUS';
            lTimingInterface.NonFcnCallPartitionName = '';
            lTimingInterface.SamplePeriod = -1;            
            lTimingInterface.TaskingMode = 'EXPLICIT_TASKING';
            obj.injectInfoFromXRelArtifact(lTimingInterface, lSymbol);
            obj.CodeInfo.TimingProperties(end+1) = lTimingInterface;
            obj.TimingForRunnable(lSymbol) = lTimingInterface;
        end
    end
end
