classdef AutosarConfigSetManager
    % AUTOSARCONFIGSETMANAGER
    %
    %   Caches config set parameters from temporary model created by
    %   arxml.importer to re-apply after the cross-release import config
    %   set manager has applied the donor config set.
    %
    %   Patches/postprocesses rather than creating a new config set
    %   manager, and does not need to carry out validation against third
    %   data source, so more straightforward to NOT inherit from 
    %   coder.internal.utils.SLConfigSetManager.
    %
    %   Copyright 2018 The MathWorks, Inc.
    %
    properties (Constant, Access = private)
        ParametersToCache (1,:) cell = { ...
                'AutosarSchemaVersion', ...         % Preserve the AUTOSAR Schema, or feature support in the temporary model may not match the features required by the original code.
                'AutosarMaxShortNameLength', ...    % Preserve the max AUTOSAR identifier length
                'EnableMultiTasking', ...           % Preserve multi-tasking (or multi-rate models will fail to build)
                'CodeInterfacePackaging', ...
                'IncludeMdlTerminateFcn', ...       % Preserve presence/absence of terminate function
                };
        ParametersToAlwaysSet (1,:) cell = { ...
            'SaveFormat', 'Array'};
    end
    
    properties (Access = private)
        ParameterNames (1,:) cell = {}
        ParameterValues (1,:) cell = {}
    end
    
    methods (Access = public)
        function obj = AutosarConfigSetManager(hModel, lParamKVPToSet)
            %   AUTOSARCONFIGSETMANAGER
            %       hModel         -  Handle or name of model, or handle to
            %                         configSet
            %       lParamKVPToSet -  Cell array of key-value pairs to set
            %                         fixed values in addition to those
            %                         hard-coded.
            validateattributes(lParamKVPToSet, {'cell'}, {});
            lParamKVPToSet = lParamKVPToSet(:)';    % Force row vec;
            
            assert(mod(numel(lParamKVPToSet), 2) == 0, 'lParamKVPToSet must contain a sequence of key-value pairs');
            
            obj.ParameterNames = [obj.ParametersToCache, obj.ParametersToAlwaysSet(1:2:end), lParamKVPToSet(1:2:end)];
            obj.ParameterValues = [ ...
                cellfun(@(x)get_param(hModel, x), obj.ParametersToCache, 'UniformOutput', false), ...
                obj.ParametersToAlwaysSet(2:2:end), ...
                lParamKVPToSet(2:2:end)];
        end
        
        
        function restoreValues(obj, hModel)
            % RESTOREVALUES
            for kParam = 1:numel(obj.ParameterNames)
                set_param(hModel, obj.ParameterNames{kParam}, obj.ParameterValues{kParam});
            end
        end
    end 
end