classdef AutosarExternalCodeImportHook < coder.internal.xrel.CodeImportHook
    % AUTOSAREXTERNALCODEIMPORTHOOK
    %
    %   Copyright 2017-2018 The MathWorks, Inc.
    
    methods (Access=public, Static)
        function entryHook(lModelName)
            import coder.internal.xrel.AUTOSAR.AutosarExternalCodeImportHook;
            if ~AutosarExternalCodeImportHook.isAutosarCodeImport(lModelName)
                return;
            end
            
            % Make sure block reduction is OFF
            set_param(lModelName, 'BlockReduction', 'off');
            
            % Make sure block IO reuse is OFF
            set_param(lModelName, 'OptimizeBlockIOStorage', 'off');
            
            % Make sure that model-to-code traceability is OFF
            lParamsToSwitchOff = { ...
                'IncludeHyperlinkInReport', ...
                'GenerateTraceInfo', ...
                'GenerateTraceReport', ...
                'GenerateTraceReportSl', ...
                'GenerateTraceReportSf', ...
                'GenerateTraceReportEml', ...
                'GenerateCodeReplacementReport'};
            
            for kParam = 1:numel(lParamsToSwitchOff)
                set_param(lModelName, lParamsToSwitchOff{kParam}, 'off');
            end
            
            set_param(lModelName, 'Dirty', 'off');
        end
        
        
        function preTlcHook(lModelName)
            % PRETLCHOOK
            import coder.internal.xrel.AUTOSAR.AutosarExternalCodeImportHook;
            if ~AutosarExternalCodeImportHook.isAutosarCodeImport(lModelName)
                return;
            end
            lRTWGenSettings = get_param(lModelName, 'RTWGenSettings');
            lModelReferenceBuild = isfield(lRTWGenSettings, 'mdlRefRelativeSimDir');
            if lModelReferenceBuild
                DAStudio.error('PIL:xrelimport:TopModelBuildOnly', lModelName);
            end
        end
        
        function postTlcHook(lModelName, lBuildInfo)
            % POSTTLCHOOK
            import coder.internal.xrel.AUTOSAR.AutosarExternalCodeImportHook;
            import coder.internal.xrel.AUTOSAR.ModelBlockSourcesHelper;
            import coder.internal.xrel.AUTOSAR.BuildInfoBuilder;
            
            [lIsAutosarCodeImport, lControlBlockHandle] = AutosarExternalCodeImportHook.isAutosarCodeImport(lModelName);
            if ~lIsAutosarCodeImport
                return;
            end
            
            % ** Info from control block ** 
            %
            % Get the cross-release information attached to UserData on the
            % control block in the temporary model.
            lUserData = get_param(lControlBlockHandle, 'UserData');
            
            % ** Top model buildInfo **
            %
            % Expunge any generated code from buildInfo
            BuildInfoBuilder.removeBuildDirAndModelSources(lBuildInfo);
            
            % Get the list of source and include files that need to be
            % added to the current BuildInfo
            bReplaceStartDir = true;
            lModelBlockSources = ModelBlockSourcesHelper.createFromBlock(lControlBlockHandle, bReplaceStartDir);
            
            % Add the imported source and include paths to buildInfo for
            % the top imported model.
            lDoCopy = true;
            lBuildFolderInfo = RTW.getbuildDir(lModelName);
            lBuildFolder = lBuildFolderInfo.BuildDirectory;
            BuildInfoBuilder.addFilesToBuild(lModelBlockSources, lBuildFolder, lBuildInfo, lModelName, lDoCopy);
            
            % ** Model Refs ** 
            %
            % Construct the build folders for the imported model refs
            % and update the top model build artifacts so that they are
            % correctly referenced during build or protection.
            lModelRefs = lUserData.ModelRefs;
            lModelRefRoot = lBuildFolderInfo.ModelRefRelativeRootTgtDir;
            lModelRefSimRoot = lBuildFolderInfo.ModelRefRelativeRootSimDir;
            lConfigSet = getActiveConfigSet(lModelName);
            lModelRefSubModelManager = coder.internal.xrel.AUTOSAR.ModelRefSubModelManager( ...
                lModelRefs, ...
                lModelRefRoot, ...
                lModelRefSimRoot, ...
                lBuildFolderInfo.CodeGenFolder, ...
                lBuildFolderInfo.SharedUtilsTgtDir, ...
                lBuildFolderInfo.ModelRefRelativeBuildDir, ...
                lDoCopy, ...
                lConfigSet);
            lModelRefSubModelManager.processModelRefs(lModelName, lBuildInfo);
            
            % ** CodeInfo **
            %
            % Ignore the CodeInfo from code generation, and build our own
            lCodeInfo = load('codeInfo');
            lCodeInfo.codeInfo = RTW.ComponentInterface;            
            m3iModel = autosar.api.Utils.m3iModel(lModelName);
            m3iComp = autosar.api.Utils.m3iMappedComponent(lModelName);
            lXrelInfo = lUserData.XrelInfo;
                        
            % Needed to write SIL header
            lCodeInfo.codeInfo.GraphicalPath = lModelName;
            lCodeInfo.codeInfo.HeaderFile = lXrelInfo.ModelHeaderFile;     % .h automatically added downstream
            lCodeInfo.codeInfo.SourceFile = lXrelInfo.ModelSourceFile;     % .c automatically added downstream
            
            % Need model name to be added BEFORE updating codeinfo.
            lCodeInfoWriter = coder.internal.xrel.AUTOSAR.mm2CodeInfo.CodeInfoBuilder(lCodeInfo.codeInfo, m3iModel, m3iComp, lXrelInfo);
            lCodeInfoWriter.updateCodeInfo;
            lCodeInfo.expInports = lCodeInfoWriter.getExpInports;
            
            save('codeInfo.mat', '-struct', 'lCodeInfo');
            
            % Patch the m3i model if there are any C-typed PIMs - this can
            % only be done after we've written CodeInfo
            nPIMs = m3iComp.Behavior.PIM.size;
            if nPIMs > 0
                lAutosarSchemaVersion = get_param(lModelName, 'AutosarSchemaVersion');
                lMaxShortNameLength = get_param(lModelName, 'AutosarMaxShortNameLength');
                m3iBehavior = m3iComp.Behavior;
                tran = M3I.Transaction(m3iModel);
                lCPimUtils = coder.internal.xrel.AUTOSAR.CTypePimUtils(lModelName, ...
                    lAutosarSchemaVersion, ...
                    lMaxShortNameLength, ...
                    m3iModel, m3iBehavior);
                for kPIM = 1:nPIMs
                    m3iPIM = m3iComp.Behavior.PIM.at(kPIM);
                    lCPimUtils.processPIM(m3iPIM, lCodeInfo.codeInfo);
                end
                tran.commit();
            end
            
            % Patch the RTE PIM Source File that was generated in TLC
            coder.internal.xrel.AUTOSAR.PIMNVRAMUtils.writePIMFileSource(lModelName, ...
                lBuildFolder, ...
                lCodeInfoWriter.getNVRAMPIMInitCondMap, ...
                m3iComp);
            
        end
                
        function accelModeSimulationCallback(lModelName)
            % Callback to indicate that accelerator mode is not permitted
            import coder.internal.xrel.AUTOSAR.AutosarExternalCodeImportHook;
            h = coder.internal.ModelCodegenMgr.getInstance(lModelName);
            if ~isempty(h)
                assert(isa(h, 'coder.internal.ModelCodegenMgr'));
                isAccelOrRaccel = ...
                    any(ismember(h.MdlRefSimModeForXRelBlock, {'Accelerator', 'accelerator', 'rapid-accelerator'}));
                if isAccelOrRaccel
                    DAStudio.error('PIL:xrelimport:SILPILOnly', lModelName);
                end
            end
        end
        
        function normalModeSimulationCallback(lModelName)
            % Called by StartFcn hook.
            DAStudio.error('PIL:xrelimport:SILPILOnly', lModelName);
        end
        
        
    end
end
