classdef AutosarImportModelManager < handle
    % AUTOSARIMPORTMODELMANAGER
    %
    %   Utility class for cross-release import of AUTOSAR models
    %
    %   Copyright 2017-2018 The MathWorks, Inc.
    properties (Constant, Access = private)
        ReadAccessKinds = {'ImplicitRead', 'ExplicitReadByArg', 'ExplicitReadByValue'};
    end
    
    methods (Access = public)
        function obj = AutosarImportModelManager()
            % Empty constructor
        end
        
        function blockHandles = doImport(obj, lXrelModelName, lXrelInfo, lOriginalFolder, lDataDictionary, lXILMode, lXrelModelManager, ...
                lSharedCodeAreaValidator, lXrelLocation, lAnchorDir, lReplacePaths)
            % DOIMPORT
            %   Main logic for cross-release import of AUTOSAR code.
            %
            
            % Read the cross release artifacts for the model references
            lModelReferences = ...
                coder.internal.xrel.ModelReferenceArtifactReader( ...
                lSharedCodeAreaValidator, ...
                fileparts(lXrelLocation), ...   % Parent of (canonical) directory containing top level import artifact
                lAnchorDir, ...
                lReplacePaths, ...
                lDataDictionary, ...
                lXrelInfo.TgtFcnLib, ...
                lXrelInfo.ModelRefArtifactDirs);                     % Cell array of names of submodel artifact directories
            
                        
            % Get an instance of the XRelArtifactUtils class (which
            % provides search methods)
            lUtils = coder.internal.xrel.AUTOSAR.XRelArtifactUtils(lXrelInfo);
            
            % Use the arxml.importer to generate a stub model
            lARXMLImporterObj = obj.getARXMLImporterFromArtifact(lXrelInfo);
            lComponentName = obj.getComponentNameForModel(lARXMLImporterObj, lXrelInfo.ComponentName);
            m3iModel = obj.getM3iModelFromArxmlImporter(lARXMLImporterObj);
            m3iComp = autosar.mm.Model.findChildByName(m3iModel, lComponentName);
            
            % Update the metamodel with parameters for NVRAM init
            % conditions where these do not already exist.
            obj.createParamsForNVRAMInitConditions(m3iModel, m3iComp);
            
            % Check whether there are any Error Status ports that are not
            % supported by createComponentAsModel, and patch the arModel to
            % ensure that they do get created.
            obj.validateAndUpdateErrorStatus(m3iModel, m3iComp, lUtils);
            
            % Need to map back from the Initialization, Reset and Terminate
            % runnable symbols to their respective runnable names.
            lInitializationRunnable = obj.getRunnableNameFromSymbol(m3iComp, lUtils.getAutosarInitRunnableSymbol);
            assert(~isempty(lInitializationRunnable), 'Could not find initialization runnable');
            
            lResetRunnables = cellfun( ...
                @(lSymbol)obj.getRunnableNameFromSymbol(m3iComp, lSymbol), ...
                lUtils.getAutosarResetRunnableSymbols, ...
                'UniformOutput', false);
            
            lTerminateRunnable = obj.getRunnableNameFromSymbol(m3iComp, lUtils.getAutosarTerminateRunnableSymbol);
            
            % Do we need explicit timing?
            lForceFunctionCall = obj.needsExplicitTiming(m3iComp, lUtils);
            
            % Now we can import the model
            hModel = obj.doModelImport(lARXMLImporterObj, ...
                'DataDictionary', lDataDictionary, ...
                'ForceFunctionCallSubsystem', lForceFunctionCall, ...
                'InitializationRunnable', lInitializationRunnable, ...
                'ResetRunnables', lResetRunnables, ...
                'TerminateRunnable', lTerminateRunnable);
            
            % Add a cleanup object to avoid model drool
            cTempModel = onCleanup(@()coder.internal.xrel.AUTOSAR.AutosarImportModelManager.safeCloseModel(hModel));
            
            % Configure the configset of the stub model
            %
            % If there is a model initialize function, then we should also
            % enable internal data zero initialization, since this is
            % required upstream.
            if isempty(lInitializationRunnable)
                lParametersToSet = {};
            else
                lParametersToSet = {'ZeroInternalMemoryAtStartup', 'on'};
            end
            lParametersToSet = [lParametersToSet, ...
                {'ERTHeaderFileRootName', [lXrelInfo.ModelHeaderFile, '$E']}];
            
            % Cache the parameters from the config set of the stub model
            % created by arxml.importer.createComponentAsModel
            lCachedParameterManager = ...
                coder.internal.xrel.AUTOSAR.AutosarConfigSetManager(hModel, ...
                lParametersToSet);
            
            % Apply the cross-release config set (from the user's harness
            % model)
            lXrelModelManager.setupConfigset(get(hModel, 'Name'));
            
            % Re-set the config set parameters from the stub model that
            % must be preserved.
            lCachedParameterManager.restoreValues(hModel);
            
            % Get the sources from the artifact
            lSourcesHelper = ...
                coder.internal.xrel.AUTOSAR.ModelBlockSourcesHelper.createFromArtifact(...
                lXrelInfo);
            
            % Re-order the ports
            obj.doPortReordering(hModel, lUtils, 'Inport');
            obj.doPortReordering(hModel, lUtils, 'Outport');
            
            % Re-name AFTER re-ordering
            obj.doPortRenaming(hModel, lUtils, 'Inport');
            obj.doPortRenaming(hModel, lUtils, 'Outport');
            
            % Re-name function callers
            obj.doFunctionRenaming(hModel, lUtils.getSimulinkFunctions, lDataDictionary);
            
            % Update dimensions on Simulink functions
            obj.updateSimulinkFunctionArgumentDimensions(hModel, lUtils);
            
            % Update task priorities
            obj.updateAsyncTaskPriorities(hModel, lUtils);
                        
            % Configure the stub model with the sources
            hBlock = obj.configureModel(hModel, ...
                lComponentName, ...
                lSourcesHelper, ...
                'ProxyModelName', lXrelModelName);
                        
            % Make sure data access mode is correct
            obj.doSRInterfaceMatching(hModel, lUtils, 'Outport');
            obj.doSRInterfaceMatching(hModel, lUtils, 'Inport');
            
            % Make sure that any "periodic" runnables that originally had
            % triggered function call trigger ports match the original
            % model
            lTriggeredPeriodicRunnableSymbols = lUtils.getTriggeredPeriodicRunnableSymbols;
            
            
            lFcnCallMap = obj.getFcnCallMap(hModel);
            for kRunnable = 1:numel(lTriggeredPeriodicRunnableSymbols)
                lRunnableName = obj.getRunnableNameFromSymbol(m3iComp, lTriggeredPeriodicRunnableSymbols{kRunnable});
                
                % We need to find the function call port that is mapped to
                % this runnable.
                lBlock = lFcnCallMap(lRunnableName);
                lDstPortHs = autosar.mm.mm2sl.SLModelBuilder.getAllDestinationPortsThroughVirtualBlocks(lBlock);
                lRunnableSubsys = get(lDstPortHs{1}, 'Parent');
                lTriggerPortHandle =find_system(get_param(lRunnableSubsys, 'handle'), 'BlockType', 'TriggerPort');                
                assert(strcmp(get(lTriggerPortHandle, 'TriggerType'), 'function-call'), 'Expected function call trigger');
                set(lTriggerPortHandle, 'SampleTimeType', 'triggered');
            end
            
            % Replace all terminators in the auto-generated ARXML model
            % with a library block containing a terminator and a signal
            % conversion block.
            %
            % This ensures that the inputs are marked as
            % DirectFeedThrough in binfo.
            obj.replaceTerminators(hModel);
            
            % Remove enums defined in base workspace that
            % conflict with session enums.
            obj.cleanUpEnums(hModel);
            
            % Remove any stateflow diagrams for queued inputs that do not
            % need a message typed port.
            obj.updateQueuedDataPorts(hModel, lUtils);
            
            % Update bus types with dimensions from artifact (unflatten)
            obj.updateStructTypes(lDataDictionary, lUtils);
            
            % C-type PIMs don't get blocks added by createComponentAsModel,
            % so we need to do this ourselves.  Need to do this after
            % updating struct types, or else we might get a mismatch.
            obj.addCTypePIMBlocks(hModel, m3iComp, lUtils, lDataDictionary);
            
            % Unflatten values in parameter objects (including in
            % bus elements) and LookupTable/Breakpoint objects
            lParameters = lUtils.getParameters;
            for kParam = 1:numel(lParameters)
                obj.updateParamDimensions(lUtils, ...
                    lDataDictionary, ...
                    lParameters(kParam));
            end
            
            % Unflatten dimensions on Ar Typed PIMs where possible.
            obj.updateArTypedPIMDimensions(hModel, m3iComp, lDataDictionary, lUtils);
            
            % Temporarily attach some stuff to UserData
            tempStruct.XrelInfo = lXrelInfo;
            tempStruct.ModelRefs = lModelReferences;
            
            set_param(hBlock, 'UserData', tempStruct);
            set_param(hBlock, 'UserDataPersistent', 'on');
            
            % Open some temporary models if we have model refs (protected
            % models need a model with a given name to exist, even when it 
            % is not actually used)
            cModelRef = ...
                coder.internal.xrel.AUTOSAR.ModelRefSubModelManager.createEmptyModelRefModels(lModelReferences);
            
            % Cache the name of the model, save it and close it - we need
            % it to be closed whilst we relocate the data dictionary -
            % which needs to be done before we package the model in the
            % protected model container.
            lModelName = get_param(hModel, 'Name');
            
            % Do a diagram update before saving, or the period for reset
            % function systems will not be serialized in the slx and will
            % throw a warning.
            save_system(hModel);
            close_system(hModel);
            
            if ~isempty(lDataDictionary) && isfile(fullfile(pwd, lDataDictionary))
                % If a data dictionary was created by createComponentAsModel we
                % need to move it into the parent folder.
                %
                % Open it
                lDDObj = Simulink.data.dictionary.open(fullfile(pwd, lDataDictionary));
                
                % Save it
                lDDObj.saveChanges;
                
                % Close it
                lDDObj.close;
                
                % If the named SLDD existed in its final location, it would
                % have been re-used rather than creating a new one in pwd.
                % But assert anyway, as a safeguard.
                lDestDDLocation = fullfile(pwd, '..', '..', '..', lDataDictionary);
                assert(~isfile(lDestDDLocation), sprintf('%s should not exist', lDestDDLocation));
                
                % Move it
                movefile(fullfile(pwd, lDataDictionary), lDestDDLocation);
            end
            
            % Create the model block
            if coder.internal.xrel.feature('ProtectedAUTOSAR')
                blockHandles = obj.doProtect( ...
                    lModelName, lOriginalFolder, lXILMode);
            else
                blockHandles = obj.createModelBlock( ...
                    lModelName, lOriginalFolder, lXILMode);
            end
            
            % Cleanup
            cModelRef.delete;
            cTempModel.delete;
        end
    end
    
    methods (Access = private, Static)
        function [hModel, bSuccess] = doModelImport(lARXMLImporterObj, varargin)
            % DOMODELIMPORT
            %   Import the required component using the arxml.import object
            %   passed in as lARXMLImporterObj.
            %
            %   Create a cleanup object so that the model is closed when
            %   the AutosarImportModelManager goes out of scope.
            %
            p = inputParser;
            p.addParameter('InitializationRunnable', '', @ischar);
            p.addParameter('ResetRunnables', '', @iscellstr);
            p.addParameter('TerminateRunnable', '', @ischar);
            p.addParameter('ComponentName', '', @ischar);
            p.addParameter('DataDictionary', '', @ischar);
            p.addParameter('ForceFunctionCallSubsystem', false, @islogical);
            p.parse(varargin{:});
            
            lInitializationRunnable = p.Results.InitializationRunnable;
            lResetRunnables = p.Results.ResetRunnables;
            lTerminateRunnable = p.Results.TerminateRunnable;
            
            lComponentName = p.Results.ComponentName;
            lDataDictionary = p.Results.DataDictionary;
            lForceFunctionCallSubsystem = p.Results.ForceFunctionCallSubsystem;
            
            if isempty(lComponentName)
                % If no component name is specified, pick the first one in
                % the ARXML files
                lComponentNameList = lARXMLImporterObj.getComponentNames;
                assert(numel(lComponentNameList) == 1);
                lComponentToImport = lComponentNameList{1};
            else
                lComponentToImport = lComponentName;
            end
            
            % Pass data dictionary through if necessary
            lVarArgs = {};
            if ~isempty(lDataDictionary)
                lVarArgs = {'DataDictionary', lDataDictionary};
            end
            
            % If client server interfaces or we have any function caller
            % inputs, then we can't have periodic runnables
            if lForceFunctionCallSubsystem
                % Call createComponentAsModel with periodic runnables
                % modeled as function call subsystems
                lVarArgs = [lVarArgs, ...
                    {'ModelPeriodicRunnablesAs', 'FunctionCallSubsystem'}];
            else
                % Call createComponentAsModel with periodic runnable
                % specifically modeled as Atomic Subsystems - this is
                % default behavior, but we don't want the information
                % message to be displayed that indicates this since we do
                % not expose this option as a manual user choice.
                lVarArgs = [lVarArgs, ...
                    {'ModelPeriodicRunnablesAs', 'AtomicSubsystem'}];
            end
            
            % Call arxml.importer
            [hModel, bSuccess] = lARXMLImporterObj.createComponentAsModel( ...
                lComponentToImport, ...
                'NameConflictAction', 'makenameunique', ...
                'OpenModel', false, ...
                'ForceLegacyWorkspaceBehavior', true, ...
                'InitializationRunnable', lInitializationRunnable, ...
                'ResetRunnable', lResetRunnables, ...
                'TerminateRunnable', lTerminateRunnable, ...
                lVarArgs{:});
        end
        
        function hBlock = configureModel(hModel, lComponentName, lSourcesHelper, varargin)
            % CONFIGUREMODEL
            %   Add the control block to the model, populated with the
            %   sources and include files contained within lSourcesHelper.
            %
            %   Set the tag so that the build hooks can identify that the
            %   model must be treated specially when building.
            %
            %   Then, save the model.
            %   Optionally re-name to ProxyModelName when saving.
            p = inputParser;
            p.addParameter('ProxyModelName', '', @ischar);
            p.parse(varargin{:});
            lProxyModelName = p.Results.ProxyModelName;
            lModelName = get_param(hModel, 'Name');
            hBlock = lSourcesHelper.addBlockToModel(lModelName);
            set_param(hBlock, 'ComponentName', lComponentName);
            set_param(lModelName, 'Tag', coder.internal.xrel.AUTOSAR.AutosarExternalCodeImportHook.XRelModelTag);
            
            % Use our custom TLC template to filter the model file from
            % being written!
            set_param(lModelName, 'ERTCustomFileTemplate', 'xrel_autosar_modelblock_customization.tlc');
            
            % Use compact file packaging - we don't want the generated
            % files anyway.
            if ~strcmp(get_param(lModelName, 'CodeInterfacePackaging'), 'Reusable function')
                set_param(lModelName, 'ERTFilePackagingFormat', 'Compact');
                
                % Set single instance model ref mode too even though we are
                % not a model ref (issues with accel build)
                set_param(lModelName, 'ModelReferenceNumInstancesAllowed', 'Single');
            end
            
            lDescription = sprintf('Cross-Release Model Block: %s', ...
                lModelName);
            set_param(lModelName, 'Description', lDescription);
            set_param(hModel, 'GenCodeOnly', 'on');
            set_param(hModel, 'UseOnlyExistingSharedCode', 'none');
            
            % Set all logging off or we have trouble when the model has
            % variants.
            set_param(hModel, 'SaveTime', 'off');
            set_param(hModel, 'SaveState', 'off');
            set_param(hModel, 'SaveOutput', 'off');
            set_param(hModel, 'SaveFinalState', 'off');
            set_param(hModel, 'SignalLogging', 'off');
            set_param(hModel, 'DSMLogging', 'off');
            set_param(hModel, 'LoggingToFile', 'off');
            set_param(hModel, 'ReturnWorkspaceOutputs', 'off');
            
            % Preserve port interface
            set_param(hModel, 'AutosarMatrixIOAsArray', 'on');

            % Make sure SimGenImportedTypeDefs is ON or Stateflow will
            % cause problems building SIM target for the stub model
            set_param(lModelName, 'SimGenImportedTypeDefs', 'on');
            
            % Don't want warnings - we know we have some unused data and
            % events with the arxml imported model.
            set_param(lModelName, 'SFUnusedDataAndEventsDiag', 'none');

            % Don't want shared constants to be generated by the stub
            % model, or we could get a phantom shared library dependency
            % which can cause issues with MSVC/Nmake on Windows.
            set_param(lModelName, 'GenerateSharedConstants', 'off');
            
            % Need to save the model or we cannot protect it
            if isempty(lProxyModelName)
                save_system(lModelName);
            else
                save_system(lModelName, lProxyModelName);
            end
        end
        
        function doPortReordering(hModel, lXrelArtifactUtils, lPortType)
            % DOPORTREORDERING
            %
            % Re-order the ports in the skeleton model imported by
            % arxml.importer based on the information from the
            % cross-release artifact (lArtifactPortInfo).
            %
            % Inputs:
            %       hModel:     Handle to the skeleton model
            %       lArtifactPortInfo:  XrelInfo.Inports or XrelInfo.Outports
            %       lPortType:  Inport or Outport
            import coder.internal.xrel.AUTOSAR.AutosarImportModelManager;
            
            % Get list of ports
            isInport = strcmp(lPortType, 'Inport');
            [lModelDataPorts, lModelDataPortNames] = AutosarImportModelManager.getModelDataPortHandlesAndNames(hModel, isInport);
            [lModelFcnPorts, lModelFcnPortNames] = AutosarImportModelManager.getModelFcnPortHandlesAndNames(hModel, isInport);
            
            % Use the AUTOSAR Simulink mapping to match the ports to avoid
            % making assumptions about the naming scheme used in the
            % skeleton model.
            slMap = autosar.api.getSimulinkMapping(hModel);
            arProps = autosar.api.getAUTOSARProperties(hModel);
            lDataPortMap = AutosarImportModelManager.getDataPortSlMapping(slMap, lModelDataPorts, lModelDataPortNames, isInport);
            lFcnPortMap = AutosarImportModelManager.getFcnPortSlMapping(slMap, arProps, lModelFcnPorts, lModelFcnPortNames);
            
            % Can compute these up front
            lIsUpdated = strcmp('IsUpdated', {lDataPortMap.DataAccessMode});
            lIsErrorStatus = strcmp('ErrorStatus', {lDataPortMap.DataAccessMode});
            
            % Log which ports we've visited
            lModelDataPortWasInArtifact = false(1, numel(lModelDataPorts));
            lModelFcnPortWasInArtifact = false(1, numel(lModelFcnPorts));
            
            % Renumber each port in turn
            for kPort = 1:lXrelArtifactUtils.getNumPorts(lPortType)
                lcurrPortNum = num2str(kPort);
                if lXrelArtifactUtils.isFunctionCall(lPortType, kPort)
                    lExpectSymbol = lXrelArtifactUtils.getPortSymbol(lPortType, kPort);
                    lIdx = find(strcmp(lExpectSymbol, lFcnPortMap), 1, 'first');
                    assert(~isempty(lIdx), 'Could not find function port for runnable with symbol %s', lExpectSymbol);
                    set_param(lModelFcnPorts(lIdx), 'Port', lcurrPortNum);
                    lModelFcnPortWasInArtifact(lIdx) = true;
                else
                    % For the graphical port of lPortType and index kPort,
                    % get the AUTOSAR port name, data element name and data
                    % access mode from the cross-release artifact.
                    [lAutosarPort, lAutosarDataElement, lDataAccessMode] = ...
                        lXrelArtifactUtils.getAutosarPortMapping(lPortType, kPort);
                    
                    lHasMatch = strcmp(lAutosarPort, {lDataPortMap.PortName}) ...
                        & strcmp(lAutosarDataElement, {lDataPortMap.DataElementName});
                    if strcmp(lDataAccessMode, 'IsUpdated')    
                        lHasMatch = lHasMatch & lIsUpdated;
                    elseif strcmp(lDataAccessMode, 'ErrorStatus')
                        lHasMatch = lHasMatch & lIsErrorStatus;
                    else
                        lHasMatch = lHasMatch & ~lIsUpdated & ~lIsErrorStatus;
                    end
                    
                    lIdx = find(lHasMatch);
                    assert(numel(lIdx) <= 1, 'Found multiple matches for %s: %s, DataElement: %s, AccessMode: %s', lPortType, lAutosarPort, lAutosarDataElement, lDataAccessMode);
                    assert(numel(lIdx) == 1, 'Found no matches for %s: %s, DataElement: %s, AccessMode: %s', lPortType, lAutosarPort, lAutosarDataElement, lDataAccessMode);
                    set_param(lModelDataPorts(lIdx), 'Port', lcurrPortNum);
                    lModelDataPortWasInArtifact(lIdx) = true;
                end
            end
            
            % Assert that we've not missed any function ports
            assert(all(lModelFcnPortWasInArtifact), 'Not all function call ports in stub model were mapped to function call ports from cross release artifact');
            
            % We might get unwanted ErrorStatus ports when triggered by
            % error events
            for kDataPort = 1:numel(lModelDataPorts)
                if isInport
                    if ~lModelDataPortWasInArtifact(kDataPort)
                        if lIsErrorStatus(kDataPort)
                            % Remove it
                            lTag = 'PORT_TO_REMOVE';
                            set_param(lModelDataPorts(kDataPort), 'Tag', lTag);
                            
                            % Should be able to just swap with a
                            % ground block?
                            replace_block(hModel, 'Tag', lTag, 'built-in/Ground', 'noprompt');
                            
                            % We know about it and have dealt with it.
                            % Mark as true.
                            lModelDataPortWasInArtifact(kDataPort) = true;
                        end
                    end
                end
            end
            assert(all(lModelDataPortWasInArtifact), 'Not all data ports in stub model were mapped to function call ports from cross release artifact');
        end
        
        function [lModelDataPorts, lModelDataPortNames]  = getModelDataPortHandlesAndNames(hModel, isInport)
            % GETMODELDATAPORTHANDLESANDNAMES
            if isInport
                % Ports with a signal output
                lModelDataPorts = find_system(hModel, 'SearchDepth', 1, 'BlockType', 'Inport', 'OutputFunctionCall', 'off');
                lModelDataPortNames = get_param(lModelDataPorts, 'Name');
            else
                % Assume all ports take a signal
                lModelDataPorts = find_system(hModel, 'SearchDepth', 1, 'BlockType', 'Outport');
                lModelDataPortNames = get_param(lModelDataPorts, 'Name');
            end
            
            % Single element returns from get_param won't be cells, so wrap them.
            if ~iscell(lModelDataPortNames)
                lModelDataPortNames = {lModelDataPortNames};
            end
        end
        
        function [lModelFcnPorts, lModelFcnPortNames]  = getModelFcnPortHandlesAndNames(hModel, isInport)
            % GETMODELFCNPORTHANDLESANDNAMES
            if isInport
                % Ports with a function call output
                lModelFcnPorts = find_system(hModel, 'SearchDepth', 1, 'BlockType', 'Inport', 'OutputFunctionCall', 'on');
                lModelFcnPortNames = get_param(lModelFcnPorts, 'Name');
            else
                lModelFcnPorts = [];
                lModelFcnPortNames = {};
            end
            if ~iscell(lModelFcnPortNames)
                lModelFcnPortNames = {lModelFcnPortNames};
            end
        end
        
        function lDataPortMap = getDataPortSlMapping(slMap, lModelDataPorts, lModelDataPortNames, isInport)
            % GETDATAPORTSLMAPPING
            %
            % Pre-allocate struct array populated with empty elements
            nModelDataPorts = numel(lModelDataPorts);
            lDataPortMap = struct('PortName', cell(1, nModelDataPorts), ...
                'DataElementName', cell(1, nModelDataPorts), ...
                'DataAccessMode', cell(1, nModelDataPorts));
            if isInport
                fhData = @slMap.getInport;
            else
                fhData = @slMap.getOutport;
            end
            for kDataPort = 1:nModelDataPorts
                [lDataPortMap(kDataPort).PortName, ...
                    lDataPortMap(kDataPort).DataElementName, ...
                    lDataPortMap(kDataPort).DataAccessMode] = fhData(lModelDataPortNames{kDataPort});
            end
        end
        
        function lFcnPortMap = getFcnPortSlMapping(slMap, arProps, lModelFcnPorts, lModelFcnPortNames)
            % GETFCNPORTSLMAPPING
            swc = get(arProps,'XmlOptions','ComponentQualifiedName');
            ib = get(arProps,swc,'Behavior');
            nModelFcnPorts = numel(lModelFcnPorts);
            lFcnPortMap = cell(1, nModelFcnPorts);
            for kFcnPort = 1:nModelFcnPorts
                % This gets the "Name", but not the "Symbol"
                lRunnableName = slMap.getFunction(lModelFcnPortNames{kFcnPort});
                
                % Convert the "Name" to a "Symbol", because that's what we
                % are able to guarantee from CodeInfo in the old release.
                lRunnablePath = arProps.find(ib, 'Runnable', 'Name', lRunnableName);
                assert(numel(lRunnablePath) == 1);  
                lFcnPortMap{kFcnPort} = arProps.get(lRunnablePath{1}, 'symbol');
            end
        end
        
        function doPortRenaming(hModel, lXrelArtifactUtils, lBlockType)
            % DOPORTRENAMING
            %
            %   Re-name the root input/output ports in the skeleton model
            %   with the GraphicalNames from the cross release artifact.
            
            % Desired input/output port names will not conflict with each
            % other.
            %
            % But they might conflict with the automatic ones (or another
            % block in the diagram)
            %
            import coder.internal.xrel.AUTOSAR.AutosarImportModelManager;
            
            for kPort = 1:lXrelArtifactUtils.getNumPorts(lBlockType)
                lcurrPortNum = num2str(kPort);
                lGraphicalName = lXrelArtifactUtils.getGraphicalName(lBlockType, kPort);
                hPortBlock = find_system(hModel, 'SearchDepth', 1, 'BlockType', lBlockType, 'Port', lcurrPortNum);
                lExistingBlock = find_system(hModel, 'SearchDepth', 1, 'Name', lGraphicalName);
                lPortHasConflict = ~isempty(lExistingBlock);
                if lPortHasConflict
                    kRenamed = 0;
                    lNewName = [lGraphicalName, '_RENAMED'];
                    while ~isempty(find_system(hModel, 'SearchDepth', 1, 'Name', lNewName))
                        kRenamed = kRenamed+1;
                        lNewName = [lGraphicalName, '_RENAMED', num2str(kRenamed)];
                    end
                    set_param(lExistingBlock, 'Name', lNewName);
                end
                set_param(hPortBlock, 'Name', lGraphicalName);
                
                % Verify dimensions
                if ~lXrelArtifactUtils.isFunctionCall(lBlockType, kPort)
                    [lExpectDims, lIsBus, lIsSymbolicDimension] = lXrelArtifactUtils.getDimensions(lBlockType, kPort);
                    if any(lIsSymbolicDimension)  % Need "any" because [Inf 1] and [1 Inf] can be possible
                        % lExpectDims will be inf if symbolic dimensions
                        % are being used.  In which case, we don't want to
                        % eval the PortDimensions because they might not be
                        % "evallable".
                        lCurrentDimStr = get_param(hPortBlock, 'PortDimensions');
                        lUpdatedDimStr = coder.internal.xrel.AUTOSAR.AutosarImportModelManager.substSymbolicDimension(lExpectDims, lIsSymbolicDimension, lCurrentDimStr);
                        set_param(hPortBlock, 'PortDimensions', lUpdatedDimStr);
                    else
                        lCurrentDims = eval(get_param(hPortBlock, 'PortDimensions'));
                       % Make sure the dimensions match the ones serialized 
                        assert(prod(lCurrentDims) == prod(lExpectDims), 'Dims dont match');
                        set_param(hPortBlock, 'PortDimensions', mat2str(lExpectDims));
                        % Trace back to source
                        if strcmp(lBlockType, 'Outport') && lIsBus
                            AutosarImportModelManager.updateConstant(hPortBlock, lExpectDims);
                        end
                    end
                end
            end
        end
        
        function lSymDimStr = substSymbolicDimension(lNumericDims, lIsSymbolicDimension, l1DSymDim)
            % SUBSTSYMBOLICDIMENSION
            assert(numel(find(lIsSymbolicDimension)) == 1);   % Only one symbolic dim - enforced in crossReleaseExport
            lNumericDims(lIsSymbolicDimension) = inf;   % Force this to handle the R2016a PIM initial conditions case
            lNumericDimStr = mat2str(lNumericDims);
            lSymDimStr = strrep(lNumericDimStr, 'Inf', l1DSymDim);
        end
        
        function updateConstant(hPortBlock, lExpectDims)
            % UPDATECONSTANT
            %
            %   Update the constant block driving an output port where the
            %   original model expected an n-D rather than 1D signal.
            %
            if numel(lExpectDims) > 1    
                lLineHandles = get_param(hPortBlock, 'LineHandles');
               
                lSourceBlock = get(lLineHandles.Inport, 'SrcBlockHandle');
                lSourceBlockType = get(lSourceBlock, 'BlockType');
                
                if strcmp(lSourceBlockType, 'SubSystem')
                    lSourcePort = get(lLineHandles.Inport, 'SrcPortHandle');
                    lSourcePortNum = get(lSourcePort, 'PortNumber');
                    lSourceParent = get_param(get_param(lSourcePort, 'Parent'), 'Handle');
                    
                    lPortInParent = find_system(lSourceParent, ...
                        'SearchDepth', 1, ...
                        'BlockType', 'Outport', ...
                        'Port', num2str(lSourcePortNum));
                    
                    lLinesInParent = get_param(lPortInParent, 'LineHandles');
                    lSecondSource = get_param(lLinesInParent.Inport, 'SrcBlockHandle');
                    lSecondSourceType = get_param(lSecondSource, 'BlockType');
                    
                    if strcmp(lSecondSourceType, 'Constant')
                        lStrZeros = sprintf('zeros(%s)', mat2str(lExpectDims));
                        set_param(lSecondSource, 'Value', lStrZeros);
                    end 
                end
            end
        end
        
        function doSRInterfaceMatching(hModel, lXRelArtifactUtils, lBlockType)
            % DOSRINTERFACEMATCHING
            %   Make sure DataAccessMode is correct for root input and
            %   output ports.
            slMap=autosar.api.getSimulinkMapping(hModel);
            for kPort = 1:lXRelArtifactUtils.getNumPorts(lBlockType)
                if ~lXRelArtifactUtils.isFunctionCall(lBlockType, kPort)
                    [lAutosarPort, lDataElement, lDataAccessMode, lGraphicalName] = lXRelArtifactUtils.getAutosarPortMapping(lBlockType, kPort);
                    
                    switch lBlockType
                        case 'Outport'
                            [lArPortName,lArDataElementName, lArDataAccessMode] = slMap.getOutport(lGraphicalName);
                            mapCmd = @slMap.mapOutport;
                        case 'Inport'
                            [lArPortName,lArDataElementName, lArDataAccessMode] = slMap.getInport(lGraphicalName);
                            mapCmd = @slMap.mapInport;
                        otherwise
                            assert(false, 'Should not get here');
                    end
                    
                    assert(isequal(lArPortName, lAutosarPort), 'Port mismatch');
                    assert(isequal(lArDataElementName, lDataElement), 'Data element mismatch');
                    
                    % Except that CodeInfo does not cover EndToEnd for
                    % non-queued data types, so we need to be careful
                    % and don't overwrite anything that is E2E protected.
                    % This matters for XIL when there are ErrorStatus
                    % ports, since the data type is different for E2E
                    % protected ports (uint32 vs uint8).
                    if ~ismember(lArDataAccessMode, {'EndToEndWrite', 'EndToEndRead'})
                        mapCmd(lGraphicalName, lArPortName, lArDataElementName, lDataAccessMode);
                    end
                end
            end
        end
        
        
        function doFunctionRenaming(hModel, lArtifactFunctions, lDataDictionary)
            % DOFUNCTIONRENAMING
            %   Update function caller blocks to reflect the original
            %   Simulink model.
            %
            lBlockType = 'FunctionCaller';
            lFcnBlocks = find_system(hModel, 'BlockType', lBlockType);
            lFcnPrototypes = get_param(lFcnBlocks, 'FunctionPrototype');
            if ~iscell(lFcnPrototypes)
                lFcnPrototypes = {lFcnPrototypes};
            end
            
            lProtoRegexp = '(?<=(\w*\s*=\s*)?)(\w*)(?=(\(|$))';
            lFcnNames = regexp(lFcnPrototypes, lProtoRegexp, 'match', 'once');
            
            if isempty(lArtifactFunctions)
                lExpectList = {};
            else
                lExpectList = strcat( ...
                    { lArtifactFunctions.AUTOSARPortName }, ...
                    '_', ...
                    { lArtifactFunctions.AUTOSAROperationName });
            end
            
            % Iterate through the function caller blocks that we find - not
            % all functions will be called.
            for kFcnCaller = 1:numel(lFcnBlocks)   
                lCurrentFunction = lFcnNames{kFcnCaller};
                lIsMatched = strcmp(lCurrentFunction, lExpectList);
                if any(lIsMatched)
                    lIdx = find(lIsMatched, 1, 'first');
                    
                    % Inputs and outputs are in the order prescribed by the
                    % function prototype in the code at this point.
                    %
                    % It's easier to re-write the function prototype rather
                    % than attempt to manipulate the arguments to re-order
                    % them into their graphical order.
                    [lInputArgNameCell, lInputArgIdx] = ...
                        coder.internal.xrel.AUTOSAR.AutosarImportModelManager.getSlFunctionPrototypeArgs(lArtifactFunctions(lIdx).FunctionInputs);
                    
                    lFunctionReturn = lArtifactFunctions(lIdx).FunctionReturn;
                    if ~isempty(lFunctionReturn) && strcmp(lFunctionReturn.Name, '')
                        % For AUTOSAR, if the return argument has an empty
                        % name, it will be the error return value.
                        lFunctionReturn.Name = 'ERR';
                    end
                    lFunctionOutputs = [lArtifactFunctions(lIdx).FunctionOutputs,lFunctionReturn];
                    [lOutputArgNameCell, lOutputArgIdx] = ...
                        coder.internal.xrel.AUTOSAR.AutosarImportModelManager.getSlFunctionPrototypeArgs(lFunctionOutputs);
                    
                    % lInputArgNameCell and lOutputArgNameCell are in the 
                    % graphical representation order.
                    lNewFunctionPrototype = ...
                        coder.internal.xrel.AUTOSAR.AutosarImportModelManager.buildSlFunctionPrototype( ...
                        lArtifactFunctions(lIdx).CGFunctionName, ...
                        lInputArgNameCell, ...
                        lOutputArgNameCell);
                    set_param( lFcnBlocks(kFcnCaller), 'FunctionPrototype', lNewFunctionPrototype);
                    
                    % We also need to ensure that dimensions are consistent
                    % and in the correct order.
                    lInputArgSpecs = get_param( lFcnBlocks(kFcnCaller), 'InputArgumentSpecifications');
                    lInputArgDims = {lArtifactFunctions(lIdx).FunctionInputs.Dimensions};
                    lUpdatedInputArgSpecs = coder.internal.xrel.AUTOSAR.FcnCallerUtils.processArgumentSpecifications(lInputArgSpecs, lInputArgDims, lInputArgIdx, lDataDictionary);
                    
                    set_param( lFcnBlocks(kFcnCaller), 'InputArgumentSpecifications', lUpdatedInputArgSpecs);
                    
                    lOutputArgSpecs = get_param( lFcnBlocks(kFcnCaller), 'OutputArgumentSpecifications');
                    
                    lOutputArgDims = { ...
                        lArtifactFunctions(lIdx).FunctionOutputs.Dimensions, ...
                        lArtifactFunctions(lIdx).FunctionReturn.Dimensions};
                    
                    lUpdatedOutputArgSpecs = coder.internal.xrel.AUTOSAR.FcnCallerUtils.processArgumentSpecifications(lOutputArgSpecs, lOutputArgDims, lOutputArgIdx, lDataDictionary);
                    set_param( lFcnBlocks(kFcnCaller), 'OutputArgumentSpecifications', lUpdatedOutputArgSpecs);
                end
            end
        end
        
        function [lArgNameCell, lArgIdx] = getSlFunctionPrototypeArgs(lFunctionArgs)
            % GETSLFUNCTIONPROTOTYPEARGS
            % 
            % Returns the list of input OR output arguments in the order 
            % for the graphical representation, and their corresponding indices.
            %
            %   e.g., for
            %
            %       lFunctionArgs(1).Name = 'argD'
            %       lFunctionArgs(1).ArgIndex = 'I3'
            %       ...
            %       lFunctionArgs(2).Name = 'argB'
            %       lFunctionArgs(2).ArgIndex = 'I1'
            %       ...
            %       lFunctionArgs(3).Name = 'argC'
            %       lFunctionArgs(3).ArgIndex = 'I2'
            %       ...
            %       lFunctionArgs(4).Name = 'argA'
            %       lFunctionArgs(4).ArgIndex = 'I0'
            %       ...
            %
            % Then:
            %
            %   lArgNameCell = {'argA', 'argB', 'argC', 'argD'};
            %   lArgIndex = [4,2,3,1]
            
            % Use ArgIndex to get the positions in the graphical
            % representation.
            lArgIdx = str2double(regexprep({lFunctionArgs.ArgIndex}, 'I|O', '')) + 1;
            nArgs = numel(lFunctionArgs);
            lArgNameCell = cell(1, nArgs);
            
            % Return a list of the argument names in order for the
            % graphical representation.
            lArgNameCell(lArgIdx) = {lFunctionArgs.Name};
        end
        
        function lNewFunctionPrototype = buildSlFunctionPrototype(lFunctionName, lInputArgNameCell, lOutputArgNameCell)
            % BUILDSLFUNCTIONPROTOTYPE
            
            % Build up an expression for the (graphical) Simulink
            % Function prototype from the info in the cross-release
            % artifact.
            nOutputs = numel(lOutputArgNameCell);
            if nOutputs == 1
                % Special case for one output, as [ varname ] = ...
                % is not a valid syntax.
                lOutputArgStr = sprintf('%s = ', lOutputArgNameCell{1});
            elseif nOutputs > 1
                lOutputArgStr = sprintf('[%s] = ', strjoin(lOutputArgNameCell, ', '));
            else
                lOutputArgStr = '';
            end
            lInputArgStr = sprintf('(%s)', strjoin(lInputArgNameCell, ', '));
            lNewFunctionPrototype = sprintf('%s%s%s', lOutputArgStr, lFunctionName, lInputArgStr);
        end
        
        function updateSimulinkFunctionArgumentDimensions(hModel, lUtils)
            % UPDATESIMULINKFUNCTIONARGUMENTDIMENSIONS
            hModel = get_param(hModel, 'handle');   % Make sure it's the handle
            
            lSimulinkFunctionTriggers = ...
                find_system(hModel, 'BlockType', 'TriggerPort', 'IsSimulinkFunction', 'on');
            for kSlFunc = 1:numel(lSimulinkFunctionTriggers)
                hTrigger = lSimulinkFunctionTriggers(kSlFunc);                
                lFunctionName = get(hTrigger, 'FunctionName');
                lArtifact = lUtils.getSimulinkFunctionByCGFunctionName(lFunctionName);
                hParent = get_param(get(hTrigger, 'Parent'), 'Handle');
                
                hFcnArgIn = coder.internal.xrel.AUTOSAR.AutosarImportModelManager.getArgPortHandles(hParent, 'ArgIn');
                coder.internal.xrel.AUTOSAR.AutosarImportModelManager.updateArgBlocks(hFcnArgIn, lArtifact.FunctionInputs)
                
                hFcnArgOut = coder.internal.xrel.AUTOSAR.AutosarImportModelManager.getArgPortHandles(hParent, 'ArgOut');
                coder.internal.xrel.AUTOSAR.AutosarImportModelManager.updateArgBlocks(hFcnArgOut, lArtifact.FunctionOutputs);
            end
        end
        
        function updateArgBlocks(hFcnArgs, lFunctionArgArtifacts)
            % UPDATEARGBLOCKS
            if isfield(lFunctionArgArtifacts, 'Name')
                lArgNames = {lFunctionArgArtifacts.Name};
            else
                lArgNames = {};
            end
            for kFcnArg = 1:numel(hFcnArgs)
                hCurrFcnArg = hFcnArgs(kFcnArg);
                lCurrentArgumentName = get_param(hCurrFcnArg, 'ArgumentName');
                
                % Find the artifact element by matching name.
                lCurrentFunctionArgArtifact = ...
                    lFunctionArgArtifacts(strcmp(lCurrentArgumentName, lArgNames));
                
                % Update dimensions
                lDimensions = lCurrentFunctionArgArtifact.Dimensions;
                lDimensionsStr = mat2str(lDimensions);
                set_param(hCurrFcnArg, 'PortDimensions', lDimensionsStr);
                
                % Update port index
                lArgIndex = lCurrentFunctionArgArtifact.ArgIndex;
                lPortStr = num2str(str2double(regexprep(lArgIndex, 'O|I', ''))+1);
                set_param(hCurrFcnArg, 'Port', lPortStr);
            end
        end
            
        function hArgs = getArgPortHandles(hParent, lType)
            % GETARGPORTHANDLES
            assert(ismember(lType, {'ArgIn', 'ArgOut'}));
            hArgs = find_system(hParent, 'BlockType', lType);
        end
        
        function updateAsyncTaskPriorities(hModel, lXrelArtifactUtils)
            % UPDATEASYNCTASKPRIORITIES
            %
            %   Update async task priorities to the values in the original
            %   model (or potentially the cross-release model will not be a
            %   drop-in replacement)
            %
            for kPort = 1:lXrelArtifactUtils.getNumPorts('Inport')
                if lXrelArtifactUtils.isFunctionCall('Inport', kPort)
                    lOutputFunction = lXrelArtifactUtils.getOutputFunctionFromInport('Inport', kPort);
                    if strcmp(lOutputFunction.TimingMode, 'ASYNCHRONOUS')
                        lGraphicalName = lXrelArtifactUtils.getGraphicalName('Inport', kPort);
                        lPriority = num2str(lOutputFunction.Priority);
                        hPortBlock = find_system(hModel, 'SearchDepth', 1, 'Name', lGraphicalName);
                        hLine = get(hPortBlock, 'LineHandles');
                        hAsyncBlock = get(hLine.Outport, 'DstBlockHandle');
                        
                        % There might not be an async task block always
                        if ~isempty(hAsyncBlock) && ...
                                strcmp(get(hAsyncBlock, 'BlockType'), 'AsynchronousTaskSpecification')
                            set_param(hAsyncBlock, 'TaskPriority', lPriority);
                        end
                    end
                end
            end
        end
        
        function addCTypePIMBlocks(hModel, m3iComp, lUtils, lDataDictionary)
            % ADDCTYPEPIMBLOCKS
            
            % Instantitate Type Utils class
            lTypeUtils = coder.internal.xrel.AUTOSAR.TypeUtils(lUtils, lDataDictionary);
            
            % Iterate through C-typed PIMS
            lDWorks = cellfun(@lUtils.getStatesDWorkName, ...
                {'Block states', 'Block signals', 'Zero crossing states'}, ...
                'UniformOutput', false);
            
            % Delete entries that are empty
            lDWorks(cellfun(@isempty, lDWorks)) = [];
            
            [~, lControlBlockHandle] = coder.internal.xrel.AUTOSAR.AutosarExternalCodeImportHook.isAutosarCodeImport(hModel);
            assert(~isempty(lControlBlockHandle), 'Could not find AUTOSAR cross-release control block');
            lControlBlockPosition = double(get_param(lControlBlockHandle, 'Position'));
            lModelName = get_param(hModel, 'Name');
            for kPim = 1:double(m3iComp.Behavior.PIM.size)
                m3iPim = m3iComp.Behavior.PIM.at(kPim);
                if ~isempty(lDWorks) && ismember(m3iPim.Name, lDWorks)
                    % Skip if this is one of the model DWorks
                    continue;
                end
                    
                lNewPos0 = [lControlBlockPosition(3)+60*kPim, lControlBlockPosition(2)+30];
                lNewPos = [lNewPos0, lNewPos0+[30 30]];
                
                % Get data type and dimensions from cross-release artifact
                lDataStoreInfo = lUtils.getDataStoreByGraphicalName(m3iPim.Name);
                if ~isempty(lDataStoreInfo)
                    if lDataStoreInfo.IsBus
                        lDataType = ['Bus: ', lDataStoreInfo.TypeName];
                    elseif lDataStoreInfo.IsEnum
                        lDataType = ['Enum: ', lDataStoreInfo.TypeName];
                    else
                        lDataType = lDataStoreInfo.TypeName;
                    end
                    
                    % Create the type if necessary
                    lTypeUtils.createTypeObjectsIfNecessary(lDataStoreInfo.TypeName);
                    
                    if lDataStoreInfo.IsEnum
                       % Go for the default value
                       lTypeArtifact = lUtils.getTypeByName(lDataStoreInfo.TypeName);
                       lInitialValue = [lDataStoreInfo.TypeName, '.', lTypeArtifact.Strings{lTypeArtifact.DefaultMember}];
                    else
                       lInitialValue = '0'; 
                    end
                    
                    
                    add_block('built-in/DataStoreMemory', ...
                        [lModelName, '/', sprintf('CPIM_%d', kPim)], ...
                        'Position', lNewPos, ...
                        'DataStoreName', m3iPim.Name, ...
                        'OutDataTypeStr', lDataType, ...
                        'Dimensions', mat2str(lDataStoreInfo.Dimensions), ...
                        'StateSignalObject', AUTOSAR.Signal, ...
                        'StateStorageClass', 'PerInstanceMemory (Custom)', ...
                        'InitialValue', lInitialValue);
                end
            end
        end
        
        function lRunnableName = getRunnableNameFromSymbol(m3iComp, lRunnableSymbol)
            % GETRUNNABLENAMEFROMSYMBOL
            lRunnableName = '';
            for kRunnable = 1:m3iComp.Behavior.Runnables.size
                m3iRunnable = m3iComp.Behavior.Runnables.at(kRunnable);
                if strcmp(m3iRunnable.symbol, lRunnableSymbol)
                    lRunnableName = m3iRunnable.Name;
                    break;
                end
            end
            
        end
        
        function lForceFunctionCall = needsExplicitTiming(m3iComp, lUtils)
            % NEEDSEXPLICITTIMING
            lPeriodicTimes = ...
                autosar.mm.mm2sl.PeriodicRunnablesModelingStyleDeterminer.collectPeriodicRunnableSampleTimes(m3iComp.Behavior.Runnables);
            lNoPeriodic = isempty(lPeriodicTimes);
            lPeriodicByFcnCall = lUtils.hasExplicitPeriodicTasking;
            lDefinesSimulinkFunctions = lUtils.definesSimulinkFunctions;
            lForceFunctionCall = lDefinesSimulinkFunctions || lPeriodicByFcnCall || lNoPeriodic;
        end
        
        function updateStructTypes(lDataDictionary, lUtils)
            % UPDATESTRUCTTYPES
            lXRelTypeUtils = coder.internal.xrel.XrelTypeUtils(lDataDictionary);
            
            lStructs = lUtils.getStructTypes;
            for kStruct = 1:numel(lStructs)
                [slObj, wasFound] = lXRelTypeUtils.getSimulinkObject(lStructs(kStruct).Name);
                if wasFound
                    % Check the dimensions of each element
                    assert(isa(slObj, 'Simulink.Bus'), 'Expected %s to be a Simulink.Bus object', lStructs(kStruct).Name);
                    lDidChange = false;
                    for kElem = 1:numel(slObj.Elements)
                        lFoundDims = slObj.Elements(kElem).Dimensions;
                        lExpectDims = lStructs(kStruct).Elements(kElem).Dimensions;
                        [lNewDimensions, ~, lMustDoChange] = ...
                            coder.internal.xrel.AUTOSAR.AutosarImportModelManager.updateDimensions( ...
                            lFoundDims, ...
                            lExpectDims);
                        if lMustDoChange
                            slObj.Elements(kElem).Dimensions = lNewDimensions;                            
                            lDidChange = true;
                        end
                    end
                    if lDidChange
                        lXRelTypeUtils.setSimulinkObject(lStructs(kStruct).Name, slObj);
                    end                    
                end
            end
        end

        function ret = hasNonScalarBusElement(lUtils, lName)
            % HASNONSCALARBUSELEMENT
            lArtifact = lUtils.getStructTypeByName(lName);
            if isempty(lArtifact)
                ret = false;
                return;
            end
            
            for kElem = 1:numel(lArtifact.Elements)
                if any(lArtifact.Elements(kElem).Dimensions ~= 1)
                    ret = true;
                    return;
                else
                    % Recurse
                    lElemType = lArtifact.Elements(kElem).DataType;
                    ret = coder.internal.xrel.AUTOSAR.AutosarImportModelManager.hasNonScalarBusElement(lUtils, lElemType);
                end
            end
        end
        
        function lName = getParameterGraphicalName(lAutosarPort, lAutosarDataElement, lDataAccessMode)
            % GETPARAMETERGRAPHICALNAME
            if strcmp(lDataAccessMode, 'Calibration')
                lName = [lAutosarPort, '_', lAutosarDataElement];
            else
                lName = lAutosarPort;
            end
            if numel(lName) > 63
                lName = lName(1:63);
            end
        end
        
        
        function [lNewDimensions, lIsSymbolicDimension, lMustDoChange] = updateDimensions(lOrigDimensions, lArtifactDimensions)
            % UPDATEDIMENSIONS
            lIsSymbolicDimension = isinf(lArtifactDimensions);
            if any(lIsSymbolicDimension)
                if isnumeric(lOrigDimensions)
                    % For bus elements, we might lose the symbolic
                    % dimension
                    assert(numel(find(lIsSymbolicDimension)) == 1);
                    lNewDimensions = lArtifactDimensions;
                    lNewDimensions(lIsSymbolicDimension) = lOrigDimensions;
                else
                    % Expecting the original dimensions to be of the form
                    % [mySymDim1,1] or [1,mySymDim1] - need to get the
                    % non-scalar dimension symbol.
                    l1DSymDim = regexp(lOrigDimensions, '[a-zA-Z]\w*', 'match');
                    l1DSymDim = l1DSymDim{1};
                    lNewDimensions = ...
                        coder.internal.xrel.AUTOSAR.AutosarImportModelManager.substSymbolicDimension( ...
                        lArtifactDimensions, lIsSymbolicDimension, l1DSymDim);
                end
                lMustDoChange = ~isequal(lNewDimensions, lOrigDimensions);
            else
                % For PIM initial conditions, symbolic dims might be frozen
                % in the artifact if generated in R2016a, but remain symbolic
                % in the parameter objects from createComponentAsModel
                if isnumeric(lOrigDimensions)
                    assert(prod(lArtifactDimensions) == prod(lOrigDimensions), 'Expected number of elements to be the same');
                    lNewDimensions = lArtifactDimensions;
                    lMustDoChange = ~isequal(lNewDimensions, lOrigDimensions);
                else
                    l1DSymDim = regexp(lOrigDimensions, '[a-zA-Z]\w*', 'match');
                    l1DSymDim = l1DSymDim{1};
                    % Support only transposition of 1D vectors
                    lIsSymbolicDimension = (lArtifactDimensions ~= 1);
                    lNewDimensions = ...
                        coder.internal.xrel.AUTOSAR.AutosarImportModelManager.substSymbolicDimension( ...
                        lArtifactDimensions, lIsSymbolicDimension, l1DSymDim);
                end
            end
        end
        
        function lNewValue = reshapeSingleValue(lOrigValue, lArtifactDimensions)
            % RESHAPESINGLEVALUE
            lIsSymbolicDimension = isinf(lArtifactDimensions);
            if any(lIsSymbolicDimension)
                lNewNumericDimensions = ones(1, numel(lIsSymbolicDimension));
                lNewNumericDimensions(lIsSymbolicDimension) = numel(lOrigValue);
                lNewValue = reshape(lOrigValue, lNewNumericDimensions);
            else
                lNewValue = reshape(lOrigValue, lArtifactDimensions);
            end
        end
        
        function [lNewValue, lNewDimensions] = reshapeSingleParam(lOrigValue, lOrigDimensions, lArtifactDimensions)
            % RESHAPESINGLEPARAM
            %
            lNewDimensions = ...
                coder.internal.xrel.AUTOSAR.AutosarImportModelManager.updateDimensions( ...
                lOrigDimensions, ...
                lArtifactDimensions);
            lNewValue = coder.internal.xrel.AUTOSAR.AutosarImportModelManager.reshapeSingleValue( ...
                lOrigValue, ...
                lArtifactDimensions);
        end
            
        function updateParamDimensions(lUtils, lDataDictionary, lArtifact)
            % UPDATEPARAMDIMENSIONS
            %
            % Determine the name of the parameter based on the Data Access
            % Mode.
            lName = coder.internal.xrel.AUTOSAR.AutosarImportModelManager.getParameterGraphicalName( ...
                lArtifact.AutosarPort, ...
                lArtifact.AutosarDataElement, ...
                lArtifact.DataAccessMode);
            
            lXRelTypeUtils = coder.internal.xrel.XrelTypeUtils(lDataDictionary);
            
            % Unflatten any data that was marked as a matrix in the
            % Cross-Release Artifact.
            lIsMatrix = numel(lArtifact.Dimensions) > 1 && any(lArtifact.Dimensions ~=1);
            if lIsMatrix
                lObj = lXRelTypeUtils.getSimulinkObject(lName);
                if ~isempty(lObj)
                    if isa(lObj, 'Simulink.Parameter')   % Lookup tables don't derive from Parameter
                        [lObj.Value, ...
                            lObj.Dimensions] = ...
                            coder.internal.xrel.AUTOSAR.AutosarImportModelManager.reshapeSingleParam(...
                            lObj.Value, lObj.Dimensions, lArtifact.Dimensions);
                        lXRelTypeUtils.setSimulinkObject(lName, lObj);
                    elseif isa(lObj, 'Simulink.Breakpoint')
                        [lObj.Breakpoints.Value, ...
                            lObj.Breakpoints.Dimensions] = ...
                            coder.internal.xrel.AUTOSAR.AutosarImportModelManager.reshapeSingleParam(...
                            lObj.Breakpoints.Value, lObj.Breakpoints.Dimensions, lArtifact.Dimensions);
                        lXRelTypeUtils.setSimulinkObject(lName, lObj);
                    elseif isa(lObj, 'Simulink.LookupTable')
                        [lObj.Table.Value, ...
                            lObj.Table.Dimensions] = ...
                            coder.internal.xrel.AUTOSAR.AutosarImportModelManager.reshapeSingleParam(...
                            lObj.Table.Value, lObj.Table.Dimensions, lArtifact.Dimensions);
                        lXRelTypeUtils.setSimulinkObject(lName, lObj);
                    end
                end
            end
            
            if lArtifact.IsBus
                % Walk through bus elements and unflatten any data that was
                % marked as matrix valued in the Cross-Release Artifact.
                lHasNonScalar = coder.internal.xrel.AUTOSAR.AutosarImportModelManager.hasNonScalarBusElement(lUtils, lArtifact.BusTypeName);
                if lHasNonScalar
                    % Then we need to do something
                    lObj = lXRelTypeUtils.getSimulinkObject(lName);
                    if ~isempty(lObj)
                        if isa(lObj, 'Simulink.Parameter')
                            lVal = lObj.Value;
                            lVal = coder.internal.xrel.AUTOSAR.AutosarImportModelManager.reshapeVal( ...
                                lUtils, ...
                                lVal, ...
                                lArtifact.BusTypeName);
                            lObj.Value = lVal;
                            lXRelTypeUtils.setSimulinkObject(lName, lObj);
                        elseif isa(lObj, 'Simulink.Breakpoint')
                            lStructTypeName = lObj.StructTypeInfo.Name;
                            lStructTypeArtifact = lUtils.getStructTypeByName(lStructTypeName);
                            lObj = coder.internal.xrel.AUTOSAR.AutosarImportModelManager.reshapeLookupValue(lObj, lStructTypeArtifact, 'Breakpoints');
                            lXRelTypeUtils.setSimulinkObject(lName, lObj);
                        elseif isa(lObj, 'Simulink.LookupTable')
                            % We need to get the name of the struct type,
                            % and the fieldname
                            if strcmp(lObj.BreakpointsSpecification, 'Explicit values')
                                lStructTypeName = lObj.StructTypeInfo.Name;
                                lStructTypeArtifact = lUtils.getStructTypeByName(lStructTypeName);
                                lObj = coder.internal.xrel.AUTOSAR.AutosarImportModelManager.reshapeLookupValue(lObj, lStructTypeArtifact, 'Table');
                                lObj = coder.internal.xrel.AUTOSAR.AutosarImportModelManager.reshapeLookupValue(lObj, lStructTypeArtifact, 'Breakpoints');
                                lXRelTypeUtils.setSimulinkObject(lName, lObj);
                            elseif strcmp(lObj.BreakpointsSpecification, 'Even spacing')
                                lStructTypeName = lObj.StructTypeInfo.Name;
                                lStructTypeArtifact = lUtils.getStructTypeByName(lStructTypeName);
                                lObj = coder.internal.xrel.AUTOSAR.AutosarImportModelManager.reshapeLookupValue(lObj, lStructTypeArtifact, 'Table');
                                lXRelTypeUtils.setSimulinkObject(lName, lObj);
                            end
                        end
                    end
                end
            end
        end
        
        function lObj = reshapeLookupValue(lObj, lStructTypeArtifact, lLookupTablePart)
            % RESHAPELOOKUPVALUE
            for kElem = 1:numel(lObj.(lLookupTablePart))
                lFieldName = lObj.(lLookupTablePart)(kElem).FieldName;
                lElemArtifact = lStructTypeArtifact.Elements(strcmp({lStructTypeArtifact.Elements.Identifier}, lFieldName));
                if ~isempty(lElemArtifact)
                    assert(prod(lElemArtifact.Dimensions) == prod(lObj.(lLookupTablePart)(kElem).Dimensions), 'Expected number of elements to be the same');
                    lObj.(lLookupTablePart)(kElem).Value = reshape(lObj.(lLookupTablePart)(kElem).Value, lElemArtifact.Dimensions);
                end
            end
        end
            
        function lVal = reshapeVal(lUtils, lVal, lBusTypeName)
            % RESHAPEVAL
            lTypeArtifact = lUtils.getStructTypeByName(lBusTypeName);
            if isempty(lTypeArtifact) || ~isstruct(lVal)
                return;
            end
            for kElem = 1:numel(lTypeArtifact.Elements)
                if numel(lTypeArtifact.Elements(kElem).Dimensions) > 1 && ...
                        any(lTypeArtifact.Elements(kElem).Dimensions ~= 1)
                    lVal.(lTypeArtifact.Elements(kElem).Identifier) = ...
                        coder.internal.xrel.AUTOSAR.AutosarImportModelManager.reshapeSingleValue( ...
                        lVal.(lTypeArtifact.Elements(kElem).Identifier), ...% Original value  
                        lTypeArtifact.Elements(kElem).Dimensions ...  % Artifact dimensions
                        );          
                end
                lElemTypeName = lTypeArtifact.Elements(kElem).DataType;
                lVal.(lTypeArtifact.Elements(kElem).Identifier) = ...
                    coder.internal.xrel.AUTOSAR.AutosarImportModelManager.reshapeVal( ...
                    lUtils, ...
                    lVal.(lTypeArtifact.Elements(kElem).Identifier), ...
                    lElemTypeName);
            end
        end
        
                
        function validateAndUpdateErrorStatus(lM3iModel, lM3iComp, lXRelArtifactUtils)
            % VALIDATEANDUPDATEERRORSTATUS
            %
            % * We can accommodate cases where mm2sl would create an error
            %   status port, but one did not already exist.
            % * Where an error status port existed but mm2sl would 
            %   not create one by default, we need to adjust the comSpec
            %   in the arModel to signal that an error status port should
            %   be created in the temporary model.
            %
            
            tran = M3I.Transaction(lM3iModel);
            
            % Iterate through runnables
            nRunnables = lM3iComp.Behavior.Runnables.size;
            for kRunnable = 1:nRunnables
                lM3iRunnable = lM3iComp.Behavior.Runnables.at(kRunnable);
                
                % Iterate through data accesses for each runnable
                nDataFlow = lM3iRunnable.dataAccess.size;
                for kDataFlow = 1:nDataFlow
                    lM3iDataFlow = lM3iRunnable.dataAccess.at(kDataFlow);
                    lKindStr = lM3iDataFlow.Kind.toString;
                    lRef = lM3iDataFlow.instanceRef;
                    
                    % Only care about error status for reads
                    if ismember(lKindStr, coder.internal.xrel.AUTOSAR.AutosarImportModelManager.ReadAccessKinds) ...
                            && ~autosar.api.Utils.isErrorStatusPortElement(lRef.Port, lRef.DataElements, lKindStr)
                        % If createComponentAsModel is not going to
                        % create an ErrorStatus port, then we need to
                        % coerce it to do so if there was a port
                        % in the model from which the code was
                        % generated in the old release.
                        %
                        % This might be because the old release didn't
                        % include a comspec or because the ErrorStatus
                        % port corresponds to a condition that is not
                        % representable by the comspec in the release
                        % that generated the code.
                        %
                        % So, firstly, check if an error status port
                        % existed in the old release.  If not, we don't
                        % need to do anything.
                        lPortNameStr = lRef.Port.Name;
                        lElemNameStr = lRef.DataElements.Name;
                        lXRelPortStruct = lXRelArtifactUtils.getXRelInfoFromPortAndDataElement('Inports', ...
                            lPortNameStr, lElemNameStr, 'ErrorStatus');
                        if ~isempty(lXRelPortStruct)
                            % In this case there was an error status
                            % port, so we have to accommodate it during
                            % import.
                            
                            % We never generate new ARXML in a cross-release
                            % workflow, and do not expose the temporary
                            % model to the user, so we do not need to change
                            % this back - the arModel here is "throwaway"
                            % and not stored in the protected model
                            % artifact.
                            coder.internal.xrel.AUTOSAR.AutosarImportModelManager.setHandleNeverReceived( ...
                                lRef.Port, lRef.DataElements);
                            
                            % This might not be sufficient if there is a
                            % PR-Port, in which case, throw an error.
                            if ~autosar.api.Utils.isErrorStatusPortElement(lRef.Port, lRef.DataElements, lKindStr)
                                 DAStudio.error('autosarstandard:xrelimport:ErrorStatusIncompatible', ...
                                    lXRelPortStruct.GraphicalName, lElemNameStr, lPortNameStr);
                            end
                        end
                    end
                end
            end
            tran.commit();
        end
        
        function createParamsForNVRAMInitConditions(m3iModel, m3iComp)
            % CREATEPARAMSFORNVRAMINITCONDITIONS
            %
            %   Ar-Typed PIMs that access NVRAM need initial conditions to
            %   be set.
            %
            %   When the ServiceDependency has a UsedParameterElement set
            %   then this parameter can be used to provide the initial
            %   conditions.
            %
            %   When UsedParameterElement is empty, we don't know what the
            %   initial conditions should be.  In this case, what we can do
            %   is create a UsedParameterElement.
            %
            import coder.internal.xrel.AUTOSAR.PIMNVRAMUtils;
            tran = M3I.Transaction(m3iModel);
            
            m3iBehavior = m3iComp.Behavior;
            m3iArTypedPIM = m3iBehavior.ArTypedPIM;
            
            for kPIM = 1:m3iArTypedPIM.size
                m3iPIM = m3iArTypedPIM.at(kPIM);
                lPIMName = m3iPIM.Name;
                m3iServiceDependency = ...
                    PIMNVRAMUtils.getServiceDependencyForArTypedPimByName( ...
                    m3iBehavior, ...
                    lPIMName);
                
                % If there is a service dependency...
                if ~isempty(m3iServiceDependency)
                    % ... and there isn't a parameter element being used
                    if isempty(m3iServiceDependency.UsedParameterElement)
                        
                        % ... then we need one
                        lParamName = PIMNVRAMUtils.getSubstituteInitialValueParamName(lPIMName);
                        
                        lNewParam = Simulink.metamodel.arplatform.interface.ParameterData(m3iModel);
                        lNewParam.Kind = Simulink.metamodel.arplatform.behavior.ParameterKind.Shared;
                        lNewParam.SwCalibrationAccess = Simulink.metamodel.arplatform.common.SwCalibrationAccessKind.ReadWrite;
                        lNewParam.Type = m3iPIM.Type;
                        lNewParam.Name = lParamName;
                        m3iBehavior.Parameters.append(lNewParam);
                        m3iServiceDependency.UsedParameterElement = lNewParam;
                        
                        DAStudio.warning('autosarstandard:xrelimport:NVRAMInitCondParamCreated', ...
                            lParamName, lPIMName);
                        
                    end
                end
            end
            
            tran.commit();
        end
        
        
        function updateArTypedPIMDimensions(hModel, m3iComp, lDataDictionary, lUtils)
            % UPDATEARTYPEDPIMDIMENSIONS
            %
            %   CodeInfo has flattened dimensions for data stores.  But the
            %   un-flattened dimensions are important when a parameter
            %   object has been used to initialize.
            %
            lXRelTypeUtils = coder.internal.xrel.XrelTypeUtils(lDataDictionary);
            m3iBehavior = m3iComp.Behavior;
            for kPIM = 1:m3iBehavior.ArTypedPIM.size
                lName = m3iBehavior.ArTypedPIM.at(kPIM).Name;
                lParamName = ...
                    coder.internal.xrel.AUTOSAR.PIMNVRAMUtils.getParamNameFromServiceDepForArTypedPimByName( ...
                    m3iBehavior, lName);
                lParamArtifact = lUtils.getParameterByName(lParamName);
                if ~isempty(lParamArtifact)
                    hBlk = find_system(hModel, 'BlockType', 'DataStoreMemory', 'DataStoreName', lName);
                    if numel(hBlk) == 1
                        set_param(hBlk, 'Dimensions', mat2str(lParamArtifact.Dimensions));
                    end
                    
                    [slObj, lWasFound] = lXRelTypeUtils.getSimulinkObject(lName);
                    if lWasFound
                        slObj.Dimensions = lParamArtifact.Dimensions;
                        lXRelTypeUtils.setSimulinkObject(lName, slObj);
                    end
                end
            end
        end
        
        function setHandleNeverReceived(lM3iPort, lM3iData)
            % SETHANDLENEVERRECEIVED
            lComSpec = autosar.mm.Model.findComSpecForDataElement(lM3iPort, lM3iData.Name, true);  % Always setting this on a read
            if ~isempty(lComSpec)
                lComSpec.HandleNeverReceived = true;
            end
        end
        
        function replaceTerminators(hModel)
            % REPLACETERMINATORS
            %   Replace terminators with a terminator connected to a signal
            %   conversion block - avoids being optimized away.
            lBlockType = 'Terminator';
            lToReplaceTag = '###TerminatorsToReplace';
            lReplacementBlock = 'pil_xrel_autosar_lib/SignalConvertingTerminator';
            load_system('pil_xrel_autosar_lib');
            coder.internal.xrel.AUTOSAR.AutosarImportModelManager.replaceBlocks(hModel, lBlockType, lToReplaceTag, lReplacementBlock);
        end

        
        function cleanUpEnums(hModel)
            % CLEANUPENUMS
            lBlocksWithEnum = find_system(hModel, 'Regexp', 'on', 'OutDataTypeStr', 'Enum:*');
            lBaseVars = evalin('base', 'who');
            for kPort = 1:numel(lBlocksWithEnum)
                lCurrPort = lBlocksWithEnum(kPort);
                lDataType = get_param(lCurrPort, 'OutDataTypeStr');
                if startsWith(lDataType, 'Enum:')
                    lEnumName = regexprep(lDataType, '^Enum:\s*', '');
                    if ismember(lEnumName, lBaseVars)
                        evalin('base', ['clearvars ', lEnumName]);
                    end        
                end
            end
        end
        
        function hXrelBlock = createModelBlock(lModelName, lDstFolder, lXILMode)
            % CREATEMODELBLOCK
            %
            %   Package the proxy model as a model block (do not protect)
            %
            lExistingSharedCode = get_param(lModelName, 'ExistingSharedCode');
            if ~isempty(lExistingSharedCode)
                set_param(lModelName, 'UtilityFuncGeneration', 'Shared Location');
            end
            save_system(lModelName);
            lModelSavedName = get_param(lModelName, 'FileName');
            
            % Create a model block in a blank system.
            hUntitledModel = new_system;
            hUntitledModelName = get_param(hUntitledModel, 'Name');
            
            hXrelBlock = add_block('built-in/ModelReference', ...
                [hUntitledModelName, '/', lModelName]);
            set_param(hXrelBlock, 'ModelName', lModelName);
            
            switch lXILMode
                case {'SIL'}
                    set_param(hXrelBlock, 'SimulationMode', 'Software-in-the-loop (SIL)');
                case 'PIL'
                    set_param(hXrelBlock, 'SimulationMode', 'Processor-in-the-loop (PIL)');
                otherwise
                    assert(false, 'lXILMode should be SIL or PIL');
            end
            set_param(hXrelBlock, 'CodeInterface', 'Top Model');
            
            % Close the model
            bdclose(lModelName);
            
            % Move the file
            movefile(lModelSavedName, lDstFolder);
        end
        
        function hXrelBlock = doProtect(lModelName, lDstFolder, lXILMode)
            % DOPROTECT
            %   Package the proxy model as a protected model
            %
            
            lWarnState = warning('off', 'MATLAB:class:RequireScalar');  % g1726957
            lWarnCleanup = onCleanup(@()warning(lWarnState));
            load_system(lModelName);
            lWarnCleanup.delete;
            
            lExistingSharedCode = get_param(lModelName, 'ExistingSharedCode');
            if ~isempty(lExistingSharedCode)
                set_param(lModelName, 'UtilityFuncGeneration', 'Shared Location');
            end
            save_system(lModelName);
            fhProtect = @()Simulink.ModelReference.protect(lModelName, ...
                'Mode', 'CodeGeneration', ...
                'CodeInterface', 'Top model', ...
                'OutputFormat', 'MinimalCode', ...
                'Harness', true, ...
                'Webview', false, ...   % This needs a Simulink Report Generator license, so must default to false
                'ObfuscateCode', false, ...
                'Path', lDstFolder);
            
            % Suppress model build information, consistent with ERT
            % workflow
            hProtHarness = ...
                coder.internal.xrel.AUTOSAR.AutosarImportModelManager.runCmdWithNoOutput(fhProtect);
            
            hXrelBlock = find_system(hProtHarness, 'BlockType', 'ModelReference');
            switch lXILMode
                case {'SIL'}
                    set_param(hXrelBlock, 'SimulationMode', 'Software-in-the-loop (SIL)');
                case 'PIL'
                    set_param(hXrelBlock, 'SimulationMode', 'Processor-in-the-loop (PIL)');
                otherwise
                    assert(false, 'lXILMode should be SIL or PIL');
            end
            bdclose(lModelName);
        end
        
        function lComponentName = getComponentNameForModel(lARXMLImporterObj, lXRelInfoComponentName)
            % GETCOMPONENTNAMEFORMODEL
            %   Get the fully qualified component name for the model
            %   described by the cross release artifact
            lComponents = lARXMLImporterObj.getComponentNames;
            lComponentName = lComponents{ ...
                strcmp(lComponents, lXRelInfoComponentName) ...             % Earlier releases serialize the fully qualified component name in CodeInfo.Name
                | endsWith(lComponents, ['/', lXRelInfoComponentName]) };   % More recently just the final part is used
            assert(~isempty(lComponentName), 'Could not find component %s', lXRelInfoComponentName);
        end
        
        function m3iModel = getM3iModelFromArxmlImporter(lARXMLImporterObj)
            % GETM3IMODELFROMARXMLIMPORTER
            %
            % Ensure that the m3iModel is loaded then get it.
            lARXMLImporterObj.getComponentNames;
            m3iModel = lARXMLImporterObj.getM3IModel;
        end
        
        function lARXMLImporterObj = getARXMLImporterFromArtifact(xrel_info)
            % GETARXMLIMPORTERFROMARTIFACT
            %   Process the cross-release artifact and find the arxml files
            %   in the build folder.  Create an ARXML importer object
            %   associated with these arxml files.
            %
            startDir = xrel_info.BuildDir;
            lARXMLImporterObj = coder.internal.xrel.AUTOSAR.AutosarImportModelManager.getARXMLImporterFromBuildDir(startDir);
            
        end
    end
       
    methods (Static, Access = public)
        function lARXMLImporterObj = getARXMLImporterFromBuildDir(startDir)
            % GETARXMLIMPORTERFROMBUILDDIR
            %
            % Searches over the arxml files in startDir and creates an
            % arxml.importer object from them.
            
            % Some interface arxml files live in the stub folder
            lDirARXMLFiles = [dir(fullfile(startDir, '*.arxml'));  ...
                dir(fullfile(startDir, 'stub', '*.arxml'))];
            % Don't want any folders ending in .arxml!
            assert(~isempty(lDirARXMLFiles));
            lDirARXMLFiles([lDirARXMLFiles.isdir]) = [];
            assert(~isempty(lDirARXMLFiles));
            
            % Convert to a cell array
            fhDirStructToFullPath = @(x)fullfile(x.folder, x.name);
            lARXMLFiles = arrayfun(fhDirStructToFullPath, lDirARXMLFiles, 'UniformOutput', false);
            lARXMLImporterObj = arxml.importer(lARXMLFiles);
        end         
            
        function isE2E = isE2E(m3iPort, m3iData)
            % ISE2E
            isE2E = false;
            lComSpec = autosar.mm.Model.findComSpecForDataElement(m3iPort, m3iData.Name, true);
            if ~isempty(lComSpec)
                isE2E = lComSpec.UsesEndToEndProtection;
            end
        end
    end
    
    
    methods (Access = private, Static)
        % Internal helper methods
        function replaceBlocks(hModel, lBlockType, lToReplaceTag, lReplacementBlock)
            % REPLACEBLOCKS
            lModelHandle = get_param(hModel, 'handle'); % Ensure is handle
            lModelName = get_param(hModel, 'Name');     % Ensure is name
            lToReplace = find_system(lModelHandle, 'LookUnderMasks', 'all', 'BlockType', lBlockType);
            for hBlock = lToReplace(:)'
                set_param(hBlock, 'Tag', lToReplaceTag);
            end
            replace_block(lModelName, 'Tag', lToReplaceTag, lReplacementBlock, 'noprompt');
        end
        
        function retVal = runCmdWithNoOutput(cmd) %#ok<INUSD>
            % runCmdWithNoOutput
            ME = [];
            retVal = [];
            output = evalc('try retVal = feval(cmd), catch ME, end');
            if ~isempty(ME)
                disp(output);
                throwAsCaller(ME);
            end
        end
    end
    
    methods (Access = public, Static)
        function safeCloseModel(hModel)
            % SAFECLOSEMODEL
            %   The model name changes when we re-save it at various
            %   points, so we end up with two potential modes of failure if
            %   we have a cleanup object that closes the model by name:
            %       * The model remains open upon failure (drool)
            %       * A new model is closed incorrectly if it
            %       shares the name of the original model (e.g., takes up
            %       the position in sequence of "untitled" models.
            %
            % bdclose does not return so we cannot chain using the &&
            % operator.  And bdclose will error if hModel is not a valid
            % diagram handle -- i.e. the model was already closed.
            %
            % Make static and public to avoid issues with destruction
            % order.
            if ishandle(hModel) && strcmp(get(hModel, 'Type'), 'block_diagram')
                bdclose(hModel);
            end 
        end
        
        function updateQueuedDataPorts(hModel, lUtils)
            % UPDATEQUEUEDDATAPORTS
            hBlocks = coder.internal.xrel.AUTOSAR.AutosarImportModelManager.getStateFlowBlockHandles(hModel);
            
            % Looks like this is only a problem for inports.  Queued
            % outports always need a message type.
            lPortNameList = coder.internal.xrel.AUTOSAR.AutosarImportModelManager.getNamesForQueuedPortsWithNonQueuedData(lUtils, 'Inports');
            for kBlock = 1:numel(hBlocks)
                [lBlockName, lBlockType] = ...
                    coder.internal.xrel.AUTOSAR.AutosarImportModelManager.getConnectedBlock( ...
                    hBlocks(kBlock), ...
                    'Inport', 1);
                if strcmp(lBlockType, 'Inport')  % As opposed to terminator or empty
                    if ismember(lBlockName, lPortNameList)
                        % Replace the stateflow blocks that do not
                        % correspond to message-typed data with signal
                        % conversion blocks.
                        lBlockToReplace = get(hBlocks(kBlock), 'Name');
                        lParent = get(hBlocks(kBlock), 'Parent');
                        replace_block(lParent, 'Name', lBlockToReplace, 'built-in/SignalConversion', 'noprompt');
                    end                    
                end
            end
        end
        
        function hBlocks = getStateFlowBlockHandles(hModel)
            % GETSTATEFLOWBLOCKHANDLES
            assert(ishandle(hModel));
            lModelName = get_param(hModel, 'Name');
            lSFRoot = sfroot;
            lSFModel = lSFRoot.find('-isa', 'Simulink.BlockDiagram', ...
                '-and','Name',lModelName);
            lBlocks = lSFModel.find('-isa','Stateflow.Chart');
            lBlockPaths = arrayfun(@(x)x.Path, lBlocks, 'UniformOutput', false);
            hBlocks = cellfun(@(x)get_param(x, 'handle'), lBlockPaths);
        end
        
        function [lBlockName, lBlockType] = getConnectedBlock(hBlock, lPortType, lPortNum)
            % GETCONNECTEDBLOCK
            lLineHandles = get_param(hBlock, 'LineHandles');
            lBlockName = '';
            lBlockType = '';
            switch lower(lPortType)
                case 'inport'
                    if lPortNum > numel(lLineHandles.Inport)
                        return;
                    end
                    hBlock = get(lLineHandles.Inport(lPortNum), 'SrcBlockHandle');
                case 'outport'
                    if lPortNum > numel(lLineHandles.Outport)
                        return;
                    end
                    hBlock = get(lLineHandles.Outport(lPortNum), 'DstBlockHandle');
                otherwise
                    assert(false, 'lPortType must be inport or outport')
            end
            lBlockName = get(hBlock, 'Name');
            lBlockType = get(hBlock, 'BlockType');
        end
        
        function lPortNames = getNamesForQueuedPortsWithNonQueuedData(lUtils, lPortType)
            % GETNAMESFORQUEUEDPORTSWITHNONQUEUEDDATA
            %
            %   Get the port name on the subsystem that models the
            %   runnable for ports that are Queued AUTOSAR data access
            %   modes, but don't correspond to a queued Simulink data type.
            %
            %   Note, we make an assumption about the naming scheme used by
            %   createComponentAsModel.
            %
            switch lPortType
                case 'Inports'
                    lArtifact = lUtils.getNonQueuedAutosarQueuedPorts('Inports');
                    if isempty(lArtifact)
                        lPortNames = {};
                    else
                        lPortNames = strcat({lArtifact.AutosarPort}, '_', {lArtifact.AutosarDataElement}, '_read');
                    end
                case 'Outports'
                    lArtifact = lUtils.getNonQueuedAutosarQueuedPorts('Outports');
                    if isempty(lArtifact)
                        lPortNames = {};
                    else
                        lPortNames = strcat({lArtifact.AutosarPort}, '_', {lArtifact.AutosarDataElement}, '_write');
                    end
                otherwise
                    assert(false, 'lPortType must be Inports or Outports');
            end
        end
       
        function lFcnCallMap = getFcnCallMap(hModel)
            % GETFCNCALLMAP
            lModelMapping = autosar.api.Utils.modelMapping(hModel);
            lFcnCallBlocks    = arrayfun(@(x)x.Block, lModelMapping.FcnCallInports, 'UniformOutput', false);
            lFcnCallRunnables = arrayfun(@(x)x.MappedTo.Runnable, lModelMapping.FcnCallInports, 'UniformOutput', false);
            if isempty(lFcnCallRunnables)
                lFcnCallMap = containers.Map.empty;
            else
                lFcnCallMap = containers.Map(lFcnCallRunnables, lFcnCallBlocks);
            end
        end
    end
end
