classdef BuildInfoBuilder < handle
    % BUILDINFOBUILDER
    %
    %   Helper class to handle population of RTW.BuildInfo objects for
    %   cross-release import of AUTOSAR models.
    %
    %   Copyright 2017-2018 The MathWorks, Inc.
   
    methods (Access = public, Static)
        
        function removeBuildDirAndModelSources(lBuildInfo)
            % REMOVEBUILDDIRANDMODELSOURCES
            for lKind = {'Src', 'Inc'}
                lGroup = get(lBuildInfo.(lKind{:}).Files, 'Group');
                lFileBuildDir = strcmp(lGroup, 'BuildDir') | strcmp(lGroup, 'ModelSources');
                lBuildInfo.(lKind{:}).Files(lFileBuildDir) = [];
            end
            
            % Suppress any linkobj sources
            lLinkObjGroup = get(lBuildInfo.LinkObj, 'Group');
            if isempty(lLinkObjGroup)
                lLinkObjGroup = {};
            elseif ~iscell(lLinkObjGroup)
                lLinkObjGroup = {lLinkObjGroup};
            end
            [lHasShared, lSharedIdx] = ismember(lLinkObjGroup, 'SHARED_SRC_LIB');
            
            if lHasShared
                lBuildInfo.LinkObj(lSharedIdx) = [];
            end
            
        end
            
        function lModelRefBuildInfo = getBuildInfoForModelRef(lRefModelName, lCodeGenFolder, lModelRefDir, lSharedCodeDir)
            % GETBUILDINFOFORMODELREF
            lModelRefBuildInfo = RTW.BuildInfo;
            lModelRefBuildInfo.ModelName = lRefModelName;
            lModelRefBuildInfo.Settings.LocalAnchorDir = lCodeGenFolder;
            lModelRefBuildInfo.addSourcePaths(lCodeGenFolder, 'StartDir');
            lModelRefBuildInfo.addSourcePaths(lModelRefDir, 'BuildDir');
            lModelRefBuildInfo.addIncludePaths(lSharedCodeDir, 'Standard');
        end
        
        function addBuildArgsForModelRef(lModelRefBuildInfo, lModelRefRelativePathToAnchor, lSharedUtilsTgtDir)
            % ADDBUILDARGSFORMODELREF
            lModelRefBuildInfo.addBuildArgs('SHARED_BIN_DIR', fullfile(lModelRefRelativePathToAnchor, lSharedUtilsTgtDir));
            lModelRefBuildInfo.addBuildArgs('SHARED_SRC_DIR', fullfile(lModelRefRelativePathToAnchor, lSharedUtilsTgtDir));
            lModelRefBuildInfo.addBuildArgs('SHARED_LIB', '');
            lModelRefBuildInfo.addBuildArgs('SHARED_SRC', '');
        end
        
        function addModelRefToTopModel(lBuildInfo, lModelRefBuildInfo, lModelRefDir)
            % ADDMODELREFTOTOPMODEL
            %
            % Paths
            lBuildInfo.addIncludePaths(lModelRefDir, 'BuildDir');
            
            % Then, we can look for ModelLib
            lBuildArgKey = get(lModelRefBuildInfo.BuildArgs, 'Key');
            lBuildArgValue = get(lModelRefBuildInfo.BuildArgs, 'Value');
            lModelRefLib = lBuildArgValue{strcmp(lBuildArgKey, 'MODELLIB')};
            lBuildInfo.addLinkObjects(lModelRefLib, lModelRefDir, [], [], [], 'MDLREF_LIB');
        end
            
        function addFilesToBuild(lModelBlockSources, lBuildFolder, lBuildInfo, lModelName, lDoCopy)
            % ADDFILESTOBUILD
            import coder.internal.xrel.AUTOSAR.BuildInfoBuilder
            
            lStubFolder = fullfile(lBuildFolder, 'stub');
            [lSourcePaths, lIncludePaths, ...
                lSourceFiles, lIncludeFiles, ...
                lSourceStubs, lIncludeStubs, ...
                lIncludeProxies] = lModelBlockSources.getData;
            
            % Write proxy include files to the RTE headers for files where
            % the custom code is expecting a header file with a different
            % name.
            for kFile = 1:numel(lIncludeProxies)
                BuildInfoBuilder.writeProxy(lIncludeProxies{kFile}, lStubFolder);
            end
            
            if lDoCopy
                % Copy source files and add to BuildInfo
                BuildInfoBuilder.copyToBuildFolder(lSourcePaths, lSourceFiles, lBuildFolder, lModelName);
                BuildInfoBuilder.copyToBuildFolder(lIncludePaths, lIncludeFiles, lBuildFolder, lModelName);
                BuildInfoBuilder.copyToBuildFolder(lSourcePaths, lSourceStubs, lStubFolder, lModelName);
                BuildInfoBuilder.copyToBuildFolder(lIncludePaths, lIncludeStubs, lStubFolder, lModelName);
                lBuildInfo.addSourceFiles(lSourceFiles, lBuildFolder, 'BuildDir');
                lBuildInfo.addSourceFiles(lSourceStubs, lStubFolder, 'BuildDir');
                lBuildInfo.addIncludeFiles(lIncludeFiles, lBuildFolder, 'BuildDir');
                lBuildInfo.addIncludeFiles(lIncludeStubs, lStubFolder, 'BuildDir');
            else
                lSourceFileFolders = BuildInfoBuilder.getFilePaths(lSourcePaths, lSourceFiles, lModelName);
                lIncludeFileFolders = BuildInfoBuilder.getFilePaths(lIncludePaths, lIncludeFiles, lModelName);
                lSourceStubFolders = BuildInfoBuilder.getFilePaths(lSourcePaths, lSourceStubs, lModelName);
                lIncludeStubFolders = BuildInfoBuilder.getFilePaths(lIncludePaths, lIncludeStubs, lModelName);
                lBuildInfo.addSourceFiles(lSourceFiles, lSourceFileFolders, 'BuildDir');
                lBuildInfo.addSourceFiles(lSourceStubs, lSourceStubFolders, 'BuildDir');
                lBuildInfo.addIncludeFiles(lIncludeFiles, lIncludeFileFolders, 'BuildDir');
                lBuildInfo.addIncludeFiles(lIncludeStubs, lIncludeStubFolders, 'BuildDir');
                lBuildInfo.addIncludePaths(lIncludePaths, 'BuildDir');
                lBuildInfo.addSourcePaths(lSourcePaths, 'BuildDir');
            end
            lBuildInfo.addIncludeFiles(lIncludeProxies, lStubFolder, 'BuildDir');
        end
        
        function copyToBuildFolder(lPaths, lFiles, lDest, lModelName)
            % Copy lFiles (cellstr, name only) from search path lPaths
            % (cellstr) to folder lDest (char)
            fhOpToPerform = @(lDirElem)coder.internal.xrel.AUTOSAR.BuildInfoBuilder.copyAndSetWriteable(lDirElem.folder, lDirElem.name, lDest);
            fhElseIfNotFound = @(currName)~(exist(fullfile(lDest, currName), 'file') == 2);
            coder.internal.xrel.AUTOSAR.BuildInfoBuilder.processFilePaths(lPaths, lFiles, lModelName, fhOpToPerform, fhElseIfNotFound);
        end
        
        function lRetVal = copyAndSetWriteable(lSrcFolder, lSrcName, lDestFolder)
            % COPYANDSETWRITEABLE
            lRetVal = copyfile(fullfile(lSrcFolder, lSrcName), lDestFolder);
            fileattrib(fullfile(lDestFolder, lSrcName), '+w');
        end
        
        function lRetVal = processFilePaths(lPaths, lFiles, lModelName, fhOpToPerform, fhElseIfNotFound)
            % Look over files and run fhOpToPerform on the dir entry
            lDir = coder.internal.xrel.AUTOSAR.BuildInfoBuilder.cellDir(lPaths);
            lDirNames = {lDir.name};
            nFiles = numel(lFiles);
            lRetVal = cell(1, nFiles);
            for kFile = 1:nFiles
                currName = lFiles{kFile};
                idx = find(strcmp(lDirNames, currName), 1, 'first');
                if ~isempty(idx)
                    lRetVal = fhOpToPerform(lDir(idx));
                elseif fhElseIfNotFound(currName)
                    DAStudio.error('PIL:xrelimport:RequiredFileNotFound', currName, lModelName);
                end
            end
        end
        
        function lFilePaths = getFilePaths(lPaths, lFiles, lModelName)
            % Get the path for each of the files and return as a cell array
            fhOpToPerform = @(lDirElem)lDirElem.folder;
            fhElseIfNotFound = @(currName)true;
            lFilePaths = coder.internal.xrel.AUTOSAR.BuildInfoBuilder.processFilePaths(lPaths, lFiles, lModelName, fhOpToPerform, fhElseIfNotFound);
        end
        
        function writeProxy(lFile, lFolder)
            % Write proxy header file lFile in folder lFolder
            lFileName = fullfile(lFolder, lFile);
            lWriter = rtw.connectivity.CodeWriter.create(...
                'language', 'C', ...
                'filename', lFileName, ...
                'callCBeautifier', true);
            lWriter.wLine('#include "Rte_Type.h"');
            lWriter.close;
        end
        
        function lDir = cellDir(lPathCell)
            % CELLDIR
            % Concatenated directory listing for all folders in cellstr
            % lPathCell
            lDir = struct('name', {}, 'folder', {}, 'date', {}, 'bytes', {}, 'isdir', {}, 'datenum', {});
            for kElem = 1:numel(lPathCell)
                lDir = [lDir; dir(lPathCell{kElem})]; %#ok<AGROW>
            end
        end
    end
end
