classdef EnumWrapper < handle
    % ENUMWRAPPER
    %   Wrapper to use with AUTOSAR cross-release import for C-Typed PIMs
    %   with enum types.
    %
    %   Stores the typedef information de-serialized from the cross-release
    %   artifact and can materialize the enum either as a
    %   Simulink.data.dictionary.EnumTypeDefinition to use in a Simulink
    %   Data Dictionary, or a session-based enum using
    %   Simulink.defineIntEnumType.
    %
    %   Copyright 2018 The MathWorks, Inc.
    properties (SetAccess = private)
        Strings;
        Values;
        DefaultMember;
        StorageType;
        HeaderFile;
    end
    
    methods (Access = public)
        function obj = EnumWrapper(lStrings, lValues, lDefaultMember, lStorageType, lHeaderFile)
            % ENUMWRAPPER
            obj.Strings = lStrings;
            obj.Values = lValues;
            obj.DefaultMember = lDefaultMember;
            obj.StorageType = lStorageType;
            obj.HeaderFile = lHeaderFile;
        end
        
        function materializeAsIntEnumType(obj, lName)
            % MATERIALIZEASINTENUMTYPE
            
            expandArgs = {};
            if ~isempty(obj.DefaultMember)
                expandArgs = [expandArgs, 'DefaultValue', obj.Strings{obj.DefaultMember}];
            end
            
            if ~isempty(obj.StorageType) && ~strcmp(obj.StorageType, 'uint32')
                % uint32 is the default and cannot be explicitly requested
                expandArgs = [expandArgs, 'StorageType', obj.StorageType];
            end
            
            if ~isempty(obj.HeaderFile)
                expandArgs = [expandArgs, 'HeaderFile', obj.HeaderFile];
            end
            
            Simulink.defineIntEnumType( ...
                lName, ...
                obj.Strings, ...
                obj.Values, ...
                expandArgs{:})
        end
        
        function lEnumDef = materializeAsEnumTypeDefinition(obj)
            % MATERIALIZEASENUMTYPEDEFINITION
            lEnumDef = Simulink.data.dictionary.EnumTypeDefinition;
            lEnumDef.clearEnumerals;
            
            for kVal = 1:numel(obj.Values)
                lEnumDef.appendEnumeral(char(obj.Strings{kVal}), obj.Values(kVal));
            end
            lEnumDef.HeaderFile = obj.HeaderFile;
            lEnumDef.DefaultValue = obj.Strings{obj.DefaultMember};
            lEnumDef.DataScope = 'Auto';
            
            if strcmp(obj.StorageType, 'int32')
                lEnumDef.StorageType = '';
            else
                lEnumDef.StorageType = obj.StorageType;
            end
        end
        
        function ret = getDifference(obj1, obj2)
            % GETDIFFERENCE
            %   Return a struct describing the difference between obj1 and
            %   obj2.
            ret = struct('Property', {}, ...
                'Expected', {}, ...
                'Found', {});
            
            if ~isequal(obj1.Strings(:)', obj2.Strings(:)')
                ret(end+1) = struct('Property', 'Strings', ...
                    'Expected', ['{', strjoin(obj2.Strings, ', '), '}'], ...
                    'Found', ['{', strjoin(obj1.Strings, ', '), '}']);
            end
            
            if ~isequal(obj1.Values(:)', obj2.Values(:)')
                ret(end+1) = struct('Property', 'Values', ...
                    'Expected', mat2str(obj2.Values), ...
                    'Found', mat2str(obj1.Values));
            end
            
            if ~isequal(obj1.DefaultMember, obj2.DefaultMember)
                ret(end+1) = struct('Property', 'DefaultMember', ...
                    'Expected', num2str(obj2.DefaultMember), ...
                    'Found', num2str(obj1.DefaultMember));
            end
            
            if ~isequal(obj1.StorageType, obj2.StorageType)
                ret(end+1) = struct('Property', 'StorageType', ...
                    'Expected', obj2.StorageType, ...
                    'Found', obj1.StorageType);
            end
            
            if ~isequal(obj1.HeaderFile, obj2.HeaderFile)
                ret(end+1) = struct('Property', 'HeaderFile', ...
                    'Expected', obj2.HeaderFile, ...
                    'Found', obj1.HeaderFile);
            end
            
        end
    end
    
    methods (Static)
        function obj = createFromStruct(lStruct, lHeaderFile)
            % CREATEFROMSTRUCT
            obj = coder.internal.xrel.AUTOSAR.EnumWrapper(lStruct.Strings, ...
                lStruct.Values, ...
                lStruct.DefaultMember, ...
                lStruct.StorageType, ...
                lHeaderFile);
        end
        
        function obj = createFromMetaClass(lMeta)
            % CREATEFROMMETACLASS
            
            assert(lMeta.Enumeration, 'lMeta must be an enumeration meta class')
            
            lEnumerators = enumeration(lMeta.Name);
            
            lStrings = arrayfun(@(x)x.char, lEnumerators, 'UniformOutput', false);
            lValues = arrayfun(@(x)x.double, lEnumerators, 'UniformOutput', true);
            
            if ismember('getDefaultValue', {lMeta.MethodList.Name})
                lDefaultValue = char(lEnumerators(1).getDefaultValue);
                lDefaultMember = find(strcmp(lStrings, lDefaultValue));
            else
                lDefaultMember = 1;
            end
            
            if ismember('getHeaderFile', {lMeta.MethodList.Name})
                lHeaderFile = char(lEnumerators(1).getHeaderFile);
            else
                lHeaderFile = '';
            end
            
            lSuperClasses = superclasses(lMeta.Name);
            
            lPermittedBase = {'int8', 'uint8', 'int16', 'uint16', 'int32', 'uint32'};
            lStorageType = 'int32';
            for lCurrPermittedBase = lPermittedBase
                if ismember(lCurrPermittedBase{:}, lSuperClasses)
                    lStorageType = lCurrPermittedBase{:};
                    break;
                end
            end            
            obj = coder.internal.xrel.AUTOSAR.EnumWrapper(lStrings(:)', ...
                lValues(:)', ...
                lDefaultMember, ...
                lStorageType, ...
                lHeaderFile);
        end
        
        function obj = createFromEnumTypeDefinition(lEnumTypeDefinition)
            % CREATEFROMENUMTYPEDEFINITION
            lStrings = {lEnumTypeDefinition.Enumerals.Name};
            lValues = cellfun(@str2double, {lEnumTypeDefinition.Enumerals.Value});
            lDefaultValue = lEnumTypeDefinition.DefaultValue;
            lDefaultMember = find(strcmp(lStrings, lDefaultValue));
            if isempty(lDefaultMember)
                lDefaultMember = 1;
            end
            lHeaderFile = lEnumTypeDefinition.HeaderFile;
            
            if isempty(lEnumTypeDefinition.StorageType)
                lStorageType = 'int32';
            else
                lStorageType = lEnumTypeDefinition.StorageType;
            end
            obj = coder.internal.xrel.AUTOSAR.EnumWrapper(lStrings(:)', ...
                lValues(:)', ...
                lDefaultMember, ...
                lStorageType, ...
                lHeaderFile); %#ok<FNDSB>
            
        end
    end
end