classdef FcnCallerUtils
    % FCNCALLERUTILS
    %
    %   Utilities to handle unflattening dimensions in Function Caller
    %   blocks in cross-release AUTOSAR workflows
    %
    %   Copyright 2018 The MathWorks, Inc.
    
    properties (Constant, Access = private)
        MatchOnes = 'ones\(\s*\S+\s*,\s*((?!\))\S)+\s*\)';
        MatchParam = '(?<=\s*)P_\S*';
        MatchEnum  = 'repmat\(\s*\S*\(\s*\S*\s*\)\s*,\s*((?!\))\S)+\s*,\s*((?!\))\S)+\s*\)';
        MatchEnumDim = '(?<=(repmat\(\s*\S*\(\s*\S*\s*\)\s*,))\s*((?!\))\S)+\s*,\s*((?!\))\S)+\s*\)';
    end
    
    methods (Static)
        function argSpecOut = processArgumentSpecifications(argSpecIn, dimList, argIdx, lDataDictionary)
            
            % argSpecIn is the arg spec, as generated by createComponentAsModel
            % dimList is the list of dimensions in the same order as in CodeInfo and the function prototype in the generated code
            % argIdx is an array containing the "graphical" position in the Simulink Function for each arg in the function prototype
            
            import coder.internal.xrel.AUTOSAR.FcnCallerUtils
            
            lXRelTypeUtils = coder.internal.xrel.XrelTypeUtils(lDataDictionary);
            
            argToks = FcnCallerUtils.tokenizeArgumentSpecifications(argSpecIn);
            assert(numel(dimList) == numel(argToks), 'Expected %d arg specs but found %d', numel(dimList), numel(argToks));
            for kArg = 1:numel(argToks)
                if FcnCallerUtils.containsExplicitDim(argToks{kArg})
                    % If the spec is of form "ones(....)" then replace the
                    % dimensions in place.
                    argToks{kArg} = FcnCallerUtils.replaceDims(argToks{kArg}, dimList{kArg});
                elseif FcnCallerUtils.isParamName(argToks{kArg})
                    argToks{kArg} = FcnCallerUtils.updateParam(argToks{kArg}, dimList{kArg}, lXRelTypeUtils);
                elseif FcnCallerUtils.isEnum(argToks{kArg})
                    argToks{kArg} = FcnCallerUtils.updateEnum(argToks{kArg}, dimList{kArg});
                end
            end            
            argSpecOut = FcnCallerUtils.reassembleArgumentSpecifications(argToks, argIdx);
            
        end
    end
    
    methods (Static, Access = public)
        function lArgToks = tokenizeArgumentSpecifications(argSpecIn)
            % Split the argument spec string into a cell array of the
            % individual argument specs.
            %
            % We want to split on commas that are not inside parentheses.
            % This is not straightforward with a single regexp, so do it
            % iteratively.
            %
            % Start by splitting on every comma
            lDelim = ',';
            lSubTokens = split(argSpecIn, lDelim);
            
            % Re-join Sub-Tokens that are between unmatched parentheses.
            lArgToks = {};
            lCurrTok = '';
            for kSubTok = 1:numel(lSubTokens)
                lCurrTok = [lCurrTok, lSubTokens{kSubTok}]; %#ok<AGROW>
                lNumOpen = numel(strfind(lCurrTok, '('));
                lNumClose = numel(strfind(lCurrTok, ')'));
                if lNumOpen == lNumClose
                    lArgToks{end+1} = lCurrTok; %#ok<AGROW>
                    lCurrTok = '';
                else
                    lCurrTok = [lCurrTok, lDelim]; %#ok<AGROW>
                end
            end
            
            if numel(lArgToks) == 1 && isempty(lArgToks{1})
                lArgToks = {};
            end   
        end
        
        function ret = containsExplicitDim(argTok)
            import coder.internal.xrel.AUTOSAR.FcnCallerUtils
            ret = ~isempty(regexp(argTok, FcnCallerUtils.MatchOnes, 'once'));
        end
        
        function ret = isParamName(argTok)
            import coder.internal.xrel.AUTOSAR.FcnCallerUtils
            ret = ~isempty(regexp(argTok, FcnCallerUtils.MatchParam, 'once'));
        end
        
        function ret = isEnum(argTok)
            import coder.internal.xrel.AUTOSAR.FcnCallerUtils
            ret = ~isempty(regexp(argTok, FcnCallerUtils.MatchEnum, 'once'));
        end
        
        function ret = replaceDims(argTok, dims)
            % Get a string containing a comma separated list of the
            % dimensions.
            import coder.internal.xrel.AUTOSAR.FcnCallerUtils
            dimStr = strjoin( ...
                arrayfun(@(x)sprintf('%d', x), dims, 'UniformOutput', false), ...
                ',');
            onesStr = sprintf('ones(%s)', dimStr);
            ret = regexprep(argTok, FcnCallerUtils.MatchOnes, onesStr);
        end
        
        function ret = updateParam(argTok, dims, lXRelTypeUtils)
            import coder.internal.xrel.AUTOSAR.FcnCallerUtils
            
            % The spec might be a reference to a parameter.
            % createComponentAsModel prefixes these with P.
            
            % We don't need to do anything if the correct dimension
            % specification is a 1D array, or a 1D column vector.
            if (numel(dims) == 1) || (numel(dims) == 2 && dims(2) == 1)
                ret = argTok;
                return;
            end
            
            % Otherwise, we need to reshape, but we need to
            % rename too, because there is no guarantee that
            % this is the only instance of the parameter!
            
            % Get a deblanked instance of the parameter name
            lParamName = regexp(argTok, FcnCallerUtils.MatchParam, 'match', 'once');
            [lObj, lWasFound] = lXRelTypeUtils.getSimulinkObject(lParamName);
            if ~lWasFound || ~isa(lObj, 'Simulink.Parameter')
                % We can't do anything
                return;
            end
            
            % Copy the existing object
            lNewObj = lObj.copy;
            
            % Update the dimensions
            lNewObj.Value = reshape(lNewObj.Value, dims);
            lDimStr = strjoin( ...
                arrayfun(@(x)sprintf('%d', x), dims, 'UniformOutput', false), ...
                '_');
            
            % Create a new name
            lNewName = sprintf('%s_DIM_%s', lParamName, lDimStr);
            
            % Assign in workspace or SLDD
            lXRelTypeUtils.setSimulinkObject(lNewName, lNewObj);
            
            ret = lNewName;
        end
        
        
        function ret = updateEnum(argTok, dims)
            % Replace the string [1 1 1 1 1] etc... with ones(....)
            import coder.internal.xrel.AUTOSAR.FcnCallerUtils
            dimStr = strjoin( ...
                arrayfun(@(x)sprintf('%d', x), dims, 'UniformOutput', false), ...
                ',');
            
            replaceStr = sprintf(' %s)', dimStr);
            ret = regexprep(argTok, FcnCallerUtils.MatchEnumDim, replaceStr);
        end
        
        function argSpecOut = reassembleArgumentSpecifications(argToksIn, argIdx)
            
            % Re-order the args when re-assembling into a string
            argToks = cell(1, numel(argToksIn));
            argToks(argIdx) = argToksIn;
            
            if numel(argToks) == 0
                argSpecOut = '';
            elseif numel(argToks) == 1
                argSpecOut = argToks{1};
            else
                argSpecOut = strjoin(argToks, ',');
            end
        end
    end
    
end