classdef ModelRefSubModelManager < handle
    % MODELREFSUBMODELMANAGER

%   Copyright 2017-2018 The MathWorks, Inc.
    
    properties (Access = private)
        ModelRefs
        ModelRefRoot
        ModelRefSimRoot
        CodeGenFolder
        SharedUtilsTgtDir
        ConfigSet
        DoCopy
    end
    
    properties (SetAccess = private)
        IsPortableWordSizesEnabled
        ModelRefRelativePathToAnchor
        NumModelRefs
    end
    
    properties (Constant, Access = private)
        RTEHeaderPath = fullfile(autosarroot, 'rte');
        RTEHeaderFiles = {'Platform_Types.h', 'Std_Types.h', 'Compiler.h'};
    end
    
    methods
        function obj = ModelRefSubModelManager(lModelRefs, ...
                lModelRefRoot, ...
                lModelRefSimRoot, ...
                lCodeGenFolder, ...
                lSharedUtilsTgtDir, ...
                lModelRefRelativeBuildDir, ...
                lDoCopy, ...
                lConfigSet)
            % MODELREFSUBMODELMANAGER
            obj.ModelRefs = lModelRefs;
            obj.ModelRefRoot = lModelRefRoot;
            obj.ModelRefSimRoot = lModelRefSimRoot;
            obj.CodeGenFolder = lCodeGenFolder;
            obj.SharedUtilsTgtDir = lSharedUtilsTgtDir;
            obj.DoCopy = lDoCopy;
            obj.ConfigSet = lConfigSet;
            
            obj.IsPortableWordSizesEnabled = strcmp(get_param(lConfigSet, 'PortableWordSizes'), 'on');
            
            lCellOfDots = ...
                repmat({'..'}, 1, numel(strfind(lModelRefRelativeBuildDir, filesep))+1);
            obj.ModelRefRelativePathToAnchor = fullfile(lCellOfDots{:});            
            obj.NumModelRefs = numel(obj.ModelRefs);
        end
        
        function processModelRefs(obj, lTopModelName, lTopModelBuildInfo)
            % PROCESSMODELREFS
            [lModelRefNameList, ...
                lModelRefRelDirList] = obj.prepareModelRefs(lTopModelBuildInfo);
            obj.addModelRefIncludesToBuildInfo(obj.ModelRefs, lTopModelBuildInfo);
            obj.updateProtectingModelWithModelRefs( ...
                lTopModelName, ...
                lModelRefNameList, ...
                lModelRefRelDirList)
        end
        
        
        function [lModelRefNameList, lModelRefRelDirList] = prepareModelRefs(obj, lTopModelBuildInfo)
            % PREPAREMODELREFS
            lModelRefNameList = cell(1, obj.NumModelRefs);
            lModelRefRelDirList = cell(1, obj.NumModelRefs);
            for kModelRef = 1:obj.NumModelRefs
                [lModelRefNameList{kModelRef}, ...
                    lModelRefRelDirList{kModelRef}] = ...
                    obj.prepareSingleModelRef(obj.ModelRefs(kModelRef), lTopModelBuildInfo);
            end
        end
    end
    
    methods (Access = private)
        
        function [lRefName, lRelModelRefDir] = prepareSingleModelRef(obj, lModelRef, lTopModelBuildInfo)
            % PREPARESINGLEMODELREF
            import coder.internal.xrel.AUTOSAR.ModelBlockSourcesHelper;
            import coder.internal.xrel.AUTOSAR.BuildInfoBuilder;
            
            lRefName = obj.getImportedModelRefName(lModelRef);
            lModelRefDir = obj.getModelRefDir(lRefName);
            lSharedCodeDir = fullfile(obj.CodeGenFolder, obj.SharedUtilsTgtDir);
            if ~isfolder(lModelRefDir)
                mkdir(lModelRefDir);
            end
            lModelRefSources = ModelBlockSourcesHelper.createFromModelRefArtifactReader(lModelRef);
            
            % Initialize an RTW.BuildInfo for the referenced model
            lModelRefBuildInfo = BuildInfoBuilder.getBuildInfoForModelRef( ...
                lRefName, ...
                obj.CodeGenFolder, ...
                lModelRefDir, ...
                lSharedCodeDir);
            
            % Set important BuildArgs that need to exist for CodeInstrBuildArgs
            % to succeed.
            BuildInfoBuilder.addBuildArgsForModelRef(lModelRefBuildInfo, ...
                obj.ModelRefRelativePathToAnchor, ...
                obj.SharedUtilsTgtDir);
            
            % Now add the cross-release source files.
            BuildInfoBuilder.addFilesToBuild( ...
                lModelRefSources, lModelRefDir, ...
                lModelRefBuildInfo, lRefName, obj.DoCopy);
            
            % We need to add includes for any child model refs
            %
            obj.addModelRefIncludesToBuildInfo(lModelRef.ChildSubModelHandles, lModelRefBuildInfo);
            
            % Add path to static stub RTE header files to BuildInfo
            obj.addStaticRTEStubHeaderFilePath(lModelRefBuildInfo);
            
            % Use coder.internal.doCompile to create the build artifacts
            % that will need to be packaged.
            coder.internal.doCompile(lModelRefBuildInfo, ...
                obj.ConfigSet, lModelRefDir, ...
                'ModelReferenceTargetType', 'RTW', ...
                'generateCodeOnly', true);
            
            % Sort out the project file
            coder.internal.generateRtwProjFile...
                (lRefName, '', ...
                lModelRefDir, ...
                pwd, ...
                '..', ...
                'InfoMATFileName', 'tmwinternal/binfo_mdlref.mat');
            
            % Create a stubbed binfo file for the model ref
            lRelModelRefDir = fullfile(obj.ModelRefRoot, lRefName);
            lRelModelRefSimDir = fullfile(obj.ModelRefSimRoot, lRefName);
            coder.internal.xrel.AUTOSAR.BInfoUtils.writeStubBinfoForSubModel( ...
                obj.IsPortableWordSizesEnabled, ...
                obj.ModelRefRelativePathToAnchor, ...
                obj.CodeGenFolder, ...
                lRelModelRefDir, ...
                lRelModelRefSimDir);
            
            % We now need to make sure we can include the header file
            % when building the top model.
            BuildInfoBuilder.addModelRefToTopModel(lTopModelBuildInfo, lModelRefBuildInfo, lModelRefDir);
        end
        
        function lModelRefDir = getModelRefDir(obj, lRefName)
            % GETMODELREFDIR
            lModelRefDir = fullfile(obj.CodeGenFolder, obj.ModelRefRoot, lRefName);
        end
        
        function addModelRefIncludesToBuildInfo(obj, lModelRefs, lBuildInfo)
            % ADDMODELREFINCLUDESTOBUILDINFO
            
            % Derive the model ref names from the cross-release artifacts
            lChildModelNames = arrayfun(@(x)obj.getImportedModelRefName(x),  ...
                lModelRefs, ...
                'UniformOutput', false);
            
            % Derive the model ref dir from the model ref names
            lChildModelDirs = cellfun(@(x)obj.getModelRefDir(x), ...
                lChildModelNames, ...
                'UniformOutput', false);
            
            % Add to BuildInfo
            if ~isempty(lChildModelDirs)
                lBuildInfo.addIncludePaths(lChildModelDirs, 'Standard');
            end
        end
        
        function addStaticRTEStubHeaderFilePath(obj, lModelRefBuildInfo)
            % ADDSTATICRTESTUBHEADERFILEPATH
            lModelRefBuildInfo.addIncludePaths(obj.RTEHeaderPath, '');
            lModelRefBuildInfo.addIncludeFiles(obj.RTEHeaderFiles, obj.RTEHeaderPath, '');
        end
    end
    
    methods (Static, Access = private)
        function updateProtectingModelWithModelRefs(lModelName, lModelRefNameList, lModelRefRelDirList)
            % UPDATEPROTECTINGMODELWITHMODELREFS
            if Simulink.ModelReference.ProtectedModel.protectingModel(lModelName)
                % This is really not an example of good architecture, but I
                % have no choice.
                pmCreator = get_param(lModelName, 'ProtectedModelCreator');
                
                % Refs must come first to avoid link errors
                pmCreator.SubModels = [lModelRefNameList, pmCreator.SubModels];
            end
            
            % We need to patch infoStruct with the modelrefs
            coder.internal.xrel.AUTOSAR.BInfoUtils.setModelRefs( ...
                lModelName, ...
                lModelRefNameList, ...
                lModelRefRelDirList);
        end
    end
    
    methods (Static, Access = public)
        function lRefName = getImportedModelRefName(lModelRef)
            % GETIMPORTEDMODELREFNAME
            [~, lRefName] = fileparts(lModelRef.XrelArtifactSource);
        end
        
        function lCleanupObjs =  createEmptyModelRefModels(lModelRefs)
            % CREATESTUBMODELREFS
            lCleanupObjs = onCleanup.empty;
            for kModelRef = 1:numel(lModelRefs)
                lRefName = coder.internal.xrel.AUTOSAR.ModelRefSubModelManager.getImportedModelRefName(lModelRefs(kModelRef));
                if bdIsLoaded(lRefName)
                    MException(message('autosarstandard:xrelimport:ModelAlreadyExists', lRefName)).throw;
                end
                hTemp = new_system(lRefName);
                set_param(hTemp, 'SystemTargetFile', 'autosar.tlc');
                lCleanupObjs(kModelRef) = onCleanup(@()bdclose(lRefName));
            end
        end
    end
end
