classdef PIMNVRAMUtils < handle
    % PIMNVRAMUTILS
    %
    %   Helper class for dealing with per-instance memories that access
    %   NVRAM.
    %
    %   Copyright 2018 The MathWorks, Inc.
    methods (Static, Access = public)
        
        function writePIMFileSource(lModelName, lBuildFolder, lNVRAMPIMInitCondMap, m3iComp)
            % Write the RTE source stub for the cross-release XIL
            % interface.
            lASWCName = m3iComp.Name;
            lPimFileSource = coder.internal.xrel.AUTOSAR.PIMNVRAMUtils.getPIMFileSource(lASWCName, lBuildFolder);
            
            if isfile(lPimFileSource)
                if coder.internal.xrel.AUTOSAR.PIMNVRAMUtils.isMultiInstance(lModelName)
                    lFcnArgs = 'Rte_Instance self';
                else
                    lFcnArgs = 'void';
                end
                
                % Replace what was originally in the file generated by TLC
                lWriter = rtw.connectivity.CodeWriter.create(...
                    'callCBeautifier', true,...
                    'filename', lPimFileSource,...
                    'append', false);
                
                m3iBehavior = m3iComp.Behavior;
                for kPIM = 1:m3iBehavior.ArTypedPIM.size
                    m3iArTypedPIM = m3iBehavior.ArTypedPIM.at(kPIM);
                    lPIMName = m3iArTypedPIM.Name;
                    if lNVRAMPIMInitCondMap.isKey(lPIMName)
                        
                        % Get variable and function names
                        lParamName = lNVRAMPIMInitCondMap(lPIMName);
                        lParamDataVarName = coder.internal.xrel.AUTOSAR.PIMNVRAMUtils.getRTEInternalCalPrmDataVar(lParamName);
                        lPIMDataVarName = coder.internal.xrel.AUTOSAR.PIMNVRAMUtils.getRTEPIMDataVar(lPIMName);
                        lInitFcnName = coder.internal.xrel.AUTOSAR.PIMNVRAMUtils.getRTEPIMInitFcnName(lModelName, lPIMName);
                        
                        m3iParameterData = coder.internal.xrel.AUTOSAR.PIMNVRAMUtils.getParameterByName(m3iBehavior.Parameters, lParamName);
                        [lDims, lIsMatrix] = coder.internal.xrel.AUTOSAR.PIMNVRAMUtils.getMatrixDims(m3iParameterData);
                        lSize = prod(lDims);
                        
                        % Write the initialization function
                        lWriter.wBlockStart('void %s(%s)', lInitFcnName, lFcnArgs);
                        if lIsMatrix
                            for kElem = 1:lSize
                                lWriter.wLine('%s[%d] = %s[%d];', lPIMDataVarName, kElem-1, lParamDataVarName, kElem-1);
                            end
                        else
                            lWriter.wLine('%s = %s;', lPIMDataVarName, lParamDataVarName);
                        end
                        lWriter.wBlockEnd;
                    end
                end
                lWriter.close;
            end
        end
        
        function lParamName = getParamNameFromServiceDepForArTypedPimByName(m3iBehavior, lName)
            % Returns the parameter used for the initial conditions of the
            % Ar Typed PIM with name lName
            m3iServiceDependency = coder.internal.xrel.AUTOSAR.PIMNVRAMUtils.getServiceDependencyForArTypedPimByName(m3iBehavior, lName);
            if ~isempty(m3iServiceDependency) && ~isempty(m3iServiceDependency.UsedParameterElement)
                lParamName = m3iServiceDependency.UsedParameterElement.Name;
            else
                lParamName = '';
            end
        end
        
        function ret = getServiceDependencyForArTypedPimByName(m3iBehavior, lName)
            % Return the m3i service dependency object for the Ar Typed PIM
            % with Name lName.
            ret = [];
            for kServiceDependency = 1:m3iBehavior.ServiceDependency.size()
                m3iServiceDependency = m3iBehavior.ServiceDependency.at(kServiceDependency);
                if ~isempty(m3iServiceDependency.UsedDataElement) && strcmp(m3iServiceDependency.UsedDataElement.Name, lName)
                    ret = m3iServiceDependency;
                    return;
                end
            end
        end
        
        function ret = getSubstituteInitialValueParamName(lPIMName)
            ret = [lPIMName, '_XRel_NvM_Init'];
        end
    end
    
    methods (Access = private, Static)
        function lPimFileSource = getPIMFileSource(lASWCName, lBuildFolder)
            % Return the full path to the RTE source file for NVRAM PIM
            % initialization, that would normally be generated in TLC
            lRteDir = autosar.mm.mm2rte.RTEGenerator.getRTEFilesFolder(lBuildFolder);
            lPimFileSource = autosar.mm.mm2rte.ASWCWriter.getRtePimSourceFileName(...
                lRteDir, lASWCName);
        end
        
        function ret = isMultiInstance(lModelName)
            % Returns true if the model is multi-instance
            ret = strcmp(get_param(lModelName, 'CodeInterfacePackaging'), 'Reusable function');
        end
        
        function ret = getRTEInternalCalPrmDataVar(lParamName)
            % Return the name of the internal variable used for the
            % InternalCalPrm % with name paramName in the stub RTE
            ret = ['Rte_CData_' lParamName '_data'];
        end
        
        function ret = getRTEPIMDataVar(lPIMName)
            % Return the name of the internal variable used for the
            % per-instance memory data in the stub RTE
            ret = ['Rte_Pim_' lPIMName '_data'];
        end
        
        function ret = getRTEPIMInitFcnName(lModelName, lPIMName)
            % Return the function name for the PIM initialization function
            % in the stub RTE.
            ret = [lModelName, '_Init_Pim_', lPIMName];
        end
        
        function [lDims, lIsMatrix] = getMatrixDims(m3iParameterData)
            % Returns the dimensions and if the parameter matrix typed
            lType = m3iParameterData.Type;
            lIsMatrix = isa(lType, 'Simulink.metamodel.types.Matrix');
            if lIsMatrix
                lDims = ones(1, lType.Dimensions.size);
                for kDim = 1:lType.Dimensions.size
                    lDims(kDim) = double(lType.Dimensions.at(kDim));
                end
            else
                lDims = 1;
            end
        end
        
        function ret = getParameterByName(m3iParameters, lName)
            % Return the parameter with name lName
            assert(isa(m3iParameters, 'Simulink.metamodel.arplatform.interface.SequenceOfParameterData'));
            for kParam = 1:m3iParameters.size
                if strcmp(m3iParameters.at(kParam).Name, lName)
                    ret = m3iParameters.at(kParam);
                    return;
                end
            end
            
            % When this method is called, it will already be known that
            % a parameter with name lName does exist.
            assert(false, 'Parameter %s not found', lName);
        end
    end
end