classdef TypeUtils < handle
    % TYPEUTILS
    %
    %   Helper class for cross-release AUTOSAR workflows to instantiate
    %   type objects for C Typed PIMs, when these do not already exist,
    %   since unlike for other types, they are not created by createComponentAsModel.
    %   
    %   Copyright 2018 The MathWorks, Inc.
    
    properties (Constant, Access = private)
        BuiltIns = {'double', 'single', 'int8', 'uint8', 'int16', 'uint16', 'int32', 'uint32', 'boolean'};
        DefaultHeaderFile = 'Rte_Type.h';
    end
    
    properties (Access = private)
        XRelArtifactUtils
        XRelTypeUtils
    end
    
    
    methods (Access = public)
        function obj = TypeUtils(lXRelArtifactUtils, lDataDictionary)
            % TYPEUTILS
            obj.XRelArtifactUtils = lXRelArtifactUtils;
            obj.XRelTypeUtils = coder.internal.xrel.XrelTypeUtils(lDataDictionary);
        end
        
        function createTypeObjectsIfNecessary(obj, lTypeName)
            % CREATETYPEOBJECTIFNECESSARY
            % Build the type anyway - in case there are dependent type
            % names
            [lNewTypeObj, lDependentTypeNames] = obj.buildTypeObject(lTypeName);
            
            if isempty(lNewTypeObj)
                % No new type object created - type is a built-in or
                % undefined
                return;
            end
            
            [lTypeExists, lOrigTypeObj] = obj.validateTypeExists(lTypeName);
            if lTypeExists
                % We probably need to validate the type if it exists.
                obj.validateType(lTypeName, lOrigTypeObj, lNewTypeObj);
            else
                if isa(lNewTypeObj, 'coder.internal.xrel.AUTOSAR.EnumWrapper')
                    if isempty(obj.XRelTypeUtils.DataDictionary)
                        lNewTypeObj.materializeAsIntEnumType(lTypeName);
                    else
                        obj.XRelTypeUtils.setSimulinkObject(lTypeName, lNewTypeObj.materializeAsEnumTypeDefinition);
                    end
                else
                    obj.XRelTypeUtils.setSimulinkObject(lTypeName, lNewTypeObj);
                end
            end
            
            % Make dependent type names unique
            lDependentTypeNames = unique(lDependentTypeNames);
            
            % Recurse dependent types
            for kDepType = 1:numel(lDependentTypeNames)
                obj.createTypeObjectsIfNecessary(lDependentTypeNames{kDepType});
            end
        end
        
        function validateType(obj, lTypeName, lOrigTypeObj, lNewTypeObj)
            % VALIDATETYPE
            %
            % The existing object might be in a hierarchy.
            lAliasedThruToName = obj.getAliasedThruToName(lTypeName);
            [~, lAliasedThruToTypeObj] = obj.validateTypeExists(lAliasedThruToName);
            
            if isa(lNewTypeObj, 'Simulink.AliasType')
                % If the new type object is an alias type then:
                %   (1) The "new" hierarchy will only be one level deep;
                %   (2) The base type will be a built-in
                %
                % Therefore, we must validate that:
                %   (1) The "old" type is a Simulink.AliasType or a
                %       Simulink.NumericType;
                if ~obj.isAliasOrNumeric(lOrigTypeObj)
                    MException( ...
                        message('autosarstandard:xrelimport:ExpectedNumericOrAliasDataTypeObject', ...
                        lTypeName, ...
                        class(lOrigTypeObj))).throw;
                end
                
                %   (2) It is aliased through to the same built-in base
                %       type as we are expecting.
                if ~strcmp(lAliasedThruToName, lNewTypeObj.BaseType)
                    MException( ...
                        message('autosarstandard:xrelimport:ExpectedAliasDataTypeObjectAliasedThruTo', ...
                        lTypeName, ...
                        lAliasedThruToName, ...
                        lNewTypeObj.BaseType)).throw;
                end
            elseif isa(lNewTypeObj, 'Simulink.NumericType')
                % The new type object is a fixed point type.
                %
                % We must validate that the "old" type is:
                %   (1) Defined by the correct type of object, or
                %   aliased through to the correct type of object
                obj.validateAliasedThruToObjectClass( ...
                    lTypeName, ...
                    lOrigTypeObj, ...
                    lAliasedThruToName, ...
                    lAliasedThruToTypeObj, ...
                    'Simulink.NumericType');
                
                %   (2) an equivalent fixed point type.
                obj.validateFixedTypeEquivalence(lAliasedThruToTypeObj, lNewTypeObj, lTypeName);
                
            elseif isa(lNewTypeObj, 'Simulink.Bus')
                % The new type object is a bus type
                %
                % We must validate that the "old" type is:
                %   (1) a bus type
                obj.validateAliasedThruToObjectClass( ...
                    lTypeName, ...
                    lOrigTypeObj, ...
                    lAliasedThruToName, ...
                    lAliasedThruToTypeObj, ...
                    'Simulink.Bus');
                
                %   (2) an equivalent bus type
                obj.validateBusTypeEquivalence(lAliasedThruToTypeObj, lNewTypeObj, lTypeName);
            elseif isa(lNewTypeObj, 'coder.internal.xrel.AUTOSAR.EnumWrapper')
                if isa(lOrigTypeObj, 'Simulink.AliasType')
                    lTypeObjToCompare = lAliasedThruToTypeObj;
                else
                    lTypeObjToCompare = lOrigTypeObj;
                end
                
                % This could be the result of following through the alias
                if isa(lTypeObjToCompare, 'Simulink.data.dictionary.EnumTypeDefinition')
                    lTypeObjToCompare = coder.internal.xrel.AUTOSAR.EnumWrapper.createFromEnumTypeDefinition(lTypeObjToCompare);
                end
                
                if ~isa(lTypeObjToCompare, 'coder.internal.xrel.AUTOSAR.EnumWrapper')
                    MException( ...
                        message('autosarstandard:xrelimport:MismatchedEnumerationClass', ...
                        lTypeName, ...
                        class(lTypeObjToCompare))).throw;
                end
                
                lMismatch = getDifference(lTypeObjToCompare, ...   % Found
                    lNewTypeObj ...                           % Expected
                    );
                obj.throwExceptionWithCausesIfMismatch( ...
                    lMismatch, ...
                    'autosarstandard:xrelimport:MismatchedEnumerationDefinition', ...
                    lTypeName); 
            end
        end
    end
    
    methods (Access = private)    
        function lAliasedThruToName = getAliasedThruToName(obj, lTypeName)
            % GETALIASEDTHRUTONAME
            %   Follow through Simulink aliased types to a built-in or
            %   undefined base type.
            lTypeName = obj.XRelTypeUtils.stripPrefixes(lTypeName);
            [slObj, wasFound] = obj.XRelTypeUtils.getSimulinkObject(lTypeName);
            if wasFound && isa(slObj, 'Simulink.AliasType') && ~strcmp(slObj.BaseType, lTypeName)
                % If it's an alias, then check a further level (as long as
                % the alias is not an alias to itself - which is not valid,
                % but could lead to an infinite recursion)
                lAliasedThruToName = obj.getAliasedThruToName(slObj.BaseType);
            elseif wasFound && ...
                    isa(slObj, 'Simulink.NumericType') && ...
                    ismember(slObj.DataTypeMode, {'Double', 'Single', 'Boolean'})
                    % Then pass through the base built-in type
                    lAliasedThruToName = lower(slObj.DataTypeMode);
            else
                lAliasedThruToName = lTypeName;
            end
        end
        
        function [lTypeObj, lDependentTypeNames] = buildTypeObject(obj, lTypeName)
            % BUILDTYPEOBJECT
            lDependentTypeNames = {};
            [lTypeSpec, lIsAlias, lIsFix, lIsStruct, lIsEnum] = obj.XRelArtifactUtils.getTypeByName(lTypeName);
            lTypeObj = [];
            if isempty(lTypeSpec)
                return;
            end
            if lIsAlias
                lTypeObj = obj.buildAliasTypeObject(lTypeSpec);
            elseif lIsStruct
                [lTypeObj,lDependentTypeNames] = obj.buildStructTypeObject(lTypeSpec);
            elseif lIsFix
                lTypeObj = obj.buildFixTypeObject(lTypeSpec);
            elseif lIsEnum
                lTypeObj = obj.buildEnumTypeDefinitionObject(lTypeSpec);
            end
        end
        
        function lTypeObj = buildAliasTypeObject(obj, lTypeSpec)
            % BUILDALIASTYPEOBJECT
            lTypeObj = Simulink.AliasType;
            lTypeObj.BaseType = lTypeSpec.StorageType;
            lTypeObj.HeaderFile = obj.DefaultHeaderFile;
        end
        
        function [lTypeObj, lDependentTypeNames] = buildStructTypeObject(obj, lTypeSpec)
            % BUILDSTRUCTTYPEOBJECT
            lDependentTypeNames = {};
            lTypeObj = Simulink.Bus;
            for kElem = 1:numel(lTypeSpec.Elements)
                lTempElem = Simulink.BusElement;
                lTempElem.Name = lTypeSpec.Elements(kElem).Identifier;
                lTempElem.Complexity = lTypeSpec.Elements(kElem).Complexity;
                lTempElem.Dimensions = lTypeSpec.Elements(kElem).Dimensions;
                lElemDataType = lTypeSpec.Elements(kElem).DataType;
                [~, lElemIsAlias, lElemIsFix,lElemIsStruct, lElemIsEnum] = obj.XRelArtifactUtils.getTypeByName(lElemDataType);
                if lElemIsStruct
                    lUpdatedElemDataType = ['Bus: ', lElemDataType];
                elseif lElemIsEnum
                    lUpdatedElemDataType = ['Enum: ', lElemDataType];
                else
                    lUpdatedElemDataType = lElemDataType;
                end
                lTempElem.DataType = lUpdatedElemDataType;
                if lElemIsAlias || lElemIsFix || lElemIsStruct || lElemIsEnum
                    lDependentTypeNames{end+1} = lElemDataType; %#ok<AGROW>
                end
                lTypeObj.Elements(end+1) = lTempElem;
                lTypeObj.HeaderFile = obj.DefaultHeaderFile;
            end
        end
        
        function lTypeObj = buildFixTypeObject(obj, lTypeSpec)
            % BUILDFIXTYPEOBJECT
            lTypeObj = eval(lTypeSpec.StringReadyForEval);
            lTypeObj.IsAlias = true;
            lTypeObj.HeaderFile = obj.DefaultHeaderFile;
        end
        
        
        function lTypeObj = buildEnumTypeDefinitionObject(obj, lTypeSpec)
            % BUILDENUMTYPEDEFINITIONOBJECT
            %   Create a wrapper for our Enum type, because none of the
            %   existing Enum descriptor objects do quite what we need.
            lTypeObj = coder.internal.xrel.AUTOSAR.EnumWrapper.createFromStruct(lTypeSpec, obj.DefaultHeaderFile);
        end
        
        function [lTypeExists, lTypeObj] = validateTypeExists(obj, lTypeName)
            % VALIDATETYPEEXISTS
            % Is the type a built-in?
            if ismember(lTypeName, obj.BuiltIns)
                % If so then all is good.
                lTypeObj = [];
                lTypeExists = true;
                return;
            end
            
            % Eval the type in the model's workspace
            lTypeObj = obj.XRelTypeUtils.getSimulinkObject(lTypeName);
            if isempty(lTypeObj)
                if isempty(obj.XRelTypeUtils.DataDictionary)
                    % If no data dictionary, look for enumerations -
                    % enumerationg in a data dictionary will have an object
                    % that will be returned
                    if isempty(enumeration(lTypeName))
                        lTypeExists = false;
                        return;
                    else
                        lTypeObj = coder.internal.xrel.AUTOSAR.EnumWrapper.createFromMetaClass(meta.class.fromName(lTypeName));
                        lTypeExists = true;
                        return;
                    end
                else
                    lTypeExists = false;
                end
            else
                lTypeExists = true;
            end
        end
        
        function validateFixedTypeEquivalence(obj, lOrigTypeObj, lNewTypeObj, lTypeName)
            % VALIDATEFIXEDTYPEEQUIVALENCE
            lPropertiesToValidate = { ...
                'WordLength', '%d'
                'FractionLength', '%d'
                'Signedness', '%s'
                'Slope', '%5.4f'
                'Bias', '%5.4f'
                };
            lMismatch = obj.validateProperties(lOrigTypeObj, lNewTypeObj, lPropertiesToValidate);
            obj.throwExceptionWithCausesIfMismatch( ...
                lMismatch, ...
                'autosarstandard:xrelimport:MismatchedFixedPointType', ...
                lTypeName);
        end
        
        function validateBusTypeEquivalence(obj, lOrigTypeObj, lNewTypeObj, lTypeName)
            % VALIDATEBUSTYPEEQUIVALENCE
            
            % Validate list of bus element names
            lOrigNames = {lOrigTypeObj.Elements.Name};
            lNewNames  = {lNewTypeObj.Elements.Name};
            
            if ~isequal(lOrigNames, lNewNames)
                lException = MException( ...
                    message('autosarstandard:xrelimport:MismatchedBusElementNames', ...
                    lTypeName, ...
                    strjoin(lOrigNames, ', '), ...
                    strjoin(lNewNames, ', ')));
                lException.throw;
            end
            
            lPropertiesToValidate = { ...
                'Complexity', '%s'
                'Dimensions', 'mat2str'
                'DataType', 'datatype'            % Type name must be exact - not just an equivalent/alias
                };
            
            for kElem = 1:numel(lOrigTypeObj.Elements)
                lOrigElem = lOrigTypeObj.Elements(kElem);
                lNewElem  = lNewTypeObj.Elements(kElem);
                lMismatch = obj.validateProperties(lOrigElem, lNewElem, lPropertiesToValidate);
                obj.throwExceptionWithCausesIfMismatch( ...
                    lMismatch, ...
                    'autosarstandard:xrelimport:MismatchedBusElement', ...
                    lOrigElem.Name, ...
                    lTypeName);
            end
        end
    
        function lMismatch = validateProperties(obj, lOrigTypeObj, lNewTypeObj, lPropertiesToValidate)
            % VALIDATEPROPERTIES
            lMismatch = struct('Property', {}, ...
                'Expected', {}, ...
                'Found', {});
            for kProp = 1:size(lPropertiesToValidate, 1)
                lCurrProp = lPropertiesToValidate{kProp, 1};
                lCurrFormatString = lPropertiesToValidate{kProp, 2};
                
                % Take copies, because we will be post-processing enum,
                % bus, etc.
                lOrigPropToCompare = lOrigTypeObj.(lCurrProp);
                lNewPropToCompare  = lNewTypeObj.(lCurrProp);
                
                if strcmp(lCurrFormatString, 'datatype')
                    % Format Bus: and Enum: prefix in standard way
                    lOrigPropToCompare = coder.internal.xrel.AUTOSAR.TypeUtils.standardizeBusEnumPrefix(lOrigPropToCompare);
                    
                    % If the string is a datatype then we need to see if
                    % it's been aliased through to something else
                    if ~startsWith(lOrigPropToCompare, {'Enum:', 'Bus:'})
                        lOrigPropToCompare = obj.getAliasedThruToName(lOrigPropToCompare);
                    end
                    if ~startsWith(lNewPropToCompare, {'Enum:', 'Bus:'})
                        lNewPropToCompare = obj.getAliasedThruToName(lNewPropToCompare);
                    end
                end
                
                if ~isequal(lOrigPropToCompare, lNewPropToCompare)
                    if strcmp(lCurrFormatString, 'mat2str')
                        lExpected = mat2str(lNewTypeObj.(lCurrProp));
                        lFound = mat2str(lOrigTypeObj.(lCurrProp));
                    elseif strcmp(lCurrFormatString, 'datatype')
                        lExpected = lNewTypeObj.(lCurrProp);
                        lFound = lOrigTypeObj.(lCurrProp);
                    else
                        lExpected = sprintf(lCurrFormatString, lNewTypeObj.(lCurrProp));
                        lFound = sprintf(lCurrFormatString, lOrigTypeObj.(lCurrProp));
                    end
                    
                    lMismatch(end+1) = struct( ...
                        'Property', lCurrProp, ...
                        'Expected', lExpected, ...
                        'Found', lFound); %#ok<AGROW>
                end
            end
        end
    end
    
    
    methods (Access = private, Static)
        function bRet = isAliasOrNumeric(lTypeObj)
            % ISALIASORNUMERIC
            bRet = isa(lTypeObj, 'Simulink.AliasType') || isa(lTypeObj, 'Simulink.NumericType');
        end
        
        
        function throwExceptionWithCausesIfMismatch(lMismatch, lCatalogId, varargin)
            % THROWEXCEPTIONIFMISMATCH
            if ~isempty(lMismatch)
                lException = MException(message(lCatalogId, varargin{:}));
                for kCause = 1:numel(lMismatch)
                    lCause = MException( ...
                        message('autosarstandard:xrelimport:MismatchedTypeProperty', ...
                        lMismatch(kCause).Property, ...
                        lMismatch(kCause).Expected, ...
                        lMismatch(kCause).Found));
                    lException = lException.addCause(lCause);
                end
                lException.throw;
            end
        end
        
        function validateAliasedThruToObjectClass( ...
                lTypeName, ...
                lOrigTypeObj, ...
                lAliasedThruToName, ...
                lAliasedThruToTypeObj, ...
                lPermittedClass)
            % VALIDATEALIASEDTHRUTOOBJECTCLASS
            
            if isa(lOrigTypeObj, 'Simulink.AliasType')
                if ~isa(lAliasedThruToTypeObj, lPermittedClass)
                    MException( ...
                        message('autosarstandard:xrelimport:ExpectedAliasDataTypeObjectAliasedThruToClass', ...
                        lTypeName, ...
                        lAliasedThruToName, ...
                        class(lAliasedThruToTypeObj), ...
                        lPermittedClass)).throw;
                end
            elseif ~isa(lOrigTypeObj, lPermittedClass)
                MException( ...
                        message('autosarstandard:xrelimport:ExpectedDataTypeObject', ...
                        lTypeName, ...
                        class(lOrigTypeObj), ...
                        lPermittedClass)).throw;
            end
        end
    end
    
    methods (Access = public, Static)
        function lDataTypeName = standardizeBusEnumPrefix(lDataTypeName)
            % STANDARDIZEBUSENUMPREFIX
            % Format Bus: and Enum: prefix in standard way
            lDataTypeName = regexprep(lDataTypeName, '^\s*Bus:\s*', 'Bus: ');
            lDataTypeName = regexprep(lDataTypeName, '^\s*Enum:\s*', 'Enum: ');
            lDataTypeName = regexprep(lDataTypeName, '^\s*\?\s*', 'Enum: ');     
        end
    end
end