classdef XRelArtifactUtils < handle
    % XRELARTIFACTUTILS
    %
    %   Wrapper around cross-release artifact providing search/find
    %   functionality needed when building CodeInfo for AUTOSAR.
    %
    %   Copyright 2017-2018 The MathWorks, Inc.
    %
    properties (Access = private)
        XRelInfo;   % Raw cross-release artifact
    end
    
    methods (Access = public)
        function obj = XRelArtifactUtils(xrel_info)
            % XRELARTIFACTUTILS
            %   Wrap the cross-release artifact
            obj.XRelInfo = xrel_info;
            
            % For compatibility with artifacts created with earlier
            % versions of crossReleaseExport.
            if isfield(obj.XRelInfo, 'SimulinkFunctions') && ~isempty(obj.XRelInfo.SimulinkFunctions)
                if isfield(obj.XRelInfo.SimulinkFunctions, 'FunctionReturn')
                    for kFunction = 1:numel(obj.XRelInfo.SimulinkFunctions)
                        if ~isfield(obj.XRelInfo.SimulinkFunctions(kFunction).FunctionReturn, 'ArgIndex')
                            nOutputs = numel(obj.XRelInfo.SimulinkFunctions(kFunction).FunctionOutputs);
                            for kReturn = 1:numel(obj.XRelInfo.SimulinkFunctions(kFunction).FunctionReturn)
                                % If there's a return argument and the
                                % artifact does not contain ArgIndex,
                                % insert the default return arg position as
                                % O0.  (0-indexed).
                                obj.XRelInfo.SimulinkFunctions(kFunction).FunctionReturn(kReturn).ArgIndex = sprintf('O%d', nOutputs + kReturn-1);
                            end
                        end
                    end
                else
                    for kFunction = 1:numel(obj.XRelInfo.SimulinkFunctions)
                        obj.XRelInfo.SimulinkFunctions(kFunction).FunctionReturn = [];                        
                    end
                end
                
                % Handle case where artifact didn't have function arg
                % indices serialized - assume no FPC active
                %
                % In this case, insert the default argument positions.
                for kFunction = 1:numel(obj.XRelInfo.SimulinkFunctions)
                    if ~isfield(obj.XRelInfo.SimulinkFunctions(kFunction).FunctionInputs, 'ArgIndex')
                        for kInput = 1:numel(obj.XRelInfo.SimulinkFunctions(kFunction).FunctionInputs)
                            obj.XRelInfo.SimulinkFunctions(kFunction).FunctionInputs(kInput).ArgIndex = sprintf('I%d', kInput-1);
                        end
                    end
                end
                
                
                for kFunction = 1:numel(obj.XRelInfo.SimulinkFunctions)
                    if ~isfield(obj.XRelInfo.SimulinkFunctions(kFunction).FunctionOutputs, 'ArgIndex')
                        for kOutput = 1:numel(obj.XRelInfo.SimulinkFunctions(kFunction).FunctionOutputs)
                            obj.XRelInfo.SimulinkFunctions(kFunction).FunctionOutputs(kOutput).ArgIndex = sprintf('O%d', kOutput-1);
                        end
                    end
                end
                
            end
        end
        
        function ret = isExplicitTaskingForRunnable(obj, lRunnableSymbol)
            % ISEXPLICITTASKINGFORRUNNABLE
            % Does this runnable symbol have a function caller input associated
            % with it?
            fhIsTask = @(x)strcmp(x.Symbol, lRunnableSymbol);
            lElem = obj.searchArtifact('OutputFunctions', fhIsTask);
            ret = strcmp(lElem.TaskingMode, 'EXPLICIT_TASKING') || ...
                strcmp(lElem.TimingMode, 'ASYNCHRONOUS');
        end
        
        function ret = getSampleOffsetForRunnable(obj, lRunnableSymbol)
            % GETSAMPLEOFFSETFORRUNNABLE
            fhIsTask = @(x)strcmp(x.Symbol, lRunnableSymbol);
            lElem = obj.searchArtifact('OutputFunctions', fhIsTask);
            assert(~isempty(lElem), 'Could not find runnable with symbol %s', lRunnableSymbol);
            ret = lElem.SampleOffset;
        end
        
        function ret = getPriorityForRunnable(obj, lRunnableSymbol)
            % GETPRIORITYFORRUNNABLE
            fhIsTask = @(x)strcmp(x.Symbol, lRunnableSymbol);
            lElem = obj.searchArtifact('OutputFunctions', fhIsTask);
            assert(~isempty(lElem), 'Could not find runnable with symbol %s', lRunnableSymbol);
            ret = lElem.Priority;
        end
        
        function ret = getInternalCalPrmDimensions(obj, lName)
            % GETINTERNALCALPRMDIMENSIONS
            fh = @(x)isfield(x, 'AutosarPort') && isfield(x, 'DataAccessMode') ...
                && strcmp(x.AutosarPort, lName) && strcmp(x.DataAccessMode, 'InternalCalPrm');
            lElem = obj.searchArtifact('Parameters', fh);
            assert(~isempty(lElem));
            ret = lElem.Dimensions;
        end
        
        function ret = getDataStoreByGraphicalName(obj, lName)
            % GETDATASTOREBYNAME
            fh = @(x)isfield(x, 'GraphicalName') && strcmp(x.GraphicalName, lName);
            ret = obj.searchArtifact('DataStores', fh);
        end
        
        function [lXrelPortStruct, lIdx] = getXRelInfoFromPortAndDataElement(obj, lField, lPort, lDataElement, lDataAccess)
            % GETXRELINFOFROMPORTANDDATALEMENT
            if ~iscell(lDataAccess)
                lDataAccess = {lDataAccess};
            end
            
            % Special case for error status ports where we need a level of
            % indirection, since the error status port only has the
            % original receiver port index to identify it.
            if numel(lDataAccess) == 1 && strcmp(lDataAccess{1}, 'ErrorStatus')
                fhIsMatch = @(x)strcmp(x.AutosarPort, lPort) && ...
                    strcmp(x.AutosarDataElement, lDataElement);
                lReceiverPortStruct = obj.searchArtifact(lField, fhIsMatch);
                assert(~isempty(lReceiverPortStruct), 'Could not find port');
                
                % Now find the port index by graphical name
                lIdx = getExpInportIdxFromGraphicalName(obj, lReceiverPortStruct.GraphicalName);
                fhIsMatch = @(x)strcmp(x.AutosarReceiverPortNumber, num2str(lIdx));
                [lXrelPortStruct, lIdx] = obj.searchArtifact(lField, fhIsMatch);
                
            else
                fhIsMatch = @(x)strcmp(x.AutosarPort, lPort) && ...
                    strcmp(x.AutosarDataElement, lDataElement) && ...
                    ismember(x.DataAccessMode, lDataAccess);
                [lXrelPortStruct, lIdx] = obj.searchArtifact(lField, fhIsMatch);
            end
        end
        
        function lIdx = getExpInportIdxFromGraphicalName(obj, lGraphicalName)
            % GETEXPINPORTIDXFROMGRAPHICALNAME
            fhIsMatch = @(x)strcmp(x.GraphicalName, lGraphicalName);
            [~, lIdx] = obj.searchArtifact('ExternalInportMapping', fhIsMatch);
            assert(~isempty(lIdx), 'Could not find port');
        end
        
        function lInportNames = getDataInportGraphicalNames(obj)
            % GETDATAINPORTGRAPHICALNAMES
            %   Get data input ports (i.e., not the function callers)
            lInports = obj.XRelInfo.Inports;
            if isfield(lInports, 'GraphicalName')
                lInportNames = {lInports.GraphicalName};
            else
                lInportNames = {};
            end
        end
        
        function lOutportNames = getDataOutportGraphicalNames(obj)
            % GETDATAOUTPORTGRAPHICALNAMES
            lOutports = obj.XRelInfo.Outports;
            if isfield(lOutports, 'GraphicalName')
                lOutportNames = {lOutports.GraphicalName};
            else
                lOutportNames = {};
            end
        end
        
        function ret = getNumPorts(obj, lPortType)
            switch lPortType
                case 'Inport'
                    ret = numel(obj.XRelInfo.ExternalInportMapping);
                case 'Outport'
                    ret = numel(obj.XRelInfo.Outports);
                otherwise
                    assert(false, 'lPortType must be Inport or Outport')
            end
        end
        
        function ret = isFunctionCall(obj, lPortType, kPort)
            % ISFUNCTIONCALL
            switch lPortType
                case 'Inport'
                    ret = obj.XRelInfo.ExternalInportMapping(kPort).IsFunctionCall;
                case 'Outport'
                    ret = false;
                otherwise
                    assert(false, 'lPortType must be Inport or Outport')
            end
        end
        
        function ret = getPortSymbol(obj, lPortType, kPort)
            % GETPORTSYMBOL
            assert(strcmp(lPortType, 'Inport'), 'getPortSymbol only for Inports');
            ret = obj.XRelInfo.ExternalInportMapping(kPort).Symbol;
        end
        
        
        function lElem = getOutputFunctionFromInport(obj, lPortType, kPort)
            % GETOUTPUTFUNCTIONFROMINPORT
            assert(strcmp(lPortType, 'Inport'), ...
                'lPortType must be Inport for getting runnable name');
            lSymbol = obj.getPortSymbol(lPortType, kPort);
            
            fh = @(x)strcmp(x.Symbol, lSymbol);
            lElem = obj.searchArtifact('OutputFunctions', fh);
            assert(~isempty(lElem), 'Could not find OutputFunction with Symbol %s', lSymbol);
        end
        
        function lElem = getDataPortElemWithIndirection(obj, lPortType, kPort)
            % GETDATAPORTELEMWITHINDIRECTION
            switch lPortType
                case 'Inport'
                    assert(kPort <= numel(obj.XRelInfo.ExternalInportMapping), ...
                        'Index %d is greater than the number of ExternalInportMapping (%d) in the artifact', ...
                        kPort, numel(obj.XRelInfo.ExternalInportMapping));
                    assert(~obj.XRelInfo.ExternalInportMapping(kPort).IsFunctionCall, 'getAutosarPortMapping does not apply to function call ports');
                    lGraphicalName = obj.XRelInfo.ExternalInportMapping(kPort).GraphicalName;
                    
                    % Look up by graphical name
                    lElem = obj.getInportFromGraphicalName(lGraphicalName);
                case 'Outport'
                    assert(kPort <= numel(obj.XRelInfo.Outports), ...
                        'Index %d is greater than the number of Outports (%d) in the artifact', ...
                        kPort, numel(obj.XRelInfo.Outports));
                    lElem = obj.XRelInfo.Outports(kPort);
                otherwise
                    assert(false, 'lPortType must be Inport or Outport');
            end
        end
        
        function [lPort, lDataElement, lDataAccessMode, lGraphicalName] = getAutosarPortMapping(obj, lPortType, kPort)
            % GETAUTOSARPORTMAPPING
            lElem = obj.getDataPortElemWithIndirection(lPortType, kPort);
            if strcmp(lElem.DataAccessMode, 'ErrorStatus')
                % For error status ports, return the port and data element of
                % the corresponding receiver port
                %
                % Error status ports will have an AutosarReceiverPortNumber
                % that we need to cross-reference
                lParentPort = str2double(lElem.AutosarReceiverPortNumber);
                lParentElem = obj.getDataPortElemWithIndirection(lPortType, lParentPort);
                lPort = lParentElem.AutosarPort;
                lDataElement = lParentElem.AutosarDataElement;
                lDataAccessMode = lElem.DataAccessMode; % The ErrorStatus port, not its parent
                lGraphicalName = lElem.GraphicalName;   % The ErrorStatus graphical name, not its parent
            else
                lPort = lElem.AutosarPort;
                lDataElement = lElem.AutosarDataElement;
                lDataAccessMode = lElem.DataAccessMode;
                lGraphicalName = lElem.GraphicalName;
            end
        end
        
        function ret = getNonQueuedAutosarQueuedPorts(obj, lPortType)
            % GETNONQUEUEDAUTOSARQUEUEDPORTS
            %   Get the ports that:
            %       * Have a Queued AUTOSAR Data Access mode
            %       * Nave a non-queued Simulink data type
            fh = @(x)isfield(x, 'DataAccessMode') && contains(x.DataAccessMode, 'Queued') && ~x.IsQueued;
            ret = obj.XRelInfo.(lPortType)(arrayfun(fh, obj.XRelInfo.(lPortType)));
        end
        
        function [lDimensions, lIsBus, lIsSymbolicDimension] = getDimensions(obj, lPortType, kPort)
            % GETDIMENSIONS
            lElem = obj.getDataPortElemWithIndirection(lPortType, kPort);
            lDimensions = lElem.Dimensions;
            lIsBus = lElem.IsBus;
            lIsSymbolicDimension = isinf(lDimensions);
        end

        
        function lGraphicalName = getGraphicalName(obj, lPortType, kPort)
            % GETGRAPHICALNAME
            switch lPortType
                case 'Inport'
                    assert(kPort <= numel(obj.XRelInfo.ExternalInportMapping), ...
                        'Index %d is greater than the number of ExternalInportMapping (%d) in the artifact', ...
                        kPort, numel(obj.XRelInfo.ExternalInportMapping));
                    lGraphicalName = obj.XRelInfo.ExternalInportMapping(kPort).GraphicalName;
                case 'Outport'
                    assert(kPort <= numel(obj.XRelInfo.Outports), ...
                        'Index %d is greater than the number of Outports (%d) in the artifact', ...
                        kPort, numel(obj.XRelInfo.Outports));
                    lGraphicalName = obj.XRelInfo.Outports(kPort).GraphicalName;
                otherwise
                    assert(false, 'lPortType must be Inport or Outport')
            end
        end
        
        function lInportElem = getInportFromGraphicalName(obj, lGraphicalName)
            % GETINPORTFROMGRAPHICALNAME
            fh = @(x)strcmp(x.GraphicalName, lGraphicalName);
            lInportElem = obj.searchArtifact('Inports', fh);
            assert(~isempty(lInportElem), 'Could not find Inport with GraphicalName %s', lGraphicalName);
        end
        
        function ret = getStructTypes(obj)
            % GETSTRUCTTYPES
            ret = obj.XRelInfo.StructTypes;
        end
        
        function ret = getParameters(obj)
            % GETPARAMETERS
            ret = obj.XRelInfo.Parameters;
        end
        
        function ret = getParameterByName(obj, lName)
            % GETPARAMETERBYNAME
            fh = @(x)strcmp(x.GraphicalName, lName);
            ret = obj.searchArtifact('Parameters', fh);
        end
        
        function ret = getParameterByImplementation(obj, lPort, lElem, lDataAccess)
            % GETPARAMETERBYIMPLEMENTATION
            fh = @(x)strcmp(x.AutosarPort, lPort) && ...
                strcmp(x.AutosarDataElement, lElem) && ...
                strcmp(x.DataAccessMode, lDataAccess);
            ret = obj.searchArtifact('Parameters', fh);
        end
        
        function ret = getStructTypeByName(obj, lName)
            % GETSTRUCTTYPEBYNAME
            fh = @(x)isfield(x, 'Name') && strcmp(x.Name, lName);
            ret = obj.searchArtifact('StructTypes', fh);
        end
        
        function [ret, isAlias, isFix, isStruct, isEnum] = getTypeByName(obj, lName)
             % GETTYPEBYNAME
             fh = @(x)isfield(x, 'Name') && strcmp(x.Name, lName);
             
             isAlias = false;
             isFix = false;
             isStruct = false;
             isEnum = false;
             
             ret = obj.searchArtifact('AliasTypes', fh);
             if ~isempty(ret)
                 isAlias = true;
                 return;
             end
             ret = obj.searchArtifact('FixTypes', fh);
             if ~isempty(ret)
                 isFix = true;
                 return;
             end
             ret = obj.searchArtifact('StructTypes', fh);
             if ~isempty(ret)
                 isStruct = true;
                 return;
             end
             ret = obj.searchArtifact('EnumTypes', fh);
             if ~isempty(ret)
                 isEnum = true;
             end
        end
        
        function ret = getStatesDWorkName(obj, lStateType)
            % GETSTATESDWORKNAME
            switch lStateType
                case 'Block states'
                    ret = obj.XRelInfo.BlockStatesDWorkName;
                case 'Block signals'
                    ret = obj.XRelInfo.SignalStatesDWorkName;
                case 'Zero crossing states'
                    ret = obj.XRelInfo.ZCStatesDWorkName;
                otherwise
                    assert(false, 'Invalid lStateType - should be Block states, Block signals or Zero crossing states');
            end
        end
        
        function ret = getStatesDWorkType(obj, lStateType)
            % GETSTATESDWORKTYPE
            switch lStateType
                case 'Block states'
                    ret = obj.XRelInfo.BlockStatesDWorkType;
                case 'Block signals'
                    ret = obj.XRelInfo.SignalStatesDWorkType;
                case 'Zero crossing states'
                    ret = obj.XRelInfo.ZCStatesDWorkType;
                otherwise
                    assert(false, 'Invalid lStateType - should be Block states, Block signals or Zero crossing states');
            end
        end
        
        function ret = getStatesDWorkTag(obj, lStateType)
            % GETSTATESDWORKTAG
            switch lStateType
                case 'Block states'
                    ret = obj.XRelInfo.BlockStatesDWorkTag;
                case 'Block signals'
                    ret = obj.XRelInfo.SignalStatesDWorkTag;
                case 'Zero crossing states'
                    ret = obj.XRelInfo.ZCStatesDWorkTag;
                otherwise
                    assert(false, 'Invalid lStateType - should be Block states, Block signals or Zero crossing states');
            end
        end
        
        function ret = definesSimulinkFunctions(obj)
            % DEFINESSIMULINKFUNCTIONS
            ret = isfield(obj.XRelInfo, 'SimulinkFunctions') && ...
                ~isempty(obj.XRelInfo.SimulinkFunctions) && ...
                any(strcmp({obj.XRelInfo.SimulinkFunctions.IsDefined}, 'yes'));
        end
        
        
        function lPeriodicByFcnCall = hasExplicitPeriodicTasking(obj)
            % HASEXPLICITPERIODICTASKING
            lPeriodicByFcnCall = false;
            % In releases <= R2016a, periodic triggers could be
            % registered with ASYNCHRONOUS timing mode, but would have
            % a sample period > 0.
            for kRunnable = 1:numel(obj.XRelInfo.OutputFunctions)
                lRunnable = obj.XRelInfo.OutputFunctions(kRunnable);
                if strcmp(lRunnable.TimingMode, 'PERIODIC') && strcmp(lRunnable.TaskingMode, 'EXPLICIT_TASKING')
                    lPeriodicByFcnCall = true;
                    break;
                elseif strcmp(lRunnable.TimingMode, 'ASYNCHRONOUS') && (lRunnable.SamplePeriod ~= -1)
                    lPeriodicByFcnCall = true;
                    break;
                end
            end
        end
        
        function ret = getAutosarInitRunnableSymbol(obj)
            % GETAUTOSARINITRUNNABLESYMBOL
            if isfield(obj.XRelInfo, 'AutosarInitRunnableSymbol')
                ret = obj.XRelInfo.AutosarInitRunnableSymbol;
            else
                ret = '';
            end
        end
        
        function ret = getAutosarTerminateRunnableSymbol(obj)
            % GETAUTOSARTERMINATERUNNABLESYMBOL
            if isfield(obj.XRelInfo, 'AutosarTerminateRunnableSymbol')
                ret = obj.XRelInfo.AutosarTerminateRunnableSymbol;
            else
                ret = '';
            end
        end
        
        function ret = getAutosarResetRunnableSymbols(obj)
            % GETAUTOSARRESETRUNNABLESYMBOLS
            lIsReset = arrayfun(@(x)strcmp(x.TimingMode, 'RESET'), obj.XRelInfo.OutputFunctions);
            ret = arrayfun(@(x)x.Symbol, ...
                obj.XRelInfo.OutputFunctions(lIsReset), ...
                'UniformOutput', false);
        end
        
        function ret = hasErrorStatusPort(obj, lPortType, lPortName, lDataElement)
            % HASERRORSTATUSPORT
            lDataAccess = 'ErrorStatus';
            lXrelPortStruct = obj.getXRelInfoFromPortAndDataElement(lPortType, lPortName, lDataElement, lDataAccess);
            
            % Return true if an error status port was found for the
            % PortName/DataElement combination
            ret = ~isempty(lXrelPortStruct);
        end
        
        function ret = getSimulinkFunctions(obj)
            % GETSIMULINKFUNCTIONS
            if isfield(obj.XRelInfo, 'SimulinkFunctions')
                ret = obj.XRelInfo.SimulinkFunctions;
            else
                ret = [];
            end
        end
        
        function ret = getSimulinkFunctionByCGFunctionName(obj, lCGFunctionName)
            % GETSIMULINKFUNCTIONBYCGFUNCTIONNAME
            ret = [];
            if isfield(obj.XRelInfo, 'SimulinkFunctions')
                fhCriterion = @(x)strcmp(x.CGFunctionName, lCGFunctionName);
                ret = obj.searchArtifact('SimulinkFunctions', fhCriterion);
            end
        end
        
        function ret = getTriggeredPeriodicRunnableSymbols(obj)
            % GETTRIGGEREDPERIODICRUNNABLESYMBOLS
            if isfield(obj.XRelInfo, 'TriggeredPeriodicRunnableSymbols')
                ret = obj.XRelInfo.TriggeredPeriodicRunnableSymbols;
            else
                ret = {};
            end
        end
        
    end
    
    methods (Access = private)
        function [ret, idx] = searchArtifact(obj, lField, fhCriterion)
            % SEARCHARTIFACT
            %   Private utility method for searching through the fields of
            %   the cross-release artifact.
            ret = [];
            idx = 0;
            nElem = numel(obj.XRelInfo.(lField));
            for kElem = 1:nElem
                lCurrElem = obj.XRelInfo.(lField)(kElem);
                if fhCriterion(lCurrElem)
                    ret = lCurrElem;
                    idx = kElem;
                    break;
                end
            end
        end
    end
end
