#ifndef _MAIN_UTILS_HPP_
#define _MAIN_UTILS_HPP_

#include <thread>
#include <semaphore.h>
#include <signal.h>

#include <ara/log/logging.hpp>

/* Helper for calling a pointer to member function. */
#define MEMBER_FN_CALL(obj, member) ((obj).*(member))

/* StepFunction handles executing a given step function from a different
 * different thread.  This class manages creating the thread and 
 * communicating with it.
 * 
 * The class interface allows step()'ing the step function once at a time
 * and stopping the thread.
 */
template <typename Model, typename StepFun>
class StepFunction
{
    Model * const model;
    StepFun func;
    std::shared_ptr<sem_t> sem;
    std::thread thread;
    std::atomic_bool halt;

    StepFunction(StepFunction const &) = delete;
    StepFunction & operator=(StepFunction const &) = delete;

    void step_wrapper()
    {
        while(true) {
            int err = sem_wait(sem.get());
            if (halt) {
                ara::log::LogInfo() << "Thread " << pthread_self() << "exiting\n";
                return;
            }

            if (err != 0) {
                ara::log::LogError() << "sem_wait failed: " << std::strerr(errorno);
                continue;
            }

            int sval;
            if (0 == sem_getvalue(sem.get(), &sval) && 1 < sval) {
                ara::log::LogInfo() << "Thread " << pthread_self() << " is behind schedule by " << sval << "\n";
            }

            try {
                MEMBER_FN_CALL(*model, func)();
            } catch (std::exception const & e) {
                ara::log::LogError() << "Error executing step: " << e.what();
            }
        }
    }

public:
    StepFunction(Model * model, StepFun func):
        model(model),
        func(func),
        sem(std::make_shared<sem_t>()),
        halt(false)
    {
        int err = sem_init(sem.get(), 0, 0);
        if (err != 0) {
            ara::log::LogError() << "Unable to initialize semaphore: " << std::strerr(errorno);
            std::exit(EXIT_FAILURE);
        }

        thread = std::thread(&StepFunction::step_wrapper, this);
    }

    ~StepFunction()
    {
        thread.join();
        if (sem_destroy(sem.get()) == -1) {
            ara::log::LogError() << "Destroying semaphore failed: " << std::strerror(errorno);
        }
    }

    void step()
    {
        if (sem_post(sem.get()) == -1) {
            ara::log::LogError() << "Semaphore post failed: " << std::strerror(errorno);
            return;
        }
    }

    void stop()
    {
        halt = true;
        step();
    }
};


class Timer
{
    timer_t timerid;
    double base_rate;

    public:

    Timer(double base_rate, void (*SigAction)(int)):
        base_rate(base_rate)
    {
        struct sigaction action;
        sigemptyset(&action.sa_mask);
        action.sa_handler = SigAction;
        action.sa_flags = SA_RESTART;
        if (sigaction(SIGRTMIN, &action, NULL) == -1) {
            ara::log::LogError() << "Registering signal handler failed: " << std::strerror(errorno);
            std::exit(EXIT_FAILURE);
        }

        struct sigevent event;
        event.sigev_notify = SIGEV_SIGNAL;
        event.sigev_signo = SIGRTMIN;
        event.sigev_value.sival_ptr = 0;
        if (timer_create(CLOCK_REALTIME, &event, &timerid) == -1) {
            ara::log::LogError() << "Timer registration failed: " << std::strerror(errorno);
            std::exit(EXIT_FAILURE);
        }
    }

    void start()
    {
        struct itimerspec timerspec;
        timerspec.it_value.tv_sec = 0;
        timerspec.it_value.tv_nsec = base_rate * 1000000000;
        timerspec.it_interval.tv_sec = 0;
        timerspec.it_interval.tv_nsec = base_rate * 1000000000;
        if (timer_settime(timerid, 0, &timerspec, nullptr) == -1)
            ara::log::LogError() << "Arming the timer failed: " << std::strerror(errorno);
            std::exit(EXIT_FAILURE);
    }

    ~Timer()
    {
        if (timer_delete(timerid) != 0) {
            ara::log::LogError() << "Deleting timer failed: " << std::strerror(errorno);
        }
    }
};

void block_signals()
{
    sigset_t sigset;
    sigemptyset(&sigset);
    sigaddset(&sigset, SIGRTMIN);
    sigaddset(&sigset, SIGTERM);
    if (pthread_sigmask(SIG_SETMASK, &sigset, NULL) == -1) {
        ara::log::LogError() << "Blocking signals failed: " << std::strerror(errorno);
    }
}

void unblock_signals()
{
    sigset_t sigset;
    sigemptyset(&sigset);
    sigaddset(&sigset, SIGRTMIN);
    sigaddset(&sigset, SIGTERM);
    if (pthread_sigmask(SIG_UNBLOCK, &sigset, NULL) == -1) {
        ara::log::LogError() << "Blocking signals failed: " << std::strerror(errorno);
    }
}

#endif // _MAIN_UTILS_HPP_
