classdef (Hidden = true) AUTOSARBuildInfoWrapper < coder.internal.xrelexport.BuildInfoWrapper
    % AUTOSARBUILDINFOWRAPPER
    %
    % Copyright 2017-2018 The MathWorks, Inc.
    %
    properties (Access = private, Constant)
        RteSubFolder = 'toolbox[\\|/]coder[\\|/]autosar[\\|/]rte[\\|/]'
    end
    
    methods (Access = public)
        function obj = AUTOSARBuildInfoWrapper(build_info_contents, xrel_msg)
            % AUTOSARBUILDINFOWRAPPER
            obj = obj@coder.internal.xrelexport.BuildInfoWrapper(build_info_contents, xrel_msg);
        end
        
        function include_files = getBuildDirIncludeFiles(obj)
            % GETBUILDDIRINCLUDEFILES
            
            % We will want to exclude anything that's already handled by
            % the shared utils folder because it will be imported using the
            % ExistingSharedCode mechanism.
            %
            % We need to also exclude any static header files provided by
            % the support package.  These can be installed anywhere, so we
            % cannot assume that they will be underneath $(MATLAB_ROOT).
            %
            % We also cannot assume that we are running in the same context
            % that the code was originally generated from, so cannot use
            % autosarroot here.
            lAdditionalFilesToRemove = [obj.AdditionalFilesToRemove, ...
                {'rtwtypes.h$'}, ...
                {[obj.RteSubFolder, 'Platform_Types.h$']}, ...
                {[obj.RteSubFolder, 'Std_Types.h$']}, ...
                {[obj.RteSubFolder, 'Compiler.h$']}, ...
                {'^\$\(START_DIR\).*[\\|/]stub[\\|/]'}, ...
                {'^\.[\\|/]stub[\\|/]'} ...
                ];
                        
            % To export from releases R2018a and earlier, the way we find
            % the shared utility folder is to look for the SHARED_SRC_DIR
            % BuildArg (if it exists).
            lSharedCodePathFromBuildArg = obj.getSharedCodePathFromBuildArg;
            if ~isempty(lSharedCodePathFromBuildArg)
                lAdditionalFilesToRemove{end+1} = ['^', obj.escapePath(lSharedCodePathFromBuildArg)];
            end
            
            % Get include files from BuildInfo and filter out those that
            % should not be included in the cross-release artifact.
            excludeGroupIDs = {'Legacy','SkipForInTheLoop','SkipForSil'};
            include_files = obj.RTWBuildInfo.getIncludeFiles(true, false, {}, excludeGroupIDs);
            include_files = obj.filter(include_files, lAdditionalFilesToRemove);
            include_files = obj.filter(include_files, obj.AdditionalPathsToRemove);
            [~, fNames, fExts] = cellfun(@fileparts, include_files, 'UniformOutput', false);
            include_files = strcat(fNames, fExts);
            include_files = obj.normalizeList(include_files);
        end
    end
    
    
    methods (Access = protected)
        function lPaths = getBuildInfoPaths(obj, type)
            % GETBUILDINFOPATHS
            lPaths = getBuildInfoPaths@coder.internal.xrelexport.BuildInfoWrapper(obj, type);
            lPaths = obj.filter(lPaths, {obj.RteSubFolder});
            
            if strcmp(type, 'include')
                % Deal with AUTOSAR stub folders from R2011b and earlier
                % These are encoded as ./stub
                % We need to correct this to (e.g.) $(START_DIR)/model_autosar_rtw/stub
                lRelStubFolderRegexp = '\.[\\|/]stub';
                lModelTopSource = [obj.RTWBuildInfo.ModelName,'.c'];
                lModelTopSourcePath = obj.SourceFilePaths(strcmp(lModelTopSource, obj.SourceFiles));
                if ~isempty(lModelTopSourcePath)
                    % In theory we could have a special case where the top
                    % model .c file has been removed/renamed.
                    for kPath = 1:numel(lPaths)
                        if ~isempty(regexp(lPaths{kPath}, lRelStubFolderRegexp, 'once'))
                            lPaths{kPath} = fullfile(lModelTopSourcePath{1}, 'stub');
                        end
                    end
                end
            end
        end
    end
    
    methods (Access = private)
        
        function lDir = getMainBuildFolder(obj)
            % GETMAINBUILDFOLDER
            %
            %   Get the first Source Path with group BuildDir
            lDirCell = obj.RTWBuildInfo.getSourcePaths(false, {'BuildDir'}, {});
            if isempty(lDirCell)
                lDir = '';
            else
                lDir = lDirCell{1};
            end
        end
        
        function lSharedCodePath = getSharedCodePathFromBuildArg(obj)
            % GETSHAREDCODEPATHFROMBUILDARG
            % The shared utility folder might be used, even if there is no
            % shared utility binary - e.g., if rtwtypes.h,
            % zero_crossing_types.h is referenced but no actual utility
            % code was generated.
            %
            % But, we should be able to rely on the BuildArg SHARED_SRC_DIR
            % but this could be relative or absolute depending on the
            % platform (absolute on Windows with nmake-based toolchain).
            lSharedCodePath = '';   % Empty value if early return
            lMainBuildFolder = obj.getMainBuildFolder;
            if isempty(lMainBuildFolder)
                return;
            end
            
            baSharedSrcDir = obj.getBuildArg('SHARED_SRC_DIR');
            if isempty(baSharedSrcDir)
                return;
            end
            
            lRawSharedCodePath = baSharedSrcDir.Value;
            lSharedCodePath = obj.catRelPaths(lMainBuildFolder, lRawSharedCodePath);
        end
        
        function lRet = getBuildArg(obj, lBuildArgKey)
            % GETBUILDARG
            %
            %   RTW.BuildInfo.findKeyValuePair is not present in R2011a and
            %   earlier.
            lRet = [];
            lBuildArgs = obj.RTWBuildInfo.BuildArgs;
            if ~isempty(lBuildArgs)
                lKeys = get(lBuildArgs, 'Key');
                if ~iscell(lKeys)
                    lKeys = {lKeys};
                end
                [lFound, lIdx] = ismember(lBuildArgKey, lKeys);
                if lFound
                    lRet = lBuildArgs(lIdx);
                end
            end
        end
    end
    
    methods (Access = private, Static)
        function lPath = escapePath(lPath)
            % ESCAPEPATH
            lPath = regexprep(lPath, '[\\|/]', '\[\\\\\|\/\]');
            lPath = strrep(lPath, '$(START_DIR)', regexptranslate('escape', '$(START_DIR)'));
        end
    end
    
    methods (Access = private, Static)
        function lFullPath = catRelPaths(lBaseFolder, lRawPath)
            % CATRELPATHS
            %
            % Computes fullfile(lBaseFolder, lRawPath) where lRawPath has
            % '..' in.  If lRawPath is absolute, lFullPath == lRawPath.
            
            % Can't use startsWith as this needs to run all the way
            % back to R2010a
            
            % Detect relative paths based on a leading "." - could be "../"
            % or "./" (or "..\" or ".\").  But hidden folders on 
            % UNIX - e.g. ".myfolder/" must not falsely flag as relative!
            lRawPath = strtrim(lRawPath);   % Just in case string is padded with spaces
            lRawSharedCodePathIsAbsolute = isempty(regexp(lRawPath, '^\.+[\\|/]', 'once'));
            if lRawSharedCodePathIsAbsolute
                lFullPath = lRawPath;
            else
                lNonReducedFullPath = fullfile(lBaseFolder, lRawPath);
                lFullPath = RTW.reduceRelativePath(lNonReducedFullPath);
            end
        end
    end
end