% AUTOSARMODELREFERENCEVALIDATION Implements the validation on a build directory
% containing generated code with Model reference interface
%   See coder.internal.xrelexport.ValidationStrategy and
%   coder.internal.xrelexport.TopModelValidation

%   Copyright 2017 The MathWorks, Inc.

classdef (Hidden = true) AUTOSARModelReferenceValidation < coder.internal.xrelexport.ValidationStrategy
    
    properties (Dependent, GetAccess = public)
        CodeInterface;
    end
       
    properties (GetAccess = public)
        ValidatesModelRefDir = true;
    end
    
    properties (Constant, Access = protected)
        % Needs to be set to satisfy superclass
        GlobalIOErrorID = 'Coder:xrelexport:DoesntSupportGlobalIO';
    end
    
    methods (Access = public)
        function obj = AUTOSARModelReferenceValidation(lXrelMsg, needsInterfaceArtifact)
            obj = obj@coder.internal.xrelexport.ValidationStrategy(lXrelMsg);
            className = mfilename('class');
            validateattributes(needsInterfaceArtifact, {'logical'}, {'nonempty'}, className, 'needsInterfaceArtifact');
            
            % This validation strategy does not apply if an interface
            % artifact is required
            obj.NeedsInterfaceArtifact = needsInterfaceArtifact;
            if needsInterfaceArtifact
                % Error out if needs interface artifact - not yet
                % supported.
                obj.XrelMsg.error('Coder:xrelexport:DoesNotSupportAUTOSARModelRefInterface');
            end
        end
        
        function validate(obj, config_set, CodeInfoWrapper, InfoStructWrapper)
            % do the initial part of the common validation
            obj.doInitialCommonValidation(config_set, CodeInfoWrapper, InfoStructWrapper);
                                    
            % Model reference interface not supported for models using CAPI
            if obj.UsesCAPI(config_set)
                obj.subModelError('Coder:xrelexport:DoesntSupportCAPI');
            end

            if obj.hasCodeVariantControl(InfoStructWrapper)
                % Code variants in referenced models cause Rte_Cfg.h to be
                % included - and that means that the referenced model will
                % not compile.
                obj.XrelMsg.error('autosarstandard:xrelexport:DoesNotSupportModelRefCodeVariants');
            end
                
            if obj.hasTunableCalPrm(CodeInfoWrapper)
                % Tunable calibration parameters cause Rte_Types.h to be
                % included - and that means that the referenced model will
                % not compile.
                obj.XrelMsg.error('autosarstandard:xrelexport:DoesNotSupportModelRefTunableParameters');
            end
            
            if obj.hasCodeVariantParameters(InfoStructWrapper)
                % Code variant parameters cause Rte_Cfg.h to be included -
                % which means that the referenced model will not compile.
                obj.XrelMsg.error('autosarstandard:xrelexport:DoesNotSupportModelRefVariantParameters');
            end
                
            
            % do the final part of the common validation
            obj.doFinalCommonValidation(config_set, CodeInfoWrapper);
        end
    end
    
    methods
        function codeInterface = get.CodeInterface(~)
            % SUBMODEL ONLY
            codeInterface = coder.internal.xrelexport.CodeInterface.SubModel;
        end
    end
    
    methods (Access = private, Static)
        function lHasVariantControl = hasCodeVariantControl(lInfoStructWrapper)
            % HASCODEVARIANTCONTROL
            lModelInterface = lInfoStructWrapper.ModelInterface;
            lHasVariantControl = isfield(lModelInterface, 'CodeVariants') && ...
                    isfield(lModelInterface.CodeVariants, 'VariantControl') && ...
                    isfield(lModelInterface.CodeVariants.VariantControl, 'Variant') && ...
                    ~isempty(lModelInterface.CodeVariants.VariantControl.Variant);
        end
        
        function lHasVariantParameters = hasCodeVariantParameters(lInfoStructWrapper)
            % HASCODEVARIANTPARAMETERS
            lModelInterface = lInfoStructWrapper.ModelInterface;
            lHasVariantParameters = isfield(lModelInterface, 'CodeVariants') && ...
                    isfield(lModelInterface.CodeVariants, 'VariantParameter') && ...
                    isfield(lModelInterface.CodeVariants.VariantParameter, 'Pair') && ...
                    ~isempty(lModelInterface.CodeVariants.VariantParameter.Pair);
        end
        
        function lHasTunableCalPrm = hasTunableCalPrm(lCodeInfoWrapper)
            % HASTUNABLECALPRM
            fh = @(x)x.IsAUTOSAR && ~x.HasImplementation;
            lHasTunableCalPrm = false;
            % Cannot use arrayfun on MCOS arrays in R2010a
            for lCurrParam = lCodeInfoWrapper.Parameters
                if fh(lCurrParam)
                    lHasTunableCalPrm = true;
                    break;
                end
            end
        end
    end
end
