% AUTOSARMODELVALIDATION
%
% Copyright 2017-2018 The MathWorks, Inc.

classdef (Hidden = true) AUTOSARModelValidation < coder.internal.xrelexport.ValidationStrategy
    
    properties (GetAccess = public)
        CodeInterface = coder.internal.xrelexport.CodeInterface.AUTOSAR;
        ValidatesModelRefDir = false;
    end
    
    properties (Constant, Access = protected)
        GlobalIOErrorID = 'Coder:xrelexport:RequiresFunctionPrototypeControl';
    end
    
    methods (Access = public)
        function obj = AUTOSARModelValidation(lXrelMsg)
            obj = obj@coder.internal.xrelexport.ValidationStrategy(lXrelMsg);
        end
        
        function validate(obj, config_set, CodeInfoWrapper, InfoStructWrapper)
            % do the initial part of the common validation
            obj.doInitialCommonValidation(config_set, CodeInfoWrapper, InfoStructWrapper);
            
            % Error status ports for queued data ports are not supported
            for kPort = 1:CodeInfoWrapper.NumMdlInputs
                if CodeInfoWrapper.Inports(kPort).IsAUTOSAR && ...
                        CodeInfoWrapper.Inports(kPort).AUTOSARData.IsAutosarErrorStatus
                    lOriginalPort = obj.getDataPortForErrorStatusPort(CodeInfoWrapper.Inports(kPort), ...
                        CodeInfoWrapper.Inports, ...
                        CodeInfoWrapper.ExternalInportMapping);
                    lIsQueued = lOriginalPort.AUTOSARData.IsQueued;
                    if lIsQueued
                        obj.XrelMsg.error('autosarstandard:xrelexport:DoesNotSupportQueuedErrorStatus');
                    end
                end
            end
            
            % Tunable parameters that do not use AUTOSAR storage classes
            % cannot be permitted if they are imported
            for kParam = 1:CodeInfoWrapper.NumMdlParameters
                if CodeInfoWrapper.Parameters(kParam).IsAUTOSAR && ...
                        ~CodeInfoWrapper.Parameters(kParam).AUTOSARData.IsAutosarCalibration && ...
                        CodeInfoWrapper.Parameters(kParam).IsImported
                    obj.XrelMsg.error('autosarstandard:xrelexport:DoesNotSupportParameterImportedStorageClass');
                end
            end
            
            % Check for multiple symbolic dimensions - can't use arrayfun
            % in R2010a
            fhIsMultiSym = @(x)(sum(isinf(x.DataType.getDimensions)) > 1);
            for lCurrDataInterface = [CodeInfoWrapper.Inports(:)', CodeInfoWrapper.Outports(:)', ...
                    CodeInfoWrapper.Parameters(:)', CodeInfoWrapper.DataStores(:)']
                if fhIsMultiSym(lCurrDataInterface)
                    obj.XrelMsg.error('autosarstandard:xrelexport:DoesNotSupportMultiDimensionalSymbolicDimensions');
                end
            end
                        
            % Can't support obsolete Server Operation modeling style
            if ~isempty(CodeInfoWrapper.getRealTimeModelDW)
                obj.XrelMsg.error('Coder:xrelexport:DoesNotSupportReusableServerOperation');
            end
            
            % Cannot support per-instance memories that need NVRAM Access
            if ~isempty(InfoStructWrapper.GeneralDataFromTLC)
                lGeneralDataFromTLC = InfoStructWrapper.GeneralDataFromTLC;
                if isfield(lGeneralDataFromTLC, 'AutosarPIMDSM')
                    lAutosarPIMDSM = lGeneralDataFromTLC.AutosarPIMDSM;
                    if ~isempty(lAutosarPIMDSM)
                        if iscell(lAutosarPIMDSM)
                            % Re-interpret as struct array
                            lAutosarPIMDSM = [lAutosarPIMDSM{:}];
                        end
                        
                        lNeedsNVRAM = [lAutosarPIMDSM.NeedsNVRAMAccess];
                        if isfield(lAutosarPIMDSM, 'IsArTypedPerInstanceMemory')
                            lIsCTyped = ~[lAutosarPIMDSM.IsArTypedPerInstanceMemory];
                        else
                            lIsCTyped = ones(1, numel(lNeedsNVRAM));
                        end
                            
                        if any( (lNeedsNVRAM & lIsCTyped ))
                            obj.XrelMsg.error('autosarstandard:xrelexport:DoesNotSupportAUTOSARCTypedPerInstanceMemoriesNVRAM');
                        end
                    end
                end
            end
            
            % Cannot support lookup tables in R2015a or earlier
            if CodeInfoWrapper.HasAutosarLookupTables &&  verLessThan('simulink', '8.6')
                obj.XrelMsg.error('autosarstandard:xrelexport:LookupTablesNotSupportedInR2015aOrEarlier');
            end
            
            % Cannot support server call points with non-scalar bus input
            % or output arguments
            for kClientCaller = 1:numel(CodeInfoWrapper.ServerCallPoints)
                lPrototype = CodeInfoWrapper.ServerCallPoints(kClientCaller).Prototype;
                lArgTypes = [lPrototype.Arguments.Type];
                for kArgType = 1:numel(lArgTypes)
                    lCurrType = lArgTypes(kArgType);
                    if lCurrType.isMatrix && prod(lCurrType.Dimensions) > 1 && lCurrType.BaseType.isStructure
                        obj.XrelMsg.error('autosarstandard:xrelexport:DoesNotSupportFunctionCallerArgsWithNonScalarBus', ...
                            lPrototype.Arguments(kArgType).Name, ...    % Argument name
                            lPrototype.Name, ...    % Function name
                            lCurrType.BaseType.Name, ... % Type Name
                            mat2str(lCurrType.Dimensions) ...
                            );
                    end
                end 
            end
            
            % Cannot support Reset functions with side IO in R2016b
            if strcmp(version('-release'), '2016b')
                lModelInterface = InfoStructWrapper.ModelInterface;
                lPortFieldList = {'Inports', 'Outports'};
                for lCurrPortField = lPortFieldList    % For each in cell array
                    if isfield(lModelInterface, lCurrPortField{:})
                        lPorts = obj.standardizeInfoStructPort(lModelInterface.(lCurrPortField{:}));
                    else
                        lPorts = {};
                    end
                    lPortIsReset = cellfun(@(x)obj.infoStructPortHasResetTiming(x), lPorts);
                    if any(lPortIsReset)
                        obj.XrelMsg.error('autosarstandard:xrelexport:DoesNotSupportResetFunctionWithSideIOInR2016b');
                    end
                end
            end
            
            % do the final part of the common validation
            obj.doFinalCommonValidation(config_set, ...
                CodeInfoWrapper);
        end 
    end
    
    methods (Access = private, Static)
        
        
        function lIsReset = infoStructPortHasResetTiming(lCurrPort)
            % INFOSTRUCTPORTHASRESETTIMING
            lIsReset = false;
            
            % Structs rather than objects - so cannot assume fields always
            % exist.
            if isfield(lCurrPort, 'SampleTime') && isfield(lCurrPort.SampleTime, 'Period')
                lPeriod = lCurrPort.SampleTime.Period; 
                % IRT events have a period of rtInf
                if strcmp(lPeriod, 'rtInf')
                    % Initialize/Terminate events have offsets of 1.0 and 2.0
                    % respectively.  Anything above this is assumed to be a
                    % Reset timing event.
                    lNumericOffset = str2double(lCurrPort.SampleTime.Offset);  % Offset stored as char array in binfo
                    lIsReset = (lNumericOffset >= 3.0);
                end
            end
            
        end
        
        function lOriginalPort = getDataPortForErrorStatusPort(lCurrInport, lInports, lExternalInportMapping)
            % GETDATAPORTFORERRORSTATUSPORT
            assert(lCurrInport.IsAUTOSAR && lCurrInport.AUTOSARData.IsAutosarErrorStatus, ...
                'getDataPortForErrorStatusPort only applies to ErrorStatus ports');
            lOriginalPortNumberStr = lCurrInport.AUTOSARData.AutosarReceiverPortNumber;
            assert(~isempty(lOriginalPortNumberStr), 'Empty AutosarReceiverPortNumber');
            lOriginalPortNumber = str2double(lOriginalPortNumberStr);
            lGraphicalName = lExternalInportMapping(lOriginalPortNumber).GraphicalName;
            [lFound, lIdx] = ismember(lGraphicalName, {lInports.GraphicalName});
            assert(lFound, 'Could not find Inport with GraphicalName %s', lGraphicalName);
            assert(numel(lIdx)==1, 'Multiple ports with GraphicalName %s', lGraphicalName);
            lOriginalPort = lInports(lIdx);
        end
        
        function lStandardizedPorts = standardizeInfoStructPort(lPorts)
            % STANDARDIZEINFOSTRUCTPORT
            if isempty(lPorts)
                lStandardizedPorts = {};
            elseif iscell(lPorts)
                lStandardizedPorts = lPorts;
            elseif isstruct(lPorts)
                lStandardizedPorts = {lPorts};
            else
                assert(false, 'Unrecognized storage of port information in modelInterface');
            end
        end
    end
end
