classdef AUTOSARSimulinkFunctionInfo < coder.internal.xrelexport.SimulinkFunctionInfo
    % AUTOSARSIMULINKFUNCTIONINFO
    %
    %   Copyright 2017-2018 The MathWorks, Inc.
    properties (SetAccess = immutable)
        AUTOSAROperationName;
        AUTOSARPortName;
        FunctionInputs;
        FunctionOutputs;
        FunctionReturn;
    end
    
    properties (Access = private, Constant)
        EmptyInputsOrOutputs = struct('Name', {}, 'Dimensions', {}, 'SymbolicDimensions', {}, 'SymbolicWidth', {}, 'ArgIndex', {});
    end
    
    methods (Access = public)
        function obj = AUTOSARSimulinkFunctionInfo(lSimulinkFunctionStruct, lCodeInfo)
            % AUTOSARSIMULINKFUNCTIONINFO
            obj = obj@coder.internal.xrelexport.SimulinkFunctionInfo(lSimulinkFunctionStruct);
            obj.AUTOSAROperationName = lSimulinkFunctionStruct.AUTOSAROperationName;
            obj.AUTOSARPortName = lSimulinkFunctionStruct.AUTOSARPortName;
            
            % Function prototype control can be used to change the
            % order of arguments.  The "ArgIndices" field has been
            % present since Simulink Functions were introduced in
            % R2014b.
            lArgIndices = lSimulinkFunctionStruct.ArgIndices;
            lRetArgIdx = lSimulinkFunctionStruct.ReturnArgIndex;
            
            % We need the dimensions for the inputs and outputs
            %            
            % For AUTOSAR, a simulink function is either defined or called
            % in a component, but not both.
            if strcmp(lSimulinkFunctionStruct.IsDefined, 'yes')
                % If the simulink function has been defined, then it will
                % appears in CodeInfo's list of OutputFunctions 
                
                if isfield(lSimulinkFunctionStruct, 'IsScoped') && strcmp(lSimulinkFunctionStruct.IsScoped, 'yes')
                    % If the simulink function is scoped
                    obj.FunctionInputs = obj.EmptyInputsOrOutputs;
                    obj.FunctionOutputs = obj.EmptyInputsOrOutputs;
                else
                    % We can use arrayfun here, because Simulink Functions did not
                    % exist in R2010a.
                    %
                    % For server functions, we can be confident that
                    % CGFunctionName and the function prototype name will
                    % be the same.
                    bIsMatch = arrayfun(@(x)strcmp(x.Prototype.Name, obj.CGFunctionName), lCodeInfo.AllOutputFcns);
                    lCurrPrototype = [lCodeInfo.AllOutputFcns(bIsMatch).Prototype];
                    
                    [obj.FunctionInputs, obj.FunctionOutputs] = obj.getFunctionInputsAndOutputsFromPrototype(lCurrPrototype, lArgIndices, lRetArgIdx);
                end
            else
                % If not defined, it must be called, or it would not be
                % there.
                assert(strcmp(lSimulinkFunctionStruct.IsCalled, 'yes'));
                bIsMatch = arrayfun(@(x)obj.matchServerCallPointPrototype( ...
                    obj.CGFunctionName, ...
                    obj.AUTOSAROperationName, ...
                    obj.AUTOSARPortName, ...
                    x), ...
                    lCodeInfo.ServerCallPoints);
                
                lCurrPrototype = [lCodeInfo.ServerCallPoints(bIsMatch).Prototype];
                [obj.FunctionInputs, obj.FunctionOutputs, obj.FunctionReturn] = obj.getFunctionInputsAndOutputsFromPrototype(lCurrPrototype, lArgIndices, lRetArgIdx);
            end
        end
    end
    
    methods (Access = private)
        function [lFunctionInputs, lFunctionOutputs, lFunctionReturn] = getFunctionInputsAndOutputsFromPrototype(obj, lCurrPrototype, lArgIndices, lRetArgIdxAsNum)
            % GETFUNCTIONINPUTSANDOUTPUTS
            
            % We should always be able to find the function we are
            % looking for, and it should be unique.
            assert(numel(lCurrPrototype) >= 1, 'Could not find RTW.FunctionInterface object for Simulink Function %s', obj.CGFunctionName);
            
            if numel(lCurrPrototype) > 1
                % Even if a simulink function is defined singly in
                % infoStruct, there might be multiple call instances in
                % CodeInfo.  We only need one of them.  They should all be
                % identical!
                lCurrPrototype = lCurrPrototype(1);
            end
            
            if isempty(lArgIndices)
                lArgIndexCell = {};
            else
                lArgIndexCell = mat2cell(lArgIndices, ones(1,size(lArgIndices,1)), 2);
            end
            
            bIsInput = ismember({lCurrPrototype.Arguments.IOType}, {'INPUT', 'RTW_IO_INPUT'});
            bIsOutput = ismember({lCurrPrototype.Arguments.IOType}, {'OUTPUT', 'RTW_IO_OUTPUT'});
            
            lInputArgs = lCurrPrototype.Arguments(bIsInput);
            lInputArgIndices = lArgIndexCell(bIsInput);
            
            lOutputArgs = lCurrPrototype.Arguments(bIsOutput);
            lOutputArgIndices = lArgIndexCell(bIsOutput);
            
            lReturnArgs = lCurrPrototype.Return;
            if isempty(lReturnArgs)
                lReturnArgIndex = {};
            else
                lReturnArgIndex = {sprintf('O%d', lRetArgIdxAsNum)};
            end
            
            
            if isempty(lInputArgs)
                lFunctionInputs = obj.EmptyInputsOrOutputs;
            else
                lFunctionInputs = arrayfun(@(x,y)obj.getDimensionsForArg(x,y), lInputArgs(:), lInputArgIndices(:));
            end
            
            if isempty(lOutputArgs)
                lFunctionOutputs = obj.EmptyInputsOrOutputs;
            else
                lFunctionOutputs = arrayfun(@(x,y)obj.getDimensionsForArg(x,y), lOutputArgs(:), lOutputArgIndices(:));
            end
            
            if isempty(lReturnArgs)
                lFunctionReturn = obj.EmptyInputsOrOutputs;
            else
                lFunctionReturn = arrayfun(@(x,y)obj.getDimensionsForArg(x,y), lReturnArgs(:), lReturnArgIndex(:));
            end
        end
    end
    
    methods (Access = private, Static)
        function bIsMatch = matchServerCallPointPrototype(lCGFunctionName, lAUTOSAROperationName, lAUTOSARPortName, lServerCallPoint)
            % MATCHSERVERCALLPOINTPROTOTYPE
            %
            % For regular simulink functions
            if isempty(lServerCallPoint.PortName)
                bIsMatch = strcmp(lServerCallPoint.Prototype.Name, lCGFunctionName);
            else
                % For functions mapped to a Port/Operation
                bIsMatch = strcmp(lServerCallPoint.PortName, lAUTOSARPortName) && ...
                    strcmp(lServerCallPoint.Prototype.Name, lAUTOSAROperationName);
            end
        end
        
        function ret = getDimensionsForArg(lArg, lIdxStr)
            % GETDIMENSIONSFORARG
            
            lName = lArg.Name;
            
            if lArg.Type.isMatrix
                % Get matrix dimensions
                lDimensions = lArg.Type.Dimensions;
            else
                % Scalars always have dimensions 1
                lDimensions = 1;
            end
            
            % Avoid nesting if statements to avoid having to set default
            % value of lSymbolicDimensions in multiple places.
            if lArg.Type.isMatrix && isprop(lArg.Type, 'SymbolicDimensions') && lArg.Type.HasSymbolicDimensions
                lSymbolicDimensions = lArg.Type.SymbolicDimensions;
                lSymbolicWidth = lArg.Type.SymbolicWidth;
            else
                lSymbolicDimensions = {};
                lSymbolicWidth = '';
            end
            ret = struct('Name', {lName}, ...
                'Dimensions', {lDimensions}, ...
                'SymbolicDimensions', {lSymbolicDimensions}, ...
                'SymbolicWidth', {lSymbolicWidth}, ...
                'ArgIndex', lIdxStr);
        end
    end
end
