classdef AUTOSARTriggeredPeriodicInfo < handle
    % AUTOSARTRIGGEREDPERIODICINFO
    %
    %   Abstraction identify from CodeInfo and InfoStruct the list of
    %   runnables that are "Periodic" but need the function trigger port to
    %   be marked as "Triggered".  These have a defined sample time, but do
    %   not require a parent model to issue the trigger at every sample
    %   instant.
    %
    % Copyright 2017-2018 The MathWorks, Inc.
    
    properties (Access = private)
        ExternalInportMapping;
    end
    
    properties (SetAccess = private)
            TriggeredPeriodicRunnableSymbols = {};
    end
    
    properties (Dependent, SetAccess = private)
        TriggeredPeriodicRunnableSymbolsString;
    end
    
    
    methods (Access = public)
        function obj = AUTOSARTriggeredPeriodicInfo(lCodeInfoWrapper, lInfoStructWrapper)
            
            % AUTOSARTRIGGEREDPERIODICINFO
            obj.ExternalInportMapping = lCodeInfoWrapper.ExternalInportMapping;
            
            % ExternalInportMapping will be populated in CodeInfoWrapper
            % even for non-export functions models.
            
            
            lModelInterface = lInfoStructWrapper.ModelInterface;
            
            if isfield(lModelInterface, 'ExternalPortGroups')
                lExternalPortGroups = obj.standardizeBinfoElement(lModelInterface.ExternalPortGroups);
                for kPortGroup = 1:numel(lExternalPortGroups)
                    lFcnCallInputPort_Base1 = lExternalPortGroups{kPortGroup}.FcnCallInputPort + 1;    % Change from zero-to-one-based indexing
                    
                    if isfield(lExternalPortGroups{kPortGroup}, 'IsPeriodicFcnCall')    % Recent releases
                        lIsPeriodicFcnCall = strcmp(lExternalPortGroups{kPortGroup}.IsPeriodicFcnCall, 'true');
                    elseif isfield(lExternalPortGroups{kPortGroup}, 'NeedRunTimeCheck')
                        lIsPeriodicFcnCall = strcmp(lExternalPortGroups{kPortGroup}.NeedRunTimeCheck, 'true');
                    else
                        lIsPeriodicFcnCall = true;  % Conservatively assume we need the run time check if we have no other data to go on
                    end
                    
                    if ~lIsPeriodicFcnCall
                        if lFcnCallInputPort_Base1 > 0  % Can be negative for Simulink Functions
                            lPortMapping = obj.ExternalInportMapping(lFcnCallInputPort_Base1);
                            assert(lPortMapping.IsFunctionCall, 'Tried to get function call info from a data port');
                            assert(~isempty(lPortMapping.Symbol), 'Function callers need to call an output function - no symbol found');
                            obj.TriggeredPeriodicRunnableSymbols{end+1} = lPortMapping.Symbol;
                        end
                    end
                end
            else
                % If the field does not exist, do nothing.
            end
        end
    end
    
    methods
        function ret = get.TriggeredPeriodicRunnableSymbolsString(obj)
            % GET.TRIGGEREDPERIODICRUNNABLESYMBOLSSTRING
            ret = obj.cell2str(obj.TriggeredPeriodicRunnableSymbols);
        end
    end
    
    methods (Access = private, Static)
        function lOutBinfoElement = standardizeBinfoElement(lBinfoElement)
            if isempty(lBinfoElement)
                lOutBinfoElement = {};
            elseif isstruct(lBinfoElement)
                lOutBinfoElement = {lBinfoElement};
            elseif iscell(lBinfoElement)
                lOutBinfoElement = lBinfoElement;
                return;
            else
                assert(false, 'BInfo element not of an expected form');
            end
        end
        
        function lOutStr = cell2str(lInCellStr)
            % CELL2STR
            if isempty(lInCellStr)
                % In R2010a-R2011a if we don't special case this, we will
                % end up with a cell array with a single entry containing
                % an empty char array rather than an empty cell array.
                lOutStr = '{}';
            else
                lQuotedCellStr = strcat('''', lInCellStr , '''');
                lOutStr = ['{', ...
                    coder.internal.xrelexport.strjoin( lQuotedCellStr, ', '), ...
                    '}'];
            end
        end
    end
    
end
    
